/*--------------------------------------------------------------------
 * $Id: phasetable.c,v 1.15 2008/12/22 16:22:01 bernhard Exp $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 * 
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

/* @45c@ */
/* @c45@ */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "phasetable.h"
#include "numeric.h"
#include "miecalc.h"
#include "mystic.h"

#ifndef PI
#define PI 3.14159265358979323846264338327
#endif

/* number of entries in the cumulative phase function lookup table */
#define NPHASE 20000


/* prototypes of internal functions */
static void free_pft (pft *p);


/***********************************************************************************/
/* Function: setup_Legendre_table                                         @45_30i@ */
/* Description:                                                                    */
/*  Calculate lookup-tables of phase function and cumulative phase function        */
/*  for a given wavelength index iv from the single scattering properties          */
/*  stored in ssprop                                                               */
/*                                                                                 */
/* Parameters:                                                                     */
/* Return value:                                                                   */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                        @i45_30@ */
/***********************************************************************************/

int setup_Legendre_table (phase_function_table *phase,
			  ssprop_struct *ssprop,
			  int iv,
			  double rmin, double rmax, 
			  double delta_scaling_mucut, double truncate, 
			  int quiet) /*TZ ds*/
{
  int i=0, l=0, ip=0, status=0;
  int imin=0, imax=0;

  
  double r0=0, r0_old, dr=0;
  size_t nreff=0, nreffact=0, nphase=0;

  double **mu=NULL;
  int *n=NULL;

  double **p=NULL;
  double **legen=NULL;
  
  if (rmin > rmax) {
    fprintf (stderr, "Error, found rmin > rmax; this message indicates that you told MYSTIC\n");
    fprintf (stderr, "to use a Mie phase function for a cloud of zero optical thickness;\n");
    fprintf (stderr, "Please change either of these two because this causes trouble internally!\n");
    return -1;
  }

  /* get radius intervals */
  r0     = ssprop->r0[iv];
  dr     = ssprop->dr[iv];
  nreff  = ssprop->nreff[iv];
  nphase = ssprop->nphase;


  /* check if the range of radii specified in the 3D input file is covered; */
  /* if only one radius is provided and its value is negative, the phase    */
  /* function is assumed to be constant with location.                      */
     
  if (nreff==1 && r0<0) { 
    fprintf (stderr, " ... found only one (negative) effective radius; assuming\n");
    fprintf (stderr, " ... phase function is constant with location.\n");
  }
  else {
    if (rmin < r0 || rmax > r0 + (double) (nreff-1) * dr) {
      fprintf (stderr, "Error, range of effective radii provided is not sufficient:\n");
      fprintf (stderr, " provided: %g - %g, required: %g - %g\n", 
	       r0, r0 + (double) (nreff-1) * dr, rmin, rmax);
      return -1;
    }
  }
  

  /* determine the required range of effective radii */
  imin=0;
  while (r0 + (double) imin * dr < rmin) {
    imin++;
    if (imin==nreff) {
      imin = nreff-1;
      break;
    }
  }

  imin = (imin>0?imin-1:0);
  
  imax=nreff-1;
  while (r0 + (double) imax * dr > rmax) {
    imax--;
    if (imax==-1) {
      imax = 0;
      break;
    }
  }
  
  imax = (imax<nreff-1?imax+1:nreff-1);

  r0_old = r0;
  r0 = r0_old + (double) imin * dr;

  nreffact = (imax-imin+1);

  if (!quiet)
    fprintf (stderr, " ... using %d radii between %g and %g\n", 
	     (int)nreffact, r0_old + (double) imin * dr, r0_old + (double) imax * dr);


  calloc_phase_function_table (phase, nreffact, r0, dr);
  
  for (i=0; i<nreffact; i++) {   /* calculate only those tables which are really needed */
    
    if (!quiet)
      fprintf (stderr, " ... calculating Legendre table for reff = %g\n", 
	       r0_old + (double) (imin+i) * dr);


    switch (ssprop->type) {
    case PHASE_MOMENTS:
     
      /* phase function is derived from legendre coefficients */
      /* first evaluate phase function at a predefined grid   */
      legen=calloc (nphase, sizeof(double*));

      for (ip=0; ip<nphase; ip++)
        legen[ip]=calloc(ssprop->nleg[iv][imin+i], sizeof(double) );
     
      for (ip=0; ip<nphase; ip++)
        for (l=0; l<ssprop->nleg[iv][imin+i]; l++)
          legen[ip][l] = (double) ssprop->legen[iv][imin+i][ip][l];
      
      /* calculate table of phase function p(mu) */
      status = calc_Legendre_phase (legen, ssprop->nleg[iv][imin+i], nphase, &mu, &p, &n);
      
      if (status!=0) {
	fprintf (stderr, "Error %d returned by calc_Legendre_phase()\n", status);
	return status;
      }
      
      status = calc_cumulative_table (mu, p, n, nphase, truncate, phase->iphase[i],
                                      delta_scaling_mucut, quiet);/*TZ ds*/
      if (status!=0) {
	fprintf (stderr, "Error %d calculating cumulative probability table\n", status);
	return status;
      }
      
      for(ip=0; ip<nphase; ip++){
        free(legen[ip]);
        free(mu[ip]);
        free(p[ip]);
      }
      
      free(legen); free(mu); free(p); free(n);
      break;

    case PHASE_EXPLICIT:
      /* phase function has been given explicitly    */
      /* calculate table of phase function p(mu)     */
      for (ip=0; ip<nphase; ip++)
        mu[ip]=calloc(nphase, sizeof(double *));
      
      mu[ip]=calloc(ssprop->nmu[iv][imin+i], sizeof(double));
      n=calloc(nphase, sizeof(int));
      
      for(ip=0; ip<nphase; ip++){
        for(i=0; i<ssprop->nmu[iv][imin+i]; i++)
          mu[ip][i]=ssprop->mu[iv][imin+i][i];
        n[ip]=ssprop->nmu[iv][imin+i];
      }
      
      status = calc_cumulative_table (mu, ssprop->p[iv][imin+i], 
                                      n, ssprop->nphase,
				      truncate, phase->iphase[i], delta_scaling_mucut,
				      quiet);/*TZ ds*/
      if (status!=0) {
	fprintf (stderr, "Error %d calculating cumulative probability table\n", status);
	return status;
      }
       
      for(ip=0; ip<nphase; ip++){
        free(mu[ip]);
      }
      
      free(mu); free(n);
      
      break;
      
    default:
      fprintf (stderr, "Error, unknown phase function specification %d\n", ssprop->type);
      return -1;
    }

    /* copy extinction coefficient to final destination */
    phase->ext[i]   = ssprop->extinc[iv][imin+i];

    /* copy scattering coefficient to final destination */
    phase->sca[i]   = ssprop->extinc[iv][imin+i] * ssprop->albedo[iv][imin+i];

    /* copy delta scaling factor to final destination */
    phase->f[i] = ssprop->f[iv][imin+i];

    /* copy delta scaling factor TZ to final destination */
    phase->dscale[i] = phase->iphase[i]->dscale;

    if (!quiet)
      fprintf (stderr, "DDSSCALE %f\n", phase->dscale[i]);

  }
  
  /* calculate extinction and scattering increments */
  for (i=0; i<nreffact-1; i++) {
    phase->dext   [i] = (phase->ext   [i+1] - phase->ext   [i]) / dr;
    phase->dsca   [i] = (phase->sca   [i+1] - phase->sca   [i]) / dr;
  }

  return 0;
}


/***********************************************************************************/
/* Function: free_pft                                                     @45_30i@ */
/* Description:                                                                    */
/*  Free memory of struct pft; used by free_phase_function_table().                */
/*                                                                                 */
/* Parameters:                                                                     */
/* Return value:                                                                   */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                        @i45_30@ */
/***********************************************************************************/

static void free_pft (pft *p)
{
  int j=0, ip=0;

  for (j=0; j<p->nscales; j++) {
    for (ip=0; ip<p->nphase; ip++) 
      free (p->p[j][ip]);
    free (p->p[j]);
    free (p->F[j]);
    free (p->A[j]);
    free (p->B[j]);
    free (p->C[j]);
  }

  free (p->p);
  free (p->F);
  free (p->A);
  free (p->B);
  free (p->C);

  for (ip=0; ip<p->nphase; ip++) 
    free (p->mu[ip]);
  free (p->mu);
  
}


/***********************************************************************************/
/* Function: calloc_phase_function_table                                  @45_30i@ */
/* Description:                                                                    */
/*  Allocate memory for struct phase_function_table                                */
/*                                                                                 */
/* Parameters:                                                                     */
/* Return value:                                                                   */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                        @i45_30@ */
/***********************************************************************************/

void calloc_phase_function_table (phase_function_table *p, int n, double r0, double dr)
{
  int i=0;

  p->n  = n;
  p->r0 = r0;
  p->dr = dr;

  p->iphase = calloc ( n, sizeof(pft *));

  p->ext  = calloc ( n, sizeof(double));
  p->dext = calloc ( n, sizeof(double));

  p->sca  = calloc ( n, sizeof(double));
  p->dsca = calloc ( n, sizeof(double));

  p->f   = calloc ( n, sizeof(double));
  p->dscale = calloc ( n, sizeof(double));

  for (i=0; i<n; i++)
    p->iphase[i] = calloc (1, sizeof(pft));
}


/***********************************************************************************/
/* Function: free_phase_function_table                                    @45_30i@ */
/* Description:                                                                    */
/*  Free memory of struct phase_function_table                                     */
/*                                                                                 */
/* Parameters:                                                                     */
/* Return value:                                                                   */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                        @i45_30@ */
/***********************************************************************************/

void free_phase_function_table (phase_function_table *p)
{
  int i=0;

  for (i=0; i<p->n; i++)
    free_pft (p->iphase[i]);

  free (p->iphase);

  free (p->ext);
  free (p->dext);

  free (p->sca);
  free (p->dsca);

  free (p->f);
  free (p->dscale);

}




/***********************************************************************************/
/* Function: calc_Legendre_phase                                          @45_30i@ */
/* Description:                                                                    */
/*  Calculate the phase function from Legendre coefficients                        */ 
/*  for a set of polar angles.                                                     */
/*                                                                                 */
/* Parameters:                                                                     */
/* Return value:                                                                   */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                        @i45_30@ */
/***********************************************************************************/

int calc_Legendre_phase (double **pmom, int nmom, int nphase, double ***mu, double ***p, int **n)
{
  int l=0, it=0, ip=0;

  double pl=0, plm1=0, plm2=0;
  double factor=0, sum=0, increment=0;

  int L=nmom;
  double **f=NULL;

  f=calloc(nphase, sizeof(double*));
  for (ip=0; ip<nphase; ip++)
    f[ip]=calloc(nmom, sizeof(double) );

  *n=calloc (nphase, sizeof(int));      
 
  /* multiply by (2l+1) to convert from 'moments of the phase function' */
  /* to Legendre coefficients;                                          */
  for (ip=0; ip<nphase; ip++)
    for (l=0; l<L; l++)
      f[ip][l] = pmom[ip][l] * (double) (2*l+1);
  
  /* normalize P11 to 1 */
  factor = 1.0 / f[0][0];

  for (ip=0; ip<nphase; ip++){
    for (l=0; l<L; l++)
      f[ip][l] *= factor;
  
  /* number of entries in the cumulative phase function lookup table */
  /* default: all phase function elemens have the same number of entries*/
    (*n)[ip] = NPHASE;
  }
  *mu=calloc (nphase, sizeof(double*)); 
  *p=calloc (nphase, sizeof(double*));
  
  for (ip=0; ip<nphase; ip++){
    (*mu)[ip] = calloc ((*n)[ip], sizeof(double));
    (*p)[ip] = calloc ((*n)[ip], sizeof(double));
  }

  
  for (ip=0; ip<nphase; ip++){
    increment = PI/(double) ((*n)[ip]-1);
  
    for (it=0; it<(*n)[ip]; it++) {
      (*mu)[ip][it] = -cos((double) it * increment);
      
      plm2 = 1;
      plm1 = (*mu)[ip][it];
      
      sum = plm2*f[ip][0] + plm1*f[ip][1];
      
      for (l=2; l<L; l++) {
        pl = ( (double) (2*l-1) * (*mu)[ip][it] * plm1 - (double) (l-1) * plm2) / (double) l;
        sum += f[ip][l]*pl;
        
        plm2=plm1;
        plm1=pl;
      }
      
      (*p)[ip][it] = sum;
      /* fprintf(stderr, "YYY %d %g %g \n", ip, (*mu)[ip][it], (*p)[ip][it] ); */
      
      if ((*p)[ip][it]<0 && ip==0) {
        fprintf (stderr, "Warning, phase function %g negative at mu = %g!\n", 
                 (*p)[ip][it], (*mu)[ip][it]);
      }
      
    }
    
  }

  
  /* free memory */
  for (ip=0; ip<nphase; ip++)
    free(f[ip]); 
  free(f);
 
  return 0;
}


/***********************************************************************************/
/* Function: calc_cumulative_table                                        @62_30i@ */
/* Description:                                                                    */
/*  Calculate a table of cumulative probabilities F1[], directions mu1[],          */
/*  and slopes D1[] for use with sc_mu(); the table is calculated from the         */
/*  phase function itself, not from its Legendre coefficients.                     */
/*                                                                                 */
/* Parameters:                                                                     */
/* Return value:                                                                   */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                        @i62_30@ */
/***********************************************************************************/

int calc_cumulative_table (double **mu, double **p, int *n, int nphase, double truncate, 
			   pft *iphase, double delta_scaling_mucut, int quiet) /*TZ ds*/
{
  
  int i=0, j=0, it=0, ip=0;
  double *F=NULL;
  double alpha=0, beta=0, factor=0;
  int itcut=0; /*TZ ds*/
  double delta_scaling_factor=0;/*TZ ds*/
  float scr=0;

  if (mu[0][0]!=-1.0) {
    fprintf (stderr, "Error, first mu in phase function must be -1.0!\n");
    fprintf (stderr, "Found %.10f\n", mu[0][0]);
    return -1;
  }

  if (mu[0][n[0]-1]!=1.0) {
    fprintf (stderr, "Error, last mu in phase function must be +1.0!\n");
    fprintf (stderr, "Found %.10f\n", mu[0][n[0]-1]);
    return -1;
  }
  
  /* calculate cumulative phase function by linear integration */
  F = calloc (n[0], sizeof(double));
  F[0] = 0;
  for (i=1; i<n[0]; i++)
    F[i] = F[i-1] + (mu[0][i]-mu[0][i-1])*(p[0][i]+p[0][i-1])/2.0;

  /* normalize to 2.0 */
  if (F[n[0]-1]!=2.0) {
    if (!quiet)
      fprintf (stderr, " ... renormalizing phase function by factor %f\n", 2.0/F[n[0]-1]);
    factor = 2.0/F[n[0]-1];

    for (i=0; i<n[0]; i++) 
      F[i] *= factor;
    
    for (ip=0; ip<nphase; ip++)
      for (i=0; i<n[ip]; i++) 
        p[ip][i] *= factor;
  }

  iphase->nphase=nphase; 
  
  if (delta_scaling_mucut>-1.0)
    iphase->nscales = 2;
  else
    iphase->nscales = 1;
  
  /* copy to final destination */
  iphase->n = calloc ((size_t) nphase, sizeof(int));
  for (ip=0; ip<nphase; ip++)
    iphase->n[ip] = n[ip];

  iphase->mu = calloc ((size_t) nphase, sizeof(double*));
  for (ip=0; ip<nphase; ip++)
    iphase->mu[ip] = calloc ((size_t) iphase->n[ip], sizeof(double));
      
  iphase->p  = calloc ((size_t) iphase->nscales, sizeof(double**));
  iphase->F  = calloc ((size_t) iphase->nscales, sizeof(double*));
  iphase->A  = calloc ((size_t) iphase->nscales, sizeof(double*));
  iphase->B  = calloc ((size_t) iphase->nscales, sizeof(double*));
  iphase->C  = calloc ((size_t) iphase->nscales, sizeof(double*));

  for (j=0; j<iphase->nscales; j++) {
    iphase->p[j]  = calloc ((size_t) nphase, sizeof(double*));
    iphase->F[j]  = calloc ((size_t) iphase->n[0], sizeof(double));
    iphase->A[j]  = calloc ((size_t) iphase->n[0], sizeof(double));
    iphase->B[j]  = calloc ((size_t) iphase->n[0], sizeof(double));
    iphase->C[j]  = calloc ((size_t) iphase->n[0], sizeof(double));
    
    for (ip=0; ip<nphase; ip++) 
      iphase->p[j][ip]  = calloc ((size_t) iphase->n[ip], sizeof(double));
  }
  
  for (ip=0; ip<nphase; ip++) 
    for (i=0; i<iphase->n[ip]; i++) {
      iphase->mu[ip][i] = mu[ip][i];
      iphase->p[MCSC_MODE_NORMAL][ip][i]  = p[ip][i];
    }
  
  for (i=0; i<iphase->n[0]; i++)    
    iphase->F[MCSC_MODE_NORMAL][i]  = F[i];
  

  /* calculate coefficients for linear interpolation of the table */
  for (i=0; i<iphase->n[0]-1; i++) {
    beta  = (p[0][i+1]-p[0][i]) / (mu[0][i+1]-mu[0][i]);
    alpha = p[0][i] - beta * mu[0][i];

    iphase->C[MCSC_MODE_NORMAL][i] = F[i] - alpha*mu[0][i] - 0.5*beta*mu[0][i]*mu[0][i];
    iphase->B[MCSC_MODE_NORMAL][i] = alpha;
    iphase->A[MCSC_MODE_NORMAL][i] = beta/2.0;
  }

  /* determine whether table is linear in theta */
  scr = (float) ( acos(iphase->mu[0][1]) - acos(iphase->mu[0][0]) );
  iphase->is_linear = 1;
  iphase->theta_0 = acos(iphase->mu[0][0]);
  for (i=2; i<iphase->n[0]; i++) {
    if (scr != (float) ( acos(iphase->mu[0][i]) - acos(iphase->mu[0][i-1]) ) ) {
      iphase->is_linear = 0;
      break;
    }
  }
  if (iphase->is_linear) {
    iphase->dthetainv = 1.0 / ( acos(iphase->mu[0][1]) - acos(iphase->mu[0][0]) );
    if (!quiet)
      fprintf(stderr,"phase function table is linear in theta, switching to fast table lookup!\n");
  }
  
  /**************************************************************************/
  /* von hier TZ ds ... */

  /* initialize delta-scaling factor to 0 = no delta-scaling */
  iphase->dscale = 0.0;

  if (delta_scaling_mucut>-1.0) {
    itcut=n[0];
    double dif=99999.;
    for (i=0; i<iphase->n[0]; i++) {
      if (fabs(mu[0][i]-delta_scaling_mucut) < dif) { /*find closest mu in tabulated values*/
	dif=fabs(mu[0][i]-delta_scaling_mucut);
	itcut=i;
      }
    }

    fprintf (stderr, " ... delta-scaling ... truncating phase function at mucut = %g\n", mu[0][itcut]);
        
    delta_scaling_factor = 2.0/F[itcut]; /* F, the integral of p, is 2 */
    iphase->dscale = 1.0-1.0/delta_scaling_factor;
    
    for (i=0; i<n[0]; i++) {
      F[i] =  F[i]*delta_scaling_factor; /*TZ ds: delta-scaled integrated phase function*/
      p[0][i] =  p[0][i]*delta_scaling_factor; /*TZ ds: delta-scaled phase function*/
      if (i>itcut) F[i] = 2.;            /*TZ ds: delta-scaled integrated phase function*/
      if (i>itcut) p[0][i] = 0;             /*TZ ds: delta-scaled phase function*/
    }    

    for (i=0; i<iphase->n[0]; i++) {
      iphase->p[MCSC_MODE_DELTA_SCALE][0][i]  = p[0][i];
      iphase->F[MCSC_MODE_DELTA_SCALE][i]  = F[i];
    }  

    /* calculate coefficients for linear interpolation of the table  */
    /* for delta-scaled version of the phase function */
    for (i=0; i<iphase->n[0]-1; i++) {
      beta  = (p[0][i+1]-p[0][i]) / (mu[0][i+1]-mu[0][i]);
      alpha = p[0][i] - beta * mu[0][i];
      
      iphase->C[MCSC_MODE_DELTA_SCALE][i] = F[i] - alpha*mu[0][i] - 0.5*beta*mu[0][i]*mu[0][i];
      iphase->B[MCSC_MODE_DELTA_SCALE][i] = alpha;
      iphase->A[MCSC_MODE_DELTA_SCALE][i] = beta/2.0;
    } 
  }
  /* ... bis hier TZ ds*/  
  /**************************************************************************/

  if (truncate>-1.0) {
    iphase->truncate = 1;
    iphase->mutrnc = truncate;
    
    it = locate (iphase->mu[0], iphase->n[0], iphase->mutrnc);

    if (it==iphase->n[0]-1 || iphase->mutrnc >= 1.0) {
      fprintf (stderr, " ... no need to truncate at %f\n", iphase->mutrnc);
      iphase->truncate = 0;
    }
    else {
      /* value of the phase function at the truncation angle */
      iphase->ptrnc = iphase->p[MCSC_MODE_NORMAL][0][it] + (iphase->mutrnc - iphase->mu[0][it]) / 
	(iphase->mu[0][it+1] - iphase->mu[0][it]) * (iphase->p[MCSC_MODE_NORMAL][0][it+1] - 
                                                     iphase->p[MCSC_MODE_NORMAL][0][it]);

      iphase->Ftrnc = iphase->F[MCSC_MODE_NORMAL][it] + (iphase->mutrnc - mu[0][it]) * 
        (iphase->ptrnc+p[0][it])/2.0;
      iphase->dtrnc = 2.0/(1.0-iphase->mutrnc) * ((2.0-iphase->Ftrnc)/(1.0-iphase->mutrnc) - 
                                                  iphase->ptrnc);
    }
  }
  

  /* free memory */
  free (F);

  return 0;
}


/***********************************************************************************/
/* Function: read_phase_function                                          @62_30i@ */
/* Description:                                                                    */
/*  Read phase function from a 2-column file.                                      */
/*                                                                                 */
/* Parameters:                                                                     */
/* Return value:                                                                   */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                        @i62_30@ */
/***********************************************************************************/

int read_phase_function (char *filename, 
			 double *r0, double *dr, size_t *nreff, size_t *nphase,
			 double **extinc, double **albedo, int **nmu, 
			 double ***mu, double ****phase, int *type, int quiet)
{
  int ir=0, imu=0, status=0;

  *nreff  = 1;  /* number of effective radii       */ 
  *nphase = 1;  /* number of phase matrix elements */
  
  /* allocate memory for fields */
  *extinc = calloc ((size_t) *nreff, sizeof (double));
  *albedo = calloc ((size_t) *nreff, sizeof (double));
  
  *r0 = -999;
  *dr = -999;
  
  *nmu    = calloc ((size_t) *nreff, sizeof (int));
  *mu     = calloc ((size_t) *nreff, sizeof (double *));
  *phase  = calloc ((size_t) *nreff, sizeof (double **));

  for (ir=0; ir<*nreff; ir++)
    *phase[ir] = calloc (*nphase, sizeof (double *));
 
  if (!quiet) 
    fprintf (stderr, " ... reading two columns of data from %s\n", filename);

  status = read_2c_file (filename, &((*mu)[0]), &((*phase)[0][0]), &((*nmu)[0]));
  
  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }
  
  /* check if the user defined mu or theta */
  for (imu=0; imu<(*nmu)[0]; imu++) {
    if ((*mu)[0][imu]<-1 || (*mu)[0][imu]>1) {
      fprintf (stderr, "Error in %s: Found mu=%f, expecting mu=cos(theta), p(mu)\n", filename, (*mu)[0][imu]);
      return -1;
    }
  }

  *type = PHASE_EXPLICIT;

  if (!quiet)
    fprintf (stderr, " ... read %d coefficients from %s\n", (*nmu)[0], filename);

  return 0;
}


