/*--------------------------------------------------------------------
 * $Id: cloud.c,v 1.149 2008/12/21 01:34:02 bernhard Exp $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 *  
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

#include <stdio.h>							    
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <math.h>
#include <glob.h>
#include <float.h>

#include "uvspec.h"
#include "ckdfu.h"
#include "cloud.h"
#include "ipa.h"
#include "ascii.h"
#include "numeric.h"
#include "miecalc.h"
#include "f77-uscore.h"
#include "wcloud3d.h"

#if HAVE_KEY56
#include "yang56.h"
#endif
 
#if HAVE_LIBNETCDF
#include <netcdf.h>
#endif

#define EPSILON 1E-6

/************************************/
/* prototypes of internal functions */
/************************************/

static int calloc_cld_microphysics (cld_out_struct *out, int nlambda, int nlev);

static int ic_yang (float wavelength, int nlyr, int habit,
		    char *path,
		    float *iwc, float *reff, 
		    float *dtau, 
		    float *f, float *g1, float *g2, float *ssa, 
		    float *zd, int layer, int newkey);

int read_cloudfraction_from_netCDF_file (char *filename, float lat, float lon, struct tm UTC, int time_interpolate, 
                                         float **z, float **cf, int *nlev,
                                         int verbose, int quiet);

int read_ECMWF_clouds (char *filename, float lat, float lon, struct tm UTC, int time_interpolate, char *cloud, 
                       int reff_prop, float reff_fixed, float altitude, float *press_atm, int cloud_overlap, int verbose, int quiet,
                       int *rows, int *max_columns, int *min_columns, float ***cld_data, cf_out_struct *cf);

int read_ECHAM_clouds (char *filename, float lat, float lon, struct tm UTC, int time_interpolate, 
                       char *cloud, float *zd,
                       int verbose, int quiet,
                       int *rows, int *max_columns, int *min_columns,
                       float ***cld_data, cf_out_struct *cf);

static int setup_cloudfraction ( int cloud_overlap, cf_out_struct *cf, float *cloudcover, int verbose );

/**************************************************************/
/* Setup water clouds.                                        */
/**************************************************************/

int setup_wcloud (input_struct input, output_struct *output)
{
  int status=0, lc=0, iv=0, ipa=0;
  int rows=0, min_columns=0, max_columns=0, max_length=0;

  int ix=0, lev=0;
  int Nx=0, Ny=0, Nz=0;
  double delX=0, delY=0;

  float sum=0;
  float rmin=0, rmax=0;
  float ***lwc_3D_ipa=NULL, ***ref_3D_ipa=NULL, ***ext_3D_ipa=NULL;
  float ***g1_3D_ipa=NULL, ***g2_3D_ipa=NULL, ***ff_3D_ipa=NULL, ***f_3D_ipa=NULL;
  float ***ssa_3D_ipa=NULL, ***dscale_3D_ipa=NULL;
  int *threed=NULL;
  int cldproperties=0;
  char wcloud3Dfilename[FILENAME_MAX];

  float *z_atm=NULL;
  int nlev_atm = 0;


  /* read cloud fraction file if specified */
  if ( strlen(input.filename[FN_CLOUD_FRACTION])>0 || strlen(input.filename[FN_CLOUD_FRACTION_MAP])>0 ) {
    
    if (strlen(input.filename[FN_CLOUD_FRACTION])>0) {
      status = read_2c_file_float (input.filename[FN_CLOUD_FRACTION], 
	  			   &(output->cf.zd),
				   &(output->cf.cf),
				   &(output->cf.nlev));
      if (status!=0) {
        fprintf (stderr, "Error %d reading cloud fraction file %s\n", 
	         status, input.filename[FN_CLOUD_FRACTION]);
        return status;
      }
    }
    else if (strlen(input.filename[FN_CLOUD_FRACTION_MAP])>0) {

      /* !!!!!!! this is not implemented jet !!!!!! */
      read_cloudfraction_from_netCDF_file(input.filename[FN_CLOUD_FRACTION_MAP],
                                          input.latitude, input.longitude, 
                                          input.UTC, input.atm.time_interpolate,
                                          &(output->cf.zd), &(output->cf.cf), &(output->cf.nlev),
                                          input.verbose, input.quiet);

      fprintf (stderr, "Error, cloud_fraction_map not yet implemented. (line %d, function %s in %s)\n", 
                       __LINE__, __func__, __FILE__);
      return -1;
    }


    /* need to shift by one layer as cloud fractions are automatically considered layer properties */
    for (lc=0;lc<output->cf.nlev-1;lc++) {
      output->cf.cf[lc] = output->cf.cf[lc+1];
    }

    if (!input.quiet) {
      fprintf (stderr, " ... read %d layers from cloud fraction file %s:\n",
	       output->cf.nlev, input.filename[FN_CLOUD_FRACTION]);
    }

    /* copy z_grid of the atmosphere file */
    if ((z_atm = calloc(output->atm.nlev, sizeof(float)))==NULL) {
      fprintf (stderr, "Error: Allocation of z_atm (line %d, function %s in %s) \n", __LINE__, __func__, __FILE__);
      return -1;
    }

    for (lc=0; lc<output->atm.nlev; lc++)
      z_atm[lc] = output->atm.zd[lc];

    nlev_atm = output->atm.nlev;

    /* add level to common z grid */
    set_common_z_grid (output->cf.zd, output->cf.nlev, 
		       &output->atm.zd, &output->atm.nlev);

    /* interpolate the atmospheric profiles to the new z-grid */
    status = interpolate_atmosphere (z_atm,
				     &(output->atm.microphys.press),
				     &(output->atm.microphys.temper),
				     &(output->atm.microphys.dens),
				     nlev_atm,
				     output->atm.zd, output->atm.nlev,
				     input.atm.interpol_method_press,
				     input.atm.interpol_method_temper,
				     input.atm.interpol_method_gas,
				     input.quiet);

    if (status != 0) {
      fprintf (stderr, "Error %d interpolating atmosphere (line %d, function %s in %s)\n", status, __LINE__, __func__, __FILE__);
      return status; 
    }

    free(z_atm);
  }

  /* if water cloud microphysical properties profile specified */
  switch (input.wc.source) {
  /* read cloud from one file */
  case CLOUDS_FROM_FILE:
  case CLOUDS_FROM_ECMWF:
  case CLOUDS_FROM_ECHAM: 
    
    /* Read microphysical properties from input.filename[FN_WCLOUD] */
    /* and convert to optical properties.                           */

    if (input.verbose)
      fprintf (stderr, "     reading water cloud properties from %s\n", 
	       input.filename[FN_WCLOUD]);

    /* Read water cloud microphysical properties and apply */
    /* default conversion to optical properties            */
    status = read_and_convert_wc_file (input, output, 
				       input.filename[FN_WCLOUD], 
				       &(output->wc));
    if (status!=0) {
      fprintf (stderr, "Error %d returned by read_and_convert_wc_file '%s' (line %d, function %s in %s)\n", 
	       status, input.filename[FN_WCLOUD], __LINE__, __func__, __FILE__);
      return status;
    }

    /* Copy cloudcover to output variable ... */
    output->wc_cloudcover = input.wc_cloudcover;
 
    /* ... and modify it if cloud_overlap assumption is used. */
    status = setup_cloudfraction (input.cloud_overlap, &(output->cf), &(output->wc_cloudcover), input.verbose );
    if (status != 0) {
      fprintf (stderr, "Error %d returned by setup_cloudfraction (line %d, function %s in %s)\n", 
	       status, __LINE__, __func__, __FILE__);
      return status;
    }    
    
    if (output->wc_cloudcover >= 0) {
      if (!input.quiet)
        fprintf (stderr, " ... IPA with two columns, %6.2f%% cloudless and %6.2f%% overcast\n", 
                 100.0*(1.0-output->wc_cloudcover), 100.0*output->wc_cloudcover);
      
      /* allocate memory for IPA arrays */
      output->wcipa     = calloc (2, sizeof(cld_out_struct));
      output->ipaweight = calloc (2, sizeof(float));
        
      for (ipa=0; ipa<2; ipa++) {
        status = calloc_cld_out (&(output->wcipa[ipa]), 
                                 output->wl.nlambda_t, 
                                 output->wc.nlev, output->wc.optprop.nphase);
        
        if (status!=0) {
          fprintf (stderr, "Error %d allocating memory for cld_out_struct (line %d, function %s in %s)\n", 
                           status, __LINE__, __func__, __FILE__);
          return status;
        }
      }
      
      output->nipa=2;                                   /* 2 independent pixels */
      output->ipaweight[1] = 1.0-output->wc_cloudcover; /* cloudless pixel      */
      output->ipaweight[0] = output->wc_cloudcover;     /* overcast pixel       */
      
      /* copy 1D cloud to 2nd pixel */
      status = cp_cld_out (&(output->wcipa[0]), output->wc, 0, 1, input.quiet);
      if (status!=0) {
        fprintf (stderr, "Error %d copying output->wc to output->wcipa[1]\n", status);
        return status;
      }
      
      /* second pixel is cloudless; only levels need to be set */
      for (lc=0; lc<output->wc.nlev; lc++)
	output->wcipa[1].zd[lc] = output->wc.zd[lc]; 
            
    }
    break;
    
  case CLOUDS_IPA_FILES:

    /* Read wc cloud files using wc_ipa_files option */
    status = read_wc_ipa (input, output); /* in ipa.c */
    if (status!=0) {
      fprintf (stderr, "Error %d returned by read_wc_ipa (line %d, function %s in %s)\n", status, __LINE__, __func__, __FILE__);
      return status;
    }
    break;

  case CLOUDS_FROM_3D_FILE:

    /* consistence check */
    if ( !input.wcipa ) {
      fprintf (stderr, "Error, inconsistent cloud properties ((not wcipa) and 3D-clouds).\n");
      fprintf (stderr, " This is not supposed to happen and indicates a coding error. (line %d, function %s in %s)\n", 
                       __LINE__, __func__, __FILE__);
      return -1;
    }

    /* we need a least one independent pixel */
    output->nipa = 1;
    
    if (strlen(input.filename[FN_WCLOUD_IPA_3D])>0) 
      strcpy (wcloud3Dfilename, input.filename[FN_WCLOUD_IPA_3D]);
    else 
      strcpy (wcloud3Dfilename,  "./wcloud3D.dat");

    if (!input.quiet)
      fprintf (stderr, " ... reading IPA cloud data from %s\n", wcloud3Dfilename);
    
    
    /* read 3D cloud data, in libsrc_c/wcloud_3d.c */
    status = read_3D_cloud (wcloud3Dfilename,
                            &Nx, &Ny, &Nz,
                            &delX, &delY,
                            &(output->wc.z3D), 
			    &lwc_3D_ipa,
			    &ref_3D_ipa,
			    &ext_3D_ipa,
			    &g1_3D_ipa,
			    &g2_3D_ipa,
			    &ff_3D_ipa,
			    &f_3D_ipa,
			    &ssa_3D_ipa,
			    &dscale_3D_ipa,
			    &rmin, &rmax,
			    &cldproperties,
			    &threed,
                            input.quiet);
    
    if (status!=0) {
      fprintf (stderr, "error %d reading %s\n", status, wcloud3Dfilename);
      return status;
    }

    output->wc.Nz3D = Nz;

    /* comparison of 3D and 1D profile moved to end of setup_atmosphere() atmosphere.c, UH 2006-03-15 */
    /* as atmospheric profiles are not complete yet */
    
    /* allocate memory for IPA arrays */
    output->wcipa     = calloc (Nx*Ny, sizeof(cld_out_struct));
    output->ipaweight = calloc (Nx*Ny, sizeof(float));

    /* allocate memory for cloud ary and set zd */
    status = calloc_cld_out (&(output->wc), output->wl.nlambda_t, Nz+1, output->wc.optprop.nphase);
    if (status!=0) {
      fprintf (stderr, "Error %d allocating memory for cld_out struct (line %d, function %s in %s)\n", status, 
                       __LINE__, __func__, __FILE__);
      return status;
    }

    if (!input.quiet)
      fprintf (stderr, " ... identifying identical columns for IPA\n");
    
    /* ??? g2 and ff are currently ignored - common HG instead of double-HG ??? */
    /* ??? also dscale_3D_ipa is ignored - no delta-scaling in IPA               ??? */
    status = threeDtoIPA (output->wc.z3D, lwc_3D_ipa, ref_3D_ipa, ext_3D_ipa, g1_3D_ipa, ssa_3D_ipa, threed, cldproperties,
                          Nx, Ny, Nz,
                          output, input.quiet);
    if (status!=0) {
      fprintf (stderr, "Error %d creating IPA input (line %d, function %s in %s)\n", status, __LINE__, __func__, __FILE__);
      return status;
    }
    
    if (!input.quiet)
      fprintf (stderr, " ... CONGRATULATIONS! You only need to consider %d independent pixels!\n", 
               output->nipa);
    
    /*************************************************************/
    /* output->nipa independent pixels are now stored in         */
    /* output->wcipa[ip] where ip = 0 ... ntau is the index of   */
    /* the pixel and output->weight[ip] is the correspoding      */
    /* weight.                                                   */
    /*************************************************************/
    
    /* free memory */
    for (lev=0; lev<Nz; lev++) {
      for (ix=0; ix<Nx; ix++) {
	free (lwc_3D_ipa   [lev][ix]);
	free (ref_3D_ipa   [lev][ix]);
	free (ext_3D_ipa   [lev][ix]);
	free (g1_3D_ipa    [lev][ix]);
	free (g2_3D_ipa    [lev][ix]);
	free (ff_3D_ipa    [lev][ix]);
	free (f_3D_ipa  [lev][ix]);
	free (ssa_3D_ipa   [lev][ix]);
	free (dscale_3D_ipa[lev][ix]);
      }
      
      free (lwc_3D_ipa   [lev]);
      free (ref_3D_ipa   [lev]);
      free (ext_3D_ipa   [lev]);
      free (g1_3D_ipa    [lev]);
      free (g2_3D_ipa    [lev]);
      free (ff_3D_ipa    [lev]);
      free (f_3D_ipa  [lev]);
      free (ssa_3D_ipa   [lev]);
      free (dscale_3D_ipa[lev]);
    }

    free (lwc_3D_ipa);
    free (ref_3D_ipa);
    free (ext_3D_ipa);
    free (g1_3D_ipa);
    free (g2_3D_ipa);
    free (ff_3D_ipa);
    free (f_3D_ipa);
    free (ssa_3D_ipa);
    free (dscale_3D_ipa);
    
    free(threed);
    
    return 0;
    break;

  case CLOUDS_LAYER_FILES:
  
    /* Read file holding filenames of layer cloud optical properties */    

    if (input.wc.properties != WCPROP_EXPLICIT) {
      fprintf (stderr, "Error, inconsistent cloud properties (wc_files, but not wc_explicit).\n");
      fprintf (stderr, " This is not supposed to happen and indicates a coding error. (line %d, function %s in %s)\n", 
                       __LINE__, __func__, __FILE__);
      return -1;
    }

    /* just to make sure, check if filename has been defined */
    if (strlen(input.filename[FN_WCLOUD_FILES]) == 0) {
      fprintf (stderr, "Error, inconsistent cloud properties (layer files, but no filename).\n");
      fprintf (stderr, " This is not supposed to happen and indicates a coding error. (line %d, function %s in %s)\n", 
                       __LINE__, __func__, __FILE__);
      return -1;
    }

    if (output->alt.altitude != 0) {
      fprintf (stderr, "Error, option 'altitude' currently not implemented in combination\n");
      fprintf (stderr, " with 'wc_files' (line %d, function %s in %s) \n", __LINE__, __func__, __FILE__);
      return -1;
    }
      
    /* determine number of levels; required for memory allocation */
    if ((status = ASCII_checkfile (input.filename[FN_WCLOUD_FILES], 
			     &rows, &min_columns, &max_columns, &max_length)) != 0) {
      fprintf (stderr, "Error %d reading %s(line %d, function %s in %s) \n", 
                        status, input.filename[FN_WCLOUD_FILES], __LINE__, __func__, __FILE__);
      return status;
    }

    /* allocate memory */
    status = calloc_cld_out (&(output->wc), output->wl.nlambda_t, rows, 1);
    if (status!=0) {
      fprintf (stderr, "Error %d allocating memory for cld_out struct (line %d, function %s in %s)\n", 
                        status, __LINE__, __func__, __FILE__);
      return status;
    }

    /* read optical properties profile, */
    /* allocate memory for moments      */

    status = read_optprop_files (input.filename[FN_WCLOUD_FILES],
                                 output->wl.lambda_t, output->wl.nlambda_t, 
                                 output->wl.nlambda_rte_lower, output->wl.nlambda_rte_upper,
                                 output->wc.zd, output->wc.nlev,
                                 output->wc.optprop.dtau, 
				 output->wc.optprop.ssa,
				 output->wc.optprop.g1,
				 output->wc.optprop.moment, output->wc.optprop.nmom,
				 input.verbose);
    if (status != 0) {
      fprintf (stderr, "Error %d reading cloud optical property files '%s' (line %d, function %s in %s)\n", 
                        status, input.filename[FN_WCLOUD_FILES], __LINE__, __func__, __FILE__);
      return status;
    }
      
      
    /* increase number of moments if more found */
    for (iv=0; iv<output->wl.nlambda_t; iv++)
      for (lc=0; lc<output->wc.nlev-1; lc++)
	if (output->wc.optprop.nmom[iv][lc]-1 > output->atm.nmom)
	  output->atm.nmom = output->wc.optprop.nmom[iv][lc]-1;

    break;

  case NO_CLOUDS: 

    /* allocate memory, even if no cloud is defined; */
    /* it may be required by other functions;        */
    
    status = calloc_cld_out (&(output->wc), output->wl.nlambda_t, 2, 1);
    if (status!=0) {
      fprintf (stderr, "Error %d allocating memory for cld_out-struct (line %d, function %s in %s)\n", 
                       status, __LINE__, __func__, __FILE__);
      return status;
      }
    
    output->wc.zd[0] = output->atm.zd[0];
    output->wc.zd[1] = output->atm.zd[output->atm.nlev-1];
    break;
    
  default:
    fprintf (stderr, "Error, unknown wc_source %d (line %d, function %s in %s)\n", 
                     input.wc.source, __LINE__, __func__, __FILE__);
    return -1;
  }

  /* overwrite these properties with user-defined optical thickness, ssa, etc */
  status = apply_user_defined_properties_to_cld (input.wc, output->atm, output->wl, 
                                                 output->alt.altitude, &(output->wc));
  if (status!=0) {
  fprintf (stderr, "Error %d overwriting cloud properties with user-defined values (line %d, function %s in %s)\n",
           status, __LINE__, __func__, __FILE__);
  return status;
  }
  

  /* Check that ipa weights add up to one */
  if (input.wcipa)  {
    sum=0;
    for (ipa=0; ipa<output->nipa; ipa++)  
      sum += output->ipaweight[ipa];
    if (sum < 1.0-EPSILON || sum > 1.0+EPSILON) {
      fprintf (stderr, "Error, ipa weights add up to %f. But the sum should be 1.0000 ! (line %d, function %s in %s)\n", 
                       sum, __LINE__, __func__, __FILE__);
      return -1;
    }
  }

  /* if atmosphere is saturated, where clouds are, then */
  if (input.wc.wv_saturation != SATURATION_NONE) {
    /* add the water cloud levels to the common profile */
    /* (interpolation of water vapour on these levels nessesary) */

    /* copy z_grid of the atmosphere file */
    if ((z_atm = calloc(output->atm.nlev, sizeof(float)))==NULL) {
      fprintf (stderr, "Error: Allocation of z_atm (line %d, function %s in %s) \n", __LINE__, __func__, __FILE__);
      return -1;
    }
    for (lc=0; lc<output->atm.nlev; lc++)
      z_atm[lc] = output->atm.zd[lc];

    nlev_atm = output->atm.nlev;

    /* add cloud levels to atmospheric z-grid */
    if (input.wcipa) {
      for (ipa=0; ipa<output->nipa; ipa++)
        set_common_z_grid (output->wcipa[ipa].zd, output->wcipa[ipa].nlev, 
                           &output->atm.zd, &output->atm.nlev);
    }
    else
      set_common_z_grid (output->wc.zd, output->wc.nlev, 
                         &output->atm.zd, &output->atm.nlev);

    /* interpolate the atmospheric profiles to the new z-grid */
    status = interpolate_atmosphere (z_atm,
				     &(output->atm.microphys.press),
				     &(output->atm.microphys.temper),
				     &(output->atm.microphys.dens),
				     nlev_atm,
				     output->atm.zd, output->atm.nlev,
				     input.atm.interpol_method_press,
				     input.atm.interpol_method_temper,
				     input.atm.interpol_method_gas,
				     input.quiet);

    if (status != 0) {
      fprintf (stderr, "Error %d interpolating atmosphere (line %d, function %s in %s)\n", status, __LINE__, __func__, __FILE__);
      return status; 
    }

    free(z_atm);
  }
  /* otherwise these levels are added at the end of setup_atmosphere */

  /* test, whether there are 1D clouds */
  output->wc.occurrence1D = FALSE; /* true, there are clouds */
  iv=0;
  if (input.wcipa) {
    for (ipa=0; ipa<output->nipa; ipa++)
      for (lc=0; lc<output->wcipa[ipa].nlev; lc++)
	for (iv=0; iv<output->wl.nlambda_t; iv++)
	  if (output->wcipa[ipa].optprop.dtau[iv][lc] > 0.0) {
	    output->wc.occurrence1D = TRUE; /* there are clouds */
	    break;
	  }
  }
  else {
    for (lc=0; lc<output->wc.nlev; lc++)
      for (iv=0; iv<output->wl.nlambda_t; iv++)
	if (output->wc.optprop.dtau[iv][lc] > 0.0) {
	  output->wc.occurrence1D = TRUE; /* there are clouds */
	  break;
	}
  }

  return 0;
}




/**************************************************************/
/* Setup ice clouds.                                          */
/**************************************************************/

int setup_icloud (input_struct input, output_struct *output)
{
  int status=0, lc=0, iv=0, ipa=0;
  int rows=0, min_columns=0, max_columns=0, max_length=0;

  float *z_atm=NULL;
  int nlev_atm = 0;
  
  /* if ice cloud microphysical properties profile specified */

  switch (input.ic.source) {
  /* read cloud from one file */
  case CLOUDS_FROM_FILE:
  case CLOUDS_FROM_ECMWF:
  case CLOUDS_FROM_ECHAM: 
    
    /* Read microphysical properties from input.filename[FN_ICLOUD] */
    /* and convert to optical properties.                           */

    if (input.verbose)
      fprintf (stderr, " ... reading ice cloud microphysical properties from %s\n", 
	       input.filename[FN_ICLOUD]);
    
    /* Read ice cloud microphysical properties */
    /* convert to optical properties           */
    status = read_and_convert_ic_file (input, output, 
				       input.filename[FN_ICLOUD], 
				       &(output->ic));
    if (status!=0) {
      fprintf (stderr, "Error %d returned by read_and_convert_ic_file '%s' (line %d, function %s in %s)\n", 
	       status, input.filename[FN_ICLOUD], __LINE__, __func__, __FILE__);
      return status;
    }
    
    /* Copy cloudcover to output variable ... */
    output->ic_cloudcover = input.ic_cloudcover;
      
    /* ... and modify it if cloud_overlap assumption is used. */
    status = setup_cloudfraction (input.cloud_overlap, &(output->cf), &(output->ic_cloudcover), input.verbose );
    if (status != 0) {
      fprintf (stderr, "Error %d returned by setup_cloudfraction (line %d, function %s in %s)\n", 
	       status, __LINE__, __func__, __FILE__);
      return status;
    }

    if (output->ic_cloudcover >= 0.0) {
      if (!input.quiet)
        fprintf (stderr, " ... IPA with two columns, %6.2f/100 cloudless and %6.2f/100 overcast\n", 
                 100*(1-output->ic_cloudcover),100*output->ic_cloudcover);
      
      /* allocate memory for IPA arrays */
      output->icipa     = calloc (2, sizeof(cld_out_struct));
      output->ipaweight = calloc (2, sizeof(float));
      
      for (ipa=0; ipa<2; ipa++) {
        status = calloc_cld_out (&(output->icipa[ipa]), 
                                 output->wl.nlambda_t, 
                                 output->ic.nlev, output->ic.optprop.nphase);
        
        if (status!=0) {
          fprintf (stderr, "Error %d allocating memory for cld_out_struct\n", status);
          return status;
        }
      }
      
      output->nipa=2;  
      /* 2 independent pixels */
      output->ipaweight[1] = 1.0-output->ic_cloudcover; /* cloudless pixel      */
      output->ipaweight[0] = output->ic_cloudcover;     /* overcast pixel       */
      
      /* copy 1D cloud to 2nd pixel */
      status = cp_cld_out (&(output->icipa[0]), output->ic, 0, 1, input.quiet);
      if (status!=0) {
        fprintf (stderr, "Error %d copying output->ic to output->icipa[1]\n", status);
        return status;
      }
      /* second pixel is cloudless; only levels need to be set */
      for (lc=0; lc<output->ic.nlev; lc++) 
	output->icipa[1].zd[lc] = output->ic.zd[lc]; 

    }
    
    break;
    
  case CLOUDS_IPA_FILES:

    /* read ice cloud IPA file; this function contains a test */
    /* if the weights defined in the ice cloud file are       */
    /* identical to those in the water cloud file             */
    status = read_ic_ipa (input, output);
    if (status!=0) {
      fprintf (stderr, "Error %d returned by read_ic_ipa()\n", status);
      return status;
    }

    break;

  case CLOUDS_LAYER_FILES:  
  
    /* Read file holding filenames of layer cloud optical properties */    

    if (input.ic.properties != ICPROP_EXPLICIT) {
      fprintf (stderr, "Error, inconsistent cloud properties (ic_files, but not ic_explicit).\n");
      fprintf (stderr, "This is not supposed to happen and indicates a coding error.\n");
      return -1;
    }
      
      /* just to make sure, check if filename has been defined */
      if (strlen(input.filename[FN_ICLOUD_FILES]) == 0) {
	fprintf (stderr, "Error, inconsistent cloud properties. This is not\n");
	fprintf (stderr, "supposed to happen and indicates a coding error.\n");
	  return -1;
      }

      if (output->alt.altitude != 0) {
	fprintf (stderr, "Sorry, option 'altitude' currently not implemented in combination\n");
	fprintf (stderr, "with 'ic_files'\n");
	return -1;
      }
      

      /* determine number of levels; required for memory allocation */
      if ((status = ASCII_checkfile (input.filename[FN_ICLOUD_FILES], 
				     &rows, &min_columns, &max_columns, &max_length)) != 0) {
	fprintf (stderr, "Error %d reading %s\n", status, input.filename[FN_ICLOUD_FILES]);
	return status;
      }

      /* allocate memory */
      status = calloc_cld_out (&(output->ic), output->wl.nlambda_t, rows, 1);
      if (status!=0) {
	fprintf (stderr, "Error %d allocating memory for cld_out struct\n", status);
	return status;
      }
      
      /* read optical properties profile, */
      /* allocate memory for moments      */

      status = read_optprop_files (input.filename[FN_ICLOUD_FILES],
				   output->wl.lambda_t, output->wl.nlambda_t,
				   output->wl.nlambda_rte_lower,
                                   output->wl.nlambda_rte_upper,
				   output->ic.zd, output->ic.nlev,
				   output->ic.optprop.dtau, 
				   output->ic.optprop.ssa, 
                                   output->ic.optprop.g1,
				   output->ic.optprop.moment, 
                                   output->ic.optprop.nmom,
				   input.verbose);
      if (status != 0) {
	fprintf (stderr, "Error %d reading cloud optical property files: %s\n", 
		 status, input.filename[FN_ICLOUD_FILES]);
	return status;
      }
    
      /* increase number of moments if more found */
      for (iv=0; iv<output->wl.nlambda_t; iv++)
	for (lc=0; lc<output->ic.nlev-1; lc++)
	  if (output->ic.optprop.nmom[iv][lc]-1 > output->atm.nmom)
	    output->atm.nmom = output->ic.optprop.nmom[iv][lc]-1;

    break;

  case NO_CLOUDS: 
    
      /* allocate memory, even if no cloud is defined; */
      /* it may be required by other functions.        */

    status = calloc_cld_out (&(output->ic), output->wl.nlambda_t, 2, output->ic.optprop.nphase);
      if (status!=0) {
	fprintf (stderr, "Error %d allocating memory for cld_out-struct\n", status);
	return status;
      }

      output->ic.zd[0] = output->atm.zd[0];
      output->ic.zd[1] = output->atm.zd[output->atm.nlev-1];

      break;
  default:
    fprintf (stderr, "Error, no suitable wc data source number. %d \n", input.wc.source);
    return -1;
  }

  /* overwrite these properties with user-defined optical thickness, ssa, etc */
  status = apply_user_defined_properties_to_cld (input.ic, output->atm, output->wl, 
						 output->alt.altitude, &(output->ic));
  if (status!=0) {
    fprintf (stderr, "Error %d overwriting cloud properties with user-defined values\n",
	     status);
    return status;
  }
  
  if (input.ic.wv_saturation != SATURATION_NONE) {
    /* add the ice cloud levels to the common profile */
    /* (interpolation of water vapour on these levels nessesary) */

    /* copy z_grid of the atmosphere file */
    if ((z_atm = calloc(output->atm.nlev, sizeof(float)))==NULL) {
      fprintf (stderr, "Error: Allocation of z_atm in setup_clouds() \n");
      return -1;
    }
    for (lc=0; lc<output->atm.nlev; lc++)
      z_atm[lc] = output->atm.zd[lc];

    nlev_atm = output->atm.nlev;

    /* add cloud levels to atmospheric z-grid */
    if (input.icipa) {
      for (ipa=0; ipa<output->nipa; ipa++)
        set_common_z_grid (output->icipa[ipa].zd, output->icipa[ipa].nlev, 
                           &output->atm.zd, &output->atm.nlev);
    }
    else {
      set_common_z_grid (output->ic.zd, output->ic.nlev, 
                         &output->atm.zd, &output->atm.nlev);
    }

    /* interpolate the atmospheric profiles to the new z-grid */
    status = interpolate_atmosphere (z_atm,
				     &(output->atm.microphys.press),
				     &(output->atm.microphys.temper),
				     &(output->atm.microphys.dens),
				     nlev_atm,
				     output->atm.zd, output->atm.nlev,
				     input.atm.interpol_method_press,
				     input.atm.interpol_method_temper,
				     input.atm.interpol_method_gas,
				     input.quiet);

    if (status != 0) {
      fprintf (stderr, "Error %d interpolating atmosphere (line %d, function %s in %s)\n", status, __LINE__, __func__, __FILE__);
      return status; 
    }
    
    free(z_atm);
  }
  /* otherwise these levels are added at the end of setup_atmosphere */

  /* test, whether there are ice clouds or not */
  output->ic.occurrence1D = FALSE; /* initialisation with: no clouds */
  iv=0;
  if (input.icipa) {
    for (ipa=0; ipa<output->nipa; ipa++)
      for (lc=0; lc<output->icipa[ipa].nlev; lc++)
	for (iv=0; iv<output->wl.nlambda_t; iv++)
	  if (output->icipa[ipa].optprop.dtau[iv][lc] > 0.0) {
	    output->ic.occurrence1D = TRUE; /* there are clouds */
	    break;
	  }
  }
  else {
    for (lc=0; lc<output->ic.nlev; lc++)
      for (iv=0; iv<output->wl.nlambda_t; iv++)
	if (output->ic.optprop.dtau[iv][lc] > 0.0) {
	  output->ic.occurrence1D = TRUE; /* there are clouds */    
	  break;
	}
  }

  /* give total cloud fraction a resonable value, even if we do not have any overlap assumption    */
  /* Attention: this must be after setup_cloudfraction is called                                   */
  /*            as a not initialzied cf.cf_total is used there as a marker for "cf not calculated" */
  if (input.cloud_overlap == CLOUD_OVERLAP_OFF) {
    if (output->wc.occurrence1D || output->ic.occurrence1D) {
      if (input.wc_cloudcover != NOT_DEFINED_INTEGER || input.ic_cloudcover != NOT_DEFINED_INTEGER ) {
        if ( input.wc_cloudcover > input.ic_cloudcover)
          output->cf.cf_total = input.wc_cloudcover;
        else 
          output->cf.cf_total = input.ic_cloudcover;
      }
      else {
        output->cf.cf_total = 1.0;
      }
    }
    else {
      output->cf.cf_total = 0.0;
    }
  }

  return 0;
}


/***********************************************************************************/
/* Function: setup_cloudfraction                                                   */
/* Description:                                                                    */
/*  calculates the column (total) cloud fraction (cf)                              */
/*  scales the cloud fraction in the cloudy column, accordingly to the total cf    */
/*                                                                                 */
/* Parameters:                                                                     */
/*   input                                                                         */
/*     int cloud_overlap                                                           */
/*     verbose                                                                     */
/*   output                                                                        */
/*     cf_out_struct *cf                                                           */
/*     float *cloudcover                                                           */
/* Return value:                                                                   */
/*     int status         == 0, if everthing is OK                                 */
/*                        < 0, if there was an error                               */
/* Example:                                                                        */
/* Files:    cloud.c                                                               */
/* Known bugs: -                                                                   */
/* Author:                                                                         */
/*    Jun 2008   U. Hamann     Created                                             */
/*                                                                                 */
/***********************************************************************************/

static int setup_cloudfraction ( int cloud_overlap, cf_out_struct *cf, float *cloudcover, int verbose ) 
{

  int lc=0;
  float cloudfree = NOT_DEFINED_FLOAT;
  float cc_max = 0.0;

  if (cloud_overlap != CLOUD_OVERLAP_OFF) {

    if (verbose) 
      fprintf(stderr, " ... setup_cloudfraction \n");

    if (cf->cf_total == NOT_DEFINED_FLOAT) {

      /* check if there is a cloud fraction profile */
      if (cf->nlev==0) {
        fprintf(stderr, "Error, option cloud_overlap requires a cloud_fraction_file (line %d, function %s in %s)\n",
                __LINE__, __func__, __FILE__ );
        return -1;
      }
    
      /* calculate column cloud cover */
      switch (cloud_overlap) {
      case CLOUD_OVERLAP_MAX:
        /* find maximum cloud fraction */
        cf->cf_total = 0.0;
        for (lc=0;lc <cf->nlev-1;lc++) {
          if ( cf->cf[lc] > cf->cf_total )
            cf->cf_total = cf->cf[lc];
        }
        break;
      case CLOUD_OVERLAP_MAXRAND:
        cloudfree = 1.0;
        cc_max = 0.0; /* maximum cloud cover in one cloud */
        for ( lc=0; lc<cf->nlev-1; lc++ ) {
          if ( cf->cf[lc] > cc_max ) cc_max = cf->cf[lc]; /* find cc_max inside one cloud */
          if (( cf->cf[lc] == 0.0 && cf->cf[lc-1] > 0.0 ) || /* end of connected cloud */ 
              ( cf->cf[lc]  > 0.0 && lc == cf->nlev-1-1 )) { /* cloud reach until bottom */ 
            cloudfree = cloudfree * (1-cc_max);
            cc_max=0.0;
          }
        }
        cf->cf_total = 1.0 - cloudfree;
        break;
      case CLOUD_OVERLAP_RAND:
        cloudfree=1.0;
        for (lc=0;lc <cf->nlev-1;lc++) {
          cloudfree *= ( 1.0 - cf->cf[lc] );
        }
        cf->cf_total = 1.0 - cloudfree;
        break;
      case CLOUD_OVERLAP_OFF:
      default:
        fprintf (stderr, "Error, unknown cloud_overlap assumption %d. (line %d, function %s in %s)\n", 
                 cloud_overlap, __LINE__, __func__, __FILE__  );
        return -1;
      }

      /* the column is treated as one cloudy column with cf_total and one cloud free column */
      /* the cloud fraction in the cloudy column is now cf/cf_total                         */
      /* For cloud_overlap == CLOUD_OVERLAP_RAND the overlap is simulated                   */
      /* by the effective optical thickness (see optical properties)                        */
      if (cloud_overlap != CLOUD_OVERLAP_RAND) {
        if (cf->cf_total != 0.0 )
          for ( lc=0; lc<cf->nlev-1; lc++ )
            cf->cf[lc] /= cf->cf_total;
      }
      
    }

    if (cloud_overlap != CLOUD_OVERLAP_RAND) {
      /* copy column cloud fraction to cloud cover */
      (*cloudcover) = cf->cf_total;
    }
    
    if (verbose) 
      fprintf(stderr, "     total cloud fraction is %f \n", cf->cf_total );

  }

  return 0;
}

static int calloc_cld_microphys_struct (cld_microphys_struct *out, int nlev) 
{
  out->lwc  = calloc (nlev, sizeof(float));
  out->effr = calloc (nlev, sizeof(float));

  /* for safety reasons allocate nlev even for layer quantities */
  out->lwc_layer  = calloc (nlev, sizeof(float));
  out->effr_layer = calloc (nlev, sizeof(float));

  out->alloc = 1;

  return 0;
}

static int free_cld_microphys_struct (cld_microphys_struct *out) 
{
  free (out->lwc_layer); 
  free (out->effr_layer);

  free (out->lwc); 
  free (out->effr);
  
  return 0;
}


int calloc_optprop (optprop_struct *out, int nlambda, int nlev, int nphase) 
{
  int iv=0, lc=0;

  out->nlam = nlambda;
  out->nlev = nlev;
  out->nphase = nphase;

  out->dtau   = calloc (nlambda, sizeof (float *));
  out->ssa    = calloc (nlambda, sizeof (float *));
  out->f   = calloc (nlambda, sizeof (float *));

  out->dscale = calloc (nlambda, sizeof (float *)); /*TZ ds*/

  /* parameters for double HG */
  out->g1   = calloc (nlambda, sizeof (float *));
  out->g2   = calloc (nlambda, sizeof (float *));
  out->ff   = calloc (nlambda, sizeof (float *));
  
  for (iv=0; iv<nlambda; iv++) {
    out->dtau  [iv] = calloc (nlev, sizeof(float));
    out->ssa   [iv] = calloc (nlev, sizeof(float));
    out->f     [iv] = calloc (nlev, sizeof(float));
    out->dscale[iv] = calloc (nlev, sizeof(float)); /*TZ ds*/
    out->g1    [iv] = calloc (nlev, sizeof(float));
    out->g2    [iv] = calloc (nlev, sizeof(float));
    out->ff    [iv] = calloc (nlev, sizeof(float));

    for (lc=0; lc<nlev; lc++)
      out->ff[iv][lc] = 1.0;
  }  
  
  out->nmom   = calloc (nlambda, sizeof (int *));
  out->moment = calloc (nlambda, sizeof (float ***)); 
  
  for (iv=0; iv<nlambda; iv++) {
    out->nmom[iv]   = calloc (nlev, sizeof (int));
    out->moment[iv] = calloc (nlev, sizeof (float **));
    for (lc=0; lc<nlev; lc++) {
      out->nmom[iv][lc] = 0;
      out->moment[iv][lc] = calloc (nphase, sizeof (float *));
    }
  }
  
  out->alloc = 1;

  return 0;
}


int free_optprop (optprop_struct *optprop)
{
  int iv=0;
  
  for (iv=0; iv<optprop->nlam; iv++) {
    free (optprop->g1[iv]);
    free (optprop->g2[iv]);
    free (optprop->ff[iv]);
    free (optprop->dscale[iv]); /*TZ ds*/
    free (optprop->ssa[iv]);
    free (optprop->dtau[iv]);
    free (optprop->nmom[iv]);
    free (optprop->moment[iv]);
  }
  
  free (optprop->g1); 
  free (optprop->g2);
  free (optprop->ff);
  free (optprop->dscale); /*TZ ds*/
  free (optprop->ssa);
  free (optprop->dtau);
  free (optprop->nmom);
  free (optprop->moment);

  return 0;
}


int calloc_ssprop_struct (ssprop_struct *out, int nlambda) 
{
  out->extinc = calloc (nlambda, sizeof(double *));
  out->albedo = calloc (nlambda, sizeof(double *));
  out->f   = calloc (nlambda, sizeof(double *));
  
  out->r0     = calloc (nlambda, sizeof(double));
  out->dr     = calloc (nlambda, sizeof(double));

  out->nreff  = calloc (nlambda, sizeof(size_t));
  out->nphase = 1;

  out->reff   = calloc (nlambda, sizeof(double *)); 

  out->nleg   = calloc (nlambda, sizeof(int *));
  out->legen  = calloc (nlambda, sizeof(float ***));

  out->nmu    = calloc (nlambda, sizeof(int *));
  out->mu     = calloc (nlambda, sizeof(double **));
  out->p      = calloc (nlambda, sizeof(double ***));

  out->alloc          = 1;  /* basic memory (extinc[iv], ...) allocated */
  out->alloc_moments  = 0;  /* phase function moments not yet allocated */
  out->alloc_explicit = 0;  /* phase function values not yet allocated  */

  return 0;
}

int free_ssprop_struct (ssprop_struct *out) 
{
  /* ??? should we also free the fields if out->alloc_xxx is set ??? */

  if (out->alloc) {
    free (out->extinc);
    free (out->albedo);
    free (out->f);

    free (out->r0);    
    free (out->dr);    

    free (out->nreff); 

    free (out->reff);

    free (out->nleg);  
    free (out->legen); 

    free (out->nmu);
    free (out->mu);
    free (out->p);
  }

  out->alloc = 0; 

  return 0;
}



/***************************************/
/* allocate memory for a cld_out_struct */
/***************************************/

int calloc_cld_out (cld_out_struct *out, int nlambda, int nlev, int nphase) 
{
  int status=0;

  out->zd     = calloc (nlev, sizeof(float));
  out->nlev   = nlev;
  out->newsiz = 1;

  status = calloc_cld_microphys_struct (&(out->microphys), out->nlev);
  if (status!=0) {
    fprintf (stderr, "Error %d allocating memory for cloud microphysics\n", status);
    return status;
  }
  
  status = calloc_optprop (&(out->optprop), nlambda, out->nlev, nphase);
  if (status!=0) {
    fprintf (stderr, "Error %d allocating memory for cloud optical properties\n", status);
    return status;
  }

  status = calloc_ssprop_struct (&(out->ssprop), nlambda);
  if (status!=0) {
    fprintf (stderr, "Error %d allocating memory for cloud single scattering properties\n", 
	     status);
    return status;
  }

  return 0;
}


/************************************/
/* free memory for a cld_out_struct */
/************************************/

int free_cld_out (cld_out_struct *out) 
{

  free (out->zd);

  free_cld_microphys_struct (&(out->microphys));
  
  free_optprop (&(out->optprop));

  free_ssprop_struct (&(out->ssprop));

  return 0;
}



/***********************************************************/
/* allocate memory for a cld_out_struct, microphysics only */
/***********************************************************/

static int calloc_cld_microphysics (cld_out_struct *out, int nlambda, int nlev) 
{
  int status=0;

  out->zd     = calloc (nlev, sizeof(float));
  out->nlev   = nlev;
  out->newsiz = 1;
  
  status = calloc_cld_microphys_struct (&(out->microphys), out->nlev);
  if (status!=0) {
    fprintf (stderr, "Error %d allocating memory for cloud microphysics\n", status);
    return status;
  }

  return 0;
}




int cp_optprop (optprop_struct *target, optprop_struct *source, int nlev, int alloc_moment, int quiet)
{
  int iv=0, lev=0, k=0, ip=0;
  
  if (target->alloc && source->alloc) {
    
    /* check if number of wavelength bands equal */
    if (target->nlam != source->nlam) {
      fprintf (stderr, "Error: cp_optprop(), nlam differs between source and target\n");
      fprintf (stderr, "       source: %d,  target: %d\n", 
	       source->nlam, target->nlam);
      return -1;
    }
    target->nphase=source->nphase; 
    
    for (lev=0; lev<nlev-1; lev++) {
      for (iv=0; iv<source->nlam; iv++) {
	target->dtau  [iv][lev] = source->dtau  [iv][lev];
	target->ssa   [iv][lev] = source->ssa   [iv][lev];
	target->f     [iv][lev] = source->f     [iv][lev];
        target->dscale[iv][lev] = source->dscale[iv][lev]; /*TZ ds*/
	
	target->g1    [iv][lev] = source->g1    [iv][lev];
	target->g2    [iv][lev] = source->g2    [iv][lev];
	target->ff    [iv][lev] = source->ff    [iv][lev];
	
	if (source->nmom[iv][lev] > 0) {
	  
	  /* allocate memory (a) if the user wants it; or    */
	  /* (b) if the number of moments in the target is 0 */
	  if (alloc_moment || (target->nmom[iv][lev]==0 && source->nmom[iv][lev]>0)) {
            for (ip=0; ip<source->nphase; ip++)
	    target->nmom[iv][lev] = source->nmom[iv][lev];
            /* Dimension of phase matrix elements */
            target->moment[iv][lev] = calloc (source->nphase, sizeof(float *));
            for (ip=0; ip<source->nphase; ip++){
              target->moment[iv][lev][ip] = calloc (source->nmom[iv][lev], sizeof(float));
            }
          }
	  
	  /* check if source and target compatible */
	  if (target->nmom[iv][lev] != source->nmom[iv][lev]) {
	    fprintf (stderr, "Error, number of moments in source (%d) and target (%d) different\n",
		     source->nmom[iv][lev], target->nmom[iv][lev]);
	    return -1;
	  }
          
       
          for (ip=0; ip<source->nphase; ip++)
            for (k=0;k<source->nmom[iv][lev];k++){
              target->moment[iv][lev][ip][k] = source->moment[iv][lev][ip][k];
            }
        }
      }
    }
  }
  else {
    /*
    if (!quiet) {

      if (!source->alloc) {
	fprintf (stderr, "*** Warning, not copying optical properties in cp_optprop() because\n");
	fprintf (stderr, "*** memory for source not allocated\n");
      }

      if (!target->alloc) {
	fprintf (stderr, "*** Warning, not copying optical properties in cp_optprop() because\n");
	fprintf (stderr, "*** memory for target not allocated\n");
      }
    }
    */
  }

  return 0;
}
    

/*************************************************/
/* copy contents of a cld_out_struct to another; */
/* NO memory allocation is done here!!!          */
/*************************************************/

int cp_cld_out (cld_out_struct *target, cld_out_struct source, 
		int copy_ssprop, int alloc_moment, int quiet)
{
  int iv=0, ir=0, imu=0, k=0, ip=0, lev=0, status=0;

  if (target->nlev < source.nlev) {
    fprintf (stderr, "Error: cp_cld_out(), nlev differs between source and target\n");
    fprintf (stderr, "       source: %d,  target: %d\n", source.nlev, target->nlev);
    return -1;
  }

  target->newsiz = source.newsiz;

  /* copy microphysical properties */
  for (lev=0; lev<source.nlev; lev++) {
    target->zd[lev] = source.zd[lev];
    if (target->microphys.alloc) {
      target->microphys.lwc  [lev] = source.microphys.lwc  [lev];
      target->microphys.effr [lev] = source.microphys.effr [lev];

      target->microphys.lwc_layer [lev] = source.microphys.lwc_layer  [lev];
      target->microphys.effr_layer[lev] = source.microphys.effr_layer [lev];
    }
  }

  /* copy single scattering properties */
  if (copy_ssprop)  {

    if (!source.ssprop.alloc) {
      fprintf (stderr, "Error, no memory for single scattering properties allocated\n");
      fprintf (stderr, "in cp_cld_out(), source!\n");
      return -1;
    }

    if (!target->ssprop.alloc) {
      fprintf (stderr, "Error, no memory for single scattering properties allocated\n");
      fprintf (stderr, "in cp_cld_out(), target!\n");
      return -1;
    }

    target->ssprop.type = source.ssprop.type;

    if (source.ssprop.alloc_moments || source.ssprop.alloc_explicit) {
      target->ssprop.nphase = source.ssprop.nphase;
      
      for (iv=0; iv<source.optprop.nlam; iv++) {
	target->ssprop.nreff[iv] = source.ssprop.nreff[iv];
	target->ssprop.r0[iv]    = source.ssprop.r0[iv];
	target->ssprop.dr[iv]    = source.ssprop.dr[iv];
        
	/* allocate memory for moments if required */
	if (source.ssprop.alloc_moments && !target->ssprop.alloc_moments) {
	  target->ssprop.extinc[iv] = calloc (target->ssprop.nreff[iv], sizeof(double));
	  target->ssprop.albedo[iv] = calloc (target->ssprop.nreff[iv], sizeof(double));
	  target->ssprop.f[iv] = calloc (target->ssprop.nreff[iv], sizeof(double));
          
	  target->ssprop.nleg[iv]   = calloc (target->ssprop.nreff[iv], sizeof(int));
	  target->ssprop.legen[iv]  = calloc (target->ssprop.nreff[iv], sizeof(float *));

          for (ir=0; ir<target->ssprop.nreff[iv]; ir++) {
            target->ssprop.legen[iv][ir] = calloc (target->ssprop.nphase, sizeof(float *));

            for (ip=0; ip<target->ssprop.nphase; ip++)
              target->ssprop.legen[iv][ir][ip] = calloc (source.ssprop.nleg[iv][ir], sizeof(float));
          }
        }
	  
	/* allocate memory for phase function if required */
	if (source.ssprop.alloc_explicit && !target->ssprop.alloc_explicit) {
	  target->ssprop.extinc[iv] = calloc (target->ssprop.nreff[iv], sizeof(double));
	  target->ssprop.albedo[iv] = calloc (target->ssprop.nreff[iv], sizeof(double));
          target->ssprop.f[iv] = calloc (target->ssprop.nreff[iv], sizeof(double));
          
	  target->ssprop.nmu[iv]  = calloc (target->ssprop.nreff[iv], sizeof(int));
	  target->ssprop.mu [iv]  = calloc (target->ssprop.nreff[iv], sizeof(double *));
	  target->ssprop.p  [iv]  = calloc (target->ssprop.nreff[iv], sizeof(double **));

          for (ir=0; ir<target->ssprop.nreff[iv]; ir++) {
 
	    target->ssprop.mu[iv][ir] = calloc (source.ssprop.nmu[iv][ir], sizeof (double));
            target->ssprop.p [iv][ir] = calloc (source.ssprop.nphase, sizeof(double *));

            for (ip=0; ip<target->ssprop.nphase; ip++)
              target->ssprop.p [iv][ir][ip] = calloc (source.ssprop.nmu[iv][ir], sizeof(double));
	  }
	}
      }

      target->ssprop.alloc_moments  = source.ssprop.alloc_moments; 
      target->ssprop.alloc_explicit = source.ssprop.alloc_explicit;

      /* finally copy source to target */
      for (iv=0; iv<source.optprop.nlam; iv++) {
	if (target->ssprop.alloc_moments) {
	  for (ir=0; ir<target->ssprop.nreff[iv]; ir++) {
	    target->ssprop.extinc [iv][ir] = source.ssprop.extinc [iv][ir];
	    target->ssprop.albedo [iv][ir] = source.ssprop.albedo [iv][ir];
            target->ssprop.f   [iv][ir] = source.ssprop.f   [iv][ir];
	    target->ssprop.nleg   [iv][ir] = source.ssprop.nleg   [iv][ir];
	    
	    for (k=0; k<target->ssprop.nleg[iv][ir]; k++)
	      for (ip=0; ip<target->ssprop.nphase; ip++)
		target->ssprop.legen [iv][ir][ip][k] = source.ssprop.legen [iv][ir][ip][k];
	  }
	}


	if (target->ssprop.alloc_explicit) {
	  for (ir=0; ir<target->ssprop.nreff[iv]; ir++) {
	    target->ssprop.extinc [iv][ir] = source.ssprop.extinc [iv][ir];
	    target->ssprop.albedo [iv][ir] = source.ssprop.albedo [iv][ir];
            target->ssprop.f   [iv][ir] = source.ssprop.f   [iv][ir];
	    target->ssprop.nmu    [iv][ir] = source.ssprop.nmu    [iv][ir];
	    
	    for (imu=0; imu<target->ssprop.nmu[iv][ir]; imu++) {
	      target->ssprop.mu [iv][ir][imu] = source.ssprop.mu [iv][ir][imu];

	      for (ip=0; ip<target->ssprop.nphase; ip++)
		target->ssprop.p[iv][ir][ip][imu] = source.ssprop.p[iv][ir][ip][imu];
	    }
	  }
	}
      }
    }
  }
  

  /* copy optical properties */
  status = cp_optprop (&(target->optprop), &(source.optprop), source.nlev, alloc_moment, quiet);
  if (status!=0) {
    fprintf (stderr, "Error %d copying cloud optical properties\n", status);
    return status;
  }

  
  return 0;
}




/* Interpolate the Legendre coefficients which are given on a grid of  */
/* effective radii r0, r0 + dr, ..., r0 + (nreff-1) * dr to a given    */
/* radius reff; memory for the array of Legendre coefficients coeffc[] */
/* is allocated automatically.                                         */

/* CE: 24.06.2008: Modified interpolation method. Interpolation no longer linear */
/* for all optical properties. Single scattering albedo is weighted with extinction */
/* and Legendre coefficients are weighted with extinction * albedo. */
/* Also included interpolation of delta scaling factor */

int interpolate_legendre (/* Input */
                          float ***legen, int *nleg, int nphase, int nreff,
                          double r0, double dr,
                          double *extinc, double *albedo, double *fdelta,
                          double reff,
                          /* Output */
                          float ***coeffc, int *ncoeffc, 
                          float *ext, float *alb, float *f)
{
  int k=0, ip=0;
  int lower=0, upper=0;
  int min=0, max=0;
  
  double reffl=0.0;
  double alb_num=0.0;
  double ext_us=0.0, ext_us_lower=0.0, ext_us_upper=0.0;
  double alb_us=0.0, alb_us_lower=0.0, alb_us_upper=0.0;

  if (reff<r0 || reff> r0 + (double) (nreff-1) * dr) {
    fprintf (stderr, "Error, reff = %f out of range (%f - %f)\n",
	     reff, r0, r0 + (double) (nreff-1) * dr);
    return -1;
  }

  /* Indices of effective radius */
  lower = (int) ((reff - r0) / dr);
  upper = lower+1;
  
  reffl = r0 + (double) lower * dr;

  if (lower==nreff-1) {   /* upper boundary value */
    *ncoeffc = nleg[nreff-1];
    *coeffc = calloc (nphase, sizeof(float *));
    for (ip=0; ip<nphase; ip++){
      (*coeffc)[ip]  = calloc (nleg[nreff-1], sizeof(float*));
      for (k=0; k<*ncoeffc; k++)
        (*coeffc)[ip][k] = legen[nreff-1][ip][k];
    }
    
    *alb = albedo[lower];
    *ext = extinc[lower];
    *f   = fdelta[lower];
    
    return 0;
  }

  /* otherwise interpolation required */
  min = (nleg[lower] < nleg[upper] ? nleg[lower] : nleg[upper]);
  max = (nleg[lower] > nleg[upper] ? nleg[lower] : nleg[upper]);
  
  *ncoeffc = max;
  
  *coeffc = calloc (nphase, sizeof(float *));
  
  /*Interpolate extinction linearly*/
  *ext = extinc[lower] + (reff - reffl) / dr * (extinc[upper] - extinc[lower]);
  
  /* Calculate unscaled albedo*/
  alb_us_lower=albedo[lower]/(1.0+fdelta[lower]*(albedo[lower]-1.0));
  alb_us_upper=albedo[upper]/(1.0+fdelta[upper]*(albedo[upper]-1.0));

  /* extinction in data is scaled, so we need to calculate
     unscaled extinction here */
  ext_us_lower = extinc[lower] / (1.0 - alb_us_lower * fdelta[lower]);
  ext_us_upper = extinc[upper] / (1.0 - alb_us_upper * fdelta[upper]);  
  
  /*interpolate un-deltascaled extinction */
  ext_us = ext_us_lower + (reff - reffl) / dr * (ext_us_upper - ext_us_lower);
  
  /*Single scattering albedo should be weighted with extinction, this way the 
    interpolation means that the optical properties are averaged */
  /* Numerator of averaged single scattering albedo.*/
  alb_num = ((albedo[lower] * extinc[lower]) + (reff - reffl) / dr 
             * (albedo[upper]*extinc[upper] - albedo[lower]* extinc[lower]));
  *alb= alb_num/ *ext;
  
  alb_us = ((alb_us_lower * ext_us_lower) + (reff - reffl) / dr 
             * (alb_us_upper*ext_us_upper - alb_us_lower*ext_us_lower))/
    ext_us;
  
  /* The moments need to be weighted with alb*ext */ 
  for (ip=0; ip<nphase; ip++){
    (*coeffc)[ip]  = calloc (*ncoeffc, sizeof(float));
    
    for (k=0; k<min; k++)
      
      (*coeffc)[ip][k] = legen[lower][ip][k]*albedo[lower]*extinc[lower]
        /alb_num + 
        (reff - reffl) / dr * 
        (legen[upper][ip][k]*albedo[upper]*extinc[upper] - 
         legen[lower][ip][k]*albedo[lower]*extinc[lower]) /
        alb_num;
    
    for (k=min; k<*ncoeffc; k++)
      if (nleg[lower] < nleg[upper])
        (*coeffc)[ip][k] = (reff - reffl) / dr * 
          (legen[upper][ip][k] * albedo[upper]*extinc[upper]) / alb_num;
      else
        (*coeffc)[ip][k] = legen[lower][ip][k]*albedo[lower]*extinc[lower]
          /alb_num - 
          (reff - reffl) / dr *
          legen[lower][ip][k]*albedo[lower]*extinc[lower] / alb_num;
  }
  
  /* The delta scaling factor is calculated from the interpolated (averaged) extinction 
     coefficient and single scattering albedo. */
  *f = (1.0 - (*ext)/ext_us)/alb_us;
  /* To avoid rounding errors for f=0.0 */
  if (*f<1e-8)
    *f=0.0;
 
  /* debugging output 
  fprintf(stderr, "reff %g reffl %g dr %g \n", reff, reffl, dr);
  fprintf(stderr, "f_low %g, f_up %g \n", fdelta[lower], fdelta[upper]);
  fprintf(stderr, "ext_low %g, ext_up %g, alb_low %g, alb_up %g \n",
          extinc[lower], extinc[upper], albedo[lower], albedo[upper]);
  fprintf(stderr, "ext %g, alb %g, fdel %g \n", *ext, *alb, *f);
  fprintf(stderr, "unscaled ext %g, alb %g \n", ext_us, alb_us);
  */

  return 0;
}


/*********************************************************************/
/* Read cloud data from file; microphysical properties are expected  */
/* in three columns:                                                 */
/*   column WCZD:    z [km]                                          */
/*   column WCLWC:   lwc [g / m3]                                    */
/*   column WCEFFR:  effective radius [micron]                       */
/* Memory for microphysical and optical properties is allocated      */
/* automatically. This function can be used for ice and water clouds */
/* because all output is written to cld_out_struct *cld which can    */
/* either be output->wc or output->ic                                */
/*********************************************************************/

int read_cld_file (input_struct input, int nlambda,
		   char *filename, cld_out_struct *cld,
		   int optical, int layer, 
                   float altitude, float *press_atm, float *zd,
                   float *z_model_layer, 
                   cf_out_struct *cf, char *cloud) 
{
  int source    = -1;
  int reff_prop = -1;
  float reff_fixed = -1.0;
  int lev=0, rows=0, min_columns=0, max_columns=0, status=0, lc=0;
  float **cld_data=NULL;

  if      (strncasecmp(cloud,"wc",2)==0) {
    source     = input.wc.source;
    reff_prop  = input.wc.reff_prop;
    reff_fixed = input.wc.reff; 
  }
  else if (strncasecmp(cloud,"ic",2)==0) {
    source = input.ic.source;
    reff_prop = input.ic.reff_prop;
    reff_fixed = input.ic.reff; 
  }
  else {
    fprintf(stderr,"Error, unknown 'cloud option %s' (line %d, function %s in %s) \n", cloud, __LINE__, __func__, __FILE__);
    return -1; 
  }

    
  switch (source) {
  /* read cloud from one file */
  case CLOUDS_FROM_FILE:
  case CLOUDS_IPA_FILES:

    status = ASCII_file2float (filename,
	  		       &rows, &max_columns, &min_columns, 
			       &cld_data);
    if (status!=0) {
      fprintf (stderr, "Error %d opening file %s\n", status, filename);
      return status;
    }
  
    if (max_columns!=min_columns) {
      fprintf (stderr, " !! ATTENTION !! Inconsistent number of columns\n");
      fprintf (stderr, "     min = %d, max =%d\n", min_columns, max_columns);
    }

    if (min_columns<3) {
      fprintf (stderr, "Error, found less than 3 columns in %s\n",
	       filename);
      return -1;
    }

    break;

  case CLOUDS_FROM_ECMWF:
    /* read cloud from ECMWF file */
    status = read_ECMWF_clouds(filename, 
                               input.latitude, input.longitude, 
                               input.UTC, input.atm.time_interpolate, 
                               cloud, reff_prop, reff_fixed, altitude, press_atm, input.cloud_overlap, 
                               input.verbose, input.quiet,
                               &rows, &max_columns, &min_columns, 
                               &cld_data, cf);
    if (status != 0) {
      fprintf (stderr, "Error %d during read_ECMWF_clouds (line %d, function %s in %s)\n", status, __LINE__, __func__, __FILE__);
      return status;
    }
    break;
    
  case CLOUDS_FROM_ECHAM:
    status = read_ECHAM_clouds(input.filename[FN_WCLOUD], 
                               input.latitude, input.longitude,
                               input.UTC, input.atm.time_interpolate, 
                               cloud, z_model_layer, 
                               input.verbose, input.quiet, 
                               &rows, &max_columns, &min_columns,
                               &cld_data, cf);
    
    if (status != 0) {
      fprintf (stderr, "Error %d during read_ECHAM_clouds (line %d, function %s in %s)\n", status, __LINE__, __func__, __FILE__);
      return -1;
    }
    
    break;

  default:  
    fprintf (stderr, "Error, unknown 'input.wc.source = %d' (line %d, function %s in %s)\n", input.wc.source, __LINE__, __func__, __FILE__);
    fprintf (stderr, "  This is a program bug, please report it to us. \n");
    return -1;
  }
  
  /* allocate memory for 1D cloud structure */

  if (optical) {
    status = calloc_cld_out (cld, nlambda, rows, 1);
    if (status!=0) {
      fprintf (stderr, "Error %d allocating memory for cld_out_struct\n", status);
      return status;
    }
  }
  else {
    status = calloc_cld_microphysics (cld, nlambda, rows);
    if (status!=0) {

      fprintf (stderr, "Error %d allocating memory for cld_out_struct\n", status);
      return status;
    }
  }

  for (lev=0; lev<cld->nlev; lev++) {
    cld->zd            [lev] = cld_data[lev][CLD_ZD];
    cld->microphys.lwc [lev] = cld_data[lev][CLD_LWC];
    cld->microphys.effr[lev] = cld_data[lev][CLD_EFFR];
  }
    
  /* free temporary memory */
  ASCII_free_float(cld_data, rows);
  
  /* newsiz is a flag, telling the Fortran function wcloud() if a */
  /* new liquid water content and/or effective radius are given   */
  /* (newsiz=1) or if the values haven't changed since the last   */
  /* call (newsiz=0).                                             */ 
  cld->newsiz = 1;

  /* test if levels are sorted in descending order */
  for (lev=0; lev<cld->nlev-1; lev++)
    if (cld->zd[lev] <= cld->zd[lev+1]) {
      fprintf (stderr, "Error, input grid in %s not sorted in descending order!\n", filename);
      fprintf (stderr, "z[%d] = %f, z[%d] = %f\n", lev, cld->zd[lev], lev+1, cld->zd[lev+1]);
      return -1;
    }
  
  /* calculate layer microphysical properties. */
  /* needs to be done in exactly the same      */
  /* way as in wcloud.f                        */
  if (cld->microphys.alloc) {
    if (layer) {
      for (lc=0; lc<cld->nlev-1; lc++) {
	cld->microphys.lwc_layer [lc] = cld->microphys.lwc [lc+1];
	cld->microphys.effr_layer[lc] = cld->microphys.effr[lc+1];
      }
    }
    else {
      for (lc=0; lc<cld->nlev-1; lc++) {
	if (cld->microphys.lwc[lc+1]>0 && cld->microphys.lwc[lc]>0) {
	  cld->microphys.lwc_layer [lc] = 
	    0.5 * (cld->microphys.lwc [lc+1] + cld->microphys.lwc [lc]);
	  cld->microphys.effr_layer[lc] = 
	    0.5 * (cld->microphys.effr[lc+1] + cld->microphys.effr[lc]);
	}
	else {
	  cld->microphys.lwc_layer [lc] = 0.0;
	  cld->microphys.effr_layer[lc] = 0.0;
	}
      }
    }
  }


  /* determine minimum and maximum effective radii */
  /* of cloudy layers; we assume that the layer    */
  /* properties are within this range, as they are */
  /* linearely interpolated from the level prop's  */

  cld->microphys.effrmin=+FLT_MAX;
  cld->microphys.effrmax=-FLT_MAX;
  for (lev=0; lev<cld->nlev; lev++) {
    if (cld->microphys.lwc [lev] > 0) {
      if (cld->microphys.effr[lev] > cld->microphys.effrmax)  cld->microphys.effrmax=cld->microphys.effr[lev]; 
      if (cld->microphys.effr[lev] < cld->microphys.effrmin)  cld->microphys.effrmin=cld->microphys.effr[lev]; 
    }
  }

  return 0;
}





/* Read water cloud data from file using read_cld_file() and convert */
/* microphysical properties to optical properties                    */

int read_and_convert_wc_file (input_struct input, output_struct *output,
			      char *filename, cld_out_struct *wc)
{
  int iv=0, nlyr=0, status=0;
  int nstring = (int) strlen(input.filename[FN_PATH]);

  void  F77_FCN (wcloud) (float *lambda_t, int *newsiz, int *nlyr,
			  char *filepath, int *nstring, float *lwc, float *wceffr, 
			  float *tmp_wc_dtau, float *tmp_wc_gg, float *tmp_wc_ssa, float *zd,
			  int *wclyr);

  /* read wc_file */
  status = read_cld_file (input, output->wl.nlambda_t, filename, wc, 1, 
                          input.wc.layer,
                          output->alt.altitude, output->atm.microphys.press, 
                          output->atm.zd, output->atm.z_model_layer,  &output->cf,
                          "wc");
  
  if (status!=0) {
    fprintf (stderr, "Error %d reading %s (line %d, function %s in %s)\n", status, filename, __LINE__, __func__, __FILE__);
    return status;
  }
  
  /* default: assume HG cloud - flag needed by MYSTIC */
  wc->cldproperties=CLD_OPTPROP;
  wc->nonHGcld=0;

  switch (input.wc.properties) {
  case WCPROP_HU:
    
    switch (input.ck_scheme) {
    case CK_FU:
      
      /* call Fu and Liou code by Fred Rose */
      
#if HAVE_FULIOU
      status = ckdfucld (wc->zd, 
			 wc->microphys.effr_layer, 
			 wc->microphys.lwc_layer, 
			 wc->nlev,
			 wc->optprop.dtau, 
			 wc->optprop.g1, 
			 wc->optprop.ssa); /* in ckdfu.c */
     

#else
      fprintf (stderr, "Error, Fu and Liou not supported!\n");
      return -1;
#endif
      
      if (status!=0) {
	fprintf (stderr, "Error %d returned by ckdfucld()\n", status);
	return status;
      }
      
      break;
      
    default:   /* old uvspec wcloud.f for Hu and Stamnes */
      
      nlyr = wc->nlev-1;

      /* wcloud calculates lwc_layer and reff_layer internally */
      for (iv=0; iv<output->wl.nlambda_t; iv++)
	F77_FCN (wcloud) (&output->wl.lambda_t[iv], &(wc->newsiz), &nlyr,
			  input.filename[FN_PATH], &nstring, 
			  wc->microphys.lwc, wc->microphys.effr, 
			  wc->optprop.dtau[iv], 
			  wc->optprop.g1[iv], wc->optprop.ssa[iv], 
			  wc->zd, &input.wc.layer);
      
      break;
    }
    
    break;
    
  case WCPROP_ECHAM4:

    for (iv=output->wl.nlambda_rte_lower; iv<=output->wl.nlambda_rte_upper; iv++) {
      status = wc_echam4 (output->wl.lambda_t[iv], wc->nlev-1,
			  wc->microphys.lwc, wc->microphys.effr, 
			  wc->optprop.dtau[iv], wc->optprop.g1[iv], wc->optprop.ssa[iv],
			  wc->zd, input.wc.layer);
      
      if (status!=0) {
	fprintf (stderr, "Error %d returned by ic_echam4()\n", status);
	return status;
      }
    }
    break;  

    if (status!=0) {
      fprintf (stderr, "Error %d returned by wcloud_echam()\n", status);
      return status;
    }

    break;  

  case WCPROP_MIE:
  case WCPROP_FILE:

    /* Read optical properties file */

    /* we don't interpolate level properties to layer properties here */
    /* (could be done but requires some work)                         */
    if (!input.wc.layer) {
      fprintf (stderr, "Error, wc_properties requires optical properties defined per layer.\n");
      fprintf (stderr, "Please specify wc_layer and retry if that is what you mean!\n");
      return -1;
    }
    
    status = read_cloud_prop (input.filename[FN_WCLOUD_PROP], 
			      output->wl.lambda_t, output->wl.nlambda_t, 
			      output->wl.nlambda_rte_lower,
                              output->wl.nlambda_rte_upper, 
			      input.wc.interpolate,
			      &(output->wc), &(output->atm.nmom),
                              input.rte.pol_nstokes, input.wc.properties, 
                              input.verbose, input.quiet);
    
    if (status!=0) {
      fprintf (stderr, "Error %d reading water cloud property file %s\n", 
	       status, input.filename[FN_WCLOUD_PROP]);
      return status;
    }

    /* non Henyey-Greenstein - flag needed by MYSTIC */
    wc->nonHGcld=1;
    wc->cldproperties=CLD_LWCREFF;

    break;

  case WCPROP_PHASE:
    
    /* Read phase function */

    /* we don't interpolate level properties to layer properties here */
    /* (could be done but requires some work)                         */
    if (!input.wc.layer) {
      fprintf (stderr, "Error, wc_properties requires optical properties defined per layer.\n");
      fprintf (stderr, "Please specify wc_layer and retry if that is what you mean!\n");
      return -1;
    }
    
    status = read_cloud_prop_phase (input.rte.mc.filename[FN_MC_PCLD], 
				    output->wl.lambda_t, output->wl.nlambda_t, 
				    output->wl.nlambda_rte_lower, output->wl.nlambda_rte_upper, 
				    input.wc.interpolate, input.verbose, input.quiet,
				    &(output->wc), &(output->atm.nmom));
    
    if (status!=0) {
      fprintf (stderr, "Error %d reading water cloud property file %s\n", 
	       status, input.rte.mc.filename[FN_MC_PCLD]);
      return status;
    }

    /* non Henyey-Greenstein - flag needed by MYSTIC */
    wc->nonHGcld=1;
    wc->cldproperties=CLD_LWCREFF;  

    break;

  default:
    fprintf (stderr, "Error, unknown water cloud property %d. This is not\n", input.wc.properties);
    fprintf (stderr, "supposed to happen and indicates a coding error.\n");

    return -1;
    break;
  }

  return 0;
}
  




/* Read ice cloud data from file using read_cld_file() and convert */
/* microphysical properties to optical properties                  */

int read_and_convert_ic_file (input_struct input, output_struct *output,
			      char *filename, cld_out_struct *ic)
{
  int iv=0, lev=0, newkey=0, status=0;

  status = read_cld_file (input, output->wl.nlambda_t, filename, ic, 1, 
                          input.ic.layer,
                          output->alt.altitude, output->atm.microphys.press, 
                          output->atm.zd, output->atm.z_model_layer,  &output->cf,
                          "ic");
  if (status!=0) {
    fprintf (stderr, "Error %d while reading %s (line %d, function %s in %s)\n", status, filename, __LINE__, __func__, __FILE__);
    return status;
  }
  
  /* default: assume HG cloud - flag needed by MYSTIC */
  ic->cldproperties=CLD_OPTPROP;
  ic->nonHGcld=0;

  switch (input.ic.properties) {
  case ICPROP_KEY:
  case ICPROP_YANG:

    if (input.ic.properties==ICPROP_KEY) {
      if (!input.quiet)
	fprintf (stderr, " ... using Key et al. [2002] ice cloud properties\n");
      
      newkey=0;
    }

    if (input.ic.properties==ICPROP_YANG) {
      if (!input.quiet) {
	fprintf (stderr, " ... using Key et al. [2002] ice cloud properties below 3.4 micron;\n");
	fprintf (stderr, " ... new data by Yang/Mayer above 3.4 micron.\n");
      }
      
      newkey=1;
    }
    
    for (iv=output->wl.nlambda_rte_lower; iv<=output->wl.nlambda_rte_upper; iv++) {
      
      status = ic_yang (output->wl.lambda_t[iv], ic->nlev-1,
			input.ic.habit,
			input.filename[FN_PATH],
			ic->microphys.lwc_layer, ic->microphys.effr_layer, 
			ic->optprop.dtau[iv], 
			ic->optprop.ff[iv], ic->optprop.g1[iv], ic->optprop.g2[iv], 
			ic->optprop.ssa[iv], 
			ic->zd, input.ic.layer, newkey); /* in this file cloud.c */
      
      if (status!=0) {
	fprintf (stderr, "Error %d returned by ic_yang()\n", status);
	return status;
      }
    }
    break;
    
  case ICPROP_FU:

    if (!input.quiet)
      fprintf (stderr, " ... using Fu et al. [1996/98] ice cloud properties\n");
    
    if (input.ic.fu2yang) /* convert from Fu (1996/98) to Key et al. (2002) effective radius */
      for (lev=0; lev<ic->nlev; lev++)
	ic->microphys.effr[lev] /= (3.0*sqrt(3.0)/4.0);

    /* now used */
    if (input.ic.fu2yang) /* convert from Fu (1996/98) to Key et al. (2002) effective radius */
      for (lev=0; lev<ic->nlev; lev++) {
        if (ic->microphys.lwc_layer[lev]>0)
	  if (!input.quiet)
	    fprintf (stderr, "fu2yang: scaling effective radius from %f to", ic->microphys.effr_layer[lev]);

	ic->microphys.effr_layer[lev] /= (3.0*sqrt(3.0)/4.0);

        if (ic->microphys.lwc_layer[lev]>0)
	  if (!input.quiet)
	    fprintf (stderr, " %f\n", ic->microphys.effr_layer[lev]);
      }	

    /* calculate cloud optical properties */
    switch (input.ck_scheme) {
    case CK_FU:
      
#if HAVE_FULIOU
      status = ckdfuice (ic->zd, 
                         ic->microphys.effr_layer, 
                         ic->microphys.lwc_layer, 
                         ic->nlev,
			 ic->optprop.dtau, 
                         ic->optprop.g1, 
                         ic->optprop.ssa, 
                         input.ic.unscaled); /* in ckdfu.c */

#else
      fprintf (stderr, "Error, Fu and Liou not supported!\n");
      return -1;
#endif
      
      if (status!=0) {
	fprintf (stderr, "Error %d returned by ckdfuice()\n", status);
	return status;
      }
      
      break;
      
    default:
      
      for (iv=output->wl.nlambda_rte_lower; iv<=output->wl.nlambda_rte_upper; iv++) {
	
	if (output->wl.lambda_t[iv]<4000.0) {
	  
	  status = ic_fu96 (output->wl.lambda_t[iv], ic->nlev-1,
			    input.filename[FN_PATH],
			    ic->microphys.lwc_layer, ic->microphys.effr_layer, 
			    ic->optprop.dtau[iv], ic->optprop.g1[iv], ic->optprop.ssa[iv], ic->optprop.f[iv],
			    ic->zd, input.ic.layer, input.ic.unscaled); /* in this file cloud.c */

	  if (status!=0) {
	    fprintf (stderr, "Error %d returned by ic_fu96()\n", status);
	    return status;
	  }
	}
	else {
	  status = ic_fu98 (output->wl.lambda_t[iv], ic->nlev-1,
			    input.filename[FN_PATH],
			    ic->microphys.lwc_layer, ic->microphys.effr_layer, 
			    ic->optprop.dtau[iv], ic->optprop.g1[iv], ic->optprop.ssa[iv], 
			    ic->zd, input.ic.layer); /* in this file cloud.c */
	  
	  if (status!=0) {
	    fprintf (stderr, "Error %d returned by ic_fu98()\n", status);
	    return status;
	  }
	}
      }
    }
    break;

  case ICPROP_ECHAM4:

    for (iv=output->wl.nlambda_rte_lower; iv<=output->wl.nlambda_rte_upper; iv++) {
      status = ic_echam4 (output->wl.lambda_t[iv], ic->nlev-1,
			  ic->microphys.lwc, ic->microphys.effr, 
			  ic->optprop.dtau[iv], ic->optprop.g1[iv], ic->optprop.ssa[iv],
			  ic->zd, input.ic.layer);
      
      if (status!=0) {
	fprintf (stderr, "Error %d returned by ic_echam4()\n", status);
	return status;
      }
    }
    break;  

  case ICPROP_FILE:
  case ICPROP_MIE:
    
    /* Read optical properties file */
    if (!(output->ic.ssprop.alloc && output->ic.ssprop.alloc_moments)) {

      /* we don't interpolate level properties to layer properties here */
      /* (could be done but requires some work)                         */
      if (!input.ic.layer) {
	fprintf (stderr, "Error, ic_properties requires optical properties defined per layer.\n");
	fprintf (stderr, "Please specify wc_layer and retry if that is what you mean!\n");
	return -1;
      }

      status = read_cloud_prop (input.filename[FN_ICLOUD_PROP], 
				output->wl.lambda_t, output->wl.nlambda_t, 
				output->wl.nlambda_rte_lower,
                                output->wl.nlambda_rte_upper, 
				input.ic.interpolate,
				&(output->ic), &(output->atm.nmom),
                                input.rte.pol_nstokes, input.ic.properties, 
                                input.verbose, input.quiet);
      
      if (status!=0) {
	fprintf (stderr, "Error %d reading ice cloud property file %s\n", 
		 status, input.filename[FN_ICLOUD_PROP]);
	return status;
      }
    }

    /* non Henyey-Greenstein - flag needed by MYSTIC */
    ic->nonHGcld=1;
    ic->cldproperties=CLD_LWCREFF;

    break;


    /* CE: (Baum parameterization) */
  case ICPROP_BAUM:
  case ICPROP_BAUMDETAILED:  
   
    /* we don't interpolate level properties to layer properties here */
    /* (could be done but requires some work)                         */
    if (!input.ic.layer) {
      fprintf (stderr, "Error, ic_properties requires optical properties defined per layer.\n");
      fprintf (stderr, "Please specify ic_layer and retry if that is what you mean!\n");
      return -1;
    }

    /* Read optical properties file */
    status = read_cloud_prop (input.filename[FN_ICLOUD_PROP], 
                              output->wl.lambda_t, output->wl.nlambda_t, 
                              output->wl.nlambda_rte_lower,
                              output->wl.nlambda_rte_upper, 
                              input.ic.interpolate,
                              &(output->ic), &(output->atm.nmom), 
                              input.rte.pol_nstokes, input.ic.properties, 
                              input.verbose, input.quiet);
    
    if (status!=0) {
      fprintf (stderr, "Error %d reading ice cloud property file %s\n", 
	       status, input.filename[FN_ICLOUD_PROP]);
      return status;
    }
    
    ic->nonHGcld=1;               /*TZ*/
    ic->cldproperties=CLD_LWCREFF;/*TZ*/

    break;
     /* CE */
    

  default:
    fprintf (stderr, "Error, unknown ice cloud property %d. This is not\n", input.ic.properties);
    fprintf (stderr, "supposed to happen and indicates a coding error.\n");

    return -1;
    break;
  }

  return 0;
}



/* convert microphysical and single scattering properties */
/* to optical properties at model layers                  */

/* ????? the following function used lwc[lc+1] and reff[lc+1] ????? */
/* ????? which was obviously wrong; replaced by lwc_layer[lc] ????? */
/* ????? and effr_layer[lc]                                   ????? */

int ssprop2optprop (cld_out_struct *cld, int iv)
{

  int lc=0, status=0;
  
  
  for (lc=0; lc<cld->nlev-1; lc++) {  /* count layers, 0 ... nlev-2 */
    
    /* first free memory, if already allocated */
    if (cld->optprop.moment[iv][lc] != NULL) {
      free (cld->optprop.moment[iv][lc]);
      cld->optprop.moment[iv][lc] = NULL;
    }

    if (cld->microphys.lwc_layer[lc] > 0) {
      /*CE: Interpolation must only be performed if nreff in optical properties
        file is at least 2 !!!*/
      if(cld->ssprop.nreff[iv] > 1){
        /* interpolate Legendre coefficients to specified effective radius; */
        /* memory for moments is allocated here in this function.           */
        if (cld->ssprop.alloc_moments) {
          
          status = interpolate_legendre (cld->ssprop.legen[iv],  
                                         cld->ssprop.nleg[iv],
                                         cld->ssprop.nphase,
                                         cld->ssprop.nreff[iv],
                                         cld->ssprop.r0[iv], 
                                         cld->ssprop.dr[iv],
                                         cld->ssprop.extinc[iv], 
                                         cld->ssprop.albedo[iv],
                                         cld->ssprop.f[iv],
                                         cld->microphys.effr_layer[lc],
                                         &(cld->optprop.moment[iv][lc]), 
                                         &(cld->optprop.nmom[iv][lc]),
                                         &(cld->optprop.dtau[iv][lc]), 
                                         &(cld->optprop.ssa[iv][lc]),
                                         &(cld->optprop.f[iv][lc]));
          
          if (status!=0) {
            fprintf (stderr, "Error %d interpolating Legendre coefficients to reff = %f in layer %d.\n", 
                     status, cld->microphys.effr_layer[lc], lc);
            return status;
          }
        }
      }
      else{ /*no interpolation required if Nreff =1*/
        /* CE: This should probably be an error ???? */
        if(cld->microphys.effr_layer[lc] != cld->ssprop.r0[iv]){
          fprintf(stderr, "Warning: The effective radius in the optical \n");
          fprintf(stderr, "         properties file %f does not correspond to the \n", cld->ssprop.r0[iv]);
          fprintf(stderr, "         effective radius of the cloud %f in layer %d. \n", cld->microphys.effr_layer[lc], lc );
        }
        
        cld->optprop.moment[iv][lc]=cld->ssprop.legen[iv][0];
        cld->optprop.nmom[iv][lc]=cld->ssprop.nleg[iv][0];
        cld->optprop.ssa[iv][lc]=cld->ssprop.albedo[iv][0];
        cld->optprop.f[iv][lc]=cld->ssprop.f[iv][0];
        cld->optprop.dtau[iv][lc]=cld->ssprop.extinc[iv][0];
      }
      
      if (cld->ssprop.alloc_explicit) {

	/* ?????????????????????????????????????????????????????????????????????????????? */
	/* currently we have a special treatment for explicitely defined phase functions; */
	/* this is subject to change, as we want to make the explicit phase function      */
	/* compatible to WCPROP_MIE/WCPROP_FILE; that is, specify phase function in the   */
	/* same format as in the momcld_files for different effective radii plus          */
	/* extinction coefficient and single scattering albedo                            */

	fprintf (stderr, "Error, mc_pcld_file needs some more thinking!\n");
	fprintf (stderr, "In particular, do not expect mc_pcld_file\n");
	fprintf (stderr, "to work with 1D clouds!\n");

	return -1;
      }

      /* Number of phase matrix elements constant */
      cld->optprop.nphase = cld->ssprop.nphase;

      /* convert from extinction / lwc to optical thickness dtau */
      cld->optprop.dtau[iv][lc] *= cld->microphys.lwc_layer[lc] * 
	(cld->zd[lc] - cld->zd[lc+1]);
      
      if (cld->optprop.nmom[iv][lc]>1)
        cld->optprop.g1[iv][lc] = cld->optprop.moment[iv][lc][0][1];
    }
    else {
      cld->optprop.dtau[iv][lc] = 0;
      cld->optprop.ssa [iv][lc] = 0;
      cld->optprop.f[iv][lc] = 0; 
      cld->optprop.g1  [iv][lc] = 0;
      
      /* reset Legendre coefficients */
      cld->optprop.nmom[iv][lc] = 0;
    }
  }
  

  return 0;
}


/*****************************************************/
/* read cloud properties from optical property files */
/*****************************************************/

int read_cloud_prop (char *filename, float *lambda_t, int nlambda_t, 
		     int nlambda_lower, int nlambda_upper,
		     int interpolate, cld_out_struct *cld, int *nmom,
                     int nstokes, int cloud_properties,
                     int verbose, int quiet)                    
{
  int status=0, lc=0, iw=0, iv=0;
  int lambda_optprop_lower=0, lambda_optprop_upper=0;

  ssprop_struct tmp_ssprop;
  float *tmp_lambda=NULL;
  int tmp_nlambda=0;

  char momfilename[FILENAME_MAX]="";
  char tempstr[20]="";

  glob_t momfiles;

  double wavelen=0, nre=0, nim=0;

  if (verbose)
    fprintf (stderr, " ... reading cloud optical properties from %s\n", filename);
  
  if (interpolate) {
    /* read optical properties from file */
    /* try netcdf first                  */
    
    if (cloud_properties==ICPROP_MIE || cloud_properties==WCPROP_MIE)
      sprintf (momfilename, "%s???.mie.cdf", filename);
    else if (cloud_properties==ICPROP_BAUM || cloud_properties==ICPROP_BAUMDETAILED)
      sprintf (momfilename, "%s*.baum.cdf", filename);
    /* User defined */
    else{
      if (!quiet)
        fprintf(stderr, " ... user defined optical properties file . \n");
      sprintf (momfilename, "%s???.mie.cdf", filename);
    }
    
    /* get list of filenames that match the pattern momfilename???.???.cdf */
    status = glob (momfilename, 0, NULL, &momfiles);
    tmp_nlambda = momfiles.gl_pathc; /* number of files found */
    
    if (!quiet)
      fprintf (stderr, " ... reading cloud optical properties from %s\n", momfilename);

    
    /* if no netcdf files found, try ASCII */
    if (tmp_nlambda==0) {
      if (cloud_properties==ICPROP_MIE)
        sprintf (momfilename, "%s???.mie", filename);
      else if (cloud_properties==ICPROP_BAUM)
        sprintf (momfilename, "%s*.baum.cdf", filename);
      
      if (!quiet)
	fprintf (stderr, " ... no netcdf cloud optical property files found, trying ASCII instead!\n");
      
      status = glob (momfilename, 0, NULL, &momfiles);
      tmp_nlambda = momfiles.gl_pathc; /* number of files found */
    }
    
    if (tmp_nlambda==0) {
      fprintf (stderr, "Error, found neither netcdf nor ASCII optical property files.\n");
      fprintf (stderr, "Check your cloud properties and retry!\n");
      return -1;
    }

    
    /* remove extension .cdf if necessary   */
    /* because read_mie_table automatically */
    /* appends .cdf to the filename         */
    for (iw=0; iw<tmp_nlambda; iw++)
      if (strlen(momfiles.gl_pathv[iw])>=4)
	if (!strcmp(momfiles.gl_pathv[iw]+strlen(momfiles.gl_pathv[iw])-4, ".cdf"))
	  momfiles.gl_pathv[iw][strlen(momfiles.gl_pathv[iw])-4]=0;
    
    tmp_lambda = calloc (tmp_nlambda, sizeof (float));


    /* first determine required wavelength range */
    for (iw=0; iw<tmp_nlambda; iw++) {
          status = read_mie_table_lambda (momfiles.gl_pathv[iw], 
				      &wavelen, quiet);
      
      if (status!=0) {
	fprintf (stderr, "Error %d reading wavelength from %s\n", 
		 status, momfiles.gl_pathv[iw]);
	return status;
      }

      tmp_lambda[iw] = wavelen;
    }

    /* check if wavelengths are sorted in ascending order;  */
    /* we could in principle sort them, but if they are not */
    /* sorted we assume that the user did something he did  */
    /* not really want.                                     */
    
    for (iw=0; iw<tmp_nlambda-1; iw++)
      if (tmp_lambda[iw+1] <= tmp_lambda[iw]) {
	fprintf (stderr, "Error, optical properties files not sorted by ascending wavelength\n");
	fprintf (stderr, "       %f nm >= %f nm\n", tmp_lambda[iw], tmp_lambda[iw+1]);
	return 0;
      }
    
    if (!quiet)
      fprintf (stderr, " ... found %d optical properties files covering the wavelength range %f - %f nm\n",
	       tmp_nlambda, tmp_lambda[0], tmp_lambda[tmp_nlambda-1]);
    
    /* check if wavelength range is sufficient */
    if (tmp_lambda[0]>lambda_t[nlambda_lower] || 
	tmp_lambda[tmp_nlambda-1]<lambda_t[nlambda_upper]) {
      fprintf (stderr, "Error, wavelength range covered by optical properties files (%f - %f nm)\n", 
	       tmp_lambda[0], tmp_lambda[tmp_nlambda-1]);
      fprintf (stderr, "does not cover the required range (%f - %f nm)\n",
	       lambda_t[nlambda_lower], lambda_t[nlambda_upper]);
      return -1;
    }

    /* determine required wavelength range */
    for (iw=0; iw<tmp_nlambda; iw++)
      if (tmp_lambda[iw] > lambda_t[nlambda_lower])
	break;
    
    lambda_optprop_lower=iw-1;


    for (iw=0; iw<tmp_nlambda; iw++)
      if (tmp_lambda[iw] >= lambda_t[nlambda_upper])
	break;
    
    lambda_optprop_upper=iw;

    if (!quiet)
      fprintf (stderr, " ... selected wavelengths %d - %d, %f - %f nm\n",
	       lambda_optprop_lower, lambda_optprop_upper,
	       tmp_lambda[lambda_optprop_lower], tmp_lambda[lambda_optprop_upper]);


    /* now select required wavelengths from available range */
    /* and rewrite arrays                                   */
    tmp_nlambda = lambda_optprop_upper - lambda_optprop_lower + 1;
    if(lambda_optprop_lower != 0){
      for (iw=0; iw<tmp_nlambda; iw++) {
        tmp_lambda[iw] = tmp_lambda[iw+lambda_optprop_lower];
        strcpy(momfiles.gl_pathv[iw], momfiles.gl_pathv[iw+lambda_optprop_lower]);
      }
    }
    

    /* allocate memory; need only single scattering properties, */
    /* hence nlev is set to 0                                   */
    status = calloc_ssprop_struct (&tmp_ssprop, tmp_nlambda);
    if (status!=0) {
      fprintf (stderr, "Error %d allocating memory for cld_out struct\n", status);
      return status;
    }
      

    /* read mie tables */
    for (iw=0; iw<tmp_nlambda; iw++) {
     
      status = read_mie_table (momfiles.gl_pathv[iw], 
			       &(tmp_ssprop.r0[iw]), 
			       &(tmp_ssprop.dr[iw]), 
			       &(tmp_ssprop.nreff[iw]),
                               &(tmp_ssprop.reff[iw]),
			       &wavelen, &nre, &nim, 
			       &(tmp_ssprop.extinc[iw]), 
			       &(tmp_ssprop.albedo[iw]),
                               &(tmp_ssprop.f[iw]),
			       &(tmp_ssprop.nleg[iw]),
			       &(tmp_ssprop.legen[iw]),
                               &(tmp_ssprop.nphase),
                               0,
                               nstokes, 
			       quiet);
      
      if (status!=0) {
	fprintf (stderr, "Error %d reading %s\n", status, momfilename);
	return status;
      }
      
      tmp_lambda[iw] = wavelen;
      
      if (!quiet)
	fprintf (stderr, " ... read %3zd reff's from file %3d, %s, %11.4f nm\n", 
		 tmp_ssprop.nreff[iw], iw, momfiles.gl_pathv[iw], tmp_lambda[iw]);

    }
    
    tmp_ssprop.alloc_moments = 1;  /* memory has been allocated by read_mie_table */

    if (!quiet)
      fprintf (stderr, " ... interpolating cloud single scattering properties to internal wavelength grid\n");

    status = interpolate_ssprop (&tmp_ssprop, &(cld->ssprop),
				 tmp_lambda, tmp_nlambda,
				 lambda_t, nlambda_t,
				 nlambda_lower, nlambda_upper);

    if (status!=0) {
      fprintf (stderr, "Error %d returned by interpolate_ssprop()\n", status);
      return status;
    }
  }
  else {
    
    if(cloud_properties==ICPROP_BAUM){
      fprintf (stderr, "Error: Optical properties for Baum parameterization must be \n");
      fprintf (stderr, "       in netcdf format (line %d, function %s in %s)\n", __LINE__, __func__, __FILE__);
      return -1;
    }
    
    for (iv=0; iv<nlambda_t; iv++) {
      
      /* read optical properties from file */
      strcpy (momfilename, filename);
      
      if (iv+1<100)
	strcat (momfilename, "0");
      
      if (iv+1<10) 
	strcat (momfilename, "0");
      
      sprintf (tempstr, "%d", iv+1);
      
      strcat (momfilename, tempstr);
      strcat (momfilename, ".mie");
      
      status = read_mie_table (momfilename, 
			       &(cld->ssprop.r0[iv]), 
			       &(cld->ssprop.dr[iv]), 
			       &(cld->ssprop.nreff[iv]),
                               &(cld->ssprop.reff[iv]),
                               &wavelen, &nre, &nim, 
			       &(cld->ssprop.extinc[iv]), 
			       &(cld->ssprop.albedo[iv]),
                               &(cld->ssprop.f[iv]), 
			       &(cld->ssprop.nleg[iv]),
			       &(cld->ssprop.legen[iv]),
                               &(cld->ssprop.nphase),
                               0,
                               nstokes,
			       quiet);
      
      
      if (status!=0) {
	fprintf (stderr, "Error %d reading %s\n", status, momfilename);
	return status;
      }
    }	
    cld->ssprop.alloc_moments = 1;  /* memory has been allocated by read_mie_table() */
  }
 
  /* convert microphysical and single scattering properties */
  /* to optical properties                                  */
  for (iv=0; iv<nlambda_t; iv++) {
    status = ssprop2optprop (cld, iv);
    
    if (status!=0) {
      fprintf (stderr, "Error %d calculating cloud optical properties\n", status);
      return status;
    }
  }

  cld->ssprop.type = PHASE_MOMENTS; /* read Legendre moments */

  /* increase number of moments if more found */
  for (iv=0; iv<nlambda_t; iv++)
    for (lc=0; lc<cld->nlev-1; lc++)
      if (cld->optprop.nmom[iv][lc]-1 > *nmom)
	*nmom = cld->optprop.nmom[iv][lc]-1;
  
  free(tmp_lambda);
  
  return 0;
}



/*****************************************************/
/* read cloud properties from phase function file    */
/*****************************************************/

int read_cloud_prop_phase (char *filename, float *lambda_t, int nlambda_t, 
			   int nlambda_lower, int nlambda_upper,
			   int interpolate, int verbose, int quiet,
			   cld_out_struct *cld, int *nmom)
{
  int status=0, iv=0, lc=0;

  ssprop_struct tmp_ssprop;
  float *tmp_lambda=NULL;
  int tmp_nlambda=0;
  
  /* if (verbose) */
    fprintf (stderr, " ... reading cloud optical properties from %s\n", filename);

  /* allocate memory; need only single scattering properties, */
  /* hence nlev is set to 0                                   */

  tmp_nlambda=2;
  tmp_lambda = calloc (tmp_nlambda, sizeof(float));

  status = calloc_ssprop_struct (&tmp_ssprop, tmp_nlambda);
  if (status!=0) {
    fprintf (stderr, "Error %d allocating memory for cld_out struct\n", status);
    return status;
  }
  
  /* first wavelength ...                            */

  fprintf (stderr, "read_phase_function (%s)\n", filename);
 
  status = read_phase_function (filename, 
				&(tmp_ssprop.r0[0]), 
				&(tmp_ssprop.dr[0]), 
				&(tmp_ssprop.nreff[0]),
				&(tmp_ssprop.nphase),
				&(tmp_ssprop.extinc[0]), 
				&(tmp_ssprop.albedo[0]), 
				&(tmp_ssprop.nmu[0]),
				&(tmp_ssprop.mu[0]),
				&(tmp_ssprop.p[0]),
				&(tmp_ssprop.type),
				quiet);
  
  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }
  

  
  /* second wavelength ...                           */
  status = read_phase_function (filename, 
				&(tmp_ssprop.r0[1]), 
				&(tmp_ssprop.dr[1]), 
				&(tmp_ssprop.nreff[1]),
				&(tmp_ssprop.nphase),
				&(tmp_ssprop.extinc[1]), 
				&(tmp_ssprop.albedo[1]), 
				&(tmp_ssprop.nmu[1]),
				&(tmp_ssprop.mu[1]),
				&(tmp_ssprop.p[1]),
				&(tmp_ssprop.type),
				quiet);
  
  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }
  
  tmp_lambda[0] = lambda_t[nlambda_lower];
  tmp_lambda[1] = lambda_t[nlambda_upper];

  tmp_ssprop.alloc_explicit = 1;  /* memory has been allocated by read_phase_function */

  if (!quiet)
    fprintf (stderr, " ... read %3zd reff's from file %3d, %s, %11.4f nm\n", 
	     tmp_ssprop.nreff[0], 0, filename, tmp_lambda[0]);
  
  
  
  if (!quiet)
    fprintf (stderr, " ... interpolating cloud single scattering properties to internal wavelength grid\n");

  status = interpolate_ssprop (&tmp_ssprop, &(cld->ssprop),
			       tmp_lambda, tmp_nlambda,
			       lambda_t, nlambda_t,
			       nlambda_lower, nlambda_upper);
  
  if (status!=0) {
    fprintf (stderr, "Error %d returned by interpolate_ssprop()\n", status);
    return status;
  }

  /* convert microphysical and single scattering properties */
  /* to optical properties                                  */
  for (iv=0; iv<nlambda_t; iv++) {
    status = ssprop2optprop (cld, iv);
    
    if (status!=0) {
      fprintf (stderr, "Error %d calculating cloud optical properties\n", status);
      return status;
    }
  }
  

  /* increase number of moments if more found */
  for (iv=0; iv<nlambda_t; iv++)
    for (lc=0; lc<cld->nlev-1; lc++)
      if (cld->optprop.nmom[iv][lc]-1 > *nmom)
	*nmom = cld->optprop.nmom[iv][lc]-1;

  return 0;
}


/* overwrite water cloud properties with user-defined properties */

/* CE: 24.06.2008: Modified this function so that all user defined 
   properties are applied to un-deltascaled optical properties. 
   This assures that for instances *ic_properties baum* and *ic_properties baum_detailed*
   give the same result when ic_set_tau is used.
   If delta scaling is not applied, the delta scaling factor f equals 0 and 
   all options give the same result as before the modificaltion of this function.
*/

int apply_user_defined_properties_to_cld (cld_inp_struct cldin, 
					  atm_out_struct atmout,
					  wl_out_struct  wlout,
					  float altitude,
					  cld_out_struct *cldout)
{
  int iv=0, lc=0, status=0;
  float scale_factor=0.0, sum=0.0;
  double *tautot = NULL, *lambda = NULL;
  double *a0=NULL, *a1=NULL, *a2=NULL, *a3=NULL;
  double tau550=0.0;
  double unscaled_gg=0.0, unscaled_ssa=0.0;


  for (iv=0; iv<wlout.nlambda_t; iv++) {
    
    /* Set the cloud asymmetry factor to a user-defined value */
    if (cldin.gg >= -1.0) {
      for (lc=0; lc<cldout->nlev-1; lc++) {
        
        /* If delta scaling is applied, for example in Fu parameterization, we need the 
           delta scaled asymmetry parameter in optprop_struct*/
	cldout->optprop.g1[iv][lc] = (cldin.gg - cldout->optprop.f[iv][lc])/
          (1.0 - cldout->optprop.f[iv][lc]) ;

        /* To assure that the asymmetry-parameter has the correct value, g2 of the Double-
           HG-function is set to 0. */
        cldout->optprop.g2[iv][lc] = 0.0;
        cldout->optprop.ff[iv][lc] = 1.0;

	/* We don't want a user-defined asymmetry parameter if an explicit  */
	/* phase function was defined (e.g. ic_properties mie, baum);       */
	/* it is not a good idea to overwrite the first moment because then */
	/* the phase function might look quite weird: The first moment      */
	/* defines the linear term in the Legendre expansion; changing that */
	/* results basically in a change of the slope of the phase function */
	/* which easily results in negative values. An alternative would be */
	/* to replace the explicit phase function by a simple HG, but is    */
	/* that what the user really wants?                                 */

        if (cldout->optprop.nmom[iv][lc] > 1) {
	  fprintf (stderr, "Error, it does not make sense to use an explicit phase function\n");
	  fprintf (stderr, "and then to manually over-write the asymmetry parameter.\n"); 
	  return -1;
	}
      }
    }
    
    /* Scale the cloud asymmetry factor with a user-defined value */
    if (cldin.scale_gg >= 0.0) {
      for (lc=0; lc<cldout->nlev-1; lc++) {

        /* The un-deltascaled asymmetry parameter should be scaled. So first we 
           unscale the asymmetry parameter */
        unscaled_gg=cldout->optprop.g1[iv][lc]*(1.0 - cldout->optprop.f[iv][lc]) +
          cldout->optprop.f[iv][lc];
        
        /* Now we apply the scaling factor to the un-deltascaled assymetry parameter */
        unscaled_gg *= cldin.scale_gg;

        /* And deltascale the result */
	cldout->optprop.g1[iv][lc] = (unscaled_gg -  cldout->optprop.f[iv][lc])/
          (1.0 - cldout->optprop.f[iv][lc]);
	

        /* g2 is nonzero in case of Key/Yang  (not delta-scaled)*/
        cldout->optprop.g2[iv][lc] *=  cldin.scale_gg;
        
	/* check if we are still within the limits */
	if (cldout->optprop.g1[iv][lc]>1.0)
	  cldout->optprop.g1[iv][lc] = 1.0;
	
	if (cldout->optprop.g1[iv][lc]<-1.0)
	  cldout->optprop.g1[iv][lc] = -1.0;
	
        
	/* We don't want a user-defined asymmetry parameter if an explicit  */
	/* phase function was defined (e.g. ic_properties mie, baum);       */
	/* see comment above.                                               */
	if (cldout->optprop.nmom[iv][lc] > 1) {
	  fprintf (stderr, "Error, it does not make sense to use an explicit phase function\n");
	  fprintf (stderr, "and then to manually over-write the asymmetry parameter.\n"); 
	  return -1;
	}
      }
    }

    /* Set the cloud single scattering albedo to a user-defined value */
    if (cldin.ssa >= 0.0)
      for (lc=0; lc<cldout->nlev-1; lc++)
        /* We assume that the user means the un-deltascaled ssa, so ssa is here
           delta-scaled */
        cldout->optprop.ssa[iv][lc] = (1.0 - cldout->optprop.f[iv][lc]) * cldin.ssa /
          (1.0 - cldin.ssa * cldout->optprop.f[iv][lc]) ;
    
    /* Scale the cloud single scattering albedo with a user-defined value */
    if (cldin.scale_ssa >= 0.0)
      for (lc=0; lc<cldout->nlev-1; lc++) {
        
        /* Here also we assume that the user wants to scale the un-deltascaled ssa. 
           So we unscale the ssa first */
        unscaled_ssa = cldout->optprop.ssa[iv][lc]/
          (1.0 + cldout->optprop.f[iv][lc] * ( cldout->optprop.ssa[iv][lc] - 1.0));
                                 
        /* Apply scaling factor */
        unscaled_ssa *= cldin.scale_ssa;

        /* Delta scale the result */
	cldout->optprop.ssa[iv][lc] = (1.0 - cldout->optprop.f[iv][lc])*unscaled_ssa/
          (1.0 - unscaled_ssa* cldout->optprop.f[iv][lc]);
        
        if (cldout->optprop.ssa[iv][lc]>1.0)
	  cldout->optprop.ssa[iv][lc] = 1.0;
      }
  
  
    /* Set the cloud optical thickness to a constant user-defined value */
    /* (CE) Important: This needs to be done *after* ssa is modified !!!*/ 
    if (cldin.tau >= 0.0) {
      
      sum = 0.0;
      
      for (lc=0; lc<cldout->nlev-1; lc++) {
        
        unscaled_ssa = cldout->optprop.ssa[iv][lc]/
          (1.0 + cldout->optprop.f[iv][lc] * ( cldout->optprop.ssa[iv][lc] - 1.0));
        
        if (cldout->zd[lc] > altitude) {    /* layers with upper boundary above the surface */
	  if (cldout->zd[lc+1] > altitude){  /* layers completely above the surface          */
            
            /* Un-deltascale optical thickness. If no delta scaled properties are used 
               f is 0 and the optical thickness is not changed here */
            sum += cldout->optprop.dtau[iv][lc]/
              (1.0 - unscaled_ssa*cldout->optprop.f[iv][lc]) ;
          }
	  else                              /* upper boundary above, lower boundary below surface */
	    sum += cldout->optprop.dtau[iv][lc]/  
              (1.0 - unscaled_ssa*cldout->optprop.f[iv][lc]) * 
              (cldout->zd[lc]-altitude) / (cldout->zd[lc]-cldout->zd[lc+1]);
	}
      }
      
      scale_factor = 0.0;
      if (sum>0)
	scale_factor = cldin.tau/sum;
      
      for (lc=0; lc<cldout->nlev-1; lc++)
	cldout->optprop.dtau[iv][lc] *= scale_factor;
      
    } 
  }



  /* Set the cloud optical thickness to a user-defined value at 550nm */
  if (cldin.tau550 >= 0.0) {
    lambda = calloc (wlout.nlambda_t, sizeof(double));
    tautot = calloc (wlout.nlambda_t, sizeof(double));
    
    for (iv=0; iv<wlout.nlambda_t; iv++) {
      lambda[iv] = (double) wlout.lambda_t[iv];

      for (lc=0; lc<cldout->nlev-1; lc++) {

	unscaled_ssa = cldout->optprop.ssa[iv][lc]/
          (1.0 + cldout->optprop.f[iv][lc] * ( cldout->optprop.ssa[iv][lc] - 1.0));
        
        if (cldout->zd[lc] > altitude) {    /* layers with upper boundary above the surface */
	  if (cldout->zd[lc+1] > altitude)  /* layers completely above the surface          */
            
            /* un-deltascale optical thickness */
	    tautot[iv] += cldout->optprop.dtau[iv][lc] /
              (1.0 - unscaled_ssa *
               cldout->optprop.f[iv][lc]) ;
          else        /* upper boundary above, lower boundary below surface */
            tautot[iv] += cldout->optprop.dtau[iv][lc]/
              (1.0 - unscaled_ssa * cldout->optprop.f[iv][lc]) * 
	      (cldout->zd[lc]-altitude) / (cldout->zd[lc]-cldout->zd[lc+1]);
	}
      }
    }
    
    status = linear_coeffc (lambda, tautot, wlout.nlambda_t, &a0, &a1, &a2, &a3);
    if (status!=0) {
      fprintf (stderr, "Error %d calculating interpolation coefficients for cloud optical thickness.\n", 
	       status);
      return status;
    }

    status = calc_splined_value (550.0, &tau550, lambda, wlout.nlambda_t, a0, a1, a2, a3);
    if (status!=0) {
      fprintf (stderr, "Error %d interpolating cloud optical thickness to 550nm.\n", 
	       status);
      fprintf (stderr, "In the current implementation of xx_set_tau550, it is required\n");
      fprintf (stderr, "that the internal wavelength range includes 550nm.\n");
      fprintf (stderr, "Hope to change that in future!\n");
      return status;
    }
    
    if (tau550>0.0) {
      scale_factor = cldin.tau550 / tau550;
      
      for (iv=0; iv<wlout.nlambda_t; iv++)
	for (lc=0; lc<cldout->nlev-1; lc++)
	  cldout->optprop.dtau[iv][lc] *= scale_factor;
    }


    free(a0); free(a1); free(a2); free(a3);
    free(lambda); free(tautot);
  }

  return 0;
}


/*agf*/
static int ic_yang (float wavelength, int nlyr, int habit,
		    char *path,
		    float *iwc, float *reff, 
		    float *dtau, 
		    float *f, float *g1, float *g2, float *ssa, 
		    float *zd, int layer, int newkey)
{
  float ext=0;
  int lc=0, status=0;
  char mipath[256]="";
  
  sprintf(mipath,"%s/ic/yang56/",path);

  if (layer==0) {
    fprintf (stderr, "Error, Yang et al. [2000] can only be used with cloud properties\n");
    fprintf (stderr, "defined per layer; please use ic_layer!\n");
    return -1;
  }

  for (lc=0; lc<nlyr; lc++) {
  
    /* calculate dtau[lc], g1[lc], g2[lc], f[lc], ssa[lc] from */
    /* iwc[lc], reff[lc], wavelength                           */

    if (iwc[lc]>0.0) {

      if (newkey) {
#if HAVE_YANG

	status = yang56 (wavelength/1000.0, reff[lc], habit, mipath, &ext, ssa+lc, g1+lc, g2+lc, f+lc, newkey);
	
	if (status!=0) {
	  fprintf (stderr, "Error %d returned by yang56()\n", status);
	  return status;
	}
#else
	fprintf (stderr, " ***********************************************************************\n");
	fprintf (stderr, " * You have built uvspec without Yang/Key/Mayer et al support and      *\n");
	fprintf (stderr, " * hence cannot use 'ic_properties yang'.                              *\n");
	fprintf (stderr, " ***********************************************************************\n");
	return -1;
#endif
      }
      else {
#if HAVE_KEY56

	status = yang56 (wavelength/1000.0, reff[lc], habit, mipath, &ext, ssa+lc, g1+lc, g2+lc, f+lc, newkey);
	
	if (status!=0) {
	  fprintf (stderr, "Error %d returned by yang56()\n", status);
	  return status;
	}
#else
	fprintf (stderr, " ***********************************************************************\n");
	fprintf (stderr, " * You have built uvspec without Key et al support and hence cannot    *\n");
	fprintf (stderr, " * use 'ic_properties key'.                                            *\n");
	fprintf (stderr, " ***********************************************************************\n");
	return -1;
#endif
      }

      dtau[lc] = ext*iwc[lc]*(zd[lc]-zd[lc+1]);
    }
  }
  
  return 0;   /* if ok */
}




/*******************************************************************/
/* Ice cloud properties according to                               */
/* Q. Fu, An Accurate Parameterization of the Solar Radiative      */
/* Properties of Cirrus Clouds for Climate Models,                 */
/* Journal of Climate 9, 2058-2082, 1996.                          */
/*******************************************************************/

int ic_fu96 (float wavelength, int nlyr,
	     char *path,
	     float *iwc, float *reff, 
	     float *dtau, float *g, float *ssa, float *f,
	     float *zd, int layer, int unscaled)
{
  int i=0;
  int lc=0, status=0;
  char filename [FILENAME_MAX]="";

  static double *al1=NULL, *al2=NULL, *a0=NULL, *a1=NULL;
  static double *bl1=NULL, *bl2=NULL, *b0=NULL, *b1=NULL, *b2=NULL, *b3=NULL;
  static double *cl1=NULL, *cl2=NULL, *c0=NULL, *c1=NULL, *c2=NULL, *c3=NULL;
  static double *dl1=NULL, *dl2=NULL, *d0=NULL, *d1=NULL, *d2=NULL, *d3=NULL;

  static int na=0, nb=0, nc=0, nd=0;

  static int first=1;

  double D=0;
  double asy=0, ext=0, coalb=0, del=0;
  double fw=0;

  float *deff = calloc (nlyr+1, sizeof(float));

  if (layer==0) {
    fprintf (stderr, "Error, Fu [1996] can only be used with cloud properties\n");
    fprintf (stderr, "defined per layer. Please use ic_layer!\n");
    return -1;
  }

  if (first) {  /* read only once */
    first=0;
    
    /* read extinction data */
    sprintf (filename,"%s/ic/fu96/fu96.ext",path);
    status = read_4c_file (filename, &al1, &al2, &a0, &a1, &na);
    if (status!=0) {
      fprintf (stderr, "Error %d reading Fu [1996] data from %s\n", 
	       status, filename);
      return status;
    }
    
    /* read absorption data */
    sprintf (filename,"%s/ic/fu96/fu96.ssa",path);
    status = read_6c_file (filename, &bl1, &bl2, &b0, &b1, &b2, &b3, &nb);
    if (status!=0) {
      fprintf (stderr, "Error %d reading Fu [1996] data from %s\n", 
	       status, filename);
      return status;
    }
    
    /* read asymmetry parameter data */
    sprintf (filename,"%s/ic/fu96/fu96.asy",path);
    status = read_6c_file (filename, &cl1, &cl2, &c0, &c1, &c2, &c3, &nc);
    if (status!=0) {
      fprintf (stderr, "Error %d reading Fu [1996] data from %s\n", 
	       status, filename);
      return status;
    }

    /* read delta-fraction parameter data */
    sprintf (filename,"%s/ic/fu96/fu96.del",path);
    status = read_6c_file (filename, &dl1, &dl2, &d0, &d1, &d2, &d3, &nd);
    if (status!=0) {
      fprintf (stderr, "Error %d reading Fu [1996] data from %s\n", 
	       status, filename);
      return status;
    }

    /* check if wavelength grids are identical */
    if (na!=nb || na!=nc || na!=nd) {
      fprintf (stderr, "Error, something weird happened to the Fu [1996] data files.\n");
      fprintf (stderr, "Found different wavelength grids.\n");
      return -1;
    }

    for (i=0; i<na; i++)
      if (al1[i] != bl1[i] || al1[i] != cl1[i] || al1[i] != dl1[i]) {
	fprintf (stderr, "Error, something weird happened to the Fu [1996] data files.\n");
	fprintf (stderr, "Found different wavelength grids.\n");
	return -1;
      }
    
    /* convert from micron to nm */
    for (i=0; i<na; i++) {
      al1[i] *= 1000.0;
      al2[i] *= 1000.0;
    }
  }
  
  /* determine the tabulated wavelength interval containing the actual wavelength */
  if (wavelength < al1[0] || wavelength > al2[na-1]) {
    fprintf (stderr, "Error, wavelength %f nm not covered by Fu [1996]\n", wavelength);
    fprintf (stderr, "Allowed range is %.1f - %.1f nm\n", al1[0], al2[na-1]);
    return -1;
  }


  /* Fu et al. requires effective diameter */
  for (lc=0; lc<=nlyr; lc++)
    deff[lc] = 2.0 * reff[lc];

  for (i=0; i<na-1; i++)
    if (wavelength < al2[i])
      break;

  for (lc=0; lc<nlyr; lc++) {
    if (iwc[lc]>0) {
      /* calculate dtau[lc], g[lc], and ssa[lc] from */
      /* iwc[lc], deff[lc], wavelength               */
      
      D = deff[lc];

      if (D < MIN_DEFF_FU96 || D > MAX_DEFF_FU96) {
	fprintf (stderr, "Error, effective radius %f um not covered by Fu [1996]\n",D/2.);
	fprintf (stderr, "Allowed range is %7.3f - %7.3f um\n",
		 MIN_DEFF_FU96/2.0, MAX_DEFF_FU96/2.0);
	return -1;
      }
      
      /* extinction */
      ext = a0[i] + a1[i]/D;
      
      /* single scattering albedo */
      coalb = b0[i] + b1[i] * D + b2[i] * D * D + b3[i] * D * D * D;
      
      /* asymmetry parameter */
      asy = c0[i] + c1[i] * D + c2[i] * D * D + c3[i] * D * D * D;

      /* delta-fraction */
      del = d0[i] + d1[i] * D + d2[i] * D * D + d3[i] * D * D * D;
      
      dtau[lc] = ext*iwc[lc]*(zd[lc]-zd[lc+1])*1000.0;
      ssa [lc] = 1.0-coalb;
      g   [lc] = asy;
     

      /* Fu [1996], Eq. 3.8 */
      f[lc]  = 0.5 / ssa[lc] + del;
      fw     = f[lc] * ssa[lc];

      /* Fu [1996], pg. 2067, left, first paragraph */
      if (f[lc]>g[lc])
	f[lc] = g[lc];
      
      /* delta-scaling, Fu [1996], eqs. A.2a - A.2c */
      if (!unscaled) {
	dtau[lc] = (1.0 - fw) * dtau[lc];
	ssa [lc] = (1.0 - f[lc]) * ssa[lc] / (1.0 - fw);
	g   [lc] = (g[lc] - f[lc]) / (1.0 - f[lc]);
      }
      else 
        /* in this case we need to set the delta-scaling factor to 0 */
	/* because it otherwise might be applied to the data later   */
	/* which is not desired                                      */
	f[lc] = 0.0;

    }
  }
  
  free(deff);

  return 0;   /* if ok */
}

/*******************************************************************/
/* Ice cloud properties according to                               */
/* Q. Fu, P. Yang, and W.B. Sun, An Accurate Parameterization of   */
/* the Infrared Radiative Properties of Cirrus Clouds for Climate  */
/* Models, Journal of Climate 11, 2223-2237, 1998.                 */
/*******************************************************************/

int ic_fu98 (float wavelength, int nlyr,
	     char *path,
	     float *iwc, float *reff, 
	     float *dtau, float *g, float *ssa, 
	     float *zd, int layer)
{
  int i=0, lower=0, upper=0;
  int lc=0, status=0;
  char filename [FILENAME_MAX]="";

  static double *al=NULL, *a0=NULL, *a1=NULL, *a2=NULL;
  static double *bl=NULL, *b0=NULL, *b1=NULL, *b2=NULL, *b3=NULL;
  static double *cl=NULL, *c0=NULL, *c1=NULL, *c2=NULL, *c3=NULL;

  static int na=0, nb=0, nc=0;

  static int first=1;

  double D=0;
  double extlower=0, extupper=0, ext=0;
  double abslower=0, absupper=0, abs=0;
  double asylower=0, asyupper=0, asy=0;

  float *deff = calloc (nlyr+1, sizeof(float));

  if (layer==0) {
    fprintf (stderr, "Error, Fu et al. [1998] can only be used with cloud properties\n");
    fprintf (stderr, "defined per layer. Please use ic_layer!\n");
    return -1;
  }

  if (first) {  /* read only once */
    first=0;
    
    /* read extinction data */
    sprintf (filename,"%s/ic/fu98/fu98.ext",path);
    status = read_4c_file (filename, &al, &a0, &a1, &a2, &na);
    if (status!=0) {
      fprintf (stderr, "Error %d reading Fu et al. [1998] data from %s\n", 
	       status, filename);
      return status;
    }
    
    /* read absorption data */
    sprintf (filename,"%s/ic/fu98/fu98.abs",path);
    status = read_5c_file (filename, &bl, &b0, &b1, &b2, &b3, &nb);
    if (status!=0) {
      fprintf (stderr, "Error %d reading Fu et al. [1998] data from %s\n", 
	       status, filename);
      return status;
    }
    
    /* read asymmetry parameter data */
    sprintf (filename,"%s/ic/fu98/fu98.asy",path);
    status = read_5c_file (filename, &cl, &c0, &c1, &c2, &c3, &nc);
    if (status!=0) {
      fprintf (stderr, "Error %d reading Fu et al. [1998] data from %s\n", 
	       status, filename);
      return status;
    }

    /* check if wavelength grids are identical */
    if (na!=nb || na!=nc) {
      fprintf (stderr, "Error, something weird happened to the Fu et al. [1998] data files.\n");
      fprintf (stderr, "Found different wavelength grids.\n");
      return -1;
    }

    for (i=0; i<na; i++)
      if (al[i] != bl[i] || al[i] != cl[i]) {
	fprintf (stderr, "Error, something weird happened to the Fu et al. [1998] data files.\n");
	fprintf (stderr, "Found different wavelength grids.\n");
	return -1;
      }
    
    /* convert from micron to nm */
    for (i=0; i<na; i++)
      al[i] *= 1000.0;
    
  }

  
  /* determine tabulated wavelengths above and below the actual wavelength */
  if (wavelength < al[0] || wavelength > al[na-1]) {
    fprintf (stderr, "Error, wavelength %f nm not covered by Fu et al. [1998]\n", wavelength);
    fprintf (stderr, "Allowed range is %.1f - %.1f nm\n", al[0], al[na-1]);
    return -1;
  }


  /* Fu et al. requires effective diameter */
  for (lc=0; lc<=nlyr; lc++)
    deff[lc] = 2.0 * reff[lc];


  for (i=0; i<na-1; i++)
    if (wavelength < al[i])
      break;

  lower=i-1;
  upper=i;
    
  for (lc=0; lc<nlyr; lc++) {
  

    if (iwc[lc]>0) {
      /* calculate dtau[lc], g[lc], and ssa[lc] from */
      /* iwc[lc], deff[lc], wavelength               */
      
      D = deff[lc];
      
      if (D < MIN_DEFF_FU98 || D > MAX_DEFF_FU98) {
	fprintf (stderr, "Error, effective radius %f um not covered by Fu et al. [1998]\n", D/2.0);
	fprintf (stderr, "Allowed range is %6.2f - %6.2f um\n",
		 MIN_DEFF_FU98/2.0, MAX_DEFF_FU98/2.0);
	return -1;
      }

      /* extinction */
      extlower = a0[lower] + a1[lower]/D + a2[lower] / (D * D);
      extupper = a0[upper] + a1[upper]/D + a2[upper] / (D * D);
      ext = extlower + (wavelength - al[lower]) / (al[upper]-al[lower]) * (extupper-extlower);
      
      /* absorption */
      abslower = b0[lower] + b1[lower] * D + b2[lower] * D * D + b3[lower] * D * D * D;
      absupper = b0[upper] + b1[upper] * D + b2[upper] * D * D + b3[upper] * D * D * D;
      abs = (abslower + (wavelength - al[lower]) / (al[upper]-al[lower]) * (absupper-abslower)) / D;
      
      /* asymmetry parameter */
      asylower = c0[lower] + c1[lower] * D + c2[lower] * D * D + c3[lower] * D * D * D;
      asyupper = c0[upper] + c1[upper] * D + c2[upper] * D * D + c3[upper] * D * D * D;
      asy = asylower + (wavelength - al[lower]) / (al[upper]-al[lower]) * (asyupper-asylower);
      
      dtau[lc] = ext*iwc[lc]*(zd[lc]-zd[lc+1])*1000.0;
      ssa [lc] = (ext-abs)/ext;
      g   [lc] = asy;
    }
  }

  free(deff);
  
  return 0;   /* if ok */
}


/*******************************************************************/
/* ECHAM4 water cloud optical properties, see                      */
/* Roeckner et al., MPI Report 118, Table 2, page 26.              */
/*******************************************************************/

int wc_echam4 (float wavelength, int nlyr,
	       float *iwc, float *reff, 
	       float *dtau, float *g, float *ssa,
	       float *zd, int layer)
{
  int lc=0;
  double logreff=0, ext=0;

  /* check allowed wavelength range */
  if (wavelength < 200.0 || wavelength > 4900.0) {
    fprintf (stderr, "Error, wavelength %f out if range; allowed is 200 - 4000nm.\n",
	     wavelength);
    return -1;
  }

  if (layer==0) {
    fprintf (stderr, "Error, ECHAM4 can only be used with cloud properties\n");
    fprintf (stderr, "defined per layer; please use ic_layer!\n");
    return -1;
  }

  for (lc=0; lc<nlyr; lc++) {
    if (iwc[lc+1]>0) {
      /* calculate dtau[lc], g[lc], and ssa[lc] from */
      /* iwc[lc], deff[lc], wavelength               */
      
      logreff = log10(reff[lc+1]);

      /* optical thickness */
      if (wavelength<=680) 
	ext = 1.8706058417*pow(reff[lc+1],-1.0756364457);
      else
	ext = 1.9655460426*pow(reff[lc+1],-1.0778999732);

      dtau[lc] = ext*iwc[lc+1]*(zd[lc]-zd[lc+1])*1000.0;
      
      /* single scattering albedo */
      if (wavelength<=680) 
        ssa [lc] = 1.0;
      else
        ssa [lc] = 0.9854369057 
	    + 0.013584242533*logreff 
	    - 0.024856960461*logreff*logreff
	    + 0.0055918314369*logreff*logreff*logreff;

      /* asymmetry parameter */
      if (wavelength<=680) 
	g [lc] = 0.78756640717
	  + 0.10660598895*logreff
	  - 0.031012468401*logreff*logreff;
      else
	g [lc] = 0.79208639802 
	  -0.044930076174*logreff
	  + 0.18980672305*logreff*logreff
	  - 0.082590933352*logreff*logreff*logreff;
    }
  }

  return 0;
}


/*******************************************************************/
/* ECHAM4 ice cloud optical properties, see                        */
/* Roeckner et al., MPI Report 118, Table 2, page 26.              */
/*******************************************************************/

int ic_echam4 (float wavelength, int nlyr,
	       float *iwc, float *reff, 
	       float *dtau, float *g, float *ssa,
	       float *zd, int layer)
{
  int lc=0;
  double logreff=0, ext=0;
  static int first=1;

  /* check allowed wavelength range */
  if (wavelength < 200.0 || wavelength > 4900.0) {
    fprintf (stderr, "Error, wavelength %f out if range; allowed is 200 - 4000nm.\n",
	     wavelength);
    return -1;
  }

  if (layer==0) {
    fprintf (stderr, "Error, ECHAM4 can only be used with cloud properties\n");
    fprintf (stderr, "defined per layer; please use ic_layer!\n");
    return -1;
  }

  if (first) {
    fprintf (stderr, "*** WARNING: Check if the additional factor 0.91 to g should really be applied!\n");
    fprintf (stderr, "*** See Roeckner et al. 1996, MPI-Report 218, pg 26.\n");
    first=0;
  }

  for (lc=0; lc<nlyr; lc++) {
    if (iwc[lc+1]>0) {
      /* calculate dtau[lc], g[lc], and ssa[lc] from */
      /* iwc[lc], deff[lc], wavelength               */
      
      logreff = log10(reff[lc+1]);

      /* optical thickness */
      if (wavelength<=680) 
	ext = 1.9056067426*pow(reff[lc+1],-1.0318784654);
      else
	ext = 2.1666771102*pow(reff[lc+1],-1.0634702711);

      dtau[lc] = ext*iwc[lc+1]*(zd[lc]-zd[lc+1])*1000.0;
      
      /* single scattering albedo */
      if (wavelength<=680) 
        ssa [lc] = 1.0;
      else
        ssa [lc] = 0.98475089485
	    + 0.0053152066002*logreff 
	    - 0.0061150583857*logreff*logreff
	    - 0.0032775655896*logreff*logreff*logreff;

      /* asymmetry parameter */
      if (wavelength<=680) 
	g [lc] = 0.7700034985
	  + 0.19598466851*logreff
	  - 0.11836420885*logreff*logreff
	  + 0.025209205131*logreff*logreff*logreff;
      else
	g [lc] = 0.83631171237
	  - 0.19965998649*logreff
	  + 0.46130320487*logreff*logreff
	  - 0.29719270332*logreff*logreff*logreff
	  + 0.062554483594*logreff*logreff*logreff*logreff;

      /* multiply with 0.91 (see Table 2 of Roeckner et al. 1996, MPI report 218) */
      g[lc] *= 0.91;
    }
  }

  return 0;
}



int interpolate_ssprop (ssprop_struct *src, ssprop_struct *tgt,
                        float *src_lambda, int src_nlambda,
                        float *tgt_lambda, int tgt_nlambda,
                        int nlambda_lower, int nlambda_upper)
{
  int ir=0, iv=0, k=0, ip=0, imu=0, status=0;
  int maxleg=0;
  int linear=1;

  float *tmp_extinc = calloc (src_nlambda, sizeof(float));
  float *tmp_albedo = calloc (src_nlambda, sizeof(float));
  float *tmp_f = calloc (src_nlambda, sizeof(float));
  float *tmp_legen  = calloc (src_nlambda, sizeof(float));

  float *tmp_extinc_t  = calloc (tgt_nlambda, sizeof(float));
  float *tmp_albedo_t  = calloc (tgt_nlambda, sizeof(float));
  float *tmp_f_t  = calloc (tgt_nlambda, sizeof(float));
  float ***tmp_legen_t = NULL;

  float *tmp_p=NULL, *tmp_p_int=NULL;

  

  if (!(src->alloc_moments || src->alloc_explicit)) {
    fprintf (stderr, "Error, not possible to interpolate ssprop because no memory allocated\n");
    return -1;
  }

  /* check source optical properties,                        */
  /* if effective radius grid identical for all wavelengths; */
  /* otherwise cannot do interpolation                       */
  
  /* CE: In case of aerosol the humidity grid (corresponding to radius dimension) is not equidistant */
  /* Therefore check radius grid instead of r0, dr and nreff */
  
  /* original code:  */
  /*for (iv=0; iv<src_nlambda-1; iv++) {
    if ((src->r0[iv] != src->r0[iv+1])) {
      fprintf (stderr, "Error, effective radius grids in optical property files not identical:\n"); 
      fprintf (stderr, "r0[%d] = %f, r0[%d] = %f\n", 
	       iv, src->r0[iv], iv+1, src->r0[iv+1]);
      return -1;
    }
    
    if ((src->dr[iv] != src->dr[iv+1])) {
      fprintf (stderr, "Error, effective radius grids in optical property files not identical:\n"); 
      fprintf (stderr, "dr[%d] = %f, dr[%d] = %f\n", 
	       iv, src->dr[iv], iv+1, src->dr[iv+1]);
      return -1;
    }
    
    if ((src->nreff[iv] != src->nreff[iv+1])) {
      fprintf (stderr, "Error, effective radius grids in optical property files not identical:\n"); 
      fprintf (stderr, "nreff[%d] = %d, nreff[%d] = %d\n", 
	       iv, src->nreff[iv], iv+1, src->nreff[iv+1]);
      return -1;
    }
  }
  */
  
  /* new code */
  for (iv=0; iv<src_nlambda-1; iv++) {
    
    if (src->nreff[iv] != src->nreff[iv+1]) {
      fprintf (stderr, "Error, effective radius grids in optical property files not identical:\n"); 
      fprintf (stderr, "nreff[%d] = %zd, nreff[%d] = %zd\n", 
	       iv, src->nreff[iv], iv+1, src->nreff[iv+1]);
      return -1;
    }
    
    for (ir=0; ir<src->nreff[iv]; ir++){
      if( src->reff[iv][ir] != src->reff[iv+1][ir]) {
        fprintf (stderr, "Error, effective radius grids in optical property files not identical:\n"); 
        return -1;
      }
    }
  }		 
  
  /*fprintf (stderr, " ... allocating %d wavelengths for tgt\n", tgt_nlambda);*/
  
  status = calloc_ssprop_struct (tgt, tgt_nlambda);
  if (status!=0) {
    fprintf (stderr, "Error %d allocating memory for target cloud structure\n", status);
    return status;
  }
  
  tgt->nphase    = src->nphase;
  tgt->type      = src->type;

  /* copy effective radii */
  for (iv=0; iv<tgt_nlambda; iv++) {
    tgt->r0[iv]    = src->r0[0];
    tgt->dr[iv]    = src->dr[0];
    tgt->nreff[iv] = src->nreff[0];
  }
  
 
  for (iv=0; iv<tgt_nlambda; iv++) {
    tgt->extinc[iv] = calloc (tgt->nreff[iv], sizeof(double));
    tgt->albedo[iv] = calloc (tgt->nreff[iv], sizeof(double));
    tgt->f[iv] = calloc (tgt->nreff[iv], sizeof(double));
    tgt->reff[iv]  =  calloc (tgt->nreff[iv], sizeof(double));
    
    tgt->legen [iv] = calloc (tgt->nreff[iv], sizeof(float **));
    tgt->nleg  [iv] = calloc (tgt->nreff[iv], sizeof(int));

    tgt->nmu   [iv] = calloc (tgt->nreff[iv], sizeof(int));
    tgt->mu    [iv] = calloc (tgt->nreff[iv], sizeof(double *));
    tgt->p     [iv] = calloc (tgt->nreff[iv], sizeof(double **));
    
    /* copy effective radius grid */
    /* Since the radius grid should be the same for all wavelength,
       it does not need to be interpolated. */
    for (ir=0; ir<tgt->nreff[iv]; ir++)
      tgt->reff[iv][ir] = src->reff[0][ir];
  }


  /* interpolate */
  for (ir=0; ir<src->nreff[0]; ir++) {
    
    for (iv=0; iv<src_nlambda; iv++) {
      tmp_extinc[iv] = (float) (src->extinc[iv][ir]);
      tmp_albedo[iv] = (float) (src->albedo[iv][ir]);
      tmp_f[iv] = (float) (src->f[iv][ir]);
    }
    
    /* interpolate extinction */
    status = arb_wvn2 (src_nlambda, src_lambda, 
		       tmp_extinc,  
		       nlambda_lower, nlambda_upper,
		       tgt_lambda, tmp_extinc_t, linear);
    
    if (status!=0) {
      fprintf (stderr, "Error %d interpolating extinction to internal wavelengths\n", status);
      return status;
    }
    
    /* interpolate single scattering albedo */
    status = arb_wvn2 (src_nlambda, src_lambda, 
		       tmp_albedo,  
		       nlambda_lower, nlambda_upper,
		       tgt_lambda, tmp_albedo_t, linear);
    
    if (status!=0) {
      fprintf (stderr, "Error %d interpolating single scattering albedo to internal wavelengths\n", status);
      return status;
    }
    
    /* CE: interpolate delta scaling factor */
    /* ?????? It is not clear whether delta scaling factor can be interpolated*/
    /* in wavelength. Also albedo and Legendre coefficients should probably not */
    /* be interpolated linearly, but weighted with tau and ssa*tau respectively */
    
    /* interpolate single scattering f */
    status = arb_wvn2 (src_nlambda, src_lambda, 
		       tmp_f,  
		       nlambda_lower, nlambda_upper,
		       tgt_lambda, tmp_f_t, linear);
    
    if (status!=0) {
      fprintf (stderr, "Error %d interpolating delta scaling factor to internal wavelengths\n", status);
      return status;
    }
    
    for (iv=0; iv<tgt_nlambda; iv++) {
      tgt->extinc[iv][ir] = (double) (tmp_extinc_t[iv]);
      tgt->albedo[iv][ir] = (double) (tmp_albedo_t[iv]);
      tgt->f[iv][ir] = (double) (tmp_f_t[iv]);
    }
    
    if (src->alloc_moments) {

      /* determine maximum number of Legendre coefficients */
      /* for this effective radius                         */
      maxleg=0;
      for (iv=0; iv<src_nlambda; iv++)
	if (src->nleg[iv][ir]>maxleg)
	  maxleg=src->nleg[iv][ir];
      
      /* allocate memory for moments */
      tmp_legen_t = calloc (src->nphase, sizeof(float *));
      for (ip=0; ip<src->nphase; ip++){
	tmp_legen_t[ip] = calloc (maxleg, sizeof(float *));
	for (k=0; k<maxleg; k++)
	  tmp_legen_t[ip][k] = calloc (tgt_nlambda, sizeof(float));
      }
      
      /* interpolate legendre moments */
      for (ip=0; ip<src->nphase; ip++){
	for (k=0; k<maxleg; k++) {
	  for (iv=0; iv<src_nlambda; iv++) {
	    tmp_legen[iv]=0;
	    if (k<src->nleg[iv][ir])
	      tmp_legen[iv] = src->legen[iv][ir][ip][k];
	  }
	  
	  
	  
	  /* interpolate Legendre coefficients */
	  status = arb_wvn2 (src_nlambda, src_lambda, 
			     tmp_legen,  
			     nlambda_lower, nlambda_upper,
			     tgt_lambda, tmp_legen_t[ip][k], linear);
	  
	  if (status!=0) {
	    fprintf (stderr, "Error %d interpolating Legendre coefficients to internal wavelengths\n", status);
	    return status;
	  }
	}
      }
      
      /* get index of maximum nonzero moment of phase function for each wavelength */
      for (iv=0; iv<tgt_nlambda; iv++) {
	for (ip=0; ip < src->nphase; ip++){
	  for (k=maxleg-1; k>=0; k--)  
	    if (tmp_legen_t[ip][k][iv]!=0)
	      break;
	}
	
	tgt->nleg[iv][ir] = k+1;
	
	/* allocate memory for moments and copy to final destination */
	tgt->legen[iv][ir] = calloc (src->nphase, sizeof(float *));
	for (ip=0; ip<src->nphase; ip++) {
	  tgt->legen[iv][ir][ip]  = calloc (tgt->nleg[iv][ir], sizeof(float));
	  for (k=0; k<tgt->nleg[iv][ir]; k++)
	    tgt->legen[iv][ir][ip][k] = tmp_legen_t[ip][k][iv];
	}
      }
      
      /* free memory for moments */
      for (ip=0; ip<src->nphase; ip++){
	for (k=0; k<maxleg; k++)
	  free(tmp_legen_t[ip][k]);
	free(tmp_legen_t[ip]);
      }
      free(tmp_legen_t);

      tgt->alloc_moments = 1; /* memory has been allocated */
    }
    
    if (src->alloc_explicit) {

      /* first make sure that all input mu grids are equal */
      for (iv=0; iv<src_nlambda; iv++) {
	for (ir=0; ir<src->nreff[iv]; ir++) {
	  if (src->nmu[iv][ir] != src->nmu[0][0]) {
	    fprintf (stderr, "Error, mu grids different - cannot interpolate ssprop\n");
	    return -1;
	  }
	  for (imu=0; imu<src->nmu[iv][ir]; imu++) {
	    if (src->mu[iv][ir][imu] != src->mu[0][0][imu]) {
	      fprintf (stderr, "Error, mu grids different - cannot interpolate ssprop\n");
	      return -1;
	    }
	  }
	}
      }

      /* now we know that all mu grids are equal - start interpolation */

      /* alloacte memory for target */
      /* ????? check if we really need all or only nlambda_lower ... nlambda_upper */
      /* ????? in the latter case cp_cld_out needs to be adapted as well!          */
      /* for (iv=nlambda_lower; iv<=nlambda_upper; iv++) { */
      for (iv=0; iv<tgt_nlambda; iv++) {
	for (ir=0; ir<tgt->nreff[iv]; ir++) {
	  tgt->nmu [iv][ir] = src->nmu[0][0];
	  tgt->mu  [iv][ir] = calloc (tgt->nmu[iv][ir], sizeof(double));
	  tgt->p   [iv][ir] = calloc (tgt->nphase, sizeof(double *));
	  
	  for (ip=0; ip<tgt->nphase; ip++) 
	    tgt->p [iv][ir][ip] = calloc (tgt->nmu[iv][ir], sizeof(double));
	}
      }	  


      /* allocate memory for temporary phase functions */
      tmp_p     = calloc (src_nlambda, sizeof(float));
      tmp_p_int = calloc (tgt_nlambda, sizeof(float));

      for (ir=0; ir<src->nreff[0]; ir++) {
	for (imu=0; imu<src->nmu[0][ir]; imu++) {
	  for (ip=0; ip<src->nphase; ip++) {
	    for (iv=0; iv<src_nlambda; iv++) 
	      tmp_p [iv] = src->p[iv][ir][ip][imu];
	    
	    status = arb_wvn2 (src_nlambda, src_lambda, 
			       tmp_p,  
			       nlambda_lower, nlambda_upper,
			       tgt_lambda, tmp_p_int, linear);
	    
	    
	    for (iv=nlambda_lower; iv<=nlambda_upper; iv++) {
	      tgt->mu[iv][ir][imu]     = src->mu[0][ir][imu];
	      tgt->p [iv][ir][ip][imu] = tmp_p_int [iv];
	    }
	  }
	}
      }

      tgt->alloc_explicit = 1; /* memory has been allocated */

      free(tmp_p);
      free(tmp_p_int);
    }
  }    
  

  free (tmp_extinc);
  free (tmp_albedo);
  free (tmp_f);
  free (tmp_legen);
  free (tmp_f_t);
  free (tmp_albedo_t);
  free (tmp_extinc_t);
  
  
  return 0;
}


int read_isccp_reflectivity (int type, float sza, float tau, char *path, int quiet,
			     float *ref)
{
  int i=0, status=0;
  int u=0, l=0;
  int utau=0, ltau=0, usza=0, lsza=0;
  double ref1=0, ref2=0;
  static int read[2]={0,0};
  static size_t ntau=0, nsza=0;
  
  static double *szas=NULL, *taus=NULL;
  static double **refs=NULL;

  char filename[FILENAME_MAX]="";

#if HAVE_LIBNETCDF
  /* try to open the CDF file */

  int ncid=0;

  int idd_nsza=0, idd_ntau=0;
  int id_sza=0, id_tau=0, id_ref=0;

  size_t start[1] = {0};
  size_t count[1] = {0};

  size_t refstart[2] = {0,0};
  size_t refcount[2] = {0,0};
#endif
 

  if (!read[type]) {
    
    strcpy (filename, path);
    
    switch (type) {
    case ISCCP_WATER:
      strcat(filename, "wc/isccp/wc_reflectivity.cdf");
      break;
    case ISCCP_ICE:
      strcat(filename, "ic/isccp/ic_reflectivity.cdf");
      break;
    default:
      fprintf (stderr, "Error, unknown ISCCP cloud type\n");
      return -1;
    }

    /* read reflectivity file */

#if HAVE_LIBNETCDF

    /* open netcdf file */
    status = nc_open (filename, NC_NOWRITE, &ncid);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, filename);
      return -1;
    }

    /* read file */
    
    /* get dimension id for "sza" */
    status = nc_inq_dimid (ncid, "sza", &idd_nsza);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
    }
    
    /* get dimension length for "sza" */
    status = nc_inq_dimlen (ncid, idd_nsza, &nsza);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
    }
    
    /* get dimension id for "tau" */
    status = nc_inq_dimid (ncid, "tau", &idd_ntau);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
    }
    
    /* get dimension length for "tau" */
    status = nc_inq_dimlen (ncid, idd_ntau, &ntau);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
    }

    /* allocate memory for fields */
    szas    = calloc (nsza, sizeof(double));
    taus    = calloc (ntau, sizeof(double));
    
    refs    = calloc (nsza, sizeof(double *));
    for (i=0; i<nsza; i++)
      refs[i] = calloc (ntau, sizeof(double));
    
    /* get variable id for "sza" */
    status = nc_inq_varid (ncid, "sza", &id_sza);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
      }
    
    /* get variable id for "tau" */
    status = nc_inq_varid (ncid, "tau", &id_tau);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
    }
    
    /* get variable id for "ref" */
    status = nc_inq_varid (ncid, "ref", &id_ref);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
    }
    
    /* read "sza" */
    count[0] = nsza;
    status = nc_get_vara_double (ncid, id_sza, start, count, szas);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
    }
    
    /* read "tau" */
    count[0] = ntau;
    status = nc_get_vara_double (ncid, id_tau, start, count, taus);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
    }
    
    /* read reflectivities */
    refstart[1] = 0;  /* start with first reflectivity    */
    refcount[0] = 1;  /* read one reflectivity at a time  */
      
    for (i=0; i<nsza; i++) {
      
      refstart[0] = i;
      refcount[1] = ntau;
      
      status = nc_get_vara_double (ncid, id_ref, refstart, refcount, refs[i]);
      if (status!=NC_NOERR) {
	fprintf (stderr, "Error %d reading %s\n", status, filename);
	return status;
      }
    }
    
    /* close netcdf file */
    nc_close(ncid);

    if (!quiet)
      fprintf (stderr, " ... read ISCCP reflectivities from %s\n", filename);

#else
    fprintf (stderr, "Error, need netcdf for the sssi solver\n");
    fprintf (stderr, "Please install netcdf and re-compile!\n");
    return -1;
#endif

    /************************************************************/
    /* now we have the cloud reflectivity data stored in arrays */
    /* szas[isza], taus[itau], and refs[isza][itau];            */
    /* next step is the interpolation of the reflectivity to    */
    /* the user-defined (sza, tau)                              */
    /************************************************************/

    read[type] = 1;
  }  

  /* check if values out of range */
  if (tau<taus[0])  {
    fprintf (stderr, "Error, tau smaller than minimum in reflectivity table, %f\n", taus[0]);
    return -1;
  }

  if (tau>taus[ntau-1])  {
    fprintf (stderr, "Error, tau larger than maximum in reflectivity table, %f\n", taus[ntau-1]);
    return -1;
  }

  if (sza<szas[0])  {
    fprintf (stderr, "Error, sza smaller than minimum in reflectivity table, %f\n", szas[0]);
    return -1;
  }

  if (sza>szas[nsza-1])  {
    fprintf (stderr, "Error, sza larger than maximum in reflectivity table, %f\n", szas[nsza-1]);
    return -1;
  }

  /* now search for sza in sza array */
  l = 0;
  u = nsza-1;
  while (1==1) {
    if (u <= l+1)
      break;

    /* midpoint */
    i=(l+u)/2;

    if (sza==szas[i])
      u=l=i;
    else {
      if (sza<szas[i])
	u = i;
      else
	l = i;
    }
  }

  lsza=l;
  usza=u;

  /* now search for tau in tau array */
  l = 0;
  u = ntau-1;
  while (1==1) {
    if (u <= l+1)
      break;

    /* midpoint */
    i=(l+u)/2;

    if (tau==taus[i])
      u=l=i;
    else {
      if (tau<taus[i])
	u = i;
      else
	l = i;
    }
  }

  ltau=l;
  utau=u;

  /* l and u are the indices (upper and lower) of the */
  /* neighbouring data points; now do simply a linear */
  /* interpolation ...                                */

  /* first, interpolation in sza */
  if (lsza==usza) {
    ref1 =  refs[lsza][ltau];
    ref2 =  refs[lsza][utau];
  }
  else {
    ref1 =  refs[lsza][ltau] + (refs[usza][ltau]-refs[lsza][ltau]) / (szas[usza] - szas[lsza]) * (sza - szas[lsza]);
    ref2 =  refs[lsza][utau] + (refs[usza][utau]-refs[lsza][utau]) / (szas[usza] - szas[lsza]) * (sza - szas[lsza]);
  }

  /* second, interpolation in tau */
  if (ltau==utau)
    *ref = ref1;
  else
    *ref = ref1 + (ref2-ref1) / (taus[utau] - taus[ltau]) * (tau - taus[ltau]);
  
  return 0;
}

/********************************************************************************************************/

int read_cloudfraction_from_netCDF_file (char *filename, float lat, float lon, struct tm UTC, int time_interpolate, 
                                         float **z, float **cf, int *cf_nlev,
                                         int verbose, int quiet)
{

#if HAVE_MYSTIC

#if HAVE_LIBNETCDF

  int status=-1;  

/*   int ncid=NOT_DEFINED_INTEGER; */

/*   size_t nlat  = NOT_DEFINED_INTEGER; */
/*   size_t nlon  = NOT_DEFINED_INTEGER; */
/*   size_t nlev  = NOT_DEFINED_INTEGER; */

/*   double *ECMWF_lat  = NULL; */
/*   double *ECMWF_lon  = NULL; */

/*   size_t ilat  = NOT_DEFINED_INTEGER;   /\* index for lat, lon, time in netCDF file *\/ */
/*   size_t ilon  = NOT_DEFINED_INTEGER;   /\* index for lat, lon, time in netCDF file *\/ */
/*   size_t itime = NOT_DEFINED_INTEGER;   /\* index for lat, lon, time in netCDF file *\/ */

/*   float dt = NOT_DEFINED_FLOAT; */
/*   int itime1 = NOT_DEFINED_INTEGER, itime2 = NOT_DEFINED_INTEGER; */

/*   int lc = NOT_DEFINED_INTEGER; /\* loop index *\/ */
/*   int t  = NOT_DEFINED_INTEGER; /\* loop index *\/ */
/*   int nt = NOT_DEFINED_INTEGER; */

/*   float **CC=NULL; */

/*   /\* open netcdf file *\/ */
/*   status = nc_open (filename, NC_NOWRITE, &ncid); */
/*   if (status!=NC_NOERR) { */
/*     fprintf (stderr, "Error %d opening netCDF file %s\n", status, filename); */
/*     return status; */
/*   } */

/*   /\* read latitude array *\/ */
/*   alloc_and_read_netCDF_1D_double(ncid,"lat", &nlat, "lat", &(ECMWF_lat)); */

/*   /\* search correct latitude index *\/ */
/*   ilat = get_grid_index( lat, ECMWF_lat, nlat, FALSE); */
/*   if (ilat < 0) { */
/*     fprintf (stderr, "Error -1 finding index for lat=%5.2f in %s\n", lat, filename); */
/*     return -1; */
/*   } */

/*   /\* read longitude *\/ */
/*   alloc_and_read_netCDF_1D_double(ncid,"lon", &nlon, "lon", &(ECMWF_lon)); */

/*   /\* search correct latitude index *\/ */
/*   ilon = get_grid_index( lon, ECMWF_lon, nlon, TRUE); */
/*   if (ilon < 0) { */
/*     fprintf (stderr, "Error -2 finding index for lon=%5.2f in %s\n", lon, filename); */
/*     return -2; */
/*   } */

/*   /\* get time index *\/ */
/*   status = get_time_index (ncid, UTC, time_interpolate,  */
/*                            &(nt), &(itime1), &(itime2), &(dt), */
/*                            verbose, quiet); */
/*   if (status != 0) { */
/*     fprintf (stderr, "Error %d, during get_time_index in read_cloudfraction_from_netCDF_file (cloud.c)\n", status); */
/*     return status; */
/*   } */

/*   /\* alloc nt timesteps for pressure *\/ */
/*   if ((CC = calloc (nt, sizeof (float *))) == NULL) { */
/*     fprintf (stderr,"Error, allocating memory for cloudfraction CC in read_cloudfraction_from_netCDF_file (cloud.c)\n"); */
/*     return -10; */
/*   } */

/*   /\* read nt (= 1 or 2) time steps *\/ */
/*   for (t=0;t<=nt-1;t++) { */

/*     if (t == 0) */
/*       itime = itime1; */
/*     if (t == 1) */
/*       itime = itime2; */

/*     /\* allocate and read cloud cover CC, defined on layers, name in ECMWF is CC == cloud cover *\/ */
/*     status = alloc_and_read_netCDF_column_float(ncid, "CC", &(CC[t]), nlev, itime, ilat, ilon, verbose); */
/*     if (status != 0) { */
/*       fprintf (stderr, "Error %d reading cloudfraction cf from netCDF file %s\n", status, filename); */
/*       return status; */
/*     } */
/*   } */

/*   (*cf_nlev) = nlev; */

/*   /\* alloc nt timesteps for pressure *\/ */
/*   if (((*cf) = calloc ((*cf_nlev), sizeof (float))) == NULL) { */
/*     fprintf (stderr,"Error, allocating memory for cloudfraction CF in read_cloudfraction_from_netCDF_file (cloud.c)\n"); */
/*     return -11; */
/*   } */

/*   if (nt == 1) { */
/*     /\* no time interpolation needed, just copy *\/ */
/*     for (lc=0; lc<(*cf_nlev); lc++) */
/*       (*cf)[lc] = CC[0][lc]; */
/*   } */
/*   else { */
/*     /\* calculate time interpolated data *\/ */
/*     for (lc=0; lc<(*cf_nlev); lc++) */
/*       (*cf)[lc] = (1.0-dt)* CC[0][lc]  +  dt* CC[1][lc]; */
/*   } */

/*   /\* allocate and read cloud cover CC, defined on layers, name in ECMWF is CC == cloud cover *\/ */
/*   status = alloc_and_read_netCDF_column_float(ncid, "Z", z, nlev, itime, ilat, ilon, verbose); */
/*   if (status != 0) { */
/*     fprintf (stderr, "Error %d reading cloudfraction cf from netCDF file %s\n", status, filename); */
/*     return status; */
/*   } */

  fprintf (stderr, "Error %d, read_cloudfraction_from_netCDF_file not yet checked (cloud.c)\n", status);
  return status;

#else
  fprintf (stderr, " ************************************************************************\n");
  fprintf (stderr, " * You have built uvspec without libnetcdf and hence cannot             *\n");
  fprintf (stderr, " * use the netCDF cloud-fraction option. Please get netcdf and rebuild. *\n");
  fprintf (stderr, " ************************************************************************\n");
  return -1;
#endif

#else
  fprintf (stderr, " Error, cloud fraction input from netCDF is not available in this version \n");
  return -1;
#endif

}


/********************************************************************************************************/

int read_ECMWF_clouds (char *filename, float lat, float lon, struct tm UTC, int time_interpolate, char *cloud, 
                       int reff_prop, float reff_fixed, float altitude, float *press_atm, int cloud_overlap, int verbose, int quiet,
                       int *rows, int *max_columns, int *min_columns, float ***cld_data, cf_out_struct *cf)
{
  int status=0;

#if HAVE_MYSTIC

#if HAVE_LIBNETCDF

  int ncid    = NOT_DEFINED_INTEGER;
  int id_data = NOT_DEFINED_INTEGER;

  char T_name    [2]="";
  char CLWC_name [5]="";
  char CIWC_name [5]="";
  char CC_name   [4]="";

  int lc = NOT_DEFINED_INTEGER; /* loop index */
  int t  = NOT_DEFINED_INTEGER; /* loop index */
  int nt = NOT_DEFINED_INTEGER;
 
  size_t nlat  = NOT_DEFINED_INTEGER;
  size_t nlon  = NOT_DEFINED_INTEGER;
  size_t nlev  = NOT_DEFINED_INTEGER;
  size_t nlay  = NOT_DEFINED_INTEGER;  

  size_t ilat  = NOT_DEFINED_INTEGER;   /* index for lat, lon, time in netCDF file */
  size_t ilon  = NOT_DEFINED_INTEGER;   /* index for lat, lon, time in netCDF file */
  size_t itime = NOT_DEFINED_INTEGER;   /* index for lat, lon, time in netCDF file */

  int itime1 = NOT_DEFINED_INTEGER, itime2 = NOT_DEFINED_INTEGER;

  double *ECMWF_lat  = NULL;
  double *ECMWF_lon  = NULL;

  float dt = NOT_DEFINED_FLOAT;
  float dx = NOT_DEFINED_FLOAT;

  float **p_level = NULL;
  float **p_layer = NULL;
  float **T       = NULL;
  float  *T_level = NULL;
  float **CWC     = NULL; /* cloud water content, liquid or ice in kg/kg */

  float **CC      = NULL; /* cloud cover, water and ice together */

  float *z        = NULL;
  float *reff     = NULL;
  float D = NOT_DEFINED_FLOAT, L = NOT_DEFINED_FLOAT, ou2fu = NOT_DEFINED_FLOAT;
  float *z_reff   = NULL;
  float dens_air  = NOT_DEFINED_FLOAT; /* number density of air moleculs in moleculs per cm^3 */
  float factor    = NOT_DEFINED_FLOAT; /* unit conversion factor */

  float c0=-1,c1=-1,c2=-1,c3=-1;
  float T_in_C=-1;

  int first_layer_done=FALSE;


  /* open netcdf file */
  status = nc_open (filename, NC_NOWRITE, &ncid);
  if (status!=NC_NOERR) { 
    fprintf (stderr, "Error '%s' opening netCDF file '%s' (line %d, function %s in %s)\n", 
                      nc_strerror(status), filename, __LINE__, __func__, __FILE__);
    return -abs(status);
  }

  /* check format */
  if      ( (status = nc_inq_varid (ncid, "t", &id_data)) == NC_NOERR ) {
    strcpy (T_name,    "t");
    strcpy (CLWC_name, "clwc");
    strcpy (CIWC_name, "ciwc");
    strcpy (CC_name,   "cc");
  }
  else if ( (status = nc_inq_varid (ncid, "T", &id_data)) == NC_NOERR ) {
    strcpy (T_name,    "T");
    strcpy (CLWC_name, "CLWC");
    strcpy (CIWC_name, "CIWC");
    strcpy (CC_name,   "CC");
  }
  else {
    fprintf (stderr, "Error '%s' while getting id for temperature from '%s' \n", nc_strerror(status), filename);
    fprintf (stderr, "      (line %d, function %s in %s)\n", __LINE__, __func__, __FILE__);
    return -1;
  }

  /* read latitude array */
  alloc_and_read_netCDF_1D_double(ncid, "lat", &nlat, "lat", &(ECMWF_lat));

  /* search correct latitude index */
  ilat = get_grid_index( lat, ECMWF_lat, nlat, FALSE);
  if (ilat < 0) {
    fprintf (stderr, "Error -1 finding index for lat=%5.2f in %s (line %d, function %s in %s)\n", lat, filename, __LINE__, __func__, __FILE__);
    return -1;
  }

  /* read longitude */
  alloc_and_read_netCDF_1D_double(ncid,"lon", &nlon, "lon", &(ECMWF_lon));

  /* search correct latitude index */
  ilon = get_grid_index( lon, ECMWF_lon, nlon, TRUE);
  if (ilon < 0) {
    fprintf (stderr, "Error -2 finding index for lon=%5.2f in %s (line %d, function %s in %s)\n", lon, filename, __LINE__, __func__, __FILE__);
    return -2;
  }

  /* get time index */
  status = get_time_index (ncid, UTC, time_interpolate, 
                           &(nt), &(itime1), &(itime2), &(dt),
                           verbose, quiet);
  if (status != 0) {
    fprintf (stderr, "Error %d, during get_time_index (line %d, function %s in %s)\n", status, __LINE__, __func__, __FILE__);
    return status;
  }

  /* alloc nt timesteps for pressure */
  if ((p_level = calloc (nt, sizeof (float *))) == NULL) {
    fprintf (stderr,"Error, allocating memory for pressure (line %d, function %s in %s)\n", __LINE__, __func__, __FILE__);
    return -10;
  }

  /* alloc nt timesteps for pressure */
  if ((p_layer = calloc (nt, sizeof (float *))) == NULL) {
    fprintf (stderr,"Error, allocating memory for pressure (line %d, function %s in %s)\n", __LINE__, __func__, __FILE__);
    return -11;
  }

  /* alloc nt timesteps for temperature */
  if ((T = calloc (nt, sizeof (float *))) == NULL) {
    fprintf (stderr,"Error, allocating memory for temperature (line %d, function %s in %s)\n", __LINE__, __func__, __FILE__);
    return -12;
  }

  /* alloc nt timesteps for temperature */
  if ((CWC = calloc (nt, sizeof (float *))) == NULL) {
    fprintf (stderr,"Error, allocating memory for Cloud Water Content (line %d, function %s in %s)\n", __LINE__, __func__, __FILE__);
    return -13;
  }

  /* alloc nt timesteps for temperature */
  if ((CC = calloc (nt, sizeof (float *))) == NULL) {
    fprintf (stderr,"Error, allocating memory for Cloud Fraction (line %d, function %s in %s)\n", __LINE__, __func__, __FILE__);
    return -14;
  }


  /* read nt (= 1 or 2) time steps */
  for (t=0;t<=nt-1;t++) {

    if (t == 0)
      itime = itime1;
    if (t == 1)
      itime = itime2;

    /* alloc and read pressure */
    /* requires that SP, hyai, hybi, hyam, and hybm are in the ECMWF netCDF file */
    alloc_and_read_ECMWF_netCDF_pressure (ncid, &(p_level[t]), &(nlev), &(p_layer[t]), &(nlay), itime, ilat, ilon, verbose);

    /* nlev == nlay, as we throw the uppermost level (p=0.0) away, in order to combine the ECMWF data with background data */
    /* !!! not any more !!! */

    /* allocate and read temperature */
    status = alloc_and_read_netCDF_column_float(ncid, T_name, &(T[t]), nlay, itime, ilat, ilon, verbose);
    if (status != 0) {
      fprintf (stderr, "Error %d reading temperature '%s' from netCDF file %s (line %d, function %s in %s)\n", 
                        status, T_name, filename, __LINE__, __func__, __FILE__);
     return status;
    }

    /* allocate and read cloud water */
    if      (strncasecmp(cloud,"wc",2)==0) {
      /* allocate and read CLWC */
      status = alloc_and_read_netCDF_column_float(ncid, CLWC_name, &CWC[t], nlay, itime, ilat, ilon, verbose);
      if (status != 0) {
        fprintf (stderr, "Error %d reading Cloud Liquid Water Content '%s' from netCDF file %s (line %d, function %s in %s)\n", 
                         status, CLWC_name, filename, __LINE__, __func__, __FILE__);
        return status;
      }
    }
    else if (strncasecmp(cloud,"ic",2)==0) {
      /* allocate and read CIWC */
      status = alloc_and_read_netCDF_column_float(ncid, CIWC_name, &CWC[t], nlay, itime, ilat, ilon, verbose);
      if (status != 0) {
        fprintf (stderr, "Error %d reading Cloud Ice Water Content '%s' from netCDF file %s (line %d, function %s in %s)\n", 
                         status, CIWC_name, filename, __LINE__, __func__, __FILE__);
        return status;
      }
    }
    else {
      fprintf(stderr,"Unknown cloud type '%s' (line %d, function %s in %s)\n", cloud, __LINE__, __func__, __FILE__ );
      return -1; 
    }

    /* allocate and read CloudCover == CloudFraction */
    if ( cloud_overlap != CLOUD_OVERLAP_OFF ) {
      status = alloc_and_read_netCDF_column_float(ncid, CC_name, &CC[t], nlay, itime, ilat, ilon, verbose);
      if (status != 0) {
        fprintf (stderr, "Error %d reading Cloud Cover '%s' from netCDF file %s (line %d, function %s in %s)\n", 
                         status, CC_name, filename, __LINE__, __func__, __FILE__);
        return status;
      }
    }
  }

  if (nt == 1) {
    /* no time interpolation needed, do nothing */
  }
  else {
    /* write time interpolated data into the zero'th entry */
    for (lc=0; lc<nlev; lc++)
      p_level[0][lc] = (1.0-dt)* p_level[0][lc]  +  dt* p_level[1][lc];
    for (lc=0; lc<nlay; lc++) {
      p_layer[0][lc] = (1.0-dt)* p_layer[0][lc]  +  dt* p_layer[1][lc];
      T      [0][lc] = (1.0-dt)* T      [0][lc]  +  dt* T      [1][lc];
      CWC    [0][lc] = (1.0-dt)* CWC    [0][lc]  +  dt* CWC    [1][lc];
      if ( cloud_overlap != CLOUD_OVERLAP_OFF )
        CC     [0][lc] = (1.0-dt)* CC     [0][lc]  +  dt* CC     [1][lc];
    }
  }

/*   /\* additional verbose output data on layers *\/ */
/*   fprintf (stderr, " ADDITIONAL VERBOSE OUTPUT: ECMWF clouds on layers \n"); */
/*   for (lc=0; lc<nlay; lc++) { */
/*     fprintf (stderr,"p_level[%3d]=%10.4f p_layer[%3d]=%10.4f, p_level[%3d]=%10.4f, T=%10.5f, CWC=%12.6e, CC=%12.6e \n", */
/*       lc+1, p_level[0][lc+1], lc, p_layer[0][lc], lc, p_level[0][lc], T[0][lc], CWC[0][lc], CC[0][lc]); */
/*   } */

  /* allocate temperature data on levels */
  if ((T_level = calloc (nlev, sizeof (float *))) == NULL) {
    fprintf (stderr,"Error, allocating memory for T_layer (line %d, function %s in %s)\n", 
                    __LINE__, __func__, __FILE__);
    return -30;
  }

  /* inter- and extra-polate from LAYER to LEVEL values for of T on the log(p/p0) grid */
  /* interpolation */
  for (lc=1; lc<nlev-1; lc++) {
    dx = ( log ( p_level[0][lc] / p_level[0][nlev-1]) - log ( p_layer[0][lc-1] / p_level[0][nlev-1])) /
         ( log ( p_layer[0][lc] / p_level[0][nlev-1]) - log ( p_layer[0][lc-1] / p_level[0][nlev-1]) );
    T_level[lc] = (1.0-dx) *  T[0][lc-1] + dx *  T[0][lc];
  }
  /* extrapolate last layer midpoint to surface */
  dx = ( log ( p_level[0][nlev-1] / p_level[0][nlev-1]) - log ( p_layer[0][nlay-1] / p_level[0][nlev-1]) ) / 
       ( log ( p_layer[0][nlay-1] / p_level[0][nlev-1]) - log ( p_layer[0][nlay-2] / p_level[0][nlev-1]) );
  T_level[nlev-1] = (1.0+dx) *  T[0][nlay-1] - dx *  T[0][nlay-2];
  /* assume constant temperature between last layer midpoint and TOA */
  T_level[  0   ] = T [0][0];

  /* if uppermost pressure level is 0, than replace it by uppermost background atmosphere pressure */
  if ( p_level[0][0] == 0.0 )
    p_level[0][0] = press_atm[0];

  /* calculate z from T and p using hydrostatic equation (in ancillary.c) */
  status = calculate_z_from_p_and_T (p_level[0], T_level, &(z), nlev, altitude, verbose);
  if (status!=0) {
    fprintf (stderr, "Error %d calculating z-grid of ECMWF cloud data (line %d, function %s in %s) \n", 
                     status, __LINE__, __func__, __FILE__);
    return status;
  }

  /* /\* additional verbose output data on levels *\/ */
  /* fprintf (stderr, " ### ECMWF p and T on levels for integration of z ### \n"); */
  /* for (lc=0; lc<nlev; lc++) { */
  /*   fprintf (stderr,"lc=%3d, p=%10.4f, T=%10.5f, z=%10.5f \n", lc, p_level[0][lc], T_level[lc], z[lc]); */
  /* } */

  /* allocate and read cloud water */
  if      (strncasecmp(cloud,"wc",2)==0) {
    /* alloc 2 levels for height of cloud droplet effective radius */
    if ((z_reff = calloc (2, sizeof (float *))) == NULL) {
      fprintf (stderr,"Error, allocating memory for 'z_reff' (line %d, function %s in %s)\n", __LINE__, __func__, __FILE__);
      return -17;
    }
    /* alloc 2 levels for cloud droplet effective radius */
    if ((reff = calloc (2, sizeof (float *))) == NULL) {
      fprintf (stderr,"Error, allocating memory for 'reff' (line %d, function %s in %s)\n", __LINE__, __func__, __FILE__);
      return -16;
    }

    reff[0]=45.0; /* 45 micro meter at top of atmosphere */
    reff[1]=10.0; /* 10 micro meter at surface */

    /* height scale for effective radius range from surface to top of atmosphere */
    z_reff[0]=z[0];
    z_reff[1]=z[nlev-1];

    /* linear interpolation of the effective radius to the z-scale of the cloud data */
    status = interpolate_profile (z_reff, &(reff), 2, z, nlev, INTERP_METHOD_LINEAR, quiet);
    if (status!=0) {
      fprintf (stderr, "Error %d during interpolate_profile of r_eff (line %d, function %s in %s) \n", 
               status, __LINE__, __func__, __FILE__);
      return status;
    }
  }
  else if (strncasecmp(cloud,"ic",2)==0) {
    switch (reff_prop) {
    case IC_REFF_FIXED:

      if (verbose) fprintf(stderr," ... fixed r_eff_ice = %f\n", reff_fixed);

      /* alloc 2 levels for height of cloud droplet effective radius */
      if ((z_reff = calloc (2, sizeof (float *))) == NULL) {
        fprintf (stderr,"Error, allocating memory for 'z_reff' (line %d, function %s in %s)\n", __LINE__, __func__, __FILE__);
        return -17;
      }
      /* alloc 2 levels for cloud droplet effective radius */
      if ((reff = calloc (2, sizeof (float *))) == NULL) {
        fprintf (stderr,"Error, allocating memory for 'reff' (line %d, function %s in %s)\n", __LINE__, __func__, __FILE__);
        return -16;
      }

      reff[0] = reff_fixed;
      reff[1] = reff_fixed;

      /* height scale for effective radius range from surface to top of atmosphere */
      z_reff[0]=z[0];
      z_reff[1]=z[nlev-1];

      /* linear interpolation of the effective radius to the z-scale of the cloud data */
      status = interpolate_profile (z_reff, &(reff), 2, z, nlev, INTERP_METHOD_LINEAR, quiet);
      if (status!=0) {
        fprintf (stderr, "Error %d during interpolate_profile of r_eff (line %d, function %s in %s) \n", 
                 status, __LINE__, __func__, __FILE__);
        return status;
      }
      break;
    case IC_REFF_OU:
      if (verbose) fprintf(stderr," ... parametrisation for r_eff_ice of Ou and Liou, Atm. Res. 1995\n");

      /* alloc nlev levels for cloud droplet effective radius */
      if ((reff = calloc (nlev, sizeof (float *))) == NULL) {
        fprintf (stderr,"Error, allocating memory for 'reff' (line %d, function %s in %s)\n", __LINE__, __func__, __FILE__);
        return -16;
      }

      c0 = 326.3;
      c1 =  12.42;
      c2 =   0.197;
      c3 =   0.0012;

      for (lc=0; lc<nlev; lc++) {
        T_in_C = T[0][lc] - 273.16;
        /* parametrisation valid between -60 to -20 degree Celsius */
        if ( -62.0 <= T_in_C && T_in_C <= -20.0 )
          reff[lc] = 0.5 * (c0 + c1*T_in_C + c2*T_in_C*T_in_C + c3*T_in_C*T_in_C*T_in_C);
        else if ( T_in_C < -62.0 )
          reff[lc] = 0.5 * 27.534;
        else if ( T_in_C > -20.0 )
          reff[lc] = 0.5 * 147.1;

        /* different definition of the effective radius, correction of ~ */
        D = 2*reff[lc]/1000.0;    /* width in micro meter  */
        L = pow(D/0.185,1/0.53);  /* length in micro meter, see Heymmsfield, 1972, JAS 29, 1358 - 1366 */
        ou2fu = D*L / ( D*L + sqrt(3.)/4.*D*D );
        
        reff[lc] *= ou2fu;

        /* reff_max for Fu */
        if (reff[lc]>64.799)
          reff[lc]=64.799;

        /* if (verbose)                                                                               */
        /*   fprintf(stderr,"z[%3d]=%8.4f, T=%9.3f, CWC=%e, r_eff_ice=%9.4f, CC=%e, ou2fu=%9.6f \n",  */
        /*                   lc, z[lc], T_in_C, CWC[0][lc], reff[lc], CC[0][lc], ou2fu);              */
      }
      break;
    default:
      fprintf (stderr, "Error, unknown reff_property %d (line %d, function %s in %s)\n", reff_prop, __LINE__, __func__, __FILE__);
      return -1;
    }
  }
  else {
    fprintf(stderr,"Unknown cloud type '%s' (line %d, function %s in %s)\n", cloud, __LINE__, __func__, __FILE__ );
    return -1; 
  }

  /* write numbers to the output structures */
  (*rows)        = nlev;
  (*max_columns) = 3;
  (*min_columns) = 3;
  if ( cloud_overlap != CLOUD_OVERLAP_OFF )
    cf->nlev     = nlev;
  else 
    cf->nlev     = 0;

  /* allocate array, which will be returned */
  if ((status = ASCII_calloc_float (cld_data,(*rows),3)) != 0) {
    fprintf (stderr," Error allocating cld_data (line %d, function %s in %s)\n",__LINE__, __func__, __FILE__);   
    return status;
  }

  factor = 1000.0 * 1.e+6 * 1.e-3 * mol_mass_air / avogadro;
     /* (kg water)/(kg air) (ECMWF) -> (g water) / (m3 air) (libRadtran) */
     /* 1000. == convert cloud water content  from (kg/kg -> g/kg)       */
     /* 1.e+6 == convert air number density   from (1/cm3 -> 1/m3)       */
     /* mol_mass_air == weight in g / mol particles                      */
     /* avogadro == Avogadro constant == 1 mol Teilchen                  */
     /* 1.e-3 == convert air mass   density   from (   g  ->  kg )       */

  /* write uppermost level as a boundary for a possible cloud in the first layer */
  (*cld_data)[0][0] = z[0];
  (*cld_data)[0][1] = 0.0;
  (*cld_data)[0][2] = 0.0;

  for (lc=1; lc<nlev; lc++) {
    (*cld_data)[lc][0] = z[lc];                                     /* z(level) */
    dens_air = p_layer[0][lc-1] * 1E-04 / (boltzmann * T[0][lc-1]); /* INDEX SHIFT FROM LAYER TO LEVEL PROP */ 
                                                                    /* dens_air is number concentration of air moleculs per cm^3 */
    (*cld_data)[lc][1] = CWC[0][lc-1] * factor * dens_air;          /* INDEX SHIFT FROM LAYER TO LEVEL PROP */  
                                                                    /* cld_data[g/m3] = CWC[kg/kg](layer) * factor * dens_air(layer) */
    /* fprintf (stderr,"z[%3d]=%8.5f CWC=%e, factor=%e, dens_air=%11.5e, cld_data=%e\n", lc, z[lc], CWC[0][lc-1], factor, dens_air, (*cld_data)[lc][1]); */
    (*cld_data)[lc][2] = reff[lc];
  }

  if ( cloud_overlap != CLOUD_OVERLAP_OFF ) {
    if ( ( cf->cf = calloc(cf->nlev, sizeof(double)) ) == NULL) {
      fprintf (stderr,"Error, allocating memory for cloud fraction (line %d, function %s in %s)\n", __LINE__, __func__, __FILE__);
      return -10;
    }
    if ( ( cf->zd = calloc(cf->nlev, sizeof(double)) ) == NULL) {
      fprintf (stderr,"Error, allocating memory for cloud fraction (line %d, function %s in %s)\n", __LINE__, __func__, __FILE__);
      return -10;
    }

    for (lc=0; lc<nlev; lc++) {
      if (CC[0][lc] != 0)
        (*cld_data)[lc][1] /= CC[0][lc];
    }

    /* uppermost layer boundary */
    cf->zd[0] = z[0];
    cf->cf[0] = 0.0;

    for (lc=1; lc<nlev; lc++) {
      cf->zd[lc] = z[lc];
      if (lc != nlev-1)
        cf->cf[lc] = CC[0][lc-1]; /* shifted by one, as we write a layer property to a z-level */
      /* fprintf(stderr," cf->zd=%8.5f, cf->cf=%e \n", cf->zd[lc], cf->cf[lc]); */
    }
  }

  /* verbose output */ 
  if (verbose) {
    first_layer_done=FALSE;

    fprintf (stderr, " *** ECMWF %s cloud data at lat = %5.2f, lon = %5.2f \n", cloud, lat, lon);
    if ( cloud_overlap != CLOUD_OVERLAP_OFF ) {
      fprintf (stderr, "----------------------------------------------------------------\n");
      fprintf (stderr, "  lc |    z    |  p_layer  | cloud water |   r_eff   |   CC    |\n");
      fprintf (stderr, "     |  [km]   |   [hPa]   |    [g/m3]   | [micro m] |         |\n");
      fprintf (stderr, "----------------------------------------------------------------\n");
    }
    else {
      fprintf (stderr, "-------------------------------------------------------\n");
      fprintf (stderr, "  lc |    z    |   p_layer  | cloud water |   r_eff   |\n");
      fprintf (stderr, "     |  [km]   |    [hPa]   |    [g/m3]   | [micro m] |\n");
      fprintf (stderr, "------------------------------------------------------\n");
    }

    for (lc=1; lc<nlev; lc++) {
      if ( lc < nlay-1 ) { /* check also CWC one level below (lc+1) */
        if ( (*cld_data)[lc][1] != 0.0 || (*cld_data)[lc+1][1] != 0.0 ) { /* || first_layer_done == TRUE */
           fprintf (stderr, "%5d %9.5f %10.4f   %e %8.4f ",
                    lc, (*cld_data)[lc][0], p_layer[0][lc-1], (*cld_data)[lc][1], (*cld_data)[lc][2] );
           if ( cloud_overlap != CLOUD_OVERLAP_OFF )
             fprintf (stderr, " %11.7f ", cf->cf[lc] );
           fprintf (stderr, "\n");
           /* fprintf (stderr,"mass dens air [kg/m3] = %e, water content [g/kg] = %e \n",  */
           /*          dens_air[lc] * 1.e+6 * 1.e-3 * air_mol_mass / avogadro, CWC[lc]*1000.0); */
           first_layer_done=TRUE;
        }
      }
      else { /* lowermost layer */
        if ( (*cld_data)[lc][1] != 0.0 ) {  /* || first_layer_done == TRUE */
           fprintf (stderr, "%5d %9.5f %10.4f   %e %8.4f ",
                    lc, (*cld_data)[lc][0], p_layer[0][lc-1], (*cld_data)[lc][1], (*cld_data)[lc][2]);
           if ( cloud_overlap != CLOUD_OVERLAP_OFF )
             fprintf (stderr, " %11.7f ", cf->cf[lc] );
           fprintf (stderr, "\n");
           /* fprintf (stderr,"mass dens air [kg/m3] = %e, water content [g/kg] = %e \n",  */
           /*          dens_air[lc] * 1.e+6 * 1.e-3 * air_mol_mass / avogadro, CWC[lc]*1000.0); */
           first_layer_done=TRUE;
        }
      }
    }
/*     for (lc=0; lc<nlev; lc++) { */
/*       fprintf (stderr, "%5d %9.5f %10.4f   %e %8.4f ", */
/*                lc, (*cld_data)[lc][0], p_layer[0][lc-1], (*cld_data)[lc][1], (*cld_data)[lc][2] ); */
/*       if ( cloud_overlap != CLOUD_OVERLAP_OFF ) */
/*         fprintf (stderr, " %11.7f ", cf->cf[lc] ); */
/*       fprintf (stderr, "\n"); */
/*     } */
  }
  nc_close (ncid);

  for (t=0;t<=nt-1;t++) {
    free(p_level[t]);
    free(p_layer[t]);
    free(T[t]);
    free(CWC[t]);
    free(CC[t]);
  }
  free(p_level);
  free(p_layer);
  free(T);
  free(CWC);
  free(CC);

  /* need to shift by one layer as cloud fractions are automatically considered layer properties */
  for (lc=0;lc<cf->nlev-1;lc++) {
    cf->cf[lc] = cf->cf[lc+1];
  }

#else
  fprintf (stderr, " ******************************************************************\n");
  fprintf (stderr, " * You have built uvspec without libnetcdf and hence cannot       *\n");
  fprintf (stderr, " * use the ECMWF data file option. Please get netcdf and rebuild. *\n");
  fprintf (stderr, " ******************************************************************\n");
  return -1;
#endif

#else
  fprintf (stderr, " Error, ECMWF cloud input is not availeble in this verion \n");
#endif

  return status;

}

/********************************************************************************************************/

int read_ECHAM_clouds (char *filename, float lat, float lon, struct tm UTC, int time_interpolate, 
                       char *cloud, float *zd, 
                       int verbose, int quiet,
                       int *rows, int *max_columns, int *min_columns,
                       float ***cld_data, cf_out_struct *cf)
{
  int status=0;

#if HAVE_LIBNETCDF

  int ncid=NOT_DEFINED_INTEGER;

  int lc = NOT_DEFINED_INTEGER; /* loop index */
  int t  = NOT_DEFINED_INTEGER; /* loop index */
  int nt = NOT_DEFINED_INTEGER;
 
  size_t nlat  = NOT_DEFINED_INTEGER;
  size_t nlon  = NOT_DEFINED_INTEGER;
  size_t nlev  = 39; /* constant for ECHAM data files; must be changed for 
                        ECHAM5 */
  
  size_t ilat  = NOT_DEFINED_INTEGER;  
  size_t ilon  = NOT_DEFINED_INTEGER;  
  size_t itime = NOT_DEFINED_INTEGER;  

  int itime1 = NOT_DEFINED_INTEGER, itime2 = NOT_DEFINED_INTEGER;

  double *ECHAM_lat  = NULL;
  double *ECHAM_lon  = NULL;
  
  float dt = NOT_DEFINED_FLOAT;
  
  float **CWP     = NULL; /* cloud water content, liquid or ice */
  float *CWC      = NULL;
  float **reff    = NULL;
  float **splitfactor  = NULL;
  float **totcc3D = NULL;
  
  /* open netcdf file */
  status = nc_open (filename, NC_NOWRITE, &ncid);
  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d opening netCDF file %s\n", status, filename);
    return status;
  }

  /* read latitude array */
  alloc_and_read_netCDF_1D_double(ncid,"lat", &nlat, "lat", &(ECHAM_lat));

  /* search correct latitude index */
  ilat = get_grid_index( lat, ECHAM_lat, nlat, FALSE);
  if (ilat < 0) {
    fprintf (stderr, "Error -1 finding index for lat=%5.2f in %s\n", lat, filename);
    return -1;
  }
  
  /* read longitude */
  alloc_and_read_netCDF_1D_double(ncid,"lon", &nlon, "lon", &(ECHAM_lon));
  
  /* search correct latitude index */
  ilon = get_grid_index( lon, ECHAM_lon, nlon, TRUE);
  if (ilon < 0) {
    fprintf (stderr, "Error -2 finding index for lon=%5.2f in %s\n", lon, filename);
    return -2;
  }

  /* get time index */
  status = get_time_index (ncid, UTC, time_interpolate,
                           &(nt), &(itime1), &(itime2), &(dt),
                           verbose, quiet);
  if (status != 0) {
    fprintf (stderr, "Error %d, during get_time_index in get_float_from_netCDF_map (ancillary.c)\n", status);
    return status;
  }

  /* alloc nt timesteps for cloud cover profile */
  if ((totcc3D = calloc (nt, sizeof (float *))) == NULL) {
    fprintf (stderr,"Error, allocating memory for pressure in read_ECHAM_clouds() (in cloud.c)\n");
    return -10;
  }
  
  /* alloc nt timesteps for cloud water content */
  if ((CWP = calloc (nt, sizeof (float *))) == NULL) {
    fprintf (stderr,"Error, allocating memory for cloud water path in read_ECHAM_clouds() (in cloud.c)\n");
    return -13;
  }
  
  if ((splitfactor = calloc (nt, sizeof (float *))) == NULL) {
    fprintf (stderr,"Error, allocating memory for splitfactor in read_ECHAM_clouds() (in cloud.c)\n");
    return -11;
  }
  
  /* alloc nt timesteps for temperature */
  if ((reff = calloc (nt, sizeof (float *))) == NULL) {
    fprintf (stderr,"Error, allocating memory for effective radius in read_ECHAM_clouds() (in cloud.c)\n");
    return -13;
  }
  
  /* read nt (= 1 or 2) time steps */
  for (t=0;t<=nt-1;t++) {
    
    if (t == 0)
      itime = itime1;
    if (t == 1)
      itime = itime2;
    
    status = alloc_and_read_netCDF_column_float(ncid, "totcc3D", &totcc3D[t], nlev, itime, ilat, ilon, verbose);
    if (status != 0) {
      fprintf (stderr, "Error %d reading cloud cover from netCDF file %s\n", status, filename);
      return status;
    }
    
  status = alloc_and_read_netCDF_column_float(ncid, "iwp", &CWP[t], nlev, itime, ilat, ilon, verbose);
  if (status != 0) {
      fprintf (stderr, "Error %d reading CWC from netCDF file %s\n", status, filename);
      return status;
    }
    
    status = alloc_and_read_netCDF_column_float(ncid, "splitfactor", &splitfactor[t], nlev, itime, ilat, ilon, verbose);
    if (status != 0) {
      fprintf (stderr, "Error %d reading splitfactor from netCDF file %s\n", status, filename);
      return status;
    }

    if (strncasecmp(cloud,"wc",2)==0) {
      status = alloc_and_read_netCDF_column_float(ncid, "reffwat", &(reff[t]), nlev, itime, ilat, ilon, verbose);
      if (status != 0) {
        fprintf (stderr, "Error %d reading effective radius from netCDF file %s\n", status, filename);
        return status;
      } 
    }
    else if (strncasecmp(cloud,"ic",2)==0) {
      status = alloc_and_read_netCDF_column_float(ncid, "reffice", &(reff[t]), nlev, itime, ilat, ilon, verbose);
      if (status != 0) {
        fprintf (stderr, "Error %d reading effective radius from netCDF file %s\n", status, filename);
        return status;
      }
    }
    else {
      fprintf(stderr,"Unknown 'cloud option %s' in read_ECHAM_clouds \n", cloud);
      return -1; 
    }
  }
  
  /* time interpolation only if nt==2, for nt==1 do nothing */
  if (nt == 2) {
    /* write time interpolated data into the zero'th entry */
    for (lc=0; lc<nlev; lc++) {
      totcc3D[0][lc] = (1.0-dt)*totcc3D[0][lc] + dt*totcc3D[1][lc];
      CWP[0][lc] = (1.0-dt)*CWP[0][lc] + dt* CWP[1][lc];
      splitfactor[0][lc] = (1.0-dt) * splitfactor[0][lc] + 
        dt * splitfactor[1][lc];
      reff[0][lc] = (1.0-dt)*reff[0][lc] + dt*reff[1][lc];
    }
  }
 
  (*rows)        = nlev;
  (*max_columns) = 3;
  (*min_columns) = 3;
  
  if ((status = ASCII_calloc_float (cld_data,(*rows),3)) != 0) {
    fprintf (stderr," Error allocating cld_data in read_ECHAM_cloud (cloud.c) \n");   
    return status;
  }

  cf->cf = calloc(nlev, sizeof(double));
  cf->zd = calloc(nlev, sizeof(double));
  cf->zd[0]= zd[0];
  (*cld_data)[0][0] = zd[0];

  for (lc=1; lc<nlev; lc++){
    /* Cloud fraction, here already converted to layer property */
    cf->cf[lc] = totcc3D[0][lc+1];
   
    cf->zd[lc] = zd[lc];
    /* Altitude levels */
    (*cld_data)[lc][0] = zd[lc];
    /* Effective radius*/ 
    (*cld_data)[lc][2] = reff[0][lc];
    if (strncasecmp(cloud,"wc",2)==0) /*LWC*/ 
      (*cld_data)[lc][1] = (CWP[0][lc] * (1.0-splitfactor[0][lc]))/
        (zd[lc-1] - zd[lc]) / 1000.0;
    else  /* IWC*/
        (*cld_data)[lc][1] = (CWP[0][lc]*splitfactor[0][lc])/
          (zd[lc-1] - zd[lc])/1000.0;
    
  }
  cf->nlev=nlev;
 

  /* verbose output */ 
  if (verbose) {

    fprintf (stderr, " *** ECHAM %s cloud data at lat = %5.2f, lon = %5.2f \n", cloud, lat, lon);
    fprintf (stderr, "-----------------------------------------------\n");
    fprintf (stderr, "  lc |   z   | cloud water |   r_eff   |   cf  \n");
    fprintf (stderr, "     |  [km] |    [g/m3]   | [micro m] |       \n");
    fprintf (stderr, "-----------------------------------------------\n");

    for (lc=0; lc<nlev; lc++) {
      fprintf (stderr, "%5d %7.3f  %e  %7.2f  %7.2f \n",
               lc, (*cld_data)[lc][0], (*cld_data)[lc][1],
               (*cld_data)[lc][2], cf->cf[lc-1]);
    }
  }
  
  for (t=0;t<=nt-1;t++) {
    free(totcc3D[t]);
    free(reff[t]);
    free(splitfactor[t]);
    free(CWP[t]);
  }
  
  free(CWC);

#else
  fprintf (stderr, " ******************************************************************\n");
  fprintf (stderr, " * You have built uvspec without libnetcdf and hence cannot       *\n");
  fprintf (stderr, " * use the ECHAM data file option. Please get netcdf and rebuild. *\n");
  fprintf (stderr, " ******************************************************************\n");
  return -1;
#endif

  return status;

}
