/*--------------------------------------------------------------------
 * $Id: cloud3d.c,v 1.79 2009/01/27 08:43:49 robert Exp $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 * 
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

#include <stdio.h>							    
#include <stdlib.h>
#include <float.h>
#include <math.h>

#include "uvspec.h"
#include "ckdfu.h"
#include "cloud3d.h"
#include "ipa.h"
#include "wcloud3d.h"
#include "elevation2d.h"
#include "temperature2d.h"
#include "mystic.h"
#include "f77-uscore.h"
#include "ascii.h"

#ifndef PI
#define PI 3.14159265358979323846264338327
#endif


static inline int float_equal (float a, float b);


/* prototypes of internal functions */
/* static int compare_levels_df (double *z1, int n1, float  *z2, int n2); */

static int setup_sample_grid (float *zd, int nlev,
			      int nxout, int nyout, 
			      double dxout, double dyout, 
			      double xsize_cloud, double ysize_cloud, 
			      float *zout, int nzout,
			      int radpat_Nr, int radpat_Nt, float radpat_dt, 
			      float *umu, int numu, float *phi, int nphi,
			      int aziconv,
			      sample_struct *sample,
			      int escape, double escape_eps_ddis,
			      int calcstd,
			      int backward, int output,
			      int backward_islower, int backward_jslower,
			      int backward_isupper, int backward_jsupper,
			      int backward_writeallpixels,
			      int backward_writeback,
			      int spherical, 
                              int polarisation, 
                              int nstokes,
                              int bcond,
			      double relerr,
			      double le_abserr,
			      double maxscatters,
			      int locest,
			      int cloner,
			      int jacobian,
			      char *lidarfilename,
			      int quiet);

static int read_wc3D_file (char *filename, 
			   cld3d_out_struct *cld3d, 
			   input_struct input, output_struct *output);

static int read_ic3D_file (char *filename, 
			   cld3d_out_struct *cld3d, 
			   input_struct input, output_struct *output);

static float interpolate_f (phase_function_table *p, int i1, int i2, double delta_r);
static float interpolate_dscale (phase_function_table *p, int i1, int i2, double delta_r);



/***********************************************************************************/
/* Function: equal                                                        @62_30i@ */
/* Description:                                                                    */
/*  Two numbers equal?                                                             */
/*                                                                                 */
/* Parameters:                                                                     */
/* Return value:                                                                   */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                        @i62_30@ */
/***********************************************************************************/

static int equal (double x, double y)
{
  if ((float) x != (float) y)
    return 0;

  return 1;
}


/***********************************************************************************/
/* Function: setup_sample_grid                                            @62_30i@ */
/* Description:                                                                    */
/*  Initialize sample_struct.                                                      */ 
/*                                                                                 */
/* Parameters:                                                                     */
/*  float *zout:  array of altitudes to be sampled                                 */
/*  int   nzout:  number of altitudes to be sampled                                */
/*                                                                                 */
/* Return value:                                                                   */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                        @i62_30@ */
/***********************************************************************************/


static int setup_sample_grid (float *zd, int nlev,
			      int nxout, int nyout, 
			      double dxout, double dyout, 
			      double xsize_cloud, double ysize_cloud, 
			      float *zout, int nzout,
			      int radpat_Nr, int radpat_Nt, float radpat_dt, 
			      float *umu, int numu, float *phi, int nphi,
			      int aziconv,
			      sample_struct *sample,
			      int escape, double escape_eps_ddis,
			      int calcstd,
			      int backward, int output,
			      int backward_islower, int backward_jslower,
			      int backward_isupper, int backward_jsupper,
			      int backward_writeallpixels,
			      int backward_writeback,
                              int spherical, 
                              int polarisation, /* CE */
                              int nstokes, /* CE */
                              int bcond,
			      double relerr,
			      double le_abserr,
			      double maxscatters,
			      int locest,
			      int cloner,
			      int jacobian,
			      char *lidarfilename,
			      int quiet)
{
  int status=0;
  int lu=0, iu=0, j=0, id=0, kc=0, ia=0;
  double *z_hist=NULL;
  int Nd=0;

  double xr=0, yr=0, alphaact=0.0;

  double *stheta=NULL, *sphi=NULL, *salpha=NULL;

  FILE *ftmp=NULL;

#if HAVE_MYSTIC
  int ili=0, i=0;
#endif

  if (!quiet)
    fprintf (stderr, " ... reading sampling information\n");

  /* test if a sample2D.dat is in the current directory */
  if ((ftmp=fopen("./sample2D.dat","r"))!=NULL) {
    
    fprintf (stderr, "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");
    fprintf (stderr, "!!! ATTENTION: Found an old sample file sample2D.dat. Before October 25, 2008 !!!\n");
    fprintf (stderr, "!!! MYSTIC got information about sample grid, radiance directions, etc. from  !!!\n");
    fprintf (stderr, "!!! this file. Since October 25, 2008 it is ignored because all the           !!!\n");
    fprintf (stderr, "!!! information can now be directly specified in the uvspec input file.       !!!\n");
    fprintf (stderr, "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");
    
    fclose (ftmp);
  }
  
  /* if no user-defined sample grid */
  if (nxout<=0 || nyout<=0 || dxout<=0 || dyout<=0) {
    
    nxout=1;
    nyout=1;

    if (xsize_cloud>0 && ysize_cloud>0) {
      dxout=xsize_cloud / 1000.0;
      dyout=ysize_cloud / 1000.0;

      fprintf (stderr, " ... no user-defined sample grid - copying domain size from cloud grid\n");
    }
    else {
      if(!spherical){
        /* use a domain size DBL_MAX / 10.0 - the factor 1 / 10 allows some photon movement */
        /* as the photon is allowed to leave the domain before it is forced back by the     */
        /* periodic boundary conditions and we don't want an overflow in this case.         */
        
        fprintf (stderr, " ... no user-defined sample grid - defining an \"infinitely\" large domain\n");
        fprintf (stderr, " ...   of length DBL_MAX / 10.0 = %gm with one single sample pixel!\n", DBL_MAX / 10.0);
        
        dxout = DBL_MAX / 10.0 / 1000.0;  /* convert to km */
        dyout = DBL_MAX / 10.0 / 1000.0;  /* convert to km */
      }
      /*CE Numerical problems may occur if sample domain is too large in spherical
        geometry ... why??? */
      else{
        fprintf (stderr, " ... no user-defined sample grid - defining an \"infinitely\" large domain\n");
        fprintf (stderr, " ...   of length 1e8 with one single sample pixel!\n");
        dxout=100000; 
        dyout=100000; 
      }
      
    }
  }

  /* allocate memory for sample flag */
  sample->sample = calloc ((size_t) nlev, sizeof(int));
  
  sample->delX = dxout * 1000.0;  /* convert to m */
  sample->delY = dyout * 1000.0;  /* convert to m */

  sample->Nx = nxout;
  sample->Ny = nyout;
  
  if (!quiet)
    fprintf (stderr, " ... sample grid %d x %d pixels with area %gm x %gm\n",
	     sample->Nx, sample->Ny, sample->delX, sample->delY);
  
  /* ??? check if we really want that always ??? */
  sample->passback3D = 1;

  /* process user altitudes */
  if (nzout>0) {

    /* first check if zout is sorted in ascending order */
    for (lu=1; lu<nzout; lu++)
      if (zout[lu] - zout[lu-1] < 0) {
	fprintf (stderr, "Error, zout not sorted in ascending order\n");
	status--;
      }
    
    /* ??? special treatment of -999, surface ??? */

    if (!quiet)
      fprintf (stderr, " ... using %d altitudes from the uvspec input file:\n", nzout);
    
    for (lu=0; lu<nzout; lu++) {
      if (zout[lu]>=0) {
	for (kc=0; kc<nlev; kc++) {
	  if (float_equal (zd[nlev-1-kc], zout[lu])) {
	    sample->sample[kc]=1;
	    break;
	  }
	}
	if (kc==nlev) {
	  fprintf (stderr, "\nFATAL error: altitude grid does not contain level %g\n", zout[lu]);
	  fprintf (stderr, "which has been specified as output altitude\n");
	  return -1;
	}
      }
    }
  }

  if (!quiet) {
    fprintf (stderr, " ... sampling result at the following altitudes:\n");

    for (kc=0; kc<nlev; kc++)
      if (sample->sample[kc])
	fprintf (stderr, " ...  %8.3f km\n", zd[nlev-1-kc]);

    fprintf (stderr, "\n");
  }

  if (zout[0]<0)
    sample->surface = 1;
    
  /* process radiance directions */
  if (numu>0 && nphi>0) {   /* use data passed as function arguments */

    if (!quiet)
      fprintf (stderr, " ... using %d radiance angles from the uvspec input file:\n", numu * nphi);

    /* allocate memory for sample radiances */
    Nd = numu * nphi;
    sample->Nd = Nd;
    if (Nd>0)
      sample->rad  = calloc (Nd, sizeof (radang));

    for (iu=0; iu<numu; iu++) {
      for (j=0; j<nphi; j++) {

	/* copy uvspec radiance to sample array */
	sample->rad[id].theta = acos(umu[iu])/PI*180.0;
	sample->rad[id].phi   = phi[j];
	sample->rad[id].alpha = 5.0;   /* default; should change that by adding */
	                               /* another libRadtran input option       */

	if (!quiet)
	  fprintf (stderr, " ...  (%g, %g)\n", sample->rad[id].theta, sample->rad[id].phi);
	
	id++;
      }
    }
  }

  /****************************************************/
  /* Changed MYSTIC azimuth convention March 1, 2004! */
  /* From then on, the MYSTIC azimuth convention      */
  /* follows the DISORT convention, that is, as seen  */
  /* from the photon rather than from the detector.   */
  /* azimuth 0 = coming from the direction of the sun */
  /* (looking towards the sun) while azimuth 180 is   */
  /* going into the direction of the sun (or for the  */
  /* observer, sun in the back. If aziconv is set to  */
  /* MCAZIMUTH_OLD then the old convention is used.   */
  /*                                                  */
  /* Inside MYSTIC nothing is changed; azimuths       */
  /* still refer to the observing instrument.         */
  /*                                                  */
  /* To achieve this, we need to reverse the azimuths */
  /* at this location and store the original values   */
  /* for the output.                                  */
  /****************************************************/

  for (id=0; id<sample->Nd; id++) {
    sample->rad[id].externalphi = sample->rad[id].phi;

    switch (aziconv) {
    case MCAZIMUTH_NEW:
      sample->rad[id].phi = 180.0 + sample->rad[id].phi;
      break;
    case MCAZIMUTH_OLD:
      sample->rad[id].phi = sample->rad[id].phi;
      break;
    default: 
      fprintf (stderr, "Error, unknown azimuth convention %d\n", aziconv);
      return -1;
    }

    /* shift to interval [0 ... 360] */
    while (sample->rad[id].phi < 0.0)
      sample->rad[id].phi += 360.0;

    while (sample->rad[id].phi > 360.0)
      sample->rad[id].phi -= 360.0;
  }

#if HAVE_LIDAR
  sample->cloner  = cloner;
  if (cloner) {
    sample->clone_max = 3.0;
    sample->clone_min = 0.3; //0.3;
  }
  else {
    sample->clone_max = 1e14;
    sample->clone_min = 0.;
  }
#endif

  if (stheta != NULL) free (stheta); 
  if (sphi != NULL)   free(sphi); 
  if (salpha != NULL) free(salpha);

  if (sample->Nd>0) {
    sample->escape = escape;
    sample->escape_eps_ddis_upf = escape_eps_ddis;

    if (!quiet)
      fprintf (stderr, " ... calculating radiances for %d directions\n", sample->Nd);
  }

#if HAVE_LIDAR
  if (jacobian) {
    sample->jacobian = jacobian;
    sample->jac_eps  = 1e-1;
  }
#endif

  if (calcstd)
    sample->std = 1;
  else 
    sample->std = 0;
  

  /* radial pathlength sampling */
  sample->Nr = radpat_Nr;
  sample->Nt = radpat_Nt;
  sample->dt = radpat_dt;

  /* fit circle into domain */
  xr = 0.5 * (double) sample->Nx * sample->delX;
  yr = 0.5 * (double) sample->Ny * sample->delY;

  sample->dr = (xr < yr ? xr : yr) / (double) sample->Nr;

  if (!quiet && sample->Nr > 0)
    fprintf (stderr, " ... radial pathlength sampling with intervals dr = %.6e and dt = %.6e\n",
	     sample->dr, sample->dt);

  free (z_hist);

  /* number of levels to be sampled */
  for (kc=0; kc<nlev; kc++)
    sample->nzout += sample->sample[kc];

  sample->backward_writeallpixels = backward_writeallpixels;
  sample->backward_writeback = backward_writeback;

  /* backward sampling area - may not be needed but is initialized anyway */
  sample->backward_islower = backward_islower;
  sample->backward_jslower = backward_jslower;
  sample->backward_isupper = backward_isupper;
  sample->backward_jsupper = backward_jsupper;
  
  if (sample->backward_islower==-999 &&
      sample->backward_jslower==-999 &&
      sample->backward_isupper==-999 &&
      sample->backward_jsupper==-999) {
    /* special case: no user-defined backward pixels - calculating all sample pixels */
  }
  else {
    if (sample->backward_islower < 0) sample->backward_islower=0;
    if (sample->backward_jslower < 0) sample->backward_jslower=0;
    if (sample->backward_isupper < 0) sample->backward_isupper=0;
    if (sample->backward_jsupper < 0) sample->backward_jsupper=0;
    
    if (sample->backward_islower >= sample->Nx) sample->backward_islower = sample->Nx-1;
    if (sample->backward_jslower >= sample->Ny) sample->backward_jslower = sample->Ny-1;
      if (sample->backward_isupper >= sample->Nx) sample->backward_isupper = sample->Nx-1;
      if (sample->backward_jsupper >= sample->Ny) sample->backward_jsupper = sample->Ny-1;
  }
  
  sample->backward_islower_org = sample->backward_islower;
  sample->backward_jslower_org = sample->backward_jslower;
  sample->backward_isupper_org = sample->backward_isupper;
  sample->backward_jsupper_org = sample->backward_jsupper;

  /* calculate zenith and azimuth angles */ 
  if (backward) {
    if (Nd>0)
      sample->backward = MCBACKWARD_RADIANCE;
    else 
      sample->backward = output;  /* irradiance or actinic flux - only one at*/
  }

  if (sample->backward && !sample->escape) {
    sample->escape = 1;
    sample->escape_eps_ddis_upf = escape_eps_ddis;
    if (!quiet)
      fprintf (stderr, " ... switching on local estimate (default for backward calculations)\n");
  }

  if (sample->backward && (sample->Nr>0 && sample->dt>0)) {
    fprintf (stderr, "Error, radial/pathlength and backward cannot be combined!\n");
    return -1;
  }

  /* boundary conditions */
  sample->bcond = bcond;

  /* spherical geometry - always use backwards and absorbing boundary conditions! */
  sample->spherical = spherical;
  if (spherical) {
    if (Nd>0)
      sample->backward = MCBACKWARD_RADIANCE;
    else 
      sample->backward = output;  /* currently only downward diffuse is implemented */
    
    sample->escape = 1;
    sample->escape_eps_ddis_upf = escape_eps_ddis;
    sample->backward_writeallpixels = 0;

    if (sample->backward_islower!=0 || 
	sample->backward_jslower!=0 ||
	sample->backward_isupper!=0 || 
	sample->backward_jsupper!=0) {
      fprintf (stderr, " ... backward sampling area ignored in spherical geometry!\n");
      fprintf (stderr, " ... (photons are always started at the center of the domain)\n");
    }
    
    sample->backward_islower = 0;
    sample->backward_jslower = 0;
    sample->backward_isupper = 0;
    sample->backward_jsupper = 0;

    if (sample->bcond!=MCBCOND_ABSORB)
      fprintf (stderr, " ... using absorbing boundary conditions for spherical\n");
    
    sample->bcond = MCBCOND_ABSORB;
  }

  sample->polarisation=polarisation; 
  sample->nstokes=nstokes;

  /* if non-zero, MC can be stopped as soon as results have reached the following precision */
  sample->stop_at_relerr = relerr;
  /* if non-zero, photon is stopped as soon as it reaches maxscatters scatterings */
  sample->maxscatters = maxscatters;

#if HAVE_LIDAR
  sample->LE_eps_ddis_upf    = 0.0;
  sample->LE_eps_ddis_uda    = 0.0;

  /* Read laser and detector data for local estimator */
  if (locest) {

    sample->LocEst = locest;

    sample->le_stop_at_abserr = le_abserr;
    
    switch(sample->LocEst) {
    case MCLOCEST_SPACE_LIDAR:
      sample->LE_D_DIS       = 1;
      sample->LE_VIS_QIDD    = 0;
      sample->LE_VIS_FOD     = 1;
      sample->LE_RIS_MAS     = 1;
      sample->LE_isoene      = 1;
      sample->LE_sponti      = 1;
      sample->LE_turnmax     = 1;
      break;
    case MCLOCEST_ATMOS_LIDAR:
      sample->LE_D_DIS       = 1;
      sample->LE_VIS_QIDD    = 0;
      sample->LE_VIS_FOD     = 1;
      sample->LE_RIS_MAS     = 1;
      sample->LE_isoene      = 1;
      sample->LE_sponti      = 1;
      sample->LE_turnmax     = 1;
      break;
    default:
      fprintf(stderr,"Error! Unknown locest type %d\n",sample->LocEst);
      return -1;
    }

    if (sample->LE_D_DIS) {
      sample->LE_eps_ddis_upf    = 0.25; /* DDIS-UPF probability within cone  */
      sample->LE_eps_ddis_uda    = 0.25; /* DDIS-UDA probability within cone  */
      sample->LE_eps_fod_dis_phi = 0.0 ; /* DDIS-UDA probability within cone  */

      if (sample->LE_eps_ddis_uda + sample->LE_eps_ddis_upf > 1.0 ) {
	fprintf (stderr, "Error! Probability of DDIS is larger than t one!\n");
	return -1;
      }
    }


#if HAVE_LIDAR

    /* check if lidar file exists and can be opened for reading */
    ftmp = fopen (lidarfilename, "r");
    if (ftmp==NULL) {
      fprintf (stderr, "Error! Either define the lidar in the input file or provide lidar.dat\n");
      return -1;
    }
    else {
      fclose(ftmp);
  
      /* read lidar.dat */
      status = read_lidar (lidarfilename, sample, quiet);
    
      if (status!=0) {
	fprintf (stderr, "Error %d reading %s\n", status, lidarfilename);
	return status;
      }
    }
#endif


#if HAVE_MYSTIC

    /* define some variables */
    sample->maxpathlength = sample->lidar_t[sample->LocEst_Nt-1];

    for (ili = 0; ili < sample->Nli; ili++) {
      sample->laser[ili].cosalpha=cos (sample->laser[ili].alpha);
      sample->laser[ili].omega=2.0 * PI * (1.0 - sample->laser[ili].cosalpha);
      sample->laser[ili].dir.dx[0]=0.;
      sample->laser[ili].dir.dx[1]=0.;
      sample->laser[ili].dir.dx[2]=-1.;
      sample->laser[ili].area=PI * sample->laser[ili].radius * sample->laser[ili].radius;

      new_direction (cos(sample->laser[ili].theta), sample->laser[ili].phi, &(sample->laser[ili].dir));

      sample->lidar[ili].cosalpha = calloc (sample->LocEst_Na, sizeof(double));
      sample->lidar[ili].omega    = calloc (sample->LocEst_Na, sizeof(double));
      alphaact = sample->lidar[ili].alpha;
      for (ia=0; ia<sample->LocEst_Na; ia++) {
	sample->lidar[ili].cosalpha[ia] = cos (alphaact);
	sample->lidar[ili].omega[ia]    = 2.0 * PI * (1.0 - sample->lidar[ili].cosalpha[ia]);
	alphaact *= sample->LocEst_afac;
      }

      sample->lidar[ili].dir.dx[0] =  0.;
      sample->lidar[ili].dir.dx[1] =  0.;
      sample->lidar[ili].dir.dx[2] = -1.;
      sample->lidar[ili].area      = PI * sample->lidar[ili].radius * sample->lidar[ili].radius;

      new_direction (cos(sample->lidar[ili].theta), sample->lidar[ili].phi, &(sample->lidar[ili].dir));

      sample->lidar[ili].s_det = sin (sample->lidar[ili].alpha);
      sample->lidar[ili].c_det = cos (sample->lidar[ili].alpha);
      sample->lidar[ili].t_det = tan (sample->lidar[ili].alpha);
      sample->lidar[ili].z_det = - sample->lidar[ili].radius / sample->lidar[ili].t_det;
      for (i=0;i<3;i++)
	sample->lidar[ili].x_cc[i] = sample->lidar[ili].x[i] + sample->lidar[ili].dir.dx[i] * sample->lidar[ili].z_det;
    }
#endif

  }
  else {
    sample->LocEst = 0;
    sample->LE_D_DIS   = 0;
    sample->LE_RIS_MAS     = 0;
    sample->LE_VIS_QIDD    = 0;
    sample->LE_VIS_FOD     = 0;
    sample->LE_sponti      = 0;
    sample->LE_turnmax     = 0;
    /* QUICK FIX; if this is not set, the mystic loop is not started for other than locest */
    sample->Nli = 1;
    sample->LocEst_Nt = 0;
    sample->le_stop_at_abserr = 0;
  }
#endif

  return 0;
}


/***********************************************************************************/
/* Function:                                                              @62_30i@ */
/* Description:                                                                    */
/*  Setup a 3D atmosphere structure, to be used as input to MYSTIC.                */
/*                                                                                 */
/* Parameters:                                                                     */
/* Return value:                                                                   */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                        @i62_30@ */
/***********************************************************************************/

int setup_cloud3D (input_struct input, output_struct *output)
{
  int kc=0, status=0;
  char cloud3d_filename[FILENAME_MAX];
  FILE *f=NULL;
  int readwc=0;


  char function_name[]="setup_cloud3D";
  char file_name[]="cloud3d.c";

  float *zd_temp=NULL;

  /* reset counters */
  output->wc.threed.nthreed = 0;
  output->ic.threed.nthreed = 0;

  /* don't need to read cloud properties if no MYSTIC calculation */
  if (input.rte.solver != MONTECARLO)
    return 0;
  
  /*************************************/
  /* Read 3D water and ice cloud files */
  /*************************************/

  /* set default name of the wcloud3D-file and check if it exists */
  readwc=0;
  if (*input.rte.mc.filename[FN_MC_WCLOUD]) {
    strcpy (cloud3d_filename,  input.rte.mc.filename[FN_MC_WCLOUD]);
    readwc=1;
  }
  else {
    readwc=0;
    fprintf (stderr, " ... did not find 3D water cloud file - trying without!\n");
    
    /* test if a wcloud3D.dat is in the current directory */
    if ((f=fopen("./wcloud3D.dat","r"))!=NULL) {

      fprintf (stderr, "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");
      fprintf (stderr, "!!! ATTENTION: Found a 3D cloud file wcloud3D.dat. Before October 25, 2008 !!!\n");
      fprintf (stderr, "!!! MYSTIC used this file as 3D cloud file even if the user did not        !!!\n");
      fprintf (stderr, "!!! explicitely define it as \"mc_wcloud_file\" - since October 25, 2008     !!!\n");
      fprintf (stderr, "!!! it is ignored. If you want it to be used as 3D water cloud file,       !!!\n");
      fprintf (stderr, "!!! please specify \"mc_wcloud_file wcloud3D.dat\" in the input file!        !!!\n");
      fprintf (stderr, "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");
	
      fclose (f);
    }
  }

  /* copy cldproperties/nonHGcld information to 3D structure */
  /* even if no 3D cloud but only a 1D cloud is defined      */ 
  output->wc.threed.nonHGcld      = output->wc.nonHGcld;
  output->wc.threed.cldproperties = output->wc.cldproperties;

  if (readwc) {
    if (!input.quiet)
      fprintf (stderr, " ... reading 3D cloud data from %s\n", cloud3d_filename);
    

    /* read 3D cloud data */
    status = read_wc3D_file (cloud3d_filename, 
			     &(output->wc.threed), 
			     input, output);
    if (status!=0) {
      fprintf (stderr, "Error %d reading 3D cloud water file %s in %s (%s)\n", 
	       status, cloud3d_filename, function_name, file_name);
      return status;
    }
    
    /* add 3D water cloud levels to atmospheric z-grid */

    /* sort by descending altitude */
    zd_temp = calloc(output->wc.threed.nlyr+1, sizeof(float));
    for (kc=0; kc<=output->wc.threed.nlyr; kc++)
      zd_temp[kc] = output->wc.threed.zd[output->wc.threed.nlyr-kc];

    set_common_z_grid (zd_temp, output->wc.threed.nlyr+1, 
		       &output->atm.zd_common, &output->atm.nlev_common);

    if (!input.quiet)
      fprintf (stderr, " ... adding %d 3D water cloud levels to the common grid\n", output->wc.threed.nlyr+1);
  }
  else {  /* no 3D water cloud data */
    output->wc.threed.nthreed=0;
    output->wc.threed.nlyr = 0;
  }

  /************************/
  /* ice cloud properties */
  /************************/

  /* copy cldproperties/nonHGcld information to 3D structure */
  /* even if no 3D cloud but only a 1D cloud is defined      */ 
  output->ic.threed.nonHGcld      = output->ic.nonHGcld;      /*TZ*/
  output->ic.threed.cldproperties = output->ic.cldproperties; /*TZ*/

  /* don't need to read cloud properties if not a 3D calculation */
  /* 3D cloud file defined or no - need to re-think once 3D and  */
  /* IPA are homogenized ???                                     */
  if (strlen(input.rte.mc.filename[FN_MC_ICLOUD])>0) {
    strcpy (cloud3d_filename,  input.rte.mc.filename[FN_MC_ICLOUD]);
    
    if (!input.quiet)
      fprintf (stderr, " ... reading 3D ice cloud data from %s\n", cloud3d_filename);
    
    /* read 3D cloud data */
    status = read_ic3D_file (cloud3d_filename, 
			     &(output->ic.threed), 
			     input, output);
    
    if (status!=0) {
      fprintf (stderr, "Error %d reading 3D ice cloud file %s in %s (%s)\n", 
	       status, cloud3d_filename, function_name, file_name);
      return status;
    }
    
    /* add 3D ice cloud levels to atmospheric z-grid */

    /* sort by descending altitude */
    zd_temp = calloc(output->ic.threed.nlyr+1, sizeof(float));
    for (kc=0; kc<=output->ic.threed.nlyr; kc++)
      zd_temp[kc] = output->ic.threed.zd[output->ic.threed.nlyr-kc];

    set_common_z_grid (zd_temp, output->ic.threed.nlyr+1, 
		       &output->atm.zd_common, &output->atm.nlev_common);

    fprintf (stderr, " ... adding %d 3D ice cloud levels to the common grid\n", output->ic.threed.nlyr+1);
  }
  else { /* no 3D ice cloud data */
    output->ic.threed.nthreed=0;
    output->ic.threed.nlyr=0;
  }    

  /* we require that water and ice clouds are defined */
  /* on identical horizontal grids                    */
  if (output->wc.threed.nthreed>0 && output->ic.threed.nthreed>0) {
    if (output->wc.threed.delX != output->ic.threed.delX || 
	output->wc.threed.delY != output->ic.threed.delY || 
	output->wc.threed.Nx != output->ic.threed.Nx     || 
	output->wc.threed.Ny != output->ic.threed.Ny) {
      fprintf (stderr, "Error: 3D water and 3D ice clouds need to be defined on identical horizontal grids!\n");
      return -1;
    }
    
    fprintf (stderr, " ... horizontal water and ice cloud grids agree\n");
  }

  /* if only water clouds defined, copy water cloud grid to ice cloud grid */
  if (output->wc.threed.nthreed>0 && output->ic.threed.nthreed==0) {
    output->ic.threed.Nx   = output->wc.threed.Nx;
    output->ic.threed.Ny   = output->wc.threed.Ny;
    output->ic.threed.delX = output->wc.threed.delX;
    output->ic.threed.delY = output->wc.threed.delY;
  }

  /* if only ice clouds defined, copy ice cloud grid to water cloud grid */
  if (output->wc.threed.nthreed==0 && output->ic.threed.nthreed>0) {
    output->wc.threed.Nx   = output->ic.threed.Nx;
    output->wc.threed.Ny   = output->ic.threed.Ny;
    output->wc.threed.delX = output->ic.threed.delX;
    output->wc.threed.delY = output->ic.threed.delY;
  }
  
  return 0;
}




/***********************************************************************************/
/* Function:                                                              @62_30i@ */
/* Description:                                                                    */
/*  Setup a 3D atmosphere structure, to be used as input to MYSTIC.                */
/*                                                                                 */
/* Parameters:                                                                     */
/* Return value:                                                                   */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                        @i62_30@ */
/***********************************************************************************/

int setup_atmosphere3D (input_struct input, output_struct *output)
{
  int kc=0, status=0;

  double temp=0;

  char function_name[]="setup_cloud3D";
  char file_name[]="cloud3d.c";

  output->atm.nthreed=0;

  /* don't need to read cloud properties if no MYSTIC calculation */
  if (input.rte.solver != MONTECARLO)
    return 0;

  /****************************************/
  /* convert cloud altitudes from km to m */
  /****************************************/
  
  if (output->wc.threed.nthreed > 0)
    for (kc=0; kc<=output->wc.threed.nlyr; kc++)
      output->wc.threed.zd[kc]*=1000.0;

  if (output->ic.threed.nthreed > 0)
    for (kc=0; kc<=output->ic.threed.nlyr; kc++)
      output->ic.threed.zd[kc]*=1000.0;

  
  /********************************/
  /* first, setup the sample grid */
  /********************************/
  
  status = setup_sample_grid (output->atm.zd, output->atm.nlev,
			      input.rte.mc.Nx_sample, input.rte.mc.Ny_sample,
			      input.rte.mc.dx_sample, input.rte.mc.dy_sample,
			      (double) output->wc.threed.Nx * output->wc.threed.delX,
			      (double) output->wc.threed.Ny * output->wc.threed.delY,
			      output->atm.zout_sur, output->atm.nzout,
			      input.rte.mc.Nr, input.rte.mc.Nt, input.rte.mc.dt, 
			      input.rte.umu, input.rte.numu, 
			      input.rte.phi, input.rte.nphi, 
			      input.rte.mc.azimuth,
			      &(output->mc.sample), 
			      input.rte.mc.escape,
			      input.rte.mc.escape_eps_ddis,
			      input.rte.mc.std, 
			      input.rte.mc.backward.yes, input.rte.mc.backward.output,  
			      input.rte.mc.backward.islower, input.rte.mc.backward.jslower,
			      input.rte.mc.backward.isupper, input.rte.mc.backward.jsupper,
			      input.rte.mc.backward.writeallpixels,
			      input.rte.mc.backward.writeback,
			      input.rte.mc.spherical,
                              input.rte.mc.polarisation, /* CE */
                              input.rte.mc.nstokes, /* CE */
                              input.rte.mc.bcond,
			      input.rte.mc.relerr,
			      input.rte.mc.le_abserr,
			      input.rte.mc.maxscatters,
			      input.rte.mc.locest,
			      input.rte.mc.cloner,
			      input.rte.mc.jacobian,
			      input.rte.mc.filename[FN_MC_LIDAR],
			      input.quiet);
  
  if (status!=0) {
    fprintf (stderr, "Error %d returned by setup_sample_grid() in %s (%s)\n", status, function_name, file_name);
    return status;
  }

  /* for memory allocation of the MYSTIC output passed */
  /* to uvspec we need to know which pixels are        */
  /* actually calculated because we allocate memory    */
  /* only for these                                    */
  if (input.rte.mc.backward.writeallpixels || 
      (input.rte.mc.backward.islower == -999 &&
       input.rte.mc.backward.isupper == -999 &&
       input.rte.mc.backward.jslower == -999 &&
       input.rte.mc.backward.jsupper == -999))  {

    output->islower = 0;
    output->jslower = 0;
    output->isupper = output->mc.sample.Nx-1;
    output->jsupper = output->mc.sample.Ny-1;
  }
  else {
    output->islower = input.rte.mc.backward.islower;
    output->isupper = input.rte.mc.backward.isupper;
    output->jslower = input.rte.mc.backward.jslower;
    output->jsupper = input.rte.mc.backward.jsupper;

    if (output->isupper >= output->mc.sample.Nx)
      output->isupper = output->mc.sample.Nx - 1;

    if (output->jsupper >= output->mc.sample.Ny)
      output->jsupper = output->mc.sample.Ny;
  }
  
  if (output->islower<0 || output->isupper<0 || output->jslower<0 || output->jsupper<0) {
    fprintf (stderr, "Error with backward pixel definition in setup_atmosphere3D() in %s (%s):\n", 
	     function_name, file_name);
    fprintf (stderr, "islower = %d, isupper = %d, jslower = %d, jsupper = %d.\n", 
	     output->islower, output->isupper, output->jslower, output->jsupper);
    fprintf (stderr, "Please tell the programmer!\n");
    return -1;
  }

  if ( input.verbose) 
    fprintf (stderr, " ... allocating memory for 3D fields, is = %d ... %d,  js = %d ... %d\n",
	     output->islower, output->isupper, output->jslower, output->jsupper);



  /* read 2D surface elevation */
  /* read 2D elevation data */

  if (strlen (input.rte.mc.filename[FN_MC_ELEVATION]) > 0) {

    /* read elevation data from file */
    status = setup_elevation2D (input.rte.mc.filename[FN_MC_ELEVATION],
				output->mc.sample.Nx, output->mc.sample.Ny, 
				output->mc.sample.delX, output->mc.sample.delY, 
				&(output->mc.elev),
				input.rte.mc.visualize, input.quiet);
    
    if (status!=0) {
      fprintf (stderr, "Error %d returned by setup_elevation2D()\n", status);
      return status;
    }

    output->mc.elev.elev2D = 1;

    output->mc.sample.surfaceparallel = input.rte.mc.surfaceparallel;  /* surface-parallel or   */
    /* horizontal irradiance */
    
    /* user-defined sensor position */
    output->mc.sample.sensorposition = input.rte.mc.sensorposition;
    if (output->mc.sample.sensorposition) {
      output->mc.sample.sensorposition_x[0] = input.rte.mc.sensorposition_x;
      output->mc.sample.sensorposition_x[1] = input.rte.mc.sensorposition_y;
      output->mc.sample.sensorposition_x[2] = input.rte.mc.sensorposition_z;
      
      fprintf (stderr, " ... user-defined sensor position\n");
      fprintf (stderr, "     (%f, %f, %f)\n", 
	       output->mc.sample.sensorposition_x[0], 
	       output->mc.sample.sensorposition_x[1], 
	       output->mc.sample.sensorposition_x[2]);
    }
    
    /* user-defined surface normal */
    output->mc.sample.sensordirection = input.rte.mc.sensordirection;
    if (output->mc.sample.sensordirection) {
      temp = sqrt (input.rte.mc.sensordirection_dx * input.rte.mc.sensordirection_dx + 
		   input.rte.mc.sensordirection_dy * input.rte.mc.sensordirection_dy + 
		   input.rte.mc.sensordirection_dz * input.rte.mc.sensordirection_dz);
      output->mc.sample.sensordirection_dx[0] = input.rte.mc.sensordirection_dx / temp;
      output->mc.sample.sensordirection_dx[1] = input.rte.mc.sensordirection_dy / temp;
      output->mc.sample.sensordirection_dx[2] = input.rte.mc.sensordirection_dz / temp;
      
      fprintf (stderr, " ... user-defined surface normal for surface irradiance calculations:\n");
      fprintf (stderr, "     (%f, %f, %f)\n", 
	       output->mc.sample.sensordirection_dx[0], 
	       output->mc.sample.sensordirection_dx[1], 
	       output->mc.sample.sensordirection_dx[2]);
    }
    
    if (!input.quiet) {
      if (output->mc.sample.surfaceparallel)
	fprintf (stderr, " ... calculating surface parallel irradiance\n");
      else 
	fprintf (stderr, " ... calculating horizontal irradiance\n");
    }
  }


  if (strlen (input.rte.mc.filename[FN_MC_TEMPERATURE]) > 0) {

    /* read 2D surface temperature data from file */
    status = setup_temperature2D (input.rte.mc.filename[FN_MC_TEMPERATURE],
				  output->mc.sample.Nx, output->mc.sample.Ny, 
				  output->mc.sample.delX, output->mc.sample.delY, 
				  &(output->mc.surftemp),
				  input.quiet);
    
    if (status!=0) {
      fprintf (stderr, "Error %d returned by setup_temperature2D()\n", status);
      return status;
    }
    
    output->mc.surftemp.surf2D = 1;
  }
		       

  /* copy 3D water and ice cloud structures to atmosphere structure */
  if (output->wc.threed.nlyr>0) {
    output->atm.Nxcld = output->wc.threed.Nx;
    output->atm.Nycld = output->wc.threed.Ny;
    output->atm.Nzcld = output->wc.threed.nlyr;
  }

  if (output->ic.threed.nlyr>0) {
    output->atm.Nxcld = output->ic.threed.Nx;
    output->atm.Nycld = output->ic.threed.Ny;
    output->atm.Nzcld = output->ic.threed.nlyr;
  }

  /* finally determine the combined water and ice cloud 3D levels */
  output->atm.nthreed=0;
  output->atm.threed = calloc (output->atm.nlyr, sizeof(int));

  for (kc=0; kc<output->atm.nlyr; kc++) {
    output->atm.threed[kc] = 0;
    
    /* water cloud layers */
    if (output->wc.threed.nthreed > 0)
      output->atm.threed[kc] += output->wc.threed.threed[kc];

    /* ice cloud layers */
    if (output->ic.threed.nthreed > 0)
      output->atm.threed[kc] += output->ic.threed.threed[kc];

    if (output->atm.threed[kc] > 0)
      output->atm.nthreed++;
  }

  if (!input.quiet)
    fprintf (stderr, " ... found %d 3D layers (water and ice clouds combined) on a %d x %d x %d grid\n", 
	     output->atm.nthreed, output->atm.Nxcld, output->atm.Nycld, output->atm.Nzcld);
  

  /* if no 3D cloud defined, copy sampling area size to cloud area size */
  if (output->atm.nthreed<1) {
    fprintf (stderr, " ... no 3D clouds defined; copying sample area size to cloud area size\n");
    
    output->atm.Nxcld    = 1;
    output->atm.Nycld    = 1;
    output->wc.threed.Nx = 1;
    output->wc.threed.Ny = 1;
    output->ic.threed.Nx = 1;
    output->ic.threed.Ny = 1;

    output->wc.threed.delX = (double) (output->mc.sample.Nx) * (output->mc.sample.delX);
    output->wc.threed.delY = (double) (output->mc.sample.Ny) * (output->mc.sample.delY);
    output->ic.threed.delX = (double) (output->mc.sample.Nx) * (output->mc.sample.delX);
    output->ic.threed.delY = (double) (output->mc.sample.Ny) * (output->mc.sample.delY);
  }
  

  /* check if cloud grid and sample grids are compatible */

  if (output->wc.threed.nthreed>0) { /* if 3D water clouds */
    if (!equal ((double) (output->mc.sample.Nx) * (output->mc.sample.delX), 
		(double) output->wc.threed.Nx * output->wc.threed.delX)) {
      fprintf (stderr, "Error, x-size of sample domain (%.10e) does not equal x-size of water cloud domain (%.10e)\n",
	       (double) (output->mc.sample.Nx) * (output->mc.sample.delX), 
	       (double) output->wc.threed.Nx * output->wc.threed.delX);
      return -1;
    }
    
    if (!equal ((double) (output->mc.sample.Ny) * (output->mc.sample.delY), 
		(double) output->wc.threed.Ny * output->wc.threed.delY)) {
      fprintf (stderr, "Error, y-size of sample domain (%.10e) does not equal y-size of water cloud domain (%.10e)\n",
	       (double) (output->mc.sample.Ny) * (output->mc.sample.delY), 
	       (double) output->wc.threed.Ny * output->wc.threed.delY);
      return -1;
    }
  }


  if (output->ic.threed.nthreed>0) { /* if 3D ice clouds */
    if (!equal ((double) (output->mc.sample.Nx) * (output->mc.sample.delX), 
		(double) output->ic.threed.Nx * output->ic.threed.delX)) {
      fprintf (stderr, "Error, x-size of sample domain (%.10e) does not equal x-size of ice cloud domain (%.10e)\n",
	       (double) (output->mc.sample.Nx) * (output->mc.sample.delX), 
	       (double) output->ic.threed.Nx * output->ic.threed.delX);
      return -1;
    }
    
    if (!equal ((double) (output->mc.sample.Ny) * (output->mc.sample.delY), 
		(double) output->ic.threed.Ny * output->ic.threed.delY)) {
      fprintf (stderr, "Error, y-size of sample domain (%.10e) does not equal y-size of ice cloud domain (%.10e)\n",
	       (double) (output->mc.sample.Ny) * (output->mc.sample.delY), 
	       (double) output->ic.threed.Ny * output->ic.threed.delY);
      return -1;
    }
  }
  

  return 0;
}


/***********************************************************************************/
/* Function: compare_levels_df                                            @62_30i@ */
/* Description:                                                                    */
/*  Compare two altitude grids. First is expected to be double, second float       */
/*                                                                                 */
/* Parameters:                                                                     */
/* Return value:                                                                   */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                        @i62_30@ */
/***********************************************************************************/

/*
static int compare_levels_df (double *z1, int n1,
			      float *z2, int n2)
{
  int kc=0;

  if (n1!=n2) {
    fprintf (stderr, "Error, number of layers differing\n");
    return -1;
  }

  for (kc=0; kc<=n1; kc++)
    if (fabs((float) z1[kc] - z2[kc]) > MC_EPSILON * z1[kc]) {
      fprintf (stderr, "Error, difference at level %d, altitude %g vs. %g\n",	
	       kc, z1[kc], z2[kc]);
      return -1;
    }

  return 0;
}
*/      




/* Read 3D water cloud data from file */

static int read_wc3D_file (char *filename, 
			   cld3d_out_struct *cld3d, 
			   input_struct input, output_struct *output)
{
  int kc=0, status=0;

  void  F77_FCN (wcloud) (float *lambda_t, int *newsiz, int *nlyr,
			  char *filepath, int *nstring, float *lwc, float *wceffr, 
			  float *tmp_wc_dtau, float *tmp_wc_gg, float *tmp_wc_ssa, float *zd,
			  int *wclyr);

  /* read wc_file */
  status = read_3D_cloud (filename,
			  &(cld3d->Nx), 
			  &(cld3d->Ny), 
			  &(cld3d->nlyr), 
			  &(cld3d->delX), &(cld3d->delY),
			  &(cld3d->zd), 
			  &(cld3d->lwc),
			  &(cld3d->reff),
			  &(cld3d->ext),
			  &(cld3d->g1),
			  &(cld3d->g2),
			  &(cld3d->ff),
			  &(cld3d->f),
			  &(cld3d->ssa),
			  &(cld3d->dscale),
			  &(cld3d->reffmin), &(cld3d->reffmax),
			  &(cld3d->cldproperties),
			  &(cld3d->threed),
			  input.quiet);

  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }
  

  if (!input.quiet)
    fprintf (stderr, " ... read %d x %d x %d cloud data set\n", 
	     cld3d->Nx, cld3d->Ny, cld3d->nlyr);
  
  /* convert horizontal grid size from km to m */
  cld3d->delX *= 1000.0;
  cld3d->delY *= 1000.0;
  
  /* we got the 3D cloud structure in cld3d = output->wc.threed now; */
  /* this structure is passed to mystic()                            */
  
  for (kc=0; kc<cld3d->nlyr; kc++)
    if (cld3d->threed[kc])
      cld3d->nthreed++;
  
  if (!input.quiet) {
    switch (cld3d->cldproperties) {
      case CLD_OPTPROP:
	fprintf (stderr, " ... read extinction coefficient, asymmetry parameter,\n");
	fprintf (stderr, " ... and single scattering albedo 3D cloud file\n");
	break;
      case CLD_EXTREFF:
	fprintf (stderr, " ... read extinction coefficient and effective droplet radius from 3D cloud file\n");
	break;
      case CLD_LWCREFF:
	fprintf (stderr, " ... read liquid water content and effective droplet radius from 3D cloud file\n");
	break;
      default:
	fprintf (stderr, "Error, unknown cloud properties %d\n", cld3d->cldproperties);
	return -1;
    }
  }
  
  if (!input.quiet)
    fprintf (stderr, " ... found %d 3D water cloud layers\n", cld3d->nthreed);
  
  if (!input.quiet) {
    for (kc=0; kc<cld3d->nlyr; kc++)
      if (cld3d->threed[kc])
	fprintf (stderr, " ... found 3D water clouds at level %d, altitude range %g - %g km\n", 
		 kc, cld3d->zd[kc], cld3d->zd[kc+1]);
  }
  
  
  /* default: Henyey-Greenstein - flag needed by MYSTIC */
  cld3d->nonHGcld=0;
  
  switch (input.wc.properties) {
    case WCPROP_HU:
    case WCPROP_ECHAM4:
      break;  
      
    case WCPROP_MIE:
    case WCPROP_FILE:
      
      if (cld3d->cldproperties!=CLD_OPTPROP) { /* no need to read optical properies if HG anyway */
	
	/* need to read cloud property file only if not yet already done */
	if (!(output->wc.ssprop.alloc && output->wc.ssprop.alloc_moments)) {
	  /* Read optical properties file */
	  status = read_cloud_prop (input.filename[FN_WCLOUD_PROP], 
				    output->wl.lambda_t, output->wl.nlambda_t, 
				    output->wl.nlambda_rte_lower, output->wl.nlambda_rte_upper, 
				    input.wc.interpolate,
				    &(output->wc), &(output->atm.nmom),
				    input.rte.pol_nstokes, input.wc.properties, 
				    input.verbose, input.quiet);
	  
	  if (status!=0) {
	    fprintf (stderr, "Error %d reading water cloud property file %s\n", 
		     status, input.filename[FN_WCLOUD_PROP]);
	    return status;
	  }
	  
	  /* set flag for non-Henyey-Greenstein, that is, tabulated phase function */
	  cld3d->nonHGcld=1;
	}
	else {
	  fprintf (stderr, " ... no need to read cloud property file %s again\n", 
		   input.filename[FN_WCLOUD_PROP]);
	}
      }
      else {
	fprintf (stderr, " ... ignoring cloud property file %s, using Henyey-Greenstein\n", 
		 input.filename[FN_WCLOUD_PROP]);
      }
      break;
      
    case WCPROP_PHASE:
      
      if (cld3d->cldproperties!=CLD_OPTPROP) { /* no need to read optical properies if HG anyway */
	
	/* need to read cloud property file only if not yet already done */
	if (!(output->wc.ssprop.alloc && output->wc.ssprop.alloc_explicit)) {
	  /* read optical properties file */
	  
	  fprintf (stderr, "read_cloud_prop_phase (%s)\n", input.rte.mc.filename[FN_MC_PCLD]);
	  
	  status = read_cloud_prop_phase (input.rte.mc.filename[FN_MC_PCLD], 
					  output->wl.lambda_t, output->wl.nlambda_t, 
					  output->wl.nlambda_rte_lower, output->wl.nlambda_rte_upper, 
					  input.wc.interpolate, input.verbose, input.quiet,
					  &(output->wc), &(output->atm.nmom));
	  
	  if (status!=0) {
	    fprintf (stderr, "Error %d reading water cloud property file %s\n", 
		     status, input.rte.mc.filename[FN_MC_PCLD]);
	    return status;
	  }
	  
	  /* set flag for non-Henyey-Greenstein, that is, tabulated phase function */
	  cld3d->nonHGcld=1;
	}
	else {
	  fprintf (stderr, " ... no need to read cloud property file %s again\n", 
		   input.rte.mc.filename[FN_MC_PCLD]);
	}
      }
      else {
	fprintf (stderr, " ... ignoring cloud property file %s, using Henyey-Greenstein\n", 
	       input.filename[FN_WCLOUD_PROP]);
      }
      
      break;
      
    default:
      fprintf (stderr, "Error, unknown wc_properties %d in read_wc3D_cloud. This is not\n", input.wc.properties);
      fprintf (stderr, "supposed to happen and indicates a coding error.\n");
      
      return -1;
      break;
  }
  
  return 0;
}



/* Read 3D ice cloud data from file TZ*/

static int read_ic3D_file (char *filename, 
			   cld3d_out_struct *cld3d, 
			   input_struct input, output_struct *output)
{
  int status=0, ix=0, iy=0, kc=0;
 
  /* read ic_file */
  status = read_3D_cloud (filename,
			  &(cld3d->Nx), 
			  &(cld3d->Ny), 
			  &(cld3d->nlyr), 
			  &(cld3d->delX), &(cld3d->delY),
			  &(cld3d->zd), 
			  &(cld3d->lwc),
			  &(cld3d->reff),
			  &(cld3d->ext),
			  &(cld3d->g1),
			  &(cld3d->g2),
			  &(cld3d->ff),
			  &(cld3d->f),
			  &(cld3d->ssa),
			  &(cld3d->dscale),
			  &(cld3d->reffmin), &(cld3d->reffmax),
			  &(cld3d->cldproperties),
			  &(cld3d->threed),
			  input.quiet);

  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }
  

  if (!input.quiet)
    fprintf (stderr, " ... read %d x %d x %d cloud data set\n", 
	     cld3d->Nx, cld3d->Ny, cld3d->nlyr);
  
  /* convert horizontal grid size from km to m */
  cld3d->delX *= 1000.0;
  cld3d->delY *= 1000.0;

  /* we got the ice 3D cloud structure in cld3d = output->ic.threed now; */
  /* this structure is passed to mystic()                                */
  
  for (kc=0; kc<cld3d->nlyr; kc++)
    if (cld3d->threed[kc])
      cld3d->nthreed++;
  
  if (!input.quiet)
    fprintf (stderr, " ... found %d 3D ice cloud layers\n", cld3d->nthreed);
  
  if (!input.quiet) {
    for (kc=0; kc<cld3d->nlyr; kc++)
      if (cld3d->threed[kc])
	fprintf (stderr, " ... found 3D ice clouds at level %d, altitude range %g - %g km\n", 
		 kc, cld3d->zd[kc], cld3d->zd[kc+1]);
  }
  
  
  
  /* default: Henyey-Greenstein - flag needed by MYSTIC */
  cld3d->nonHGcld=0;

  switch (input.ic.properties) {/*switch ic.properties*/
  
  case ICPROP_MIE:
  case ICPROP_FILE:

    if (cld3d->cldproperties!=CLD_OPTPROP) { /* no need to read optical properies if HG anyway */
      
      /* need to read cloud property file only if not yet already done */
      if (!(output->ic.ssprop.alloc && output->ic.ssprop.alloc_moments)) {
	/* Read optical properties file */
	status = read_cloud_prop (input.filename[FN_ICLOUD_PROP], 
				  output->wl.lambda_t, output->wl.nlambda_t, 
				  output->wl.nlambda_rte_lower, output->wl.nlambda_rte_upper, 
				  input.ic.interpolate,
				  &(output->ic), &(output->atm.nmom),
                                  input.rte.pol_nstokes, input.ic.properties, 
                                  input.verbose, input.quiet);
	
	if (status!=0) {
	  fprintf (stderr, "Error %d reading ice cloud property file %s\n", 
		   status, input.filename[FN_ICLOUD_PROP]);
	return status;
	}

	/* set flag for non-Henyey-Greenstein, that is, tabulated phase function */
	cld3d->nonHGcld=1;
      }
      else {
	fprintf (stderr, " ... no need to read cloud property file %s again\n", 
		 input.filename[FN_ICLOUD_PROP]);
      }
    }
    else {
      fprintf (stderr, " ... ignoring cloud property file %s, using Henyey-Greenstein\n", 
	       input.filename[FN_ICLOUD_PROP]);
    }
    break;

  case ICPROP_BAUM:
  case ICPROP_BAUMDETAILED:

    if (cld3d->cldproperties!=CLD_OPTPROP) { /* no need to read optical properies if HG anyway */
      
      /* need to read cloud property file only if not yet already done */
      if (!(output->ic.ssprop.alloc && output->ic.ssprop.alloc_moments)) {
	/* Read optical properties file */
	status = read_cloud_prop (input.filename[FN_ICLOUD_PROP], 
                                  output->wl.lambda_t,
                                  output->wl.nlambda_t, 
                                  output->wl.nlambda_rte_lower,
                                  output->wl.nlambda_rte_upper, 
                                  input.ic.interpolate,
                                  &(output->ic), &(output->atm.nmom),
                                  input.rte.pol_nstokes, input.ic.properties, 
                                  input.verbose, input.quiet);
	
	if (status!=0) {
	  fprintf (stderr, "Error %d reading ice cloud property file %s\n", 
		   status, input.filename[FN_ICLOUD_PROP]);
	return status;
	}

	/* set flag for non-Henyey-Greenstein, that is, tabulated phase function */
	cld3d->nonHGcld=1;
      }
      else {
	fprintf (stderr, " ... no need to read cloud property file %s again\n", 
		 input.filename[FN_ICLOUD_PROP]);
      }
    }
    else {
      fprintf (stderr, " ... ignoring cloud property file %s, using Henyey-Greenstein\n", 
	       input.filename[FN_ICLOUD_PROP]);
    }
    break;
    
  case ICPROP_FU:
    
    /* check if users wants Fu or Yang definition of effective radius */

    if (input.ic.fu2yang) {
      fprintf(stderr, " ... converting from Fu (1996/98) to Key et al. (2002) definition of effective radius \n");
      for (kc=0; kc<cld3d->nlyr; kc++)
	if (cld3d->nthreed>0)  /* if at least one 3D layer */
	  if (cld3d->threed[kc])  /* 3D layer */
	    for (ix=0; ix<cld3d->Nx; ix++)
	      for (iy=0; iy<cld3d->Ny; iy++)  {
		fprintf (stderr, "fu2yang: scaling effective radius from %f to", cld3d->reff[kc][ix][iy]);
		cld3d->reff[kc][ix][iy] /= (3.0*sqrt(3.0)/4.0);		 
		fprintf (stderr, " %f\n", cld3d->reff[kc][ix][iy]);
	      }
      
      cld3d->reffmin /= (3.0*sqrt(3.0)/4.0);
      cld3d->reffmax /= (3.0*sqrt(3.0)/4.0);
    }
    
    break;

  case ICPROP_YANG:
  case ICPROP_KEY:
  case ICPROP_ECHAM4:
    break;

  default:
    fprintf (stderr, "Error, unknown ic_properties %d in read_ic3D_cloud. This is not\n", input.ic.properties);
    fprintf (stderr, "supposed to happen and indicates a coding error.\n");

    return -1;
    break;
  }
  
  return 0;
}



/********************************************************************/
/* Convert 1D and 3D water cloud data to MYSTIC optical properties  */
/* for a given wavelength index iv. In particular, need to          */
/* extinction, single scattering albedo, and delta-scaling factor   */
/* to the effective radius of each 3D grid box. Also, need to       */
/* calculate the cumulative scattering function table for           */
/* the respective wavelength.                                       */
/********************************************************************/

int convert_wc_mystic (input_struct input, output_struct *output,
		       cld3d_out_struct *wcld3d, int iv)
{
  int status=0;
  int nstring = (int) strlen(input.filename[FN_PATH]);

  float lwc [2] = {0,0};
  float reff[2] = {0,0};
  float dtau[2] = {0,0}, g1[2]={0,0}, ssa[2]={0,0};
  float **wvl_dtau = NULL;
  float **wvl_g1   = NULL;
  float **wvl_ssa  = NULL;

  float zd  [2] = {1,0};
  float rmin=0, rmax=0;
  int nlyr=1, nlev=2;
  int newsiz=1;
  int wclyr=1;
  int ix=0, iy=0, iz=0;

  int i1=0;
  double extinc=0, scattr=0, omega=0, dscale=0, delta_r=0, f=0;

  char function_name[]="convert_wc_mystic";
  char file_name[]="cloud3d.c";

  void  F77_FCN (wcloud) (float *lambda_t, int *newsiz, int *nlyr,
			  char *filepath, int *nstring, float *lwc, float *wceffr, 
			  float *tmp_wc_dtau, float *tmp_wc_gg, float *tmp_wc_ssa, float *zd,
			  int *wclyr);
  
  if ((status = ASCII_calloc_float (&wvl_dtau,18,2)) != 0) {
    fprintf(stderr,"Error in %s (%s),\n", function_name, file_name);
    fprintf(stderr,"      allocation of 'wvl_dtau'\n");
    return -1;
  } 

  if ((status = ASCII_calloc_float (&wvl_g1,18,2)) != 0) {
    fprintf(stderr,"Error in %s (%s),\n", function_name, file_name);
    fprintf(stderr,"      allocation of 'wvl_g1'\n");
    return -1;
  }

  if ((status = ASCII_calloc_float (&wvl_ssa,18,2)) != 0) {
    fprintf(stderr,"Error in %s (%s),\n", function_name, file_name);
    fprintf(stderr,"      allocation of 'wvl_ssa'\n");
    return -1;
  }

  switch (wcld3d->cldproperties) {
  case CLD_OPTPROP:

    /* ?????????????????????????????????????????????????????????????????????????????? */
    /* currently we have a special treatment for explicitely defined phase functions; */
    /* this is subject to change, as we want to make the explicit phase function      */
    /* compatible to WCPROP_MIE/WCPROP_FILE; that is, specify phase function in the   */
    /* same format as in the momcld_files for different effective radii plus          */
    /* extinction coefficient and single scattering albedo                            */

    if (input.wc.properties == WCPROP_PHASE) {

      rmin = 10; rmax = 10; /* doesn't matter anyway */

      /* generate phase function table */
      status = setup_Legendre_table (&(wcld3d->phase_cld), 
				     &(output->wc.ssprop),
				     iv,
				     rmin, rmax, 
				     input.rte.mc.delta_scaling_mucut, 
				     input.rte.mc.truncate, 
				     input.quiet);/*TZ ds*/
      if (status!=0) {
	fprintf (stderr, "Error %d returned by setup_Legendre_table()\n", 
		 status);
	return status;
      }
    }

    break;

  case CLD_EXTREFF:
  case CLD_LWCREFF:

    /* first create phase function table */
    switch (input.wc.properties) {
    case WCPROP_HU:
    case WCPROP_ECHAM4:
      break;  
      
    case WCPROP_MIE:
    case WCPROP_FILE:
    case WCPROP_PHASE:
      
      /* combine 1D and 3D rmin/rmax */
      rmax = (wcld3d->reffmax > output->wc.microphys.effrmax ? wcld3d->reffmax : output->wc.microphys.effrmax);

      /* take the smaller of both which is larger than zero */
      if (wcld3d->reffmin > 0) {
	if (output->wc.microphys.effrmin > 0)
	  rmin = (wcld3d->reffmin < output->wc.microphys.effrmin ? wcld3d->reffmin : output->wc.microphys.effrmin);
	else
	  rmin = wcld3d->reffmin;
      }
      else 
	rmin = output->wc.microphys.effrmin;
      
      if (rmin<=0) /* no need to do anything */
	break;
      
      fprintf (stderr, " ... generating phase function tables between rmin=%f and rmax=%f\n", rmin, rmax);

      /* generate phase function table */
      status = setup_Legendre_table (&(wcld3d->phase_cld), 
				     &(output->wc.ssprop),
				     iv,
				     rmin, rmax, 
				     input.rte.mc.delta_scaling_mucut, 
				     input.rte.mc.truncate, 
				     input.quiet);/*TZ ds*/

      if (status!=0) {
	fprintf (stderr, "Error %d returned by setup_Legendre_table()\n", status);
	return status;
      }
      break;
      
    default:
      fprintf (stderr, "Error, unknown water cloud property %d in convert_wc_mystic. This is not\n", input.wc.properties);
      fprintf (stderr, "supposed to happen and indicates a coding error.\n");
      
      return -1;
      break;
    }

    for (iz=0; iz<wcld3d->nlyr; iz++) {
      if (wcld3d->nthreed>0) {
	if (wcld3d->threed[iz]) { /* 3D layer */

	  for (ix=0; ix<wcld3d->Nx; ix++) {/*ix*/
	    for (iy=0; iy<wcld3d->Ny; iy++) {/*iy*/
      
	      if ((wcld3d->cldproperties==CLD_EXTREFF && wcld3d->ext[iz][ix][iy]>0) ||
		  (wcld3d->cldproperties==CLD_LWCREFF && wcld3d->lwc[iz][ix][iy]>0)) {
		
		/* use dummy LWC - it doesn't matter anyway */
		if (wcld3d->cldproperties==CLD_EXTREFF) {
		  lwc[0] = 1.0; 
		  lwc[1] = 1.0;
		}
		
		if (wcld3d->cldproperties==CLD_LWCREFF) {
		  lwc [0] = wcld3d->lwc[iz][ix][iy];
		  lwc [1] = wcld3d->lwc[iz][ix][iy];
		}
		
		reff[0] = wcld3d->reff[iz][ix][iy];
		reff[1] = wcld3d->reff[iz][ix][iy];

		switch (input.wc.properties) {
		case WCPROP_HU:
		
		  switch (input.ck_scheme) {
		  case CK_FU:
		  
		    /* call Fu and Liou code by Fred Rose */
		  
#if HAVE_FULIOU
		   
		    /* Following issue still not optimized:                */
		    /* currently ckdfuice loops over all 18 bands which is */
		    /* a waste of time and should be                       */

		    status = ckdfucld (zd, 
				       reff, 
				       lwc, 
				       nlev,
				       wvl_dtau, wvl_g1, wvl_ssa);

		    /* write results to final destination */

		    if (wcld3d->cldproperties==CLD_LWCREFF)
		      wcld3d->ext[iz][ix][iy] = wvl_dtau[iv][0] / 1000.0; 

		    wcld3d->ssa[iz][ix][iy] = wvl_ssa [iv][0];
		    wcld3d->g1 [iz][ix][iy] = wvl_g1  [iv][0];
		    wcld3d->g2 [iz][ix][iy] = 0;
		    wcld3d->ff [iz][ix][iy] = 1.0;
		  
#else
		    fprintf (stderr, "Error, Fu and Liou not supported!\n");
		    return -1;
#endif
		  
		    if (status!=0) {
		      fprintf (stderr, "Error %d returned by ckdfucld()\n", status);
		      return status;
		    }
		  
		    break;
		  
		  default:   /* old uvspec wcloud.f for Hu and Stamnes */
		  
		    /* wcloud calculates lwc_layer and reff_layer internally */
		    F77_FCN (wcloud) (&output->wl.lambda_t[iv], &newsiz, &nlyr,
				      input.filename[FN_PATH], &nstring, 
				      lwc, reff, 
				      dtau, g1, ssa, 
				      zd, &wclyr);
		  
		    /* write results to final destination */
		    wcld3d->ssa[iz][ix][iy] = ssa[0];
		    wcld3d->g1 [iz][ix][iy] = g1 [0];
		    wcld3d->g2 [iz][ix][iy] = 0;
		    wcld3d->ff [iz][ix][iy] = 1.0;
		    
		    /*
		      fprintf (stderr, "WCLOUD3D %d %d %d  %f %f %f %f\n",
		      iz, ix, iy,
		      wcld3d->ssa[iz][ix][iy],
		      wcld3d->g1 [iz][ix][iy],
		      wcld3d->g2 [iz][ix][iy],
		      wcld3d->ff [iz][ix][iy]);
		    */
 
		    if (wcld3d->cldproperties==CLD_LWCREFF)
		      wcld3d->ext[iz][ix][iy] = dtau[0] / 1000.0; 
		    
		    /* need to convert again from 1/km to 1/m */		
		    /* because wcloud() thinks in km rather   */
		    /* than m                                 */		    		    
		    break;

		  }	    	      
		  break;
		
		case WCPROP_ECHAM4:
		
		  status = wc_echam4 (output->wl.lambda_t[iv], nlev-1,
				      lwc, reff, 
				      dtau, g1, ssa,
				      zd, wclyr);
		
		  if (status!=0) {
		    fprintf (stderr, "Error %d returned by wc_echam4()\n", status);
		    return status;
		  }

		  /* write results to final destination */
		  wcld3d->ssa[iz][ix][iy] = ssa[0];
		  wcld3d->g1 [iz][ix][iy] = g1 [0];
		  wcld3d->g2 [iz][ix][iy] = 0;
		  wcld3d->ff [iz][ix][iy] = 1.0;
		
		  if (wcld3d->cldproperties==CLD_LWCREFF)
		    wcld3d->ext[iz][ix][iy] = dtau[0] / 1000.0;
		
		  /* need to convert again from 1/km to 1/m */		
		  /* because wcloud() thinks in km rather   */
		  /* than m                                 */

		  break;  
		
		case WCPROP_MIE:
		case WCPROP_FILE:
		case WCPROP_PHASE:

		  /* interpolate extinction and absorption coefficient */
		  /* of each box to given effective radius             */
		  i1 = (int) ((wcld3d->reff[iz][ix][iy] - wcld3d->phase_cld.r0) / wcld3d->phase_cld.dr);
		
		  /* careful with rounding errors at the array boundaries */
		  i1 = (i1>wcld3d->phase_cld.n-1?wcld3d->phase_cld.n-1:i1);
		  i1 = (i1<0?0:i1);
	      
		
		  /* BM, changed interpolation: now interpolating scattering coefficient */
		  /* instead of single scattering albedo                                 */

		  if (i1==wcld3d->phase_cld.n-1) {
		    extinc  = wcld3d->phase_cld.ext   [i1];
		    scattr  = wcld3d->phase_cld.sca   [i1];
		    dscale  = wcld3d->phase_cld.dscale[i1]; 
		    f       = wcld3d->phase_cld.f     [i1];
		  }		    
		  else {
		    delta_r = wcld3d->reff[iz][ix][iy] - (wcld3d->phase_cld.r0 + (double) i1 * wcld3d->phase_cld.dr);
		    
		    extinc  = wcld3d->phase_cld.ext[i1] + wcld3d->phase_cld.dext[i1] * delta_r;
		    scattr  = wcld3d->phase_cld.sca[i1] + wcld3d->phase_cld.dsca[i1] * delta_r;
		    
		    /* ??? DDDDD interpolate the delta scaling factor f   DDDDD ??? */
		    if (wcld3d->phase_cld.f[i1]!=0 || wcld3d->phase_cld.f[i1+1]!=0)
		      f = interpolate_f (&(wcld3d->phase_cld), i1, i1+1, delta_r);
		    else 
		      f = 0;

		    if (f<0 || f>1) {
		      fprintf (stderr, "Error in %s, line %d, function %s:\n", __FILE__, __LINE__, __func__);
		      fprintf (stderr, "delta-scaling factor %f out of range!\n", f);
		      fprintf (stderr, "Tried to interpolate between radii %f and %f,\n", 
			       wcld3d->phase_cld.r0 + (double) i1 * wcld3d->phase_cld.dr,
			       wcld3d->phase_cld.r0 + (double) (i1+1) * wcld3d->phase_cld.dr);
		      fprintf (stderr, "f values %f and %f\n", 
			       wcld3d->phase_cld.f[i1], wcld3d->phase_cld.f[i1+1]);
		      return -1;
		    }

		    /* ??? DDDDD interpolate the delta scaling factor dscale DDDDD ??? */
		    /* dscale  = wcld3d->phase_cld.dscale[i1] + wcld3d->phase_cld.ddscale[i1] * delta_r; */
		    if (wcld3d->phase_cld.dscale[i1]!=0 || wcld3d->phase_cld.dscale[i1+1]!=0)
		      dscale = interpolate_dscale (&(wcld3d->phase_cld), i1, i1+1, delta_r);
		    else 
		      dscale = 0;

		    if (dscale<0 || dscale>1) {
		      fprintf (stderr, "Error in %s, line %d, function %s:\n", __FILE__, __LINE__, __func__);
		      fprintf (stderr, "delta-scaling factor %f out of range!\n", dscale);
		      fprintf (stderr, "Tried to interpolate between radii %f and %f,\n", 
			       wcld3d->phase_cld.r0 + (double) i1 * wcld3d->phase_cld.dr,
			       wcld3d->phase_cld.r0 + (double) (i1+1) * wcld3d->phase_cld.dr);

		      fprintf (stderr, "f values %f and %f\n", 
			       wcld3d->phase_cld.dscale[i1], wcld3d->phase_cld.dscale[i1+1]);

		      return -1;
		    }
		  }

		  /* single scattering albedo */
		  omega   = scattr / extinc;
		
		  /* extinction coefficient */     
		  if (wcld3d->cldproperties==CLD_LWCREFF)
		    wcld3d->ext[iz][ix][iy] = wcld3d->lwc[iz][ix][iy] * extinc / 1000.0;
		 
		  /* single scattering albedo */
		  wcld3d->ssa[iz][ix][iy] = omega;
		
                  /* external delta-scaling factor CE */
                  wcld3d->f[iz][ix][iy]=f;  

		  /* internal delta-scaling factor TZ */
		  wcld3d->dscale[iz][ix][iy] = dscale;

		  break;
		
		default:
		  fprintf (stderr, "Error, unknown water cloud property %d in convert_wc_mystic (II). This is not\n", input.wc.properties);
		  fprintf (stderr, "supposed to happen and indicates a coding error.\n");
		
		  return -1;
		  break;
		}
	      }
	    } 
	  }
	} /* end if (wcld3d->threed[iz]) */
      } /* end if (wcld3d->nthreed>0) */
      

      /* for a 1D cloud we only need to interpolate the delta scaling factor;  */
      /* this is a bit unfortunate but we need to do that here because         */
      /* all other properties are already interpolated in setup_wcloud()       */
      /* where dscale is not yet known                                         */

      if (output->wc.microphys.lwc_layer[iz] > 0) { 
	
	switch (input.wc.properties) {
	case WCPROP_HU:
	case WCPROP_ECHAM4:

	  /* no need to do anything because delta scaling only for */
	  /* non-HG phase functions                                */

	  break;

        case WCPROP_MIE:
        case WCPROP_FILE:
        case WCPROP_PHASE:

	  /* determine index of effective radius */
	  i1 = (int) ((output->wc.microphys.effr_layer[iz] - wcld3d->phase_cld.r0) / wcld3d->phase_cld.dr);
	  
	  /* careful with rounding errors at the array boundaries */
	  i1 = (i1>wcld3d->phase_cld.n-1?wcld3d->phase_cld.n-1:i1);
	  i1 = (i1<0?0:i1);
	  
	  /* ??? DDDDD interpolate delta-scaling factor dscale DDDDD ??? */
	  if (wcld3d->phase_cld.dscale[i1]!=0 || wcld3d->phase_cld.dscale[i1+1]!=0) {
	    delta_r = output->wc.microphys.effr_layer[iz] - (wcld3d->phase_cld.r0 + (double) i1 * wcld3d->phase_cld.dr);
	    dscale = interpolate_dscale (&(wcld3d->phase_cld), i1, i1+1, delta_r);
	  }
	  else 
	    dscale = 0;
	  
	  if (dscale<0 || dscale>1) {
	    fprintf (stderr, "Error in %s, line %d, function %s:\n", __FILE__, __LINE__, __func__);
	    fprintf (stderr, "delta-scaling factor %f out of range!\n", dscale);
	    fprintf (stderr, "Tried to interpolate between radii %f and %f,\n", 
		     wcld3d->phase_cld.r0 + (double) i1 * wcld3d->phase_cld.dr,
		     wcld3d->phase_cld.r0 + (double) (i1+1) * wcld3d->phase_cld.dr);

	    fprintf (stderr, "f values %f and %f\n", 
		     wcld3d->phase_cld.dscale[i1], wcld3d->phase_cld.dscale[i1+1]);
	    return -1;
	  }
	  
	  output->wc.optprop.dscale[iv][iz] = dscale;

	  /*
	  output->wc.optprop.dscale[iv][iz] = wcld3d->phase_cld.dscale[i1] + wcld3d->phase_cld.ddscale[i1] * 
	      (output->wc.microphys.effr_layer[iz] - (wcld3d->phase_cld.r0 + (double) i1 * wcld3d->phase_cld.dr));
	  */

	  break;
	default:
	  fprintf (stderr, "Error, unknown cloud property %d\n", input.wc.properties);
	  return -1;
	}	
      }
    } /* for (iz=0; iz<wcld3d->nlyr; iz++) */
    break;
    
  default:
    fprintf (stderr, "Error, unknown water cloud properties %d\n", wcld3d->cldproperties);
    return -1;
  }

  if ((status = ASCII_free_float (wvl_dtau,18)) != 0) {
    fprintf(stderr,"Error in %s (%s),\n", function_name, file_name);
    fprintf(stderr,"      free memory of 'wvl_dtau'\n");
    return -1;
  }
  if ((status = ASCII_free_float (wvl_g1,18)) != 0) {
    fprintf(stderr,"Error in %s (%s),\n", function_name, file_name);
    fprintf(stderr,"      free memory of 'wvl_g1'\n");
    return -1;
  }
  if ((status = ASCII_free_float (wvl_ssa,18)) != 0) {
    fprintf(stderr,"Error in %s (%s),\n", function_name, file_name);
    fprintf(stderr,"      free memory of 'wvl_ssa'\n");
    return -1;
  }
  return 0;
}
  



/********************************************************************/
/* Convert 1D and 3D ice cloud data to MYSTIC optical properties    */
/* for a given wavelength index iv. In particular, need to          */
/* extinction, single scattering albedo, and delta-scaling factor   */
/* to the effective radius of each 3D grid box. Also, need to       */
/* calculate the cumulative scattering function table for           */
/* the respective wavelength.                                       */
/********************************************************************/

int convert_ic_mystic (input_struct input, output_struct *output,
		       cld3d_out_struct *icld3d, int iv)
{
  char path[FILENAME_MAX]="";

  int status=0;

  float lay_lwc [2] = {0,0};
  float lay_reff[2] = {0,0};
  float lay_dtau[2] = {0,0}, lay_g1[2]={0,0}, lay_ssa[2]={0,0};
  float lay_zd  [2] = {1,0};
  float lay_f[2] = {0,0};
  float **wvl_dtau;
  float **wvl_g1;
  float **wvl_ssa;

  int layer_flag=0;
  int nlev=2;
  int newkey=0;
  int ic_habit=0;

  float rmin=0, rmax=0;
  int ix=0, iy=0, iz=0;
  float  extinct=0, gg1=0, gg2=0, ff=0, oomega=0;
  float wavelength;

  int i1=0;
  double reff = 0, extinc=0, scattr=0, omega=0, dscale=0, delta_r=0, f=0;
  double ssa_us = 0; 

  char function_name[]="convert_ic_mystic";
  char file_name[]="cloud3d.c";

 
  if ((status = ASCII_calloc_float (&wvl_dtau,18,2)) != 0) {
    fprintf(stderr,"Error in %s (%s),\n", function_name, file_name);
    fprintf(stderr,"      allocation of 'wvl_dtau'\n");
    return -1;
  }

  if ((status = ASCII_calloc_float (&wvl_g1,18,2)) != 0) {
    fprintf(stderr,"Error in %s (%s),\n", function_name, file_name);
    fprintf(stderr,"      allocation of 'wvl_g1'\n");
    return -1;
  }

  if ((status = ASCII_calloc_float (&wvl_ssa,18,2)) != 0) {
    fprintf(stderr,"Error in %s (%s),\n", function_name, file_name);
    fprintf(stderr,"      allocation of 'wvl_ssa'\n");
    return -1;
  }
  
  
  switch (icld3d->cldproperties) {
  
  case CLD_OPTPROP: 
    /* ??? need special treatment for explicit ice cloud phase function, ??? */
    /* ??? once it is implemented ...                                        */
    break;

  case CLD_EXTREFF:
  case CLD_LWCREFF:
    
    /* first create phase function table */
    switch (input.ic.properties) {     

    case ICPROP_FU:
    case ICPROP_ECHAM4:
      break;

    case ICPROP_KEY:         
      newkey=0;
      ic_habit=IC_HABIT_SOLIDCOLUMN; /* ??? currently no way to define 3D habit ??? */
      break;

    case ICPROP_YANG:
      newkey=1;
      ic_habit=IC_HABIT_SOLIDCOLUMN; /* ??? currently no way to define 3D habit ??? */
      break;


    case ICPROP_BAUM:
    case ICPROP_BAUMDETAILED:
    case ICPROP_MIE:             
    case ICPROP_FILE:
      
      /* combine 1D and 3D rmin/rmax */
      rmax = (icld3d->reffmax > output->ic.microphys.effrmax ? icld3d->reffmax : output->ic.microphys.effrmax);
      
      /* take the smaller of both which is larger than zero */
      if (icld3d->reffmin > 0) {
	if (output->ic.microphys.effrmin > 0)
	  rmin = (icld3d->reffmin < output->ic.microphys.effrmin ? icld3d->reffmin : output->ic.microphys.effrmin);
	else
	  rmin = icld3d->reffmin;
      }
      else 
	rmin = output->ic.microphys.effrmin;
      
      if (rmin<=0) /* no need to do anything */
	break;
      
      fprintf (stderr, " ... generating phase function tables between rmin=%f and rmax=%f\n", rmin, rmax);
    
      /* generate phase function table */
      status = setup_Legendre_table (&(icld3d->phase_cld), 
				     &(output->ic.ssprop),
				     iv,
				     rmin, rmax, 
				     input.rte.mc.delta_scaling_mucut, 
				     input.rte.mc.truncate, 
				     input.quiet);
      if (status!=0) {
	fprintf (stderr, "Error %d returned by setup_Legendre_table()\n", 
		 status);
	return status;
      }
      break;
      
      
    default:
      fprintf (stderr, "Error, unknown ice cloud property %d in convert_ic_mystic (I). This is not\n", input.ic.properties);
      fprintf (stderr, "supposed to happen and indicates a coding error.\n");
      
      return -1;
      break;
    }
    
    
    for (iz=0; iz<icld3d->nlyr; iz++) {
    
      if (icld3d->nthreed>0) { /* if we have any 3D layers */
       
	if (icld3d->threed[iz]) { /* 3D layer */
         
	  for (ix=0; ix<icld3d->Nx; ix++) {/*ix*/
	    for (iy=0; iy<icld3d->Ny; iy++) {/*iy*/
              
	      if ((icld3d->cldproperties==CLD_EXTREFF && icld3d->ext[iz][ix][iy]>0) ||
                  (icld3d->cldproperties==CLD_LWCREFF && icld3d->lwc[iz][ix][iy]>0)) { /* not if CLD_OPTPROP or no data */
                
                /*makeshift to provide input to the following ic_fu96/ic_echam
                  routines expecting profile information not 3D fields*/
		lay_reff[0] = icld3d->reff[iz][ix][iy]; 
		lay_reff[1] = icld3d->reff[iz][ix][iy];      	    

		if (icld3d->cldproperties==CLD_LWCREFF) {
		  lay_lwc[0] = icld3d->lwc[iz][ix][iy];
		  lay_lwc[1] = icld3d->lwc[iz][ix][iy];      
		}	     
		/* use dummy LWC - it doesn't matter anyway */
		/*if (icld3d->cldproperties==CLD_EXTREFF) {*/
		else{
                  lay_lwc[0] = 1.0; 
		  lay_lwc[1] = 1.0;
		}
		
                
                /* CE: Un-deltascale extinction. This is necessary 
                   because otherwise extinction is
                   scaled again and again for each wavelength */
                if( icld3d->cldproperties!=CLD_LWCREFF){
                  ssa_us = icld3d->ssa[iz][ix][iy] /
                    ( 1.0 + icld3d->f [iz][ix][iy] * ( icld3d->ssa[iz][ix][iy] - 1.0));
                  icld3d->ext[iz][ix][iy] /= (1.0 -  ssa_us * icld3d->f [iz][ix][iy]);
                }
                
		switch (input.ic.properties) {/*  switch (input.ic.properties) */

		case ICPROP_KEY:
		case ICPROP_YANG:			
                  
                  /* call yang routine to obtain optical properties */
		  reff = icld3d->reff[iz][ix][iy];
		  strcpy (path, input.filename[FN_PATH]);
		  strcat (path, "/ic/yang56/");
		  wavelength=output->wl.lambda_t[iv];
		  
		  status = yang56 (wavelength/1000.0, reff, ic_habit, path, 
				   &extinct, &oomega, 
				   &gg1, &gg2, &ff,
				   newkey);
		  
		  if (status!=0) {
		    fprintf (stderr, "Error %d calculating Yang/Key/Mayer ice cloud properties for\n", status);
		    fprintf (stderr, " effective radius %f micron at wavelength %f nm\n", reff,wavelength);
		    return status;
		  }
		  
		  icld3d->g1[iz][ix][iy] = gg1;
		  icld3d->g2[iz][ix][iy] = gg2;
		  icld3d->ff[iz][ix][iy] = ff;
		  
		  if (icld3d->cldproperties==CLD_LWCREFF)
		    icld3d->ext[iz][ix][iy] = icld3d->lwc[iz][ix][iy] * extinct / 1000.0;

		  icld3d->ssa[iz][ix][iy] = oomega;
		  
		  break;
                  
		case ICPROP_FU:		
                  
                  /* calculate cloud optical properties */
		  switch (input.ck_scheme) {
		  case CK_FU:
                    
		    /* call Fu and Liou code by Fred Rose */
		  
#if HAVE_FULIOU

		    /* Following issue still not optimized:                */
		    /* currently ckdfuice loops over all 18 bands which is */
		    /* a waste of time and should be changed               */
                    
		    status = ckdfuice (lay_zd, 
				       lay_reff, 
				       lay_lwc, 
				       nlev,
				       wvl_dtau, 
				       wvl_g1, 
				       wvl_ssa, 
				       input.ic.unscaled);
		  
		    /* write results to final destination */
		    if (icld3d->cldproperties==CLD_LWCREFF)
		      icld3d->ext[iz][ix][iy] = wvl_dtau[iv][0] /1000.;

		    icld3d->ssa[iz][ix][iy] = wvl_ssa [iv][0];
		    icld3d->g1 [iz][ix][iy] = wvl_g1  [iv][0];
		    icld3d->g2 [iz][ix][iy] = 0;
		    icld3d->ff [iz][ix][iy] = 1.0;
		  
#else
		    fprintf (stderr, "Error, Fu and Liou not supported!\n");
		    return -1;
#endif
		  
		    if (status!=0) {
		      fprintf (stderr, "Error %d returned by ckdfuice()\n", status);
		      return status;
		    }
		  
		    break;		    
		  
		  default:
		    layer_flag=1;/* for 3D clouds this is always true*/
		  
		    if (output->wl.lambda_t[iv]<4000.0) {
		    
		      status = ic_fu96 (output->wl.lambda_t[iv], nlev-1,
					input.filename[FN_PATH],
					lay_lwc, lay_reff, 
					lay_dtau, lay_g1, lay_ssa, lay_f,
					lay_zd, layer_flag , input.ic.unscaled); /* in file cloud.c */
		    
		      if (status!=0) {
			fprintf (stderr, "Error %d returned by ic_fu96()\n", status);
			return status;
		      }
		    }
		    else {
		      status = ic_fu98 (output->wl.lambda_t[iv], nlev-1,
					input.filename[FN_PATH],
					lay_lwc, lay_reff, 
					lay_dtau, lay_g1, lay_ssa, 
					lay_zd, layer_flag); /* in file cloud.c */
		    
		      if (status!=0) {
			fprintf (stderr, "Error %d returned by ic_fu98()\n", status);
			return status;
		      }
		    
		    }
		  
		    fprintf (stderr, "FU96 %f %f %f\n", lay_dtau[0], lay_g1[0], lay_ssa[0]);

		    /* write results to final destination */
		    if (icld3d->cldproperties==CLD_LWCREFF)
		      icld3d->ext[iz][ix][iy] = lay_dtau[0] /1000.;
                    
		    icld3d->ssa [iz][ix][iy] = lay_ssa[0];
		    icld3d->g1  [iz][ix][iy] = lay_g1 [0];
		    icld3d->g2  [iz][ix][iy] = 0;
		    icld3d->ff  [iz][ix][iy] = 1.0;
		    icld3d->f[iz][ix][iy] = lay_f[0];
		  
		  } /*switch (input.ck_scheme)*/		  
		  break;
                  
		case ICPROP_ECHAM4:	              
		  layer_flag=1;/* for 3D clouds this is always true*/

		  status = ic_echam4 (output->wl.lambda_t[iv], nlev,
				      lay_lwc, lay_reff, 
				      lay_dtau, lay_g1, lay_ssa,
				      lay_zd, layer_flag);
		
		  if (status!=0) {
		    fprintf (stderr, "Error %d returned by ic_echam4()\n", status);
		    return status;
		  }		
		
		  /* write results to final destination */
		  if (icld3d->cldproperties==CLD_LWCREFF)
		    icld3d->ext[iz][ix][iy] = lay_dtau[0]/1000.;

		  icld3d->ssa[iz][ix][iy] = lay_ssa[0];
		  icld3d->g1 [iz][ix][iy] = lay_g1 [0];
		  icld3d->g2 [iz][ix][iy] = 0;
		  icld3d->ff [iz][ix][iy] = 1.0;
		  icld3d->f [iz][ix][iy] = lay_f[0];
		  break;   

		case ICPROP_BAUM:
		case ICPROP_BAUMDETAILED:  		 
		case ICPROP_MIE:
		case ICPROP_FILE:	  
		  /* interpolate extinction and absorption coefficient 
		     of each box to given effective radius             */
		  i1 = (int) ((icld3d->reff[iz][ix][iy] - icld3d->phase_cld.r0) / icld3d->phase_cld.dr);
		
		  /* careful with rounding errors at the array boundaries */
		  i1 = (i1>icld3d->phase_cld.n-1?icld3d->phase_cld.n-1:i1);
		  i1 = (i1<0?0:i1);
                  
		  /* BM, changed interpolation: now interpolating scattering coefficient */
		  /* instead of single scattering albedo                                 */

		  if (i1==icld3d->phase_cld.n-1) {
		    extinc  = icld3d->phase_cld.ext[i1];
		    scattr  = icld3d->phase_cld.sca[i1];
		    dscale  = icld3d->phase_cld.dscale[i1]; 
		    f       = icld3d->phase_cld.f[i1];
		  }		    
		  else {
		    delta_r = icld3d->reff[iz][ix][iy] - (icld3d->phase_cld.r0 + (double) i1 * icld3d->phase_cld.dr);

		    extinc  = icld3d->phase_cld.ext[i1] + icld3d->phase_cld.dext[i1] * delta_r;
		    scattr  = icld3d->phase_cld.sca[i1] + icld3d->phase_cld.dsca[i1] * delta_r;
		    
		    /* ??? DDDDD interpolate the delta scaling factor f   DDDDD ??? */
		    if (icld3d->phase_cld.f[i1]!=0 || icld3d->phase_cld.f[i1+1]!=0)
		      f = interpolate_f (&(icld3d->phase_cld), i1, i1+1, delta_r);
		    else
		      f = 0;

		    if (f<0 || f>1) {
		      fprintf (stderr, "Error in %s, line %d, function %s:\n", __FILE__, __LINE__, __func__);
		      fprintf (stderr, "delta-scaling factor %f out of range!\n", f);
		      fprintf (stderr, "Tried to interpolate between radii %f and %f,\n", 
			       icld3d->phase_cld.r0 + (double) i1 * icld3d->phase_cld.dr,
			       icld3d->phase_cld.r0 + (double) (i1+1) * icld3d->phase_cld.dr);
		      fprintf (stderr, "f values %f and %f\n", 
			       icld3d->phase_cld.f[i1], icld3d->phase_cld.f[i1+1]);
		      return -1;
		    }

		    /* ??? DDDDD interpolate the delta scaling factor dscale DDDDD ??? */
		    /* dscale  = icld3d->phase_cld.dscale[i1] + icld3d->phase_cld.ddscale[i1] * delta_r; */
		    if (icld3d->phase_cld.dscale[i1]!=0 || icld3d->phase_cld.dscale[i1+1]!=0)
		      dscale = interpolate_dscale (&(icld3d->phase_cld), i1, i1+1, delta_r);
		    else 
		      dscale = 0;

		    if (dscale<0 || dscale>1) {
		      fprintf (stderr, "Error in %s, line %d, function %s:\n", __FILE__, __LINE__, __func__);
		      fprintf (stderr, "delta-scaling factor %f out of range!\n", dscale);
		      fprintf (stderr, "Tried to interpolate between radii %f and %f,\n", 
			       icld3d->phase_cld.r0 + (double) i1 * icld3d->phase_cld.dr,
			       icld3d->phase_cld.r0 + (double) (i1+1) * icld3d->phase_cld.dr);

		      fprintf (stderr, "f values %f and %f\n", 
			       icld3d->phase_cld.dscale[i1], icld3d->phase_cld.dscale[i1+1]);
		      return -1;
		    }
		  }

		  /* single scattering albedo */
		  omega   = scattr / extinc;


		  /* extinction coefficient */     
		  if (icld3d->cldproperties==CLD_LWCREFF)
		    icld3d->ext[iz][ix][iy] = icld3d->lwc[iz][ix][iy] * extinc / 1000.0;
		 
		  /* single scattering albedo */
		  icld3d->ssa[iz][ix][iy] = omega;
                  
                  /* external delta-scaling factor CE */
                  icld3d->f[iz][ix][iy]=f;  

		  /* internal delta-scaling factor TZ */
		  icld3d->dscale[iz][ix][iy] = dscale;
                  
		  break;
                  

                default:
		  fprintf (stderr, "Error, unknown ice cloud property %d in convert_ic_mystic (II). This\n", input.ic.properties);
		  fprintf (stderr, "is not supposed to happen and indicates a coding error.\n");
		
		  return -1;
		}
                
                /* CE: Apply delta scaling to extinction. Here we assume that the specified input 
                   files correspond to unscaled extinction */
                if( icld3d->cldproperties!=CLD_LWCREFF){
                  ssa_us = icld3d->ssa[iz][ix][iy] /
                    ( 1.0 + icld3d->f [iz][ix][iy] * ( icld3d->ssa[iz][ix][iy] - 1.0));
                  icld3d->ext[iz][ix][iy] *= (1.0 -  ssa_us * icld3d->f [iz][ix][iy]);
                }
                
	      }
	    } 
	  }
	} /* if (icld3d->threed[iz]) */
      } /* if (icld3d->nthreed>0)  */
      
      
      /* for a 1D cloud we only need to interpolate the delta scaling factor;  */
      /* this is a bit unfortunate but we need to do that here because         */
      /* all other properties are already interpolated in setup_icloud()       */
      /* where dscale is not yet known                                         */
    
      if (output->ic.microphys.lwc_layer[iz] > 0) { /* if (output->ic.microphys.lwc_layer[iz] > 0) */

	switch (input.ic.properties) {
	case ICPROP_FU:
	case ICPROP_ECHAM4:
	case ICPROP_KEY:
	case ICPROP_YANG:

	  /* no need to do anything because delta scaling only for */
	  /* non-HG phase functions                                */

	  break;

	case ICPROP_MIE:
	case ICPROP_FILE:
	case ICPROP_BAUM:
	case ICPROP_BAUMDETAILED:
	
	  /* determine index of effective radius */
	  i1 = (int) ((output->ic.microphys.effr_layer[iz] - icld3d->phase_cld.r0) / icld3d->phase_cld.dr);
	
	  /* careful with rounding errors at the array boundaries */
	  i1 = (i1>icld3d->phase_cld.n-1?icld3d->phase_cld.n-1:i1);
	  i1 = (i1<0?0:i1);
	
	  /* ??? DDDDD interpolate delta-scaling factor DDDDD ??? */
	  if (icld3d->phase_cld.dscale[i1]!=0 || icld3d->phase_cld.dscale[i1+1]!=0) {
	    delta_r = output->ic.microphys.effr_layer[iz] - (icld3d->phase_cld.r0 + (double) i1 * icld3d->phase_cld.dr);
	    dscale = interpolate_dscale (&(icld3d->phase_cld), i1, i1+1, delta_r);
	  }
	  else
	    dscale = 0;

	  if (dscale<0 || dscale>1) {
	    fprintf (stderr, "Error in %s, line %d, function %s:\n", __FILE__, __LINE__, __func__);
	    fprintf (stderr, "delta-scaling factor %f out of range!\n", dscale);
	    fprintf (stderr, "Tried to interpolate between radii %f and %f,\n", 
		     icld3d->phase_cld.r0 + (double) i1 * icld3d->phase_cld.dr,
		     icld3d->phase_cld.r0 + (double) (i1+1) * icld3d->phase_cld.dr);

	    fprintf (stderr, "f values %f and %f\n", 
		     icld3d->phase_cld.dscale[i1], icld3d->phase_cld.dscale[i1+1]);
	    return -1;
	  }

	  output->ic.optprop.dscale[iv][iz] = dscale;

	  /*
	  output->ic.optprop.dscale[iv][iz] = icld3d->phase_cld.dscale[i1] + icld3d->phase_cld.ddscale[i1] * 
	      (output->ic.microphys.effr_layer[iz] - (icld3d->phase_cld.r0 + (double) i1 * icld3d->phase_cld.dr));
	  */

	  break;

	default:
	  fprintf (stderr, "Error, unknown ice cloud properties %d\n", input.ic.properties);
	  return -1;
	}
      
	break;
      }
    } /* for (iz=0; iz<icld3d->nlyr; iz++) */
    break;
    
  default:
    fprintf (stderr, "Error, unknown ice cloud properties %d\n", icld3d->cldproperties);
    return -1;
  }
  
  if ((status = ASCII_free_float (wvl_dtau,18)) != 0) {
    fprintf(stderr,"Error in %s (%s),\n", function_name, file_name);
    fprintf(stderr,"      free memory of 'wvl_dtau'\n");
    return -1;
  }

  if ((status = ASCII_free_float (wvl_g1,18)) != 0) {
    fprintf(stderr,"Error in %s (%s),\n", function_name, file_name);
    fprintf(stderr,"      free memory of 'wvl_g1'\n");
    return -1;
  }

  if ((status = ASCII_free_float (wvl_ssa,18)) != 0) {
    fprintf(stderr,"Error in %s (%s),\n", function_name, file_name);
    fprintf(stderr,"      free memory of 'wvl_ssa'\n");
    return -1;
  }
  
  
  return 0;
}





/* Free 3D water cloud optical properties */

int free_wc_mystic (input_struct input, cld3d_out_struct *wcld3d)
{
  switch (wcld3d->cldproperties) {
  case CLD_OPTPROP:
    break;
    
  case CLD_EXTREFF:
  case CLD_LWCREFF:
    
    /* free phase function table */
    switch (input.wc.properties) {
    case WCPROP_HU:
    case WCPROP_ECHAM4:
      break;  
      
    case WCPROP_MIE:
    case WCPROP_FILE:
    case WCPROP_PHASE:
      
      free_phase_function_table (&(wcld3d->phase_cld));
      break;
      
    default:
      fprintf (stderr, "Error, unknown water cloud property %d in free_wc_mystic. This is not\n", input.wc.properties);
      fprintf (stderr, "supposed to happen and indicates a coding error.\n");
      
      return -1;
      break;
    }
  }
  
  return 0;
}


/* Free 3D ice cloud optical properties */

int free_ic_mystic (input_struct input, cld3d_out_struct *icld3d)
{
  switch (icld3d->cldproperties) {
  case CLD_LWCREFF:

    /* free phase function table */
    switch (input.ic.properties) {
 
    case ICPROP_YANG:
    case ICPROP_KEY:
    case ICPROP_ECHAM4:
    case ICPROP_FU:
      break;

    case ICPROP_MIE:
    case ICPROP_FILE:
    case ICPROP_BAUM:
    case ICPROP_BAUMDETAILED:
   
      free_phase_function_table (&(icld3d->phase_cld));
      break;
      
    default:
      fprintf (stderr, "Error, unknown ice cloud property %d in free_ic_mystic. This is not\n", input.ic.properties);
      fprintf (stderr, "supposed to happen and indicates a coding error.\n");
      
      return -1;
      break;
    }
  }

  return 0;
}
  

/* check if two floats are equal - still not optmical, but how to improve? */

static inline int float_equal (float a, float b)
{
  
  if (a==0 || b==0) {
    if (fabs(a-b) < MC_EPSILON)
      return 1;
  }
  else {  /* relative difference smaller than MC_EPSILON */
    if (fabs(a-b) < MC_EPSILON * fabs(a))
      return 1;
  }

  return 0;
}


/* ??? DDDDD interpolate the delta scaling factor f     DDDDD ??? */
/* Interpolate the external delta scaling factor; the idea is to interpolate the      */
/* scaled and unscaled extinction and scattering coefficients and to calculate the    */
/* delta scaling factor from those - this seems more appropriate than a simple linear */
/* interpolation of the delta scaling factor. In the phase function table we          */
/* store the scaled optical properties (with respect to the external delta scaling    */
/* factor) and need to calculate the unscaled ones.                                   */
 
static float interpolate_f (phase_function_table *p, int i1, int i2, double delta_r)
{
  double f1=0, f2=0, f=0;
  double om1=0, om2=0, om1_unscaled=0, om2_unscaled=0, om_unscaled=0;
  double ext1=0, ext2=0, ext=0, ext1_unscaled=0, ext2_unscaled=0, ext_unscaled=0;
  double sca_unscaled=0;

  om1 = p->sca[i1] / p->ext[i1];
  om2 = p->sca[i2] / p->ext[i2];
  
  f1  = p->f[i1];
  f2  = p->f[i2];
  
  /* compare linear and improved interpolation for debugging purposes */
  fprintf (stderr, " ... f linear %.6f ", f1 + delta_r / p->dr * (f2 - f1));

  om1_unscaled = om1 / (1.0 - f1 + f1*om1);
  om2_unscaled = om2 / (1.0 - f2 + f2*om2);
  
  ext1 = p->ext[i1];
  ext2 = p->ext[i2];
  
  ext1_unscaled = ext1 / (1.0 - f1*om1_unscaled);
  ext2_unscaled = ext2 / (1.0 - f2*om2_unscaled);
  
  /* interpolate scaled and unscaled extinction as well as unscaled single scattering albedo */
  ext          = ext1          + delta_r / p->dr * (ext2 - ext1);
  ext_unscaled = ext1_unscaled + delta_r / p->dr * (ext2_unscaled - ext1_unscaled);
  sca_unscaled = ext1_unscaled*om1_unscaled + delta_r / p->dr * (ext2_unscaled*om2_unscaled - ext1_unscaled*om2_unscaled);
  om_unscaled  = sca_unscaled / ext_unscaled;
  
  f = (1.0 - ext / ext_unscaled) / om_unscaled;
  fprintf (stderr, "vs. improved %.6f interpolation\n", f);
  
  return f;
}


/* ??? DDDDD interpolate the delta scaling factor dscale   DDDDD ??? */
/* Interpolate the internal delta scaling factor; the idea is to interpolate the      */
/* scaled and unscaled extinction and scattering coefficients and to calculate the    */
/* delta scaling factor from those - this seems more appropriate than a simple linear */
/* interpolation of the delta scaling factor. In the phase function table we          */
/* store the unscaled optical properties (with respect to the internal delta scaling  */
/* factor) and need to calculate the scaled ones.                                     */

static float interpolate_dscale (phase_function_table *p, int i1, int i2, double delta_r)
{
  double ext1=0, ext2=0, ext=0, ext1_scaled=0, ext2_scaled=0, ext_scaled=0;
  double om1=0, om2=0, om=0;
  double dscale1=0, dscale2=0;
  double sca=0;
  double dscale=0;


  ext1 = p->ext[i1];
  ext2 = p->ext[i2];

  om1 = p->sca[i1] / p->ext[i1];
  om2 = p->sca[i2] / p->ext[i2];

  dscale1 = p->dscale[i1];
  dscale2 = p->dscale[i2];

  /* compare linear and improved interpolation for debugging purposes */
  fprintf (stderr, " ... dscale linear %.6f ", dscale1 + delta_r / p->dr * (dscale2 - dscale1));

  ext1_scaled = (1.0 - dscale1*om1) * ext1;
  ext2_scaled = (1.0 - dscale2*om2) * ext2;

  ext        = ext1        + delta_r / p->dr * (ext2 - ext1);
  ext_scaled = ext1_scaled + delta_r / p->dr * (ext2_scaled - ext1_scaled);
  sca = ext1*om1 + delta_r / p->dr * (ext2*om2 - ext1*om2);
  om = sca/ext;
  
  dscale = (1.0 - ext_scaled / ext) / om;
  fprintf (stderr, "vs. improved %.6f interpolation\n", dscale);

  return dscale; 
}

