/*--------------------------------------------------------------------
 * $Id: mie.c,v 1.33 2008/10/29 16:58:18 claudia Exp $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 * 
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "ascii.h"
#include "mie.h"
#include "numeric.h"
#include "getopt.h"

#if HAVE_LIBNETCDF
#include <netcdf.h>
#endif


#define PI 3.1415926535897932
#define MAXLINE 255


int    setup_wlgrid_and_read_ref_index(input_struct input, output_struct *output, 
                                       double **ri_real, double **ri_cplx, 
                                       int *lambda_i_start, int *lambda_i_end);
int    calc_distribution_radius_grid  (input_struct input, output_struct *output, int iv);
int    calc_distribution_number_dens  (input_struct input, output_struct *output, int iv, int ir);
double GAMMA_LN(float xx);
double double_min(double x, double y);
double double_max(double x, double y);
int    write_output (input_struct input, output_struct *output, int ir, int iv);
int    make_reff (float *LWC, int n, float dropconc, int distribution, float alpha, float **r_eff);
int mie_calc_sizedist_optimised (mie_inp_struct input, mie_out_struct *output, 
                                 int program, int medium, mie_complex crefin, 
                                 float wavelength, float temperature, int nstokes, 
                                 double *x_size, double *y_size, int n_size,
                                 double *beta, double *omega, double *g,
                                 mie_complex *ref, 
                                 int ir, float **qext_array, float **qsca_array, 
                                 float **gsca_array, float ****pmom_array);

int aerosol_input_data(input_struct input, output_struct *output);
int aerosol_size_distribution(input_struct input, output_struct *output, int iv, int ih);


int aerosol_refractive_index(input_struct input, output_struct *output, 
                             double wavelength, int ih, float *ri_re,
                             float *ri_im);

#if HAVE_LIBNETCDF
int calloc_netcdf_var(input_struct input, output_struct *output);
int generate_netcdf_var(input_struct input, output_struct *output, int ir,
                        int iv);
int write_output_netcdf(input_struct input, output_struct *output, int iv);
#endif

int mie (input_struct input, output_struct *output)
{
 
  int ip=0, im=0, ir=0, iv=0, status=0;
  int first=1;

  double wavelength=0;

  float *qext_array;
  float *qsca_array;
  float *gsca_array;
  float ***pmom_array;  /* Phase function moments */
  float **tmp_pmom;
  
  double *ri_real;
  double *ri_cplx;

  double x, ext, sca, vol;
  int lambda_i_start=0;
  int lambda_i_end=0;
  double norm=0.0;

  /* Maximum allowed number of radii in aerosol size distributions */
  int nmax=5000; 
  
  output->ref.re = 0.0;
  output->ref.im = 0.0; 

  /* Allocate memory for Legendre polynomials of all phase matrix elements */
  output->mie.pmom = calloc (4, sizeof(float *));
  for (ip=0; ip<4; ip++)
    output->mie.pmom[ip] = calloc (input.mie.nmom+1, sizeof(float));

  tmp_pmom     = calloc (4, sizeof(double *));
  for (ip=0; ip<4; ip++) 
    tmp_pmom [ip] = calloc (input.mie.nmom+1, sizeof(double));
  
  /* setup wavelength grid and            */
  /* read refraction index file, if given */
  /* For aerosols the refractive index is read from the OPAC datanbase later on 
     (not in this function). */
  status = setup_wlgrid_and_read_ref_index(input, output, &ri_real, &ri_cplx, &lambda_i_start, &lambda_i_end);
  if (status != 0) {
    fprintf (stderr, "Error %d returned by setup_wlgrid_and_read_ref_index()\n", status);
    return status;
  }
  

  /* determine the number of r_eff to be simulated */
  switch(input.distribution) {
  case DIST_NONE:
  case DIST_GAMMA:
  case DIST_LOGNORMAL:
    /* more than one distribution (==r_eff) can be calculated with one call */
    /* determine number of r_eff to be calculated */
    if (input.r_eff_step != 0.0)
      /* 1.E-7 rouding error */
      output->n_r_eff = (int) floor((input.r_eff_max - input.r_eff_min+1.E-7) / input.r_eff_step); 
    else
      output->n_r_eff = 0;

    /* setup r_eff grid */  
    output->r_eff = (double *) calloc (output->n_r_eff+1, sizeof(double));
    for (ir=0; ir<=output->n_r_eff; ir++)
      output->r_eff[ir] = input.r_eff_min + ir*input.r_eff_step;
    break;
  case DIST_FILE:
    /* size distribution file == one r_eff */
    output->n_r_eff = 0;
    output->r_eff = (double *) calloc (output->n_r_eff+1, sizeof(double));
    output->r_eff[0] = -999.0;
    break;
  case DIST_AER:
    /* Read aerosol data from OPAC file */
    status = aerosol_input_data(input, output);
    if (status!=0) {
      fprintf (stderr, "Error %d returned by aerosol_input_data()\n", status);
      return status;
    }
    /* Effective radius is "replaced" by relative humidity for 
       aerosols in the output data. */
    output->n_r_eff=output->aerdistr.n_hum-1;
    output->r_eff = (double *) calloc (output->n_r_eff+1, sizeof(double));
    for (ir=0; ir<=output->n_r_eff; ir++)
      output->r_eff[ir]=output->aerdistr.rel_hum[ir];
    /* Maximum allowed radii in size distribution. This is necessary because some 
       arrays are allocated before the relative humidity loop, where the size distribution
       can actually be calculated. Otherwise aerosol would require a completely 
       new function. */
    output->dist.n_dens=nmax;
    output->dist.radius = calloc(output->dist.n_dens, sizeof(double));
    output->dist.number_dens = calloc(output->dist.n_dens, sizeof(double));
    break; 
  default:
    fprintf (stderr, "Error, distribution number in mie.c %d\n", input.distribution);
    return -3;
    break;
  }
  
#if HAVE_LIBNETCDF
  if(input.output_format == OUTPUT_FORMAT_NETCDF){
    status = calloc_netcdf_var(input, output);
    if (status!=0) {
      fprintf (stderr, "Error %d returned by calloc_netcdf_var()\n", status);
      return status;
    }
  }
#else
  fprintf (stderr, " ***********************************************************************\n");
  fprintf (stderr, " * You have built mie without libnetcdf and hence cannot               *\n");
  fprintf (stderr, " * use the option *output_user netcdf*. Please get netcdf and rebuild. *\n");
  fprintf (stderr, " ***********************************************************************\n");
  return -1;
#endif
    
  output->warnings = calloc (output->wl.nlambda+1, sizeof(int *));
  for (iv=0; iv<=output->wl.nlambda; iv++)
    output->warnings[iv] = calloc (output->n_r_eff+1, sizeof(int));
  
  /* call the mie programms, which will do the main work */
  /* starting lambda-loop */
  for (iv=lambda_i_start; iv<=lambda_i_end; iv++) {
    
    fprintf (stderr, " ... start Mie-calculation for lambda = %10.3f nm  \n", 
               output->wl.lambda[iv]);

    /* change wavelength dependent properties */
    wavelength = (output->wl.lambda[iv]) * 1.e-03;  /* Convert from nm to micro meter */
    if (strlen (input.ri_filename)>0) {
      input.crefin.re = ri_real[iv];
      input.crefin.im = ri_cplx[iv];
    }
   
    /* Aerosol requires  different treatment from clouds, size distribution calculated below
       in loop over relative humudities*/
    if (input.distribution != DIST_AER){
      /* generate the grid of the droplet distribution */
      status = calc_distribution_radius_grid (input, output, iv); 
      if (status!=0) {
        fprintf (stderr, "Error %d returned by make_distribution()\n", status);
        return status;
      }
    }
    
    /* ??? qext_array = calloc(4, sizeof(float)); ??? */
    
    qext_array = (float *) calloc (output->dist.n_dens, sizeof(float));
    qsca_array = (float *) calloc (output->dist.n_dens, sizeof(float));
    gsca_array = (float *) calloc (output->dist.n_dens, sizeof(float));
    
    
    pmom_array = calloc (4, sizeof(float **));
    
    for (ip=0; ip<4; ip++) {
      pmom_array[ip] = calloc (input.mie.nmom+1, sizeof(float *));
      for (im=0; im<input.mie.nmom+1; im++)
           pmom_array[ip][im] = calloc (output->dist.n_dens, sizeof(float));
    }
    
    /* Loop over effective radii. relative humidities (aerosol)*/
    for (ir=0; ir<=output->n_r_eff; ir++) {
      
      if (input.distribution != DIST_AER){
        /* number density distribution per radius intervall,*/ 
        /* unfortunately also wavelength dependent          */
        status = calc_distribution_number_dens (input, output, iv, ir); 
        if (status!=0) {
          fprintf (stderr, "Error %d returned by make_distribution()\n", status);
          return status;
        }
      }
      else{
        /* read aerosol refractive index and interpolate on wavelength */
        status = aerosol_refractive_index(input, output, wavelength, 
                                          ir, &input.crefin.re,
                                          &input.crefin.im);
        if (status!=0) {
          fprintf (stderr, "Error %d returned by aerosol_refractive_index()\n",
                   status);
          return status;
        }
        
        fprintf(stderr, " ... aerosol refractive index %g %g \n", 
                input.crefin.re, input.crefin.im);
        /* Set up aerosol size distribution radius grid and calculate
           number densities.*/ 
        status = aerosol_size_distribution(input, output, iv, ir);
        if (status!=0) {
          fprintf (stderr, "Error %d returned by aerosol_size_distribution()\n",status);
          return status;
        }
        /* refractive index for the required relative humidity */
        output->rho_medium=output->aerdistr.rho[ir];
      } 
     
      switch(input.distribution) {
      case DIST_NONE:
          /* one single radius */
        status = mie_calc (input.mie, &(output->mie), 
                           input.program, input.medium, input.crefin,
                           wavelength, output->r_eff[ir], input.temperature, input.nstokes,
                           &(output->ref));
        if (status!=0) {
          fprintf (stderr, "Error %d returned by mie_calc()\n", status);
          return status;
        }
        /* calculate beta and omega */
        x = (2.0 * PI * output->r_eff[ir]) / wavelength; /* (micro m / micro m) */
        ext = PI * x*x * output->mie.qext;
        sca = PI * x*x * output->mie.qsca;
        vol = (4.0/3.0) * PI * pow((output->r_eff[ir]*1E-6),3.0); /* micro m -> m */
        output->beta  = ext / vol;
        output->omega = sca / ext;
        break;
      case DIST_FILE:
      case DIST_GAMMA:
      case DIST_LOGNORMAL:
        /* droplet radius distribution */
        status = mie_calc_sizedist_optimised (input.mie, &(output->mie), 
                                              input.program, input.medium, input.crefin,
                                              wavelength, input.temperature, input.nstokes, output->dist.radius, 
                                              output->dist.number_dens, output->dist.n_dens,
                                              &output->beta, &output->omega, &output->g, &(output->ref), 
                                              ir, &qext_array, &qsca_array, &gsca_array, &pmom_array);
        if (status!=0) {
          fprintf (stderr, "Error %d returned by mie_calc_sizedist_optimised()\n", status);
          return status;
        }
        break;
      case DIST_AER:
        status =  mie_calc_sizedist (input.mie, &(output->mie), 
                                     input.program, input.medium, input.crefin, 
                                     wavelength, input.temperature, input.nstokes,
                                     output->dist.radius, output->dist.number_dens,
                                     output->dist.n_dens, &output->beta, &output->omega,
                                     &output->g, &(output->ref));
        
        /* Reset number of radii in size distribution.*/
        output->dist.n_dens=nmax;
        break;
      default:
        fprintf (stderr, "Error, distribution number in mie.c %d\n", input.distribution);
        return -4;
        
        break;
      }

      if (input.mie.nmom > 0){
        for (ip=0; ip<input.nstokes; ip++) 
          for (im=0; im<input.mie.nmom +1; im++)
            tmp_pmom[ip][im]=output->mie.pmom[ip][im]; 
        
        if(input.nstokes == 4){
          /* Convert from Wiscombe definition to Evans definition 
             Allocate memory for Legendre polynomials of all phase 
             matrix elements */
          norm = tmp_pmom[0][0]+tmp_pmom[1][0];
          
          for (im=0; im<input.mie.nmom +1; im++){ 
            output->mie.pmom[0][im]=  
              (tmp_pmom[0][im]+tmp_pmom[1][im])/norm ;        
            output->mie.pmom[1][im]=  
              (tmp_pmom[1][im]-tmp_pmom[0][im])/norm;  
            output->mie.pmom[2][im]= 2.*tmp_pmom[2][im]/norm;        
            output->mie.pmom[3][im]= 2.*tmp_pmom[3][im]/norm;      
          }
        }
        else if(input.nstokes==1)
          for (im=0; im<input.mie.nmom +1; im++)
            output->mie.pmom[0][im]=tmp_pmom[0][im]/tmp_pmom[0][0];
        else{
          fprintf(stderr, "Error, nstokes must be either 1 or 4\n");
          return -1;
        }
      }
      
      
      if(input.output_format != OUTPUT_FORMAT_NETCDF){
        /* write output */
        status = write_output (input, output, ir, iv);
        if (status!=0) {
          fprintf (stderr, "Error %d returned by write_output()\n", status);
          return status;
        }
      }
#if HAVE_LIBNETCDF
      else{
        status = generate_netcdf_var(input, output, ir, iv);
        if (status!=0) {
          fprintf (stderr, "Error %d returned by generate_netcdf_var()\n", status);
          return status;
        }
      }
#endif

        /* for debugging: print cumulative probability distribution      */
        /* to 'cumulative.dat':                                          */
        /* cumulative_probability (output->mie.pmom[0], input.mie.nmom); */
      
        /* for debugging: print phase function to 'phase.dat':           */
        /* phase_function (output->mie.pmom[0], input.mie.nmom);         */
      
    } /* for (ir=0; ir<=output->n_r_eff; ir++) { */ 
    
#if HAVE_LIBNETCDF
    if(input.output_format == OUTPUT_FORMAT_NETCDF){
      write_output_netcdf(input, output, iv);
      if (status!=0) {
        fprintf (stderr, "Error %d returned by write_output_netcdf()\n", status);
        return status;
      }
    }
#endif
    
    for (ip=0; ip<4; ip++) {
      for (im=0; im<input.mie.nmom+1; im++)
	free(pmom_array[ip][im]);
      free(pmom_array[ip]);
    }
    free(pmom_array);
    
     /* free distribution dependent properties */
    free(qext_array);
    free(qsca_array);
    free(gsca_array);

  }/* loop over iv */
  
   /* for (iv=0; iv<=output->wl.nlambda; iv++) { ...*/
  for (ip=0; ip<4; ip++) 
    free(tmp_pmom[ip]);
  free(tmp_pmom);
  
  /* write warnings */
  if (input.output_format == OUTPUT_FORMAT_CLOUDPROP ||
      input.output_format == OUTPUT_FORMAT_AERPROP ||
      input.output_format == OUTPUT_FORMAT_NETCDF) {
    for (iv=0; iv<=output->wl.nlambda; iv++) {
      for (ir=0; ir<=output->n_r_eff; ir++) {
        if (output->warnings[iv][ir]==1) {
          if (first == 1) { /* true */
            fprintf (stderr, "*** Warning: maxmimum accuracy of the phase function not reached\n");
              fprintf (stderr, "*** with %4d Legendre coefficients.\n",input.mie.nmom);
              fprintf (stderr, "*** In order to get more accurate results\n"); 
              fprintf (stderr, "*** please increase 'nmom' in the mie input file!\n"); 
              first = 0;
          }
          if(input.medium !=AEROSOL)
            fprintf (stderr, "          lambda = %10.3f nm, r_eff = %8.3f micro meter\n", output->wl.lambda[iv], output->r_eff[ir]);
          else
            fprintf (stderr, "          lambda = %10.3f nm, rel_hum = %2f percent\n", output->wl.lambda[iv], output->aerdistr.rel_hum[ir]);
        }
      } /* for (ir=0; ir<=output->n_r_eff; ir++) { */ 
    } /* for (iv=0; iv<=output->wl.nlambda; iv++) { ...*/
  }
  
  /*free memory */
  for (iv=0; iv<=output->wl.nlambda; iv++)
    free(output->warnings[iv]);  
  free(output->warnings);
  
  return 0;
}


int setup_wlgrid_and_read_ref_index(input_struct input, output_struct *output, 
                                    double **ri_real, double **ri_cplx, 
                                    int *lambda_i_start, int *lambda_i_end)
{  
  int status = 0;
  int iv=0;
  char *wvl_filename=NULL; 

  int min_columns=0;
  int max_columns=0;
  int max_length=0;

  char ***string=NULL;
  char *dummy=NULL;

  int unit_factor=1;    /* specifies if wvl contain micro meter or nm */

  char function_name[]="setup_wlgrid_and_read_ref_index";
  char file_name[]="mie.c";

  if (strlen(input.data_files_path) == 0)
    input.data_files_path="../data/"; 

  /* read refractive index */
  if (strlen (input.ri_filename)>0) {
    status = read_3c_file (input.ri_filename, 
			   &output->wl.lambda, 
                           ri_real, ri_cplx, 
                           &output->wl.nlambda);
    
    if (status != 0) {
      fprintf (stderr, "error %d reading %s\n", status, input.ri_filename);
      return status;
    }

    fprintf (stderr, " ... read %d data points from %s\n", 
	     output->wl.nlambda, input.ri_filename);

    
    /* check if data points are sorted in ascending order */
    for (iv=0; iv<output->wl.nlambda-1; iv++)
      if (output->wl.lambda[iv+1] < output->wl.lambda[iv]) {
	fprintf (stderr, "error: %s not sorted in ascending order\n", 
		 input.ri_filename);
	return -1;
      }

    /* special convention here */
    output->wl.nlambda--;
  }
  else {
    if (input.ck_scheme == CK_NONE) {
      if (input.wl.wl_step != 0.0)
        output->wl.nlambda = (int) (input.wl.wl_end-input.wl.wl_start) / input.wl.wl_step;
      else
        output->wl.nlambda = 0;

      output->wl.lambda = (double *) calloc (output->wl.nlambda+1, sizeof(double));
      for (iv=0; iv<=output->wl.nlambda; iv++)
        output->wl.lambda[iv] = input.wl.wl_start + input.wl.wl_step*iv;
    }
    else {

      if (( wvl_filename = calloc (FILENAME_MAX,sizeof(char)))==NULL) {
        fprintf (stderr, "Error: Allocation of wvl_filename in %s (%s)\n", function_name, file_name);
        return -1;
      }
  
      strcpy(wvl_filename,input.data_files_path);

      switch(input.ck_scheme) {
      case CK_KATO:
        strcat(wvl_filename,"correlated_k/kato/wvl.dat");
        fprintf (stderr, " ... Kato wavelength bands\n");
        unit_factor=1;  /* Kato wvl.dat contains nano meter */
        break;
      case CK_KATO2:
        strcat(wvl_filename,"correlated_k/kato2/wvl.dat");
        fprintf (stderr, " ... Kato2 wavelength bands\n");
        unit_factor=1;  /* Kato2 wvl.dat contains nano meter */
        break;
      case CK_KATO2_96:
        strcat(wvl_filename,"correlated_k/kato2.hitran96/wvl.dat");
        fprintf (stderr, " ... Kato2_96 wavelength bands\n");
        unit_factor=1;  /* Kato2_96 wvl.dat contains nano meter */
        break;
      case CK_FU:
        strcat(wvl_filename,"correlated_k/fu/wvl.dat");
        fprintf (stderr, " ... Fu wavelength bands\n");
        unit_factor=1000;  /* Fu wvl.dat contains micro meter */
        break;
      case CK_AVHRR_KRATZ:
        strcat(wvl_filename,"correlated_k/kratz/avhrr_wvl.dat");
        fprintf (stderr, " ... AVHRR-Kratz wavelength bands\n");
        unit_factor=1000;  /* AVHRR-Kratz wvl.dat contains micro meter */
        break;
      case CK_GENERIC:
        strcpy (wvl_filename, input.filename_ck_generic);
        fprintf (stderr, " ... generic correlated_k wavelength bands\n");
	break;
      default:
        fprintf (stderr, "Error: unsupported correlated-k scheme, nr %2d\n", input.ck_scheme);
        return -1;
      }

      status = ASCII_checkfile(wvl_filename, &(output->wl.nlambda), &(min_columns), &(max_columns), &(max_length));
      if (status!=0) {
        fprintf(stderr,"Error %d reading wavelength_file %s in %s (%s) \n", status, wvl_filename, function_name, file_name);
        return status;
      }

      if ( min_columns < 2 ) {
        fprintf(stderr,"Error, wavelength_file %s must have at least two columns! \n", wvl_filename);
        return -1;
      }

      if ((status = ASCII_calloc_string(&string, output->wl.nlambda, max_columns, FILENAME_MAX+1)) != 0) {
        fprintf(stderr,"Error %d allocating string in %s (%s)\n", status, function_name, file_name);
        return status;
      }

      if ((status = ASCII_readfile(wvl_filename, string)) != 0) {
        fprintf(stderr,"Error %d reading wavelength-file: %s in %s (%s) \n", status, wvl_filename, function_name, file_name);
        return status;
      }

      if ((output->wl.lambda = (double *) calloc (output->wl.nlambda, sizeof(double))) == NULL) {
        fprintf(stderr,"Error %d allocating output->wl.lambda in %s (%s) \n", status, function_name, file_name);
        return status;
      }

      /* second line is center wavelength */
      for (iv=0; iv<output->wl.nlambda; iv++) {
        output->wl.lambda[iv] = (double) strtod (string[iv][1], &dummy) * unit_factor;
      }

      input.wl.wl_start = output->wl.lambda[0];
      input.wl.wl_end   = output->wl.lambda[output->wl.nlambda-1];

      /* special convention here */
      output->wl.nlambda--;

      for (iv=0; iv<=output->wl.nlambda; iv++)
        fprintf (stderr, "   lambda[%3d] = %10.3f nm\n", iv+1, output->wl.lambda[iv]);
    }
  }

  /* setup of wavelength range */
  if (input.lambda_i_start == -999)  
    (*lambda_i_start) = 0;
  else {
    if (input.lambda_i_start < 1) {
      fprintf(stderr,"  wavelength_index_start = %d has to be 1 or larger \n",input.lambda_i_start);
      status=-1;
    }
    if (output->wl.nlambda < input.lambda_i_start) {
      fprintf(stderr,"  wavelength_index_start = %d, but there are only %d wavelength bands \n",
                        input.lambda_i_start, output->wl.nlambda);
      status=-2;
    }
    (*lambda_i_start) = input.lambda_i_start-1;
  }

  if (input.lambda_i_end == -999)  
    (*lambda_i_end) = output->wl.nlambda;
  else {
    if (input.lambda_i_end < (*lambda_i_start)) {
      fprintf(stderr,"  wavelength_index_end = %d has to be larger than wavelength_index_start = %d\n",
                        input.lambda_i_end, (*lambda_i_start));
      status=-3;
    }
    if (output->wl.nlambda < input.lambda_i_end-1) {
      fprintf(stderr,"  wavelength_index_end = %d, but there are only %d wavelength bands \n",
                        input.lambda_i_end+1, output->wl.nlambda+1);
      status=-4;
    }
    (*lambda_i_end) = input.lambda_i_end-1;
  }

  return status;
}

int calc_distribution_radius_grid(input_struct input, output_struct *output, int iv)
{
  float radmin, radmax, rad, delta_rad;
  float x, delta_x;
  int is;
  int status=0;
  
  /* fprintf(stderr,"calculate droplet radius distribution\n"); */

  switch(input.distribution) {
  case DIST_NONE:
    /* single radius calculation, nothing to do */
    output->dist.n_dens = 1;
    break;
  case DIST_FILE:
    /* read size distribution from file */
    status = read_2c_file (input.sd_filename, &output->dist.radius, &output->dist.number_dens, &output->dist.n_dens);
    
    if (status != 0) {
      fprintf (stderr, "error %d reading %s\n", status, input.sd_filename);
      return status;
    }

    fprintf (stderr, " ... read %d data points from %s\n", 
	     output->dist.n_dens, input.sd_filename);
    break;
  case DIST_GAMMA:
  case DIST_LOGNORMAL:
    /* ????? are these reasonable values? Evans used 0.02 reff_min - 5 reff_max; ????? */
    /* ????? BM found that 5 is not enough for small particles - but somebody    ????? */
    /* ????? increased the upper boundary to 8 reff_max which is VERY EXPENSIVE  ????? */
    
    radmin = 0.02 * input.r_eff_min; /* cover the whole droplet spectrum   */
    radmax = 8.00 * input.r_eff_max; /* in order to save mie-calculations  */

    /* count number of radii */
    /* the grid of dist.radius is independent of ir, that means output->r_eff[ir] */
    rad = radmin;
    is = 0;
    while (rad <= radmax) {
      x = 2*PI*rad / (output->wl.lambda[iv]*1.e-03);  /* 1E-3 = nm -> micro m */
      delta_x = double_max(0.03,0.03*sqrt(x));
      delta_rad = delta_x * (output->wl.lambda[iv]*1.e-03) / (2*PI);
      rad = rad + delta_rad;
      is = is + 1;
    }
    output->dist.n_dens = is;

    /* callocation */
    if ((output->dist.radius = calloc(output->dist.n_dens, sizeof(double)))==NULL) {
      fprintf (stderr, "Error: Allocation of output->dist.radius (make_distribution)\n");
      return -2;
    }
    if ((output->dist.number_dens = calloc(output->dist.n_dens, sizeof(double)))==NULL) {
      fprintf (stderr, "Error: Allocation of output->dist.numb_dens (make_distribution)\n");
      return -3;
    } 

    /* calculate radius grid */
    rad = radmin;
    is = 0;
    while (rad <= radmax) {
      x = 2*PI*rad / (output->wl.lambda[iv]*1.e-03);  /* 1E-3 = nm -> micro m */
      delta_x = double_max(0.03,0.03*sqrt(x));
      delta_rad = delta_x * (output->wl.lambda[iv]*1.e-03) / (2*PI);
      rad = rad + delta_rad;
      output->dist.radius[is]      = rad;
      is = is + 1;
    }
    break; 
    
  case DIST_AER: 
    /* Do nothing here. Radius grid and number densities are calculated in
       aerosol_size_distribution. */
    break;
  default:
    fprintf (stderr, "Error in make_distribution : unrecognized distribution %d\n", input.distribution);
    break;
  } /* switch(input.distribution) {... */


  /* check if data points are sorted in ascending order */
  switch(input.distribution) {
  case DIST_NONE:
    break;
  case DIST_FILE:
  case DIST_GAMMA:
  case DIST_LOGNORMAL:
    /* for gamma and lognormal this is not nessesary, but it costs almost no time */

    for (is=0; is<output->dist.n_dens-1; is++)
      if (output->dist.radius[is+1]<output->dist.radius[is]) {
        if (input.distribution == DIST_FILE)
          fprintf (stderr, "error: %s not sorted in ascending order\n", 
	                    input.sd_filename);
        else
          fprintf (stderr, "error: radius in distribution not sorted in ascending order, program bug!!!\n");
        return -5;
      }
    
    if (output->dist.radius[0]==0) {
      output->dist.radius[0] = 0.1 * output->dist.radius[1];
      fprintf (stderr, " ... cannot handle r=0; setting first data point to r=%g\n",
	     output->dist.radius[0]);
    }
    break;
  default:
    fprintf (stderr, "Error in make_distribution : unrecognized distribution %d\n", input.distribution);
    break;
  }

  return status;
}

int calc_distribution_number_dens (input_struct input, output_struct *output, int iv, int ir)
{
  float rad;
  float A, B;
  float LWC=1.0;
  int is;
  int status=0;

  /* fprintf(stderr,"calculate droplet radius distribution\n"); */

  switch(input.distribution) {
  case DIST_NONE:
    /* single radius calculation, nothing to do */
    break;
  case DIST_FILE:
    /* only one effective radius */
    if (ir > 0) {
      fprintf(stderr,"Error in make_distribution: more than one loop in ir for distribution \n");
      fprintf(stderr,"    for distribution read from file (obviously only one reff possible)\n");
      return -1;
    }
    break;
  case DIST_GAMMA:
  case DIST_LOGNORMAL:

    if (input.distribution == DIST_GAMMA) {
      B = (input.alpha+3)/output->r_eff[ir];
      A = (0.75/PI)*pow(B,(input.alpha+4))*1000./exp(GAMMA_LN(input.alpha+4.));
    }
    else if (input.distribution == DIST_LOGNORMAL) {
      B = output->r_eff[ir]*exp(-2.5*input.alpha*input.alpha);
      A = 1000./((4./3.*PI)* sqrt(2*PI)*input.alpha * B*B*B *exp(4.5*input.alpha*input.alpha));
    }
    else{
      fprintf(stderr,"Unknown 'distribution %d' in make_distribution\n",input.distribution);
      status--;
      return status;
    }

    if (output->rho_medium <= 0.0) {
      fprintf (stderr, "Error, rho_medium = %f is 0 or negative!\n",output->rho_medium);
      return -1;
    }
    A = A*LWC/output->rho_medium;

    /* calculate number density */
    for (is=0;is<output->dist.n_dens;is++) {

      rad = output->dist.radius[is];

      /* calculate particle density for each radius */ 
      if (input.distribution == DIST_GAMMA)
        output->dist.number_dens[is] = A * pow(rad,input.alpha) * exp(-B*rad);
      else if (input.distribution == DIST_LOGNORMAL)
        output->dist.number_dens[is] = A/rad * exp(-0.5* pow(log(rad/B),2) / pow(input.alpha,2));
      else {
        fprintf(stderr,"Error in make_distribution : unrecognized distribution %d\n",input.distribution);
        return -4;
      }

      /* fprintf(stderr,"is = %3d, radius = %8.4f, number_dens = %13.6f, is=%3d \n", */
      /*           is, output->dist.radius[is], output->dist.number_dens[is], is);   */
    }
    break;
  default:
    fprintf (stderr, "Error in make_distribution : unrecognized distribution %d\n", input.distribution);
    break;
  } /* switch(input.distribution) {... */

  return status;
}


/* get the effective radius from LWC and number density, assuming a distribution */
int make_reff (float *LWC, int n, float dropconc, int distribution, float alpha, float **r_eff)
{
  int ir=-999;
  int status=0;

  for (ir=0; ir<=n; ir++) {
    if (distribution == DIST_GAMMA) {
      (*r_eff)[ir] = 100.0 * (LWC[ir] *pow((0.75*(alpha+3)*(alpha+3))/(PI*(alpha+1)*(alpha+2)*dropconc),(1.0/3.0)));
    }
    else if (distribution == DIST_LOGNORMAL) {
      (*r_eff)[ir] = 100.0 * pow((3.*LWC[ir]*exp(3.*alpha*alpha)/(4.*PI*dropconc)), (1.0/3.0));
    }
    else {
      fprintf(stderr,"Unknown 'distribution %d' in make_reff\n",distribution);
      status--;
    }
  }
  return status;
}



double GAMMA_LN(float xx)
{
  int j=0;
  double *cof;
  double stp = 2.50662827465;
  double half = 0.5, one=1.0, fpf = 5.5, x, tmp, ser;
  double gammaln = 0.0;

  cof  = (double *) calloc (6, sizeof(double));
  cof[0] =  76.18009173;
  cof[1] = -86.50532033;
  cof[2] =  24.01409822;
  cof[3] =  -1.231739516;
  cof[4] =   0.120858003;
  cof[5] =  -0.53638200;

  x = xx - one;
  tmp = x + fpf;
  tmp = (x+half)*log(tmp)-tmp;
  ser=one;
  for (j=0; j<=5; j++) {
    x = x + one;
    ser = ser + cof[j]/x; 
  }
  gammaln=tmp+log(stp*ser);
  return gammaln;
}


double double_min(double x, double y)
{
  if (x <= y)
    return x;
  else
    return y;
}

double double_max(double x, double y)
{
  if (x >= y)
    return x;
  else
    return y;
}



/***********************************************************************************/
/* Function: mie_calc_sizedist_optimised                                           */
/* Description:                                                                    */
/*  Mie calculations, using the MIEV0 and BHMIE codes by Warren Wiscombe           */
/*  (ftp://climate.gsfc.nasa.gov/pub/wiscombe) and Bohren and Huffman              */
/*  (ftp://astro.princeton.edu/draine/scat/bhmie/).                                */
/*                                                                                 */
/* Parameters:                                                                     */
/*  mie_inp_struct input:    mie input structure (see libsrc_c/miecalc.h)          */
/*  mie_out_struct *output:  mie output structure (see libsrc_c/miecalc.h)         */
/*  int program:             MIEV0 or BHMIE                                        */
/*  int medium:              WATER, ICE, or USER; if USER, the refractive index    */
/*                           is read from crefin                                   */
/*  mie_complex crefin:      Complex refractive index (both numbers positive)      */
/*  float wavelength:        wavelength [micron]                                   */
/*  float temperature:       temperature                                           */
/*  double *x_size:          size distribution, radius [um]                        */
/*  double *y_size:          size distribution, n(r)                               */
/*  int n_size:              size distribution, number of radii                    */
/*  double *beta:            extinction coefficient [km-1] per unit                */
/*                           liquid water content (returned)                       */
/*  double *omega:           Single scattering albedo (returned)                   */
/*  double *g:               Asymmetry parameter (returned)                        */
/*                                                                                 */
/* Return value:                                                                   */
/*  0  if o.k., <0 if error                                                        */
/*                                                                                 */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/*  Syntax and parameters of this function are subject to change.                  */
/* Author:                                                                         */
/*                                                                                 */
/* Remark: this is almost the same function than mie_calc_sizedist                 */
/*         but the call of the mie_calc() function is only executed,               */
/*         if this is the first call in the radius-loop                            */
/*         as the dist_radius grid and therefor the MIE-Properties does not change */
/*         but the concentration does change with r_eff and that means             */
/*         that the integration is executed every turn                             */
/***********************************************************************************/

int mie_calc_sizedist_optimised (mie_inp_struct input, mie_out_struct *output, 
                                 int program, int medium, mie_complex crefin, 
                                 float wavelength, float temperature, int nstokes,
                                 double *x_size, double *y_size, int n_size,
                                 double *beta, double *omega, double *g, 
                                 mie_complex *ref, 
                                 int ir, float **qext_array, float **qsca_array, float **gsca_array,
                                 float ****pmom_array)
{
  int status=0, is=0, ip=0, im=0;

  double xsquared=0, xcubed=0, norm=0;
  
  double ext=0, sca=0, asy=0, vol=0;


  /* allocate memory for fields */
  double *extinction = calloc (n_size, sizeof (double));
  double *scattering = calloc (n_size, sizeof (double));
  double *asymmetry  = calloc (n_size, sizeof (double));
  double *volume     = calloc (n_size, sizeof (double));


  double ***pmom=NULL;

  if (input.nmom>0) {
    pmom     = calloc (4, sizeof(double **));
   
    for (ip=0; ip<4; ip++) {
      pmom    [ip] = calloc (input.nmom+1, sizeof(double *));
      
      for (im=0; im<=input.nmom; im++)
        pmom[ip][im] = calloc (n_size, sizeof(double));
    }
  }

  
  for (is=0; is<n_size; is++) {

    if (ir == 0) {
      if (is % 4 == 0)
        fprintf (stderr, "\b|");
      else if (is % 4 == 1)
        fprintf (stderr, "\b/");
      else if (is % 4 == 2)
        fprintf (stderr, "\b-");
      else if (is % 4 == 3)
        fprintf (stderr, "\b\\");

      status = mie_calc (input, output, program, medium, crefin, 
	  	         wavelength, x_size[is], temperature, nstokes, ref);
      if (status!=0) {
        fprintf (stderr, "Error %d returned by mie_calc()\n", status);
        return status;
      }

      /* save results for further calculations */
      (*qext_array)[is] = output->qext;
      (*qsca_array)[is] = output->qsca;
      (*gsca_array)[is] = output->gsca;

      if (input.nmom>0)
        for (ip=0; ip<4; ip++)
          for (im=0; im<=input.nmom; im++)
	    (*pmom_array)[ip][im][is] = output->pmom[ip][im];
    }
    else {
      /* use the saved results */
      output->qext = (*qext_array)[is];
      output->qsca = (*qsca_array)[is];
      output->gsca = (*gsca_array)[is];

      if (input.nmom>0)
        for (ip=0; ip<4; ip++)
          for (im=0; im<=input.nmom; im++)
	    output->pmom[ip][im] = (*pmom_array)[ip][im][is];
    }
    
    /* start integration */
    xsquared = x_size[is] * x_size[is] * 1e-12;
    xcubed   = xsquared   * x_size[is] * 1e-6;
    
    extinction[is] = output->qext * xsquared * y_size[is];
    scattering[is] = output->qsca * xsquared * y_size[is];
    asymmetry [is] = output->gsca * scattering[is]; 
    volume    [is] = xcubed * y_size[is];

    /* quote from MIEV.doc, concerning the normalization of pmom[]:   */
    /*	                                                              */
    /*	The normalized moments are                                    */
    /*      4 / ( XX**2 * QSCA ) * PMOM,  but it is PMOM itself, not  */
    /*      these normalized moments, which should be integrated over */
    /*      a size distribution.                                      */

    if (input.nmom>0)
      for (ip=0; ip<4; ip++)
        for (im=0; im<=input.nmom; im++)
	  pmom[ip][im][is] = output->pmom[ip][im] * y_size[is];

  } /* for (is=0; is<n_size; is++) { */

  if (ir == 0)
    fprintf (stderr, "\n\a");
  
  norm = integrate (x_size, y_size,     n_size);
  
  ext  = integrate (x_size, extinction, n_size) / norm;
  sca  = integrate (x_size, scattering, n_size) / norm;
  asy  = integrate (x_size, asymmetry,  n_size) / norm;
  vol  = integrate (x_size, volume,     n_size) / norm;
  
  if (input.nmom>0)
    for (ip=0; ip<4; ip++)
      for (im=0; im<=input.nmom; im++)
	output->pmom[ip][im] = integrate (x_size, pmom[ip][im], n_size) / norm;
  

  *beta  = 3.0*ext/(4.0*vol)/1000.0;
  *omega = sca/ext;
  *g     = asy/sca;

  /* free memory */
  free (extinction);
  free (scattering);
  free (asymmetry);
  free (volume);

  if (input.nmom>0) {
    for (ip=0; ip<4; ip++) {
      for (im=0; im<=input.nmom; im++)
	free(pmom[ip][im]);
      
      free(pmom[ip]);
    }    
    free(pmom);
  }

  return 0;
}


/** 
 * Function: aerosol_input_data
 *
 * Reads aeosol input data from OPAC datafile 
 * data/aerosol/OPAC/size_distr.cfg. 
 * The data includes size distribution parameters 
 * and the density of the medium. 
 * 
 * @param input input structure
 * @param output output structure
 * 
 * @return status
 *
 * @author Claudia Emde
 */
int aerosol_input_data(input_struct input, output_struct *output)
{

  int i=0, ih=0;
  int status; 
  int n_dist, n_hum; 
  double *aerosol=NULL;
  double *rel_hum=NULL; 
  double *r_min=NULL;
  double *r_max=NULL; 
  double *rmod=NULL; 
  double *rho=NULL; 
  double *sigma=NULL; 
  char *aerdistr_filename=NULL; 

  if (strlen(input.data_files_path) == 0)
    input.data_files_path="../data/"; 
  
  aerdistr_filename= calloc (FILENAME_MAX,sizeof(char));
  
  strcpy(aerdistr_filename,input.data_files_path);
  strcat(aerdistr_filename,"aerosol/OPAC/size_distr.cfg");
  
  /* allocate memory for size distribution data */
  n_dist = 39; /* number of distributions in datafile */
   
  output->aerdistr.n_hum=0;
  status = read_7c_file(aerdistr_filename, 
                        &aerosol, &rel_hum, &r_min, &r_max, &rmod, &rho, &sigma, &n_dist);
  if (status != 0) {
    fprintf (stderr, "error %d reading %s\n", status, aerdistr_filename );
    return status;
  }
  
  /* Check whether aerosol type is water soluble or dry. In the first case size distributions
     for 8 relative humidities are needed; in the second case only 1. */
  if(input.aerosol_type == TYPE_INSO || input.aerosol_type == TYPE_MIAM || 
     input.aerosol_type == TYPE_MICM || input.aerosol_type == TYPE_MINM || 
     input.aerosol_type == TYPE_MITR || input.aerosol_type == TYPE_SOOT){
    output->aerdistr.n_hum= 1;
  }
  else{
    output->aerdistr.n_hum = 8; 
  }
  n_hum=output->aerdistr.n_hum;

  /* allocate aerdistr_struct */
  output->aerdistr.rel_hum = calloc(n_hum, sizeof(double));
  output->aerdistr.r_min = calloc(n_hum, sizeof(double));
  output->aerdistr.r_max = calloc(n_hum, sizeof(double));
  output->aerdistr.rmod = calloc(n_hum, sizeof(double));
  output->aerdistr.rho = calloc(n_hum, sizeof(double));
  
  output->aerdistr.type = input.aerosol_type;
  while(i<n_dist){
    if((int)aerosol[i] == input.aerosol_type) {
    output->aerdistr.rel_hum[ih]=rel_hum[i];
    output->aerdistr.r_min[ih]=r_min[i]; 
    output->aerdistr.r_max[ih]=r_max[i];
    output->aerdistr.rmod[ih]=rmod[i];
    output->aerdistr.rho[ih]=rho[i];
    if(ih==0)
      output->aerdistr.sigma=sigma[i];
    ih++;
    }
    i++;
  }
    
  /* Check, whether all relative humidities have been  read. */
  if (ih != output->aerdistr.n_hum ){
    fprintf(stderr, "Something went wrong while reading aerosol input.ih %d nhum %d \n", ih, output->aerdistr.n_hum); 
    return -1;
  }
  
  free(aerdistr_filename);
  free(aerosol); free(rel_hum); free(r_min); free(r_max);
  free(rmod); free(rho); free(sigma);
  return 0; 
}


/** 
 * Function: aerosol_size_distribution
 *
 * Aerosol size distributions are generated using the OPAC parameters. The 
 * distributions are log-normal distributions. The radius grid is generated 
 * using the method proposed by Evans (cloudprp.f), which seems to be an efficient 
 * and accurate method.
 * 
 * @param input Input structure
 * @param output  Output structure
 * @param iv Wavelength index
 * @param ih Relative humidity index
 * 
 * @return status
 *
 * @author Claudia Emde
 */
int aerosol_size_distribution(input_struct input, output_struct *output, 
                              int iv, int ih)
{

  int i; 
  double rmin, rmax, rstep, a, xstep, x; 
  
  /* Allocate memory */
  /* output->dist.radius = calloc(output->dist.n_dens, sizeof(double));      */
  /* output->dist.number_dens = calloc(output->dist.n_dens, sizeof(double)); */

  /* Generate radius grids for all relative humidities */
  rmin=output->aerdistr.r_min[ih];
  rmax=output->aerdistr.r_max[ih];
      
  /* First point in size distribution */
  output->dist.radius[0]=rmin;

  a=(log(output->dist.radius[0])-log(output->aerdistr.rmod[ih]))/
    log(output->aerdistr.sigma);
 
  output->dist.number_dens[0]=1/(sqrt(2*PI)*log(output->aerdistr.sigma)/log(10))* 
      exp(-0.5*a*a); 
  i=0;  

  /* The next points are determined iteratively */
  while (output->dist.radius[i] <= rmax && i<output->dist.n_dens &&
         /* Cutoff for small particle number densities. Assure that calculation is not stopped too early.*/ 
         ( output->dist.number_dens[i] > 1e-12 || output->dist.radius[i] <output-> aerdistr.rmod[ih] ) ) { 
    
    /* Size parameter */
    x = 2*PI* output->dist.radius[i] /
      (output->wl.lambda[iv]*1.e-03);  /* 1E-3 = nm -> micro m */ 
    xstep = double_max(0.03,0.03*sqrt(x)); /* This is the same rule as proposed by Evans in cloudprp.f */
    rstep = xstep * (output->wl.lambda[iv]*1.e-03) / (2.0*PI);
    i++;
    output->dist.radius[i]= output->dist.radius[i-1] + rstep;
    a=(log(output->dist.radius[i])-log(output->aerdistr.rmod[ih]))/
      log(output->aerdistr.sigma); 
    output->dist.number_dens[i]=1.0/
      (sqrt(2.0*PI)*log(output->aerdistr.sigma)*log(10.0 )*output->dist.radius[i])* 
      exp(-0.5*a*a); 
    /*fprintf(stderr, "%g   %g \n", output->dist.radius[i], output->dist.number_dens[i]);*/
    
  } 
  
  if (i==output->dist.n_dens-1)
    fprintf(stderr, "Warning: Aerosol size distribution cut at %g micrometer.\n",
            output->dist.radius[output->dist.n_dens-1]);
  
  /* Set number of size distribution points to i+1 to avoid calculations for r=0*/
  output->dist.n_dens=i+1;
  
  /* An alternative would be to use a logarithmic radius. This is done in OPAC.f. */
  /* The solution above should be more accurate */
 /*  int n;  */
/*   double rlog; */
/*   n=1000;  */
  
/*   rstep=(log(rmax)-log(rmin))/(double)n;  */
  
/*   output->dist.radius[0]=rmin;   */
/*   rlog=log(rmin);   */
/*   i=1;    */
/*   while (i<n+1){   */
/*     rlog+=rstep;   */
/*     /\* radius grid *\/  */
/*     output->dist.radius[i]=exp(rlog);   */
/*     /\* particle number densities, total number density is set to 1.*\/   */
/*     a=(log(output->dist.radius[i])-log(output->aerdistr.rmod[ih]))/  */
/*       log(output->aerdistr.sigma);   */
/*     output->dist.number_dens[i]=1/(sqrt(2*PI)*log(output->aerdistr.sigma)*log(10))*  */
/*       exp(-0.5*a*a);    */
/*     i++;   */
/*   }   */
/*   output->dist.n_dens=n+1;   */
  
  return 0; 
}

/** 
 * Function: aerosol_refractive_index
 *
 * This function reads the refractive index from the OPAC data 
 * in data/aerosol/OPAC for the requested aerosol type and 
 * interpolates it on the required wavelength.
 * 
 * @param input input structure
 * @param output output structure
 * @param wavelength wavelength
 * @param ih index of relative humidity
 * @param ri_re refractive index (real)
 * @param ri_im refractive index (imag)
 * 
 * @return status 
 * 
 * @author Claudia Emde
 */
int aerosol_refractive_index(input_struct input, output_struct *output, 
                             double wavelength, int ih, float *ri_re,
                             float *ri_im)
{
  int status=0;
  double *lambda=NULL; 
  double *ri_real=NULL; 
  double *ri_imag=NULL;
  int nlambda=0;
  char *type;
  char *rel_hum;
  char *filename;
  double dummy; 

  /* Coefficents for interpolation */
  double *a0=NULL, *a1=NULL, *a2=NULL, *a3=NULL;
  
  if (strlen(input.data_files_path) == 0)
    input.data_files_path="../data/"; 
  
  filename= calloc (FILENAME_MAX,sizeof(char));
  strcpy(filename, input.data_files_path);
  
  strcat(filename,"aerosol/OPAC/");

  if(input.aerosol_type == TYPE_INSO)
    type="inso"; 
  else if (input.aerosol_type == TYPE_WASO)
    type="waso";
  else if (input.aerosol_type == TYPE_SOOT)
    type="soot";
  else if (input.aerosol_type == TYPE_SSAM)
    type="ssam";
  else if (input.aerosol_type == TYPE_SSCM)
    type="sscm";
  else if (input.aerosol_type == TYPE_MINM)
    type="minm";
  else if (input.aerosol_type == TYPE_MIAM)
    type="miam";
  else if (input.aerosol_type == TYPE_MICM)
    type="micm";
  else if (input.aerosol_type == TYPE_MITR)
    type="mitr";
  else if (input.aerosol_type == TYPE_SUSO)
    type="suso";
  else{
    fprintf(stderr, "Error: Unknown aerosol type \n");
    return -1; 
  }
  
  if (ih == 0) rel_hum="00";
  else if (ih == 1) rel_hum="50";
  else if (ih == 2) rel_hum="70";
  else if (ih == 3) rel_hum="80";
  else if (ih == 4) rel_hum="90";
  else if (ih == 5) rel_hum="95";
  else if (ih == 6) rel_hum="98";
  else if (ih == 7) rel_hum="99";
  else{
    fprintf(stderr, "Error: Humidity index ih=%d is wrong. \n", ih);
    return -1; 
  } 
  
  strcat(filename, type);
  strcat(filename, rel_hum);
  strcat(filename, "_refr.dat");


  status = read_3c_file (filename, &lambda, &ri_real, &ri_imag, &nlambda);
  if (status != 0) {
    fprintf (stderr, "error %d reading %s\n", status, filename);
    return status;
  }

  /* Interpolate refractive index on the required wavelength. */
  /* real part*/
  status = linear_coeffc (lambda, ri_real, nlambda, &a0, &a1, &a2, &a3);
  status = calc_splined_value (wavelength, &dummy, lambda, nlambda,
                               a0, a1, a2, a3);
  *ri_re = (float) dummy;  
  
  if (status!=0)  {
    fprintf (stderr, "Sorry cannot interpolate refractive index\n");
    return status;
  }
  
  free(a0); free(a1); free(a2); free(a3); 
  
  /* imaginary part */
  status = linear_coeffc (lambda, ri_imag, nlambda, &a0, &a1, &a2, &a3);
  status = calc_splined_value (wavelength, &dummy, lambda, nlambda,
                               a0, a1, a2, a3);
  *ri_im = (float) dummy; 

  if (status!=0)  {
    fprintf (stderr, "Sorry cannot interpolate refractive index\n");
    return status;
  }
  
  free(a0); free(a1); free(a2); free(a3); 
  free(filename);
  free(lambda); free(ri_real); free(ri_imag);
  return 0;
}

/* write all the output to the screen or into files */
int write_output (input_struct input, output_struct *output, int ir, int iv)
{

  int  io=0;
  int  nmom=-999, nstokes=1;
  int  im=0;
  char outputfile[255];
  char ck_band_str[4];
  int  status=0;
  int  ip=0, nphase=0;
  int default_output=0;

  if (input.output_format == OUTPUT_FORMAT_NORMAL) {

    if (input.n_output_user == 0) {  
      /* default output */
      input.n_output_user = 8;
      input.output_user = calloc (input.n_output_user, sizeof (int));
      input.output_user[0]=OUTPUT_USER_WAVE;
      input.output_user[1]=OUTPUT_USER_REFRAC_REAL;
      input.output_user[2]=OUTPUT_USER_REFRAC_IMAG;
      input.output_user[3]=OUTPUT_USER_QEXT;
      input.output_user[4]=OUTPUT_USER_OMEGA;
      input.output_user[5]=OUTPUT_USER_GG;
      input.output_user[6]=OUTPUT_USER_SPIKE;
      input.output_user[7]=OUTPUT_USER_PMOM;
      default_output=1;
    }

    for (io=0; io<input.n_output_user; io++) {
      switch (input.output_user[io]) {
      case OUTPUT_USER_NONE:
        fprintf (stdout, "%9s ", " ");
        break;
      case OUTPUT_USER_WAVE:
        fprintf (stdout, "%9.3f ", output->wl.lambda[iv]);
        break;
      case OUTPUT_USER_WAVENUMBER:
        fprintf (stdout, "%9.3f ", 1E7/output->wl.lambda[iv]);
        break;
      case OUTPUT_USER_R_EFF:
        fprintf (stdout, "%9.3f ", output->r_eff[ir]);
	  break;
      case OUTPUT_USER_REFRAC_REAL:
	  fprintf (stdout, "%12.6e ", output->ref.re);
	  break;
      case OUTPUT_USER_REFRAC_IMAG:
        fprintf (stdout, "%12.6e ", output->ref.im);
        break;
      case OUTPUT_USER_QEXT:
        if (input.distribution == DIST_NONE)
          fprintf (stdout, "%12.6e ", output->mie.qext);
        else
          fprintf (stdout, "%12.6e ", output->beta);
        break;
      case OUTPUT_USER_QSCA:
        if (input.distribution == DIST_NONE)
          fprintf (stdout, "%12.6e ", output->mie.qsca);
        else
          fprintf (stdout, "%12.6e ", output->omega*output->beta);
        break;
      case OUTPUT_USER_OMEGA:
        if (input.distribution == DIST_NONE)
          fprintf (stdout, "%12.6e ", output->mie.qsca/output->mie.qext);
        else
          fprintf (stdout, "%12.6e ", output->omega);
        break;
      case OUTPUT_USER_GG:
        if (input.distribution == DIST_NONE)
          fprintf (stdout, "%12.6e ", output->mie.gsca);
        else
	    fprintf (stdout, "%12.6e ", output->g);
        break;
      case OUTPUT_USER_SPIKE:
        if (input.distribution == DIST_NONE )
          fprintf (stdout, "%12.6e ", output->mie.spike);
        else
          fprintf (stdout, "%12.6e ", -1.0);
        break;
      case OUTPUT_USER_SFORW:
	if (input.distribution == DIST_NONE)
          fprintf (stdout, "%12.6e %12.6e ", output->mie.sforw.re, output->mie.sforw.im);
        else
          fprintf (stdout, "%12.6e %12.6e ", NAN, NAN);  /* not yet implemented*/
        break;
      case OUTPUT_USER_SBACK:
	if (input.distribution == DIST_NONE)
          fprintf (stdout, "%12.6e %12.6e ", output->mie.sback.re, output->mie.sback.im);
        else
          fprintf (stdout, "%12.6e %12.6e ", NAN, NAN);  /* not yet implemented*/
        break;
      case OUTPUT_USER_PMOM:
        if (input.mie.nmom>0){
          if (input.nstokes>1 || default_output==0){
            fprintf(stdout , "\n");
            for (im=0; im<=input.mie.nmom; im++){
              for (ip=0; ip < input.nstokes; ip++)
                fprintf (stdout, " %14.6e", output->mie.pmom[ip][im]);
              fprintf(stdout , "\n");
            }
          }
          else{
            for (im=0; im<=input.mie.nmom; im++)
              fprintf (stdout, " %14.6e", output->mie.pmom[0][im]);
            fprintf(stdout , "\n");
          }
        }
        break;
      default:
        fprintf (stderr, "Error, unknown user output %d\n", input.output_user[io]);
        break;
      } /* switch (input.output_user[io]) { */
    } /* for (io=0; io<input.n_output_user; io++) { */
    fprintf (stdout, "\n");
    fflush (stdout);
  } /* if (input.output_user[0] != OUTPUT_USER_CLOUDPROP) { */

  else if (input.output_format == OUTPUT_FORMAT_CLOUDPROP || 
        input.output_format == OUTPUT_FORMAT_AERPROP   ) {
    
    if ( input.distribution == DIST_NONE ){
      fprintf(stderr, "The option output_user cloudprop works only for size distributions, not for mono-disperse particles. \n");
      return -1;
    }
      
    /* For polarized calculations, all six elements are calculated for mie files */
    if (input.nstokes > 1)
      nstokes =4; 
    
    if (ir == 0) {
      /* generate file name */
      switch (input.medium) {
      case WATER:
        strcpy (outputfile, "wc.");          
	break;
      case ICE:
        strcpy (outputfile, "ic.");   
        break;
      case USER:
        strcpy (outputfile, "uc.");   
	break;
      case AEROSOL:
        if(input.aerosol_type == TYPE_INSO)
          strcpy (outputfile, "inso."); 
        else if (input.aerosol_type == TYPE_WASO)
          strcpy (outputfile, "waso."); 
        else if (input.aerosol_type == TYPE_SOOT)
          strcpy (outputfile, "soot."); 
        else if (input.aerosol_type == TYPE_SSAM)
          strcpy (outputfile, "ssam."); 
        else if (input.aerosol_type == TYPE_SSCM)
          strcpy (outputfile, "sscm.");
        else if (input.aerosol_type == TYPE_MINM)
          strcpy (outputfile, "minm.");
        else if (input.aerosol_type == TYPE_MIAM)
          strcpy (outputfile, "miam."); 
        else if (input.aerosol_type == TYPE_MICM)
          strcpy (outputfile, "micm."); 
        else if (input.aerosol_type == TYPE_MITR)
          strcpy (outputfile, "mitr.");
        else if (input.aerosol_type == TYPE_SUSO)
          strcpy (outputfile, "suso.");
        else{
          fprintf(stderr, "Error: Unknown aerosol type \n");
          return -1; 
        }
        break;
      }
      
      sprintf(ck_band_str, "%03d",iv+1);
      strcat (outputfile, ck_band_str);
      strcat (outputfile, ".mie");
      fprintf (stderr, " ... generating file: %s\n", outputfile);
          
      if ((output->ptr_output = fopen (outputfile, "w")) == NULL) {
        fprintf (stderr, "Error, cannot open file: %s in write_output! \n", outputfile);   
        return -1;         
      }
    }
    
    if(input.medium != AEROSOL){
      fprintf(stderr,"   lambda = %10.3f nm (%3d of %3d), r_eff = %8.3f micro meter (%3d of %3d) \n", 
              output->wl.lambda[iv], iv+1,output->wl.nlambda+1, output->r_eff[ir], ir+1,output->n_r_eff+1);
    }
    else{
      fprintf(stderr,"   lambda = %10.3f nm (%3d of %3d), rel_hum = %4.0f percent (%3d of %3d) \n", 
               output->wl.lambda[iv], iv+1,output->wl.nlambda+1, output->aerdistr.rel_hum[ir], ir+1, output->aerdistr.n_hum);
    }
      
    
    if (ir == 0) {
      switch (input.medium) {
      case WATER:
        fprintf (output->ptr_output, "! Mie table vs. effective radius (LWC=1 g/m^3)\n");
	break;
      case ICE:
        fprintf (output->ptr_output, "! Mie table vs. effective radius (IWC=1 g/m^3)\n");
	break;
      case USER:
        fprintf (output->ptr_output, "! Mie table vs. effective radius (ATTENTION!!! assume rho_medium = %7.4f g/cm3)\n", 
                                      output->rho_medium);
        break;
      case AEROSOL: 
        fprintf (output->ptr_output, "! Mie table vs. relative humidity (N=1, rho_medium = %7.4f g/cm3)\n", 
                 output->rho_medium);
	break;
      }
      
      fprintf (output->ptr_output," %12.6e  wavelength (micron) \n",output->wl.lambda[iv]/1000.);
      fprintf (output->ptr_output," %12.6e  %12.6e  index of refraction \n", output->ref.re, output->ref.im);
      if (input.distribution == DIST_NONE) {
        fprintf (output->ptr_output,"S  distribution type ! one Single radius  \n");
        fprintf   (output->ptr_output,"-999.99  distribution shape parameter !(dummy to fit the format)\n");
      }
      else if (input.distribution == DIST_GAMMA) {
        fprintf (output->ptr_output,"G  distribution type\n");
        fprintf   (output->ptr_output,"%7.4f  distribution shape parameter\n", input.alpha);
      }
      else if (input.distribution == DIST_LOGNORMAL) {
        fprintf (output->ptr_output,"L  distribution type\n");
        fprintf   (output->ptr_output,"%7.4f  distribution shape parameter\n", input.alpha);
      }
      else if (input.distribution == DIST_FILE) {
        fprintf (output->ptr_output,"U  distribution type ! User defined, from file \n");
        fprintf   (output->ptr_output,"-999.99  distribution shape parameter !(dummy to fit the format)\n");
      }
      else if (input.distribution == DIST_AER) {
        fprintf (output->ptr_output,"A  distribution type \n");
        fprintf (output->ptr_output,"Size distribution parameters from OPAC database \n");
      }

      if (input.medium != AEROSOL){
        fprintf (output->ptr_output,"%4d %8.4f %8.4f  number, starting, ending effective radius \n", 
                 output->n_r_eff+1, output->r_eff[0], output->r_eff[output->n_r_eff]); 
      }
      else{
        fprintf (output->ptr_output,"%4d %8.4f %8.4f  number, starting, ending relative humidity \n", 
                 output->aerdistr.n_hum, output->aerdistr.rel_hum[0], output->aerdistr.rel_hum[output->aerdistr.n_hum-1]); 
      }
      
      
      } /* if (ir == 0) { */

    
    nmom=0;
    for (ip=0; ip<nstokes; ip++){
      for (im=0;im<=input.mie.nmom;im++){
        /* if (fabs((2*im+1)*output->mie.pmom[ip][im]/output->mie.pmom[ip][0]) < 0.000005) { */
        if (fabs((2*im+1)*output->mie.pmom[ip][im]) < 1E-5){
          if ( im > nmom )
            nmom=im;
          break;
        }
      }
    }
    
    /* remember to write a warning, if there could be more Legendre coefficients */
    if (nmom == input.mie.nmom) 
      output->warnings[iv][ir]=1;
    
    if(nstokes==1) nphase = 1;
    else nphase = 6;
    
    if (input.medium != AEROSOL){
    fprintf (output->ptr_output,"%9.4f  %12.5e  %12.5e %6d  %4d Reff  Ext  Alb  Nleg Nphase\n", output->r_eff[ir], 
             output->beta/output->rho_medium, 
             output->omega, nmom, nphase);
    }
    else{
      fprintf (output->ptr_output,"%9.4f  %12.5e  %12.5e %6d  %4d RelHum  Ext  Alb  Nleg Nphase\n", output->aerdistr.rel_hum[ir], 
               output->beta/output->rho_medium, 
               output->omega, nmom, nphase);
    }
    
    /*     if (input.distribution == DIST_GAMMA ||  */
    /*         input.distribution == DIST_LOGNORMAL || */
    /*         input.distribution == DIST_FILE) */
    /* fprintf (output->ptr_output,"%9.4f  %12.5e  %12.5e %6d  Reff  Ext  Alb  Nleg\n", output->r_eff[ir], output->beta/output->rho_medium,  */
    /*              output->omega, nmom+1); */
    /*           else  */
    
    /*     else */
    /*       fprintf (output->ptr_output,"%9.4f  %12.5e  %12.5e %6d  Reff  Ext  Alb  Nleg\n", output->r_eff[ir], output->mie.qext*1E+6*1E-3, /\* 1E+6 = cm-3/m3 -> m3/m3 *\/ */
    /*                                                                                       output->mie.qsca/output->mie.qext, nmom);       /\* 1E-3 = km -> m *\/ */
    
      
    if (input.mie.nmom>0) {
      fprintf (output->ptr_output, "  ");
      for (ip=0; ip < input.nstokes; ip++){
        for (im=0; im<=nmom; im++)
          fprintf (output->ptr_output, " %9.5f", (2*im+1) * output->mie.pmom[ip][im]);
        fprintf( output->ptr_output, "\n");
      }
      /* Print also P5 and P6, equal to P1 and P3 for spheres */
      if(nstokes == 4){
        for (im=0; im<=nmom; im++)
          fprintf (output->ptr_output, " %9.5f",(2*im+1) * output->mie.pmom[0][im]);
        fprintf (output->ptr_output, "\n");
        for (im=0; im<=nmom; im++)
          fprintf (output->ptr_output, " %9.5f", (2*im+1) *output->mie.pmom[2][im]);
        fprintf (output->ptr_output, "\n");
      }
    }
    fprintf (output->ptr_output, "\n");
    fflush (output->ptr_output);  
   
    if ((ir == output->n_r_eff && input.output_format == OUTPUT_FORMAT_CLOUDPROP) ||
        (ir == output->aerdistr.n_hum && input.output_format == OUTPUT_FORMAT_AERPROP) )
      fclose(output->ptr_output);
    
  } /* else if (input.output_format == OUTPUT_FORMAT_CLOUDPROP || OUTPUT_FORMAT_AERPROP) ... */ 
  
  return status;
}
 
 
#if HAVE_LIBNETCDF

/** 
 * calloc_netcdf_var
 *
 * Allocates memory for netcdf output. 
 * 
 * @param input mie input structure
 * @param output mie output structure
 * 
 * @return status
 *
 * @author Claudia Emde
 */
int calloc_netcdf_var(input_struct input, output_struct *output)
{
  output->miecdf.nreff=output->n_r_eff+1;
  output->miecdf.maxmom=input.mie.nmom+1;
  
  if(input.nstokes==1) 
    output->miecdf.nphase = 1;
  else output->miecdf.nphase = 6;
  
  output->miecdf.reff = calloc (output->miecdf.nreff, sizeof(double));
  output->miecdf.ext = calloc (output->miecdf.nreff, sizeof(double));
  output->miecdf.ssa = calloc (output->miecdf.nreff, sizeof(double));
  output->miecdf.refre = calloc (output->miecdf.nreff, sizeof(double));
  output->miecdf.refim = calloc (output->miecdf.nreff, sizeof(double));
  output->miecdf.rho = calloc (output->miecdf.nreff, sizeof(double));
  output->miecdf.nmom = calloc (output->miecdf.nreff, sizeof(int));
  
  ASCII_calloc_float_3D(&output->miecdf.pmom, output->miecdf.nreff, 
                        output->miecdf.maxmom, output->miecdf.nphase);
  
  return 0;
}


/** 
 * generate_netcdf_var
 * 
 * Fill netcdf output structure. 
 *
 * @param input mie input structure
 * @param output mie output structure
 * @param ir index effective radius
 * @param iv index wavelength
 * 
 * @return status
 *
 * @author Claudia Emde
 */
int generate_netcdf_var(input_struct input, output_struct *output, 
                        int ir, int iv)
{
  int ip=0, im=0;

  /* Find number of polynomials that are non-zero */
  output->miecdf.nmom[ir]=0;
  for (ip=0; ip<input.nstokes; ip++){
    for (im=0;im<=input.mie.nmom;im++){
      if (fabs((2*im+1)*output->mie.pmom[ip][im]) < 1E-5){
        if ( im > output->miecdf.nmom[ir] )
          output->miecdf.nmom[ir]=im;
        break;
      }
    }
  }
  
 /* remember to write a warning, if there could be more
     Legendre coefficients */
  if (output->miecdf.nmom[ir] == 0){
    output->miecdf.nmom[ir]=input.mie.nmom;
    output->warnings[iv][ir]=1;
  }
  
  output->miecdf.reff[ir] = output->r_eff[ir];
  output->miecdf.ext[ir] = output->beta/output->rho_medium;
  output->miecdf.ssa[ir] = output->omega;
  output->miecdf.refre[ir] = output->mie.crefin.re;
  output->miecdf.refim[ir] = output->mie.crefin.im;
  output->miecdf.rho[ir] = output->rho_medium;

  /* write Legendre polynomials */
  for (ip=0; ip<input.nstokes; ip++){
    for (im=0;im<output->miecdf.maxmom;im++)
      output->miecdf.pmom[ir][im][ip]= (2*im+1) * output->mie.pmom[ip][im];
    /* Print also P5 and P6, equal to P1 and P4 for spheres */
    if(input.nstokes==4) {
      for (im=0;im<output->miecdf.maxmom;im++){
        output->miecdf.pmom[ir][im][4]= (2*im+1) * output->mie.pmom[0][im];
        output->miecdf.pmom[ir][im][5]= (2*im+1) * output->mie.pmom[2][im];
      }     
    }
  }
  
  return 0; 
}

/** 
 * write_output_netcdf 
 *
 * Write the output of the Mie calculation in netcdf format. This can directly be
 * used as input for uvspec. 
 * 
 * @param input mie input structure
 * @param output mie output structure
 * @param iv wavelength index
 * 
 * @return status
 *
 * @author Claudia Emde
 */
int write_output_netcdf(input_struct input, output_struct *output, int iv)
{
  
  int retval=0;
  int ncid=0;
  /* Netcdf dimension IDs */
  int id_nreff=0, id_maxmom=0, id_nphase=0, id_nlam=0;
  int dimids[3];
  /* Netcdf variable IDs */
  int id_reff=0, id_ext=0, id_ssa=0, id_nmom=0, id_pmom=0,
    id_refre=0, id_refim=0, id_rho=0, id_lam=0;

  int ir=0, im=0, ip=0, i=0;
  float *pmoms=NULL;
  float fill_value[1];
  char name[MAXLINE], unit[10], varname[10];
  char outputfile[FILENAME_MAX], medium[MAXLINE];
  char distr[MAXLINE];
  char ck_band_str[4];
  double lambdadim[1], distr_param[1]; 
  double lambda=0.0;
    
  lambdadim[0]=1; 
  
  /* Create a 1D array for netcdf output */  
  pmoms = calloc(output->miecdf.nreff*output->miecdf.maxmom*
                 output->miecdf.nphase, sizeof(float));
  
  i=0;
  for (ir=0; ir<output->miecdf.nreff; ir++){
    for (ip=0; ip<output->miecdf.nphase; ip++){
      for (im=0;im<output->miecdf.maxmom;im++){
        pmoms[i]=output->miecdf.pmom[ir][im][ip];
        i++;
      }
    }
  }
  
  /* Generate filename */
  switch (input.medium) {
  case WATER:
    strcpy (outputfile, "wc.");
    strcpy (medium, "water");
    break;
  case ICE:
    strcpy (outputfile, "ic.");   
    strcpy (medium, "ice");
    break;
  case USER:
    strcpy (outputfile, "uc.");
    strcpy (medium, "user defined refractive index");
    break;
  case AEROSOL:
  
    if(input.aerosol_type == TYPE_INSO){
      strcpy (outputfile, "inso."); 
      strcpy (medium, "Aerosol: insoluble"); 
    }
    else if (input.aerosol_type == TYPE_WASO){
      strcpy (outputfile, "waso."); 
      strcpy (medium, "Aerosol: water soluble");
    }
    else if (input.aerosol_type == TYPE_SOOT){
      strcpy (outputfile, "soot."); 
      strcpy (medium, "Aerosol: soot");
    }
    else if (input.aerosol_type == TYPE_SSAM){
      strcpy (outputfile, "ssam."); 
      strcpy (medium, "Aerosol: sea salt accumulated mode");
    }    
    else if (input.aerosol_type == TYPE_SSCM){
      strcpy (outputfile, "sscm.");
      strcpy (medium, "Aerosol: sea salt coarse mode");
    }  
    else if (input.aerosol_type == TYPE_MINM){
      strcpy (outputfile, "minm.");
      strcpy (medium, "Aerosol: mineral nucleation mode");
    }
    else if (input.aerosol_type == TYPE_MIAM){
      strcpy (outputfile, "miam.");
      strcpy (medium, "Aerosol: mineral accumulated mode");
    }
    else if (input.aerosol_type == TYPE_MICM){
      strcpy (outputfile, "micm."); 
      strcpy (medium, "Aerosol: mineral coarse mode");
    }
    else if (input.aerosol_type == TYPE_MITR){
      strcpy (outputfile, "mitr.");
      strcpy (medium, "Aerosol: mineral transported");
    } 
    else if (input.aerosol_type == TYPE_SUSO){
      strcpy (outputfile, "suso.");
      strcpy (medium, "Aerosol: sulfate");
    }
    else{
      fprintf(stderr, "Error: Unknown aerosol type \n");
      return -1; 
    }
    break;
  default:
    fprintf (stderr, "Error in write_output_netcdf : unrecognized medium %d\n", input.medium); 
    return -1; 
  }
  
  switch(input.distribution){
  case DIST_NONE:
    strcpy (distr, "Calculation for single radius.");
    distr_param[0]=-999.;
    break;
  case DIST_GAMMA:
    strcpy (distr, "Gamma distribution.");
    distr_param[0]=input.alpha;
    break; 
  case DIST_LOGNORMAL:
    strcpy (distr, "Lognormal distribution.");
    distr_param[0]=input.alpha;
    break;
  case DIST_FILE:
    strcpy (distr, "User defined size distribution.");
    distr_param[0]=-999.;
    break; 
  case DIST_AER:
    strcpy (distr, 
            "Aerosol size distribution,\nlog-normal with parameters from OPAC");
    distr_param[0]=-999.;
    break;
  default: 
    fprintf (stderr, "Error in write_output_netcdf : unrecognized distribution %d\n", input.distribution); 
    return -1; 
    break; 
  }

  sprintf(ck_band_str, "%03d",iv+1);
  strcat (outputfile, ck_band_str);
  strcat (outputfile, ".mie.cdf");
  fprintf (stderr, " ... generating file: %s\n", outputfile);
  
  if ((retval = nc_create(outputfile, NC_CLOBBER, &ncid)))
    {
      fprintf(stderr, "Error: creating netcdf file");
      return -1;
    }

  strcpy(name, "Netcdf file created using libRadtran-Mie-tool");
  nc_put_att_text(ncid, NC_GLOBAL, "file_info", strlen(name), name);  
  
  nc_put_att_text(ncid, NC_GLOBAL, "medium", strlen(medium), medium);  
 
  nc_put_att_text(ncid, NC_GLOBAL, "size_distr", strlen(distr), distr);
  
  nc_put_att_double(ncid, NC_GLOBAL, "param_alpha", NC_DOUBLE, 1, distr_param);
   
  /* Define dimensions.*/
  if ((retval = nc_def_dim(ncid, "reff", output->miecdf.nreff, &id_nreff))){
       fprintf(stderr, "Error: creating dimension nreff.\n" );
       return -1;
  }
  
  if ((retval = nc_def_dim(ncid, "maxmom", output->miecdf.maxmom, &id_maxmom))){
    fprintf(stderr, "Error: creating dimension maxmom.\n" );
    return -1;
  }
  
  if ((retval = nc_def_dim(ncid, "nphase", output->miecdf.nphase, &id_nphase))){
    fprintf(stderr, "Error: creating dimension nphase.\n" );
    return -1;
  }
   
  if ((retval = nc_def_dim(ncid, "nlam", 1, &id_nlam))){
    fprintf(stderr, "Error: creating dimension nlam.\n" );
    return -1;
  }
   
  /* The dimids array is used to pass the IDs of the dimensions of
   * the variable. */
  dimids[0] = id_nreff;
  dimids[1] = id_nphase;
  dimids[2] = id_maxmom;
  
  /* Define variables */ 
  if ((retval = nc_def_var(ncid, "wavelen", NC_DOUBLE, 1, 
                           &id_nlam, &id_lam))){
    fprintf(stderr, "Error: defining variable wavelen. \n");
    return -1;
  } 
  strcpy(name,"wavelength");
  strcpy(unit,"micrometer");

  nc_put_att_text(ncid, id_lam, "long_name", strlen(name), name);  
  nc_put_att_text(ncid, id_lam, "units", strlen(unit), unit);
  
  if (input.medium == AEROSOL){
    strcpy(varname,"hum");
    strcpy(name,"relative humidity");
    strcpy(unit,"per cent");
   }
   else{
     strcpy(varname,"reff");
     strcpy(name,"effective radius");
     strcpy(unit,"micrometer");
   }
  
   if ((retval = nc_def_var(ncid, varname, NC_DOUBLE, 1, 
                            &id_nreff, &id_reff))){
     fprintf(stderr, "Error: defining variable reff. \n");
     return -1;
   } 
   nc_put_att_text(ncid, id_reff, "long_name", strlen(name), name);  
   nc_put_att_text(ncid, id_reff, "units", strlen(unit), unit);

   if ((retval = nc_def_var(ncid, "ext", NC_DOUBLE, 1, 
                            &id_nreff, &id_ext))){
     fprintf(stderr, "Error: defining variable ext. \n");
     return -1;
    } 
   strcpy(name,"extinction");
   strcpy(unit,"1/km");
   nc_put_att_text(ncid, id_ext, "long_name", strlen(name), name);  
   nc_put_att_text(ncid, id_ext, "units", strlen(unit), unit);
   
   if ((retval = nc_def_var(ncid, "ssa", NC_DOUBLE, 1, 
                            &id_nreff, &id_ssa))){
     fprintf(stderr, "Error: defining variable ssa. \n");
     return -1;
   } 
   strcpy(name,"single scattering albedo");
   strcpy(unit," ");
   nc_put_att_text(ncid, id_ssa, "long_name", strlen(name), name);  
   nc_put_att_text(ncid, id_ssa, "units", strlen(unit), unit);

   if ((retval = nc_def_var(ncid, "nmom", NC_INT, 1, 
                            &id_nreff, &id_nmom))){
     fprintf(stderr, "Error: defining variable maxmom. \n");
     return -1;
   } 
   strcpy(name,"number of Legendre moments");
   strcpy(unit," ");
   nc_put_att_text(ncid, id_nmom, "long_name", strlen(name), name);  
   nc_put_att_text(ncid, id_nmom, "units", strlen(unit), unit);

   if ((retval = nc_def_var(ncid, "refre", NC_DOUBLE, 1, 
                            &id_nreff, &id_refre))){
     fprintf(stderr, "Error: defining variable refre. \n");
     return -1;
   }
   strcpy(name,"refractive index (real)");
   strcpy(unit," ");
   nc_put_att_text(ncid, id_refre, "long_name", strlen(name), name);  
   nc_put_att_text(ncid, id_refre, "units", strlen(unit), unit);
   
   if ((retval = nc_def_var(ncid, "refim", NC_DOUBLE, 1, 
                            &id_nreff, &id_refim))){
     fprintf(stderr, "Error: defining variable refre. \n");
     return -1;
   }
   strcpy(name,"refractive index (imaginary)");
   strcpy(unit," ");
   nc_put_att_text(ncid, id_refim, "long_name", strlen(name), name);  
   nc_put_att_text(ncid, id_refim, "units", strlen(unit), unit);
   
   if ((retval = nc_def_var(ncid, "rho", NC_DOUBLE, 1, 
                            &id_nreff, &id_rho))){
     fprintf(stderr, "Error: defining variable refre. \n");
     return -1;
   }
   strcpy(name,"density of medium");
   strcpy(unit,"g/cm^3");
   nc_put_att_text(ncid, id_rho, "long_name", strlen(name), name);  
   nc_put_att_text(ncid, id_rho, "units", strlen(unit), unit);
   

   
   if ((retval = nc_def_var(ncid, "pmom", NC_FLOAT, 3, 
                            dimids, &id_pmom))){
     fprintf(stderr, "Error: defining variable pmom. \n");
     return -1;
   } 
   
   strcpy(name,"moments of the phase matrix");
   strcpy(unit," ");
   nc_put_att_text(ncid, id_pmom, "long_name", strlen(name), name);  
   nc_put_att_text(ncid, id_pmom, "units", strlen(unit), unit);
   
   fill_value[0]=0.0;
   nc_put_att_float(ncid, id_pmom, "_FillValue", NC_FLOAT, 1, fill_value);

   
   /* End define mode. This tells netCDF we are done defining
    * metadata. */ 
   if ((retval = nc_enddef(ncid))){
     fprintf(stderr, "Error: netcdf definition end. \n");
     return -1;
   } 
   
 
   /* Write variables */
   lambda=output->wl.lambda[iv]/1000.;
   nc_put_var_double(ncid, id_lam, &lambda); 
   
   if ((retval = nc_put_var_double(ncid, id_reff, &output->miecdf.reff[0]))){
     fprintf(stderr, "Error: writing variable reff.\n");
     return -1;
   } 
   
   if ((retval = nc_put_var_double(ncid, id_ext, &output->miecdf.ext[0]))){
     fprintf(stderr, "Error: writing variable ext.\n");
     return -1;
   } 

   if ((retval = nc_put_var_double(ncid, id_ssa, &output->miecdf.ssa[0]))){
     fprintf(stderr, "Error: writing variable ssa.\n");
     return -1;
   } 
   
   if ((retval = nc_put_var_double(ncid, id_refre, &output->miecdf.refre[0]))){
     fprintf(stderr, "Error: writing variable refre.\n");
     return -1;
   } 

   if ((retval = nc_put_var_double(ncid, id_refim, &output->miecdf.refim[0]))){
     fprintf(stderr, "Error: writing variable refre.\n");
     return -1;
   } 
   
   if ((retval = nc_put_var_double(ncid, id_rho, &output->miecdf.rho[0]))){
     fprintf(stderr, "Error: writing variable refre.\n");
     return -1;
   } 

   if ((retval = nc_put_var_int(ncid, id_nmom, &output->miecdf.nmom[0]))){
     fprintf(stderr, "Error: writing variable nmom.\n");
     return -1;
   } 
   
   if ((retval = nc_put_var_float(ncid, id_pmom, &pmoms[0]))){
     fprintf(stderr, "Error: writing variable pmom.\n");
     return -1;
   } 
  
   /* Close the file. This frees up any internal netCDF resources
    * associated with the file, and flushes any buffers. */
   if ((retval = nc_close(ncid))){
     fprintf(stderr, "Error: closing netcdf file.\n");
     return -1;
   } 
   
   return 0; 
}
 
#endif
