/*--------------------------------------------------------------------
 * $Id: uvspec.h,v 1.230 2009/01/08 09:40:38 arve Exp $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 * 
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

#ifndef __uvspec_h
#define __uvspec_h

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "mystic.h"
#include <time.h>


#define VERSION "1.4"

/* Indices for the reading of the atmosphere files */
#define ATMFILE_ZD       0
#define ATMFILE_PRESS    1
#define ATMFILE_TEMPER   2
#define ATMFILE_AIR      3
#define ATMFILE_O3       4
#define ATMFILE_O2       5
#define ATMFILE_H2O      6
#define ATMFILE_CO2      7
#define ATMFILE_NO2      8

/* Atmospheric molecules that may be included in spectral calculations */
#define MOL_NN        9
#define MOL_AIR       0
#define MOL_O3        1
#define MOL_O2        2
#define MOL_H2O       3
#define MOL_CO2       4
#define MOL_NO2       5
#define MOL_BRO       6
#define MOL_OCLO      7
#define MOL_HCHO      8

/* Possible units of total molecule column */
#define CM_2          1
#define DU            2

/* Possible units of density profiles */
#define CM_3          1    /* per cubic centimeter */
#define M_3           2    /* per cubic meter      */
#define MMR           3    /* mass mixing ratio    */
#define VMR           4    /* volumn mixing ratio  */
#define RH            5    /* relative humidity    */

#define RS_NO_DATA    0   /* no radiosonde data */
#define RS_FROM_FILE  1   /* radiosonde data source */
#define RS_FROM_ECMWF 2   /* ECMWF data converted with CDO */
#define RS_FROM_ECHAM 3   /* ECHAM data*/


#define FN_NN                 43  /* Number of general files, increase 
				     when adding new general files */

#define FN_PATH                0
#define FN_ALBEDO              1
#define FN_ALBEDO_LIB_PATH     2
#define FN_ALBEDO_LIB_WVL      3
#define FN_ALBEDO_MAP          4 
#define FN_SURFACE_TYPE_MAP    5
#define FN_SURFACE_TEMP_MAP    6
#define FN_RPV                 7
#define FN_RPV_LIB_PATH        8
#define FN_ALTITUDE_MAP        9
#define FN_ATMOSPHERE         10
#define FN_RADIOSONDE         11
#define FN_ECMWF              12 
#define FN_ECMWF_WIND_MAP     13
#define FN_U10_MAP            14
#define FN_SALINITY_MAP       15
#define FN_PIGMENTS_MAP       16
#define FN_ECHAM              17
#define FN_EXTRATERRESTRIAL   18 
#define FN_SZA                19
#define FN_SATGEOMETRY        20
#define FN_WLTRANS            21
#define FN_WLBANDS            22 
#define FN_WCLOUD             23
#define FN_WCLOUD_FILES       24 
#define FN_WCLOUD_PROP        25
#define FN_WCLOUD_IPA         26
#define FN_WCLOUD_IPA_3D      27
#define FN_ICLOUD             28
#define FN_ICLOUD_FILES       29 
#define FN_ICLOUD_PROP        30 
#define FN_ICLOUD_IPA         31
#define FN_ICLOUD_IPA_3D      32
#define FN_CLOUD_FRACTION     33
#define FN_CLOUD_FRACTION_MAP 34
#define FN_SLITFUNCTION       35
#define FN_FILTERFUNCTION     36
#define FN_SPLINE             37
#define FN_RAYLEIGH           38
#define FN_MOLABS             39
#define FN_REFIND             40
#define FN_CK_GENERIC         41
#define FN_NETCDF_OUTFILE     42

#define FN_AER_NN            9  /* Number of aerosol related files, increase 
				   when adding new aerosol files */
#define FN_AER_TAU           0
#define FN_AER_GG            1
#define FN_AER_SSA           2
#define FN_AER_MOM           3
#define FN_AER_REF           4
#define FN_AER_SIZ           5
#define FN_AER_FILES         6
#define FN_AER_SPECIES       7
#define FN_AER_SPECIES_LIB   8

#define FN_MC_NN            12 /* Number of mystic related files, increase
				   when adding new mystic files */
#define FN_MC_PHOTONS        0
#define FN_MC_PCLD           1
#define FN_MC_BASENAME       2
#define FN_MC_ALBEDO         3
#define FN_MC_TEMPERATURE    4
#define FN_MC_ELEVATION      5
#define FN_MC_RPV            6
#define FN_MC_RPVTYPE        7
#define FN_MC_AMBRALS        8
#define FN_MC_WCLOUD         9
#define FN_MC_ICLOUD        10
#define FN_MC_LIDAR         11


#define CK_NONE         0
#define CK_KATO         1
#define CK_KATO2        2
#define CK_KATO2_96     3
#define CK_FU           4
#define CK_AVHRR_KRATZ  5
#define CK_GENERIC      6
#define CK_LOWTRAN      7
#define CK_RAMAN        8   /* This is not a CK-distribution, but a way to do many RTE solutions for one wavelength, AK */

#define WLGRID_NONE    0
#define WLGRID_CK      1
#define WLGRID_USER    2
#define WLGRID_BANDS   3
#define WLGRID_MOLABS  4
#define WLGRID_UVSPEC  5

#define SRC_NONE    0
#define SRC_SOLAR   1
#define SRC_THERMAL 2

#define RAYLEIGH_BODHAINE  1
#define RAYLEIGH_NICOLET   2
#define RAYLEIGH_PENNDORF  3

#define O3_MOLINA    1
#define O3_DAUMONT   2
#define O3_BASSPAUR  3    

#define CO2_YOSHINO  1

#define O2_DEFAULT   1

#define NO2_GOME     1

#define BRO_IASB     1

#define OCLO_WAHNER  1

#define HCHO_IASB  1

#define RAMAN_NONE         0
#define RAMAN_CALC         1   /* Raman scattering included */
#define RAMAN_DEFAULT      1   /* The default Raman scattering cross section */

#define RAYLEIGH_NONE      0
#define RAYLEIGH_CALC      1
#define RAYLEIGH_FILE_MONO 2
#define RAYLEIGH_FILE_SPEC 3

#define MOLABS_NONE      0
#define MOLABS_CALC      1
#define MOLABS_FILE_MONO 2
#define MOLABS_FILE_SPEC 3

#define MX_NN  9  
#define MX_O2  0
#define MX_CO2 1
#define MX_H2O 2
#define MX_CH4 3
#define MX_N2O 4
#define MX_NO2 5
#define MX_F11 6
#define MX_F12 7
#define MX_F22 8

#define CLD_ZD     0    
#define CLD_LWC    1
#define CLD_EFFR   2

#define BRDF_NONE   0
#define BRDF_RPV    1  /* don't change these numbers as they are used by Fortran code */
#define BRDF_CM     2  /* which of course has no access to this header file           */
#define BRDF_RPV2D  3
#define BRDF_CM2D   4

#define RPV_CONSTANT       0
#define RPV_FROM_RPV_FILE  1
#define RPV_IGBP_LIBRARY   2
#define RPV_USER_LIBRARY   3

#define DISORT      1
#define SDISORT     2
#define TWOSTR      3
#define TWOSTRPP    4 
#define SUCORD      5
#define MONTECARLO  6
#define POLRADTRAN  7
#define DISORT2     8
#define SPSDISORT   9
#define QDISORT    10
#define TZS        11
#define SSS        12
#define SSSI       13
#define SSS2       14
#define NULLSOLVER 15

#define WCPROP_NONE         0   /* no water cloud                            */
#define WCPROP_HU           1   /* "wc_file" and Hu and Stamnes [1993]       */
#define WCPROP_ECHAM4       2   /* "wc_file" and ECHAM4                      */
#define WCPROP_MIE          3   /* "wc_file" and "wc_properties mie"         */
#define WCPROP_FILE         4   /* "wc_file" and "wc_properties <filename>"  */
#define WCPROP_PHASE        5   /* mc_pcld_file, only MYSTIC                 */
#define WCPROP_EXPLICIT     6   /* "wc_files"                                */

#define ICPROP_NONE         0   /* no ice cloud                              */
#define ICPROP_FU           1   /* "ic_file" and Fu et al. [1996/98]         */
#define ICPROP_ECHAM4       2   /* "ic_file" and ECHAM4                      */
#define ICPROP_KEY          3   /* "ic_file" and Key et al. [2002]           */
#define ICPROP_YANG         4   /* "ic_file" and Yang/Key/Mayer              */
#define ICPROP_MIE          5   /* "ic_file" and "ic_properties mie"         */
#define ICPROP_FILE         6   /* "ic_file" and "ic_properties <filename>"  */
#define ICPROP_EXPLICIT     7   /* "ic_files"                                */
#define ICPROP_BAUM         8   /* "ic_properties baum"                      */
#define ICPROP_BAUMDETAILED 9   /* "ic_properties baum_detailed"             */

#define IC_REFF_FIXED       0   /* fixed r_eff in all heights                */
#define IC_REFF_OU          1   /* Ou and Liou, 1995, Atm Res. 35, 127-138   */

#define SATURATION_NONE     0
#define SATURATION_CLOUD    1
#define SATURATION_IPA      2

#define PROCESS_NONE        0
#define PROCESS_SUM         1
#define PROCESS_RGB         2
#define PROCESS_RGBNORM     3
#define PROCESS_INT         4
#define PROCESS_RAMAN       5

#define UNIT_NOT_DEFINED      -999
#define UNIT_PER_NM              0
#define UNIT_PER_CMm1            1
#define UNIT_PER_CK_BAND         2

#define NO_HEAT       0   /* no heating rate calculation                                               */
#define HEAT_LOCAL    1   /* heating rates from (k_abs * actinic fluxes)                               */
#define HEAT_LAYER_FD 2   /* heating rates calculated with Forward  Difference of the flux over layers */
#define HEAT_LAYER_CD 3   /* heating rates calculated with Centered Difference of the flux over layers */

/* main switch for interpolation onto zout and altitude levels */
#define Z_INTERPOLATE_LOGLIN  0      /* no interpolation                   */
#define Z_INTERPOLATE_LINMIX  1      /* linear mixing ratio interpolation  */
#define Z_INTERPOLATE_SPLINE  2      /* spline interpolation               */

/* assumptions for interpolation of profiles,   */
/* calculation of effective densities for       */
/* optical properties and columns               */
#define INTERP_METHOD_SPLINE     0  /* spline interpolation (columns not consistent)               */
#define INTERP_METHOD_LINEAR     1  /* linear profile                                              */
#define INTERP_METHOD_LOG        2  /* logarithmic profile                                         */
#define INTERP_METHOD_LINMIX     3  /* linear mixing ratio (only for gas densities other than air) */
#define INTERP_METHOD_LOG_SPLINE 4  /* first take logarithm, than spline (columns not consistent)  */

#define NO_ZOUT_INTERPOLATE    0  /* used for zout_interpolate */ 
#define ZOUT_INTERPOLATE       1  /* used for zout_interpolate */

#define TIME_NEAREST_DATE      0  /* used for time_interpolate */ 
#define TIME_INTERPOLATION     1  /* used for time_interpolate */

#define TIME_UTC        0
#define TIME_LAT        1

#define ZOUT_MYSTIC  -999
#define ZOUT_SURFACE -666
#define ZOUT_TOA      666
#define ZOUT_CPT      699

#define OUTPUT_USER_NONE           0
#define OUTPUT_USER_WAVE           1
#define OUTPUT_USER_WAVE_MAX       2
#define OUTPUT_USER_WAVE_MIN       3
#define OUTPUT_USER_WAVENUMBER     4
#define OUTPUT_USER_WAVENUMBER_MAX 5
#define OUTPUT_USER_WAVENUMBER_MIN 6
#define OUTPUT_USER_ZOUT_SUR       7
#define OUTPUT_USER_ZOUT_SEA       8
#define OUTPUT_USER_Z_SUR          9
#define OUTPUT_USER_P             10
#define OUTPUT_USER_T             11
#define OUTPUT_USER_T_D           12
#define OUTPUT_USER_T_SUR         13
#define OUTPUT_USER_THETA         14
#define OUTPUT_USER_THETA_E       15
#define OUTPUT_USER_EDIR          16
#define OUTPUT_USER_EGLO          17
#define OUTPUT_USER_EDN           18
#define OUTPUT_USER_EUP           19
#define OUTPUT_USER_ENET          20
#define OUTPUT_USER_ESUM          21
#define OUTPUT_USER_FDIR          22
#define OUTPUT_USER_FGLO          23
#define OUTPUT_USER_FDN           24
#define OUTPUT_USER_FUP           25
#define OUTPUT_USER_F             26
#define OUTPUT_USER_UU            27
#define OUTPUT_USER_UDIR          28
#define OUTPUT_USER_UGLO          29
#define OUTPUT_USER_UDN           30
#define OUTPUT_USER_UUP           31
#define OUTPUT_USER_U             32
#define OUTPUT_USER_ALB           33
#define OUTPUT_USER_SZA           34
#define OUTPUT_USER_C_P           35
#define OUTPUT_USER_HEAT          36
#define OUTPUT_USER_EMIS          37 
#define OUTPUT_USER_ABS           38 
#define OUTPUT_USER_W_RAD         39
#define OUTPUT_USER_M_RAD         40
#define OUTPUT_USER_N             41
#define OUTPUT_USER_RHO           42
#define OUTPUT_USER_VMR           43
#define OUTPUT_USER_MMR           44
#define OUTPUT_USER_RH            45
#define OUTPUT_USER_RH_ICE        46
#define OUTPUT_USER_CLWC          47
#define OUTPUT_USER_CIWC          48 
#define OUTPUT_USER_TCC           49 
#define OUTPUT_USER_CLOUDS        50 
#define OUTPUT_USER_WIND_U        51
#define OUTPUT_USER_WIND_V        52
#define OUTPUT_USER_WIND_W        53
#define OUTPUT_USER_DTDX          54
#define OUTPUT_USER_DTDY          55
#define OUTPUT_USER_DTDZ          56
#define OUTPUT_USER_HEAT_AD_X     57
#define OUTPUT_USER_HEAT_AD_Y     58
#define OUTPUT_USER_HEAT_AD_Z     59
#define OUTPUT_USER_HEAT_AD       60

#define OUTPUT_FORMAT_ASCII     0
#define OUTPUT_FORMAT_MATRIX    1
#define OUTPUT_FORMAT_FLEXSTOR  2
#define OUTPUT_FORMAT_NETCDF    3
#define OUTPUT_FORMAT_SATPIC    4

#define OUTLEVEL_ZOUT_ABOVE_SUR           0
#define OUTLEVEL_ZOUT_ABOVE_SEA           1
#define OUTLEVEL_PRESS                    2
#define OUTLEVEL_ATM_LEVELS               3
#define OUTLEVEL_ALL_LEVELS               4
#define OUTLEVEL_MODEL_LEVELS             5
#define OUTLEVEL_MODEL_LAYERS             6
#define OUTLEVEL_MODEL_LEVELS_AND_LAYERS  7

#define OUTCAL_ABSOLUTE        0
#define OUTCAL_THERMAL         1
#define OUTCAL_TRANSMITTANCE   2
#define OUTCAL_REFLECTIVITY    3
#define OUTCAL_BRIGHTNESS      4

#define ISCCP_WATER     0
#define ISCCP_ICE       1

#define MAXAWVN        47

#define FALSE           0
#define NO              0
#define TRUE            1
#define YES             1

#define ALBEDO_CONSTANT             0 
#define ALBEDO_FROM_ALBEDO_FILE     1
#define ALBEDO_FROM_ALBEDO_MAP      2
#define ALBEDO_FROM_EMISSIVITY_MAP  3
#define ALBEDO_IGBP_LIBRARY         4
#define ALBEDO_USER_LIBRARY         5

#define ALT_NOT_DEFINED  0
#define ALT_DIRECT_INPUT 1
#define ALT_FROM_MAP     2
#define ALT_FROM_ATM     3

#define NO_CLOUDS           0
#define CLOUDS_FROM_FILE    1
#define CLOUDS_IPA_FILES    2
#define CLOUDS_FROM_3D_FILE 3
#define CLOUDS_LAYER_FILES  4
#define CLOUDS_FROM_ECMWF   5
#define CLOUDS_FROM_ECHAM   6

/* cloud overlap */
#define CLOUD_OVERLAP_RAND     0
#define CLOUD_OVERLAP_MAX      1
#define CLOUD_OVERLAP_MAXRAND  2
#define CLOUD_OVERLAP_OFF      3

#define SZA_BY_TIME_AND_LOCATION  0
#define SZA_DIRECT_INPUT          1
#define SZA_FROM_SZA_FILE         2
#define SZA_ECHAM                 3

#define NOT_DEFINED_FLOAT   -999.0
#define NOT_DEFINED_INTEGER -999

/* in the style of netCDF types NC_xxx */
#define TYPE_CHAR    2
#define TYPE_SHORT   3
#define TYPE_INT     4
#define TYPE_FLOAT   5
#define TYPE_DOUBLE  6

/* additional error numbers for netCDF output */
#define ERROR_READ_MISSING_VALUE  -501

/* define ascending or descending axis */
#define ASCENDING   0
#define DESCENDING  1

/* minimum number of photons for each band*/
#define MIN_MCPHOTONS 1

/* physical constants */
#define gamma 6.673E-11        /* +/- 0.010E-11  Newton's gravitational constant in N m2 kg-2                              */ 
                               /* accuracy of gamma is quite poor, as it is hard to measure                                */
                               /* recommended value from Fattori, M., Towards an atom interferometric determination of the */
                               /* Newtonian graviational constant, 2003, Physics Letters A 318 (2003), 184-191             */
#define boltzmann  1.3806504e-23 /* boltzmann constant k, form wikipedia */
#define avogadro   6.022142e23   /* 1 mol particles */
 

/* data of the planet */
#define r_earth   6378000.0                       /* radius of the earth in m */
#define m_earth   5.97412E24                      /* mass of the earth in kg  */
#define g_surface gamma*m_earth/(r_earth*r_earth) /* gravitational acceleration */

/* averaged data of "air" */
#define c_p_dry_STD  1003.8   /* specific heat of dry air at constant pressure (J kg-1 K-1), standard T = 275 K */
#define c_v_dry_STD   716.87  /* specific heat of dry air at constant volumn   (J kg-1 K-1), standard T = 275 K */ 
#define mol_mass_air 28.977   /* mass per mol particle in g / mol (from function ck.c) */
#define mol_mass_wv  18.015   /* mass per mol particle in g / mol (from function ck.c) */
#define R_air  boltzmann*avogadro*1000.0/mol_mass_air  /* gas constant for dry air */

/**********************************************/
/*            structure definitions           */
/**********************************************/

/**************************/
/* Monte Carlo STRUCTURES */
/**************************/

/* Monte Carlo backward structure */

typedef struct {
  int yes;  /* flag if backward is desired */

  int islower;   /* x-index of lower left  sample pixel */
  int jslower;   /* y-index of lower left  sample pixel */
  int isupper;   /* x-index of upper right sample pixel */
  int jsupper;   /* y-index of upper right sample pixel */

  int writeallpixels;
  int writeback;

  int output;      /* which quantity: edir, edn, eup, ...    */

  int absorption;  /* flag: calculation of absorption yes/no */
} mc_backward_struct;

/* Monte Carlo input structure */

typedef struct {
  int     photons;

  int     Nx_sample;
  int     Ny_sample;

  double  dx_sample;
  double  dy_sample;

  int     Nr;   /* number of radii, for I3RC case 7 */
  int     Nt;   /* number of times, for I3RC case 7 */
  float   dt;   /* time interval,   for I3RC case 7 */

  int     ipa;            /* MYSTIC in IPA mode                         */
  int     escape;         /* escape probabilities for radiances         */
  double  escape_eps_ddis; /* epsilon for ddis for escape               */

  int     locest;         /* local estimator                            */
  int     jacobian;

  int cloner;             /* logical, is cloning on                     */
  double  relerr;         /* result converges if relative error smaller */
  double  le_abserr;      /* result converged if absolute error smaller */
  double  maxscatters;    /* photon killed if number of scatters larger */

  float   delta_scaling_mucut;       /* TZ ds*/
  float   truncate;       /* truncate phase function                    */

  int     spherical;      /* spherical flag                             */
  
  int     polarisation;    /* polarisation flag                          */
  int     nstokes;         /* number of Stokes components              */ 
  
  int     bcond;          /* boundary conditions                        */

  int     reflectalways;  /* reflect each photon and weight with albedo */

  int     sensorposition;   /* user-defined sensor position for         */
  double  sensorposition_x; /* sensors at point locations               */ 
  double  sensorposition_y; 
  double  sensorposition_z; 

  int     surfaceparallel;  /* irradiance parallel to the surface       */
                            /* instead of horizontal                    */

  int     sensordirection;    /* user-defined sampling direction for    */
  double  sensordirection_dx; /* specifically oriented sensors          */ 
  double  sensordirection_dy; 
  double  sensordirection_dz; 

  int     absorption;     /* calculate absorption for each box          */ 
  int     abs_unit;       /* output units for absorbed radiance         */

  char    **filename;

  int     azimuth;        /* new or old azimuth convention              */

  int     std;            /* calculate standard deviation               */
  int     randomseed;     /* seed for the random number generator       */
  int     readrandomseed;   /* read seed for the random number generator  */
  int     readrandomstatus; /* read status of the random number generator */

  mc_backward_struct backward;  /* information for backward MYSTIC      */

  int visualize;          /* OpenGL visualization                       */
} mc_inp_struct;

/* Monte Carlo output structure */
/* (MYSTIC uses profiles in reverted order (the lowest level is 0)) */

typedef struct {
  float       *dtair;           /* free'ed */
  float       *dtmol;           /* free'ed */
  float       *dtaer;           /* free'ed */
  float       *dtcld;           /* free'ed */
  float       *dtice;           /* free'ed */
  float       *omaer;           /* free'ed */
  float       *omcld;           /* free'ed */
  float       *omice;           /* free'ed */

  float       *g1aer;
  float       *g2aer;
  float       *ffaer;

  float       *g1cld;
  float       *g2cld;
  float       *ffcld;

  float       *dscld; /*TZ ds*/

  float       *g1ice;
  float       *g2ice;
  float       *ffice;

  float       *dsice; /*TZ ds*/

  float       *refind;

  float       ***momaer;
  int         *nmomaer;
  int         nphaseaer;

  float       *z;
  float       *temper;

  float       *recld;
  float       *reice;

  sample_struct sample;      /* 3d sampling            */
  elevation_struct elev;     /* 2d elevation           */
  surftemp_struct surftemp;  /* 2d surface temperature */
} mc_out_struct;


/***********************************************/
/* Raditive Transfer Equation STRUCTURE        */
/* See disort.doc for description of variables */
/***********************************************/

typedef struct {
  int     deltam;
  int     fisot;
  int     ichap;  
  int     nrefrac;  
  int     nstr;  
  int     nscat;
  int     maxphi;
  int     maxumu;
  int     nphi;
  int     numu;
  int     solver;
  int     nprndis;
  int     reverse;
  int*    prndis;
  int     pol_aziorder;
  int     pol_nstokes;
  char*   pol_quad_type;
  char    pol_stokes[4];
  int     pol_src_code;
  double  pol_max_delta_tau;
  float*  phi;
  float*  umu;
  int*    cmuind; /* Indices for the cmu angles, used with Raman option */
  int*    umuind; /* Indices for the umu angles, used with Raman option */
  mc_inp_struct mc;
} rte_struct;


/* Cloud output STRUCTURE */

typedef struct {
  /* quantities as defined by the user */
  float *lwc;        /* z-profile of the liquid water content */
  float *lwc_zout;   /* z-profile of the liquid water content at zout levels */
  float *effr;       /* z-profile of the effective radius     */

  /* layer quantities */
  float *lwc_layer;  /* z-profile of the liquid water content */
  float *effr_layer; /* z-profile of the effective radius     */

  float effrmin;     /* minimum effective radius */
  float effrmax;     /* maximum effective radius */
  
  int   alloc;       /* flag, if memory is allocated          */
} cld_microphys_struct;


/********************************************************/
/* optical properties of aerosol, water, and ice clouds */
/********************************************************/

typedef struct {
  float **dtau;     /* layer optical thickness [wavelength][layer]         */

  float **ff;       /* f  for double HG [wavelength][layer]                */
  float **g1;       /* g1 for double HG [wavelength][layer]                */
  float **g2;       /* g2 for double HG [wavelength][layer]                */

  float **ssa;      /* layer single scattering albedo [wavelength][layer]  */
  float **f;     /* delta-m scaling factor                              */

  int   nlam;       /* number of wavelengths                               */
	
  float ****moment;  /* moments of the phase function for each              */
                    /* wavelength and layer                                */
  int   nlev;       /* number of levels                                    */

  int   **nmom;     /* number of moments for each wavelength and layer     */
  int   nphase;     /* number of phase matrix elements    */

  float **dscale;   /* layer delta-scaling factor                     TZds */

  int alloc;        /* flag, if memory is allocated;                       */
} optprop_struct;


/* Cloud output STRUCTURE */
typedef struct {

  int   occurrence1D; /* are there any 1D clouds at all?            */
  float *zd;          /* level altitudes                            */
  int   nlev;         /* number of levels                           */

  cld_microphys_struct microphys;  /* microphysical properties      */
  optprop_struct       optprop;    /* optical properties            */
  ssprop_struct        ssprop;     /* single scattering properties  */

  /* the following two are used by MYSTIC only                      */
  int cldproperties; /* defined optical or microphysical properties */
  int nonHGcld;      /* Henyey-Greenstein or not                    */

  cld3d_out_struct     threed;     /* 3d water cloud structure      */

  int newsiz;      /* flag for the Fortran function wcloud() */

  float *z3D;      /* checking of the 3D grid */ 
  int   Nz3D;
  
} cld_out_struct;

/* Cloud input STRUCTURE */

typedef struct {
  int     source;  /* from which file: cloud file or ECMWF */
  int     layer;

  float   tau;  
  float   gg;  
  float   ssa; 
  float   reff; 

  float   tau550;  

  float scale_gg;
  float scale_ssa;

  int no_scattering;

  int properties;
  int reff_prop;
  int unscaled;
  int fu2yang;
  int interpolate;
  int habit;
  int wv_saturation;
  float wv_sat_ratio;
} cld_inp_struct;

/**********************/
/* Aerosol structures */
/**********************/

/* Aerosol input STRUCTURE */

typedef struct {
  int     standard;  /* Shettle standard aerosol defined    */
  int     spec;      /* specific aerosol properties defined */

  char **filename;   /* array of filenames */
  int haze;
  int vulcan;
  int seasn;
  float alpha;
  float beta;
  float visibility;
  
  int   n_species;        /* number of aerosol species */
  char  **species_names;  /* aerosol species name [i_aer] */

  float re;          /* refractive index            */
  float im;          /* (independent of wavelength) */

  float tau550;

  float scale_ssa;
  float scale_tau;
  float set_gg;
  float set_ssa;
  float set_tau;
  int   no_scattering;
} aer_inp_struct;

/* Aerosol output STRUCTURE */

typedef struct {
  float   visibility;

  float *zd;       /* level altitudes                  */
  int   nlev;      /* number of levels                 */

  float *zd_alt;       /* level altitudes, starting at 'altitude'   */
  int   nlev_alt;      /* number of levels, starting at 'altitude'  */

  /* optical properties profile */
  optprop_struct optprop;
  ssprop_struct ssprop; 
  optprop_struct *optprop_ipa; /* FIXMECE This should be removed. */
  optprop_struct *optprop_opac;

  int   n_species;        /* number of aerosol species */
  char  **species_names;  /* aerosol species name [i_aer] */
  float **massdens;       /* aerosol mass density profiles [i_aer][lev] */
  float **massdens_zout;  /* aerosol mass density at zout levels */
  char  *aerosol_library; /* library for aerosol optical properties */

  double   *xsiz;    /* size distribution              */
  double   *ysiz;
  int      nsiz;

  float   *re_t;     /* real and imaginary part of the */
  float   *im_t;     /* refractive index               */
} aer_out_struct;

/*********************/
/* Albedo structures */
/*********************/

/* albedo input structure */

typedef struct {
  int    source;          /* where to get the albedo from (constant, file, library) */
  int    surface;         /* reference number for spectral albedo type inside a library */
  float  albedo;          /* constant input albedo value */
  char  *netCDF_alb_name; /* name of the albedo variable in the netCDF file */
  char  *netCDF_surf_name; /* name of the albedo variable in the netCDF file */
  float  scale_factor;    /* in order to convert data in "per cent" or other ... */
  int    surf_type_map;   /* specifies, if the surface type is picked up in a map */
} alb_inp_struct;

/* albedo output structure */

typedef struct {

  int    source;         /* where to get the albedo from (constant, file, library) */
  int    surface;        /* reference number for spectral albedo type inside a library */
  float  albedo;         /* constant input albedo value */
  int    surf_type_map;  /* specifies, if the surface type is picked up in a map */

  int    file_nlambda_albedo;  /* number of elements in the albedo file   */
  float *file_albedo;          /* albedo on the spectral grid of the file */
  float *file_lambda_albedo;   /* wavelength grid of the albedo file      */

  float *albedo_t;             /* albedo on the actual transmission grid  */
  int    nlambda_albedo;

} alb_out_struct;

/* RPV (Rahman, Pinty, Verstraete) structure */

typedef struct {
  float rho0;
  float k;
  float theta;
  float sigma;  /* sigma, t1, t2: Deguenther and Meerkoetter [2000]; */
  float t1;     /* snow; includes forward peak.                      */
  float t2;
  float scale;  /* optional scaling factor for the BRDF, to adjust albedo */
  int   source;
} rpv_inp_struct;


typedef struct {
  float iso;  /* isotropic BRDF parameter  */
  float vol;  /* volumetric BRDF parameter */
  float geo;  /* geometric BRDF parameter  */
} ambrals_inp_struct;


/* Cox and Munk structure */

typedef struct {
  float u10;
  float pcl;
  char *pcl_netCDF_name;
  float pcl_scale_factor;
  float sal;
  float uphi;
  int   solar_wind;
  char *sal_netCDF_name;
  float sal_scale_factor;
} cm_inp_struct;


/* RPV (Rahman, Pinty, Verstraete) structure */
/* to store wavelength dependent BRDF        */
typedef struct {
  int   n;

  float *lambda;
  float *rho0;     /* input wavelength grid    */
  float *k;
  float *theta;
  float *scale;

  float *rho0_t;   /* internal wavelength grid */
  float *k_t;
  float *theta_t;
  float *scale_t;

} rpv_out_struct;

/* Array of RPV (Rahman, Pinty, Verstraete) structures */
/* for different surfaces, addressed by a label        */
typedef struct {
  int   n;

  char  **label;
  char  **filename;

  rpv_out_struct *rpv;
} surfaces_out_struct;


typedef struct {
  float tautot;
  int lctop;
  int type;
  float ref;
} sssi_out_struct;


/*************************/
/* Wavelength structures */
/*************************/

/* Wavelength input grid STRUCTURE */

typedef struct {
  float start;
  float end;

  int start_index;
  int end_index;
} wl_inp_struct;

/* Wavelength output grid STRUCTURE */

typedef struct {
  float start;
  float end;

  int   type;

  int   nlambda_h;
  int   nlambda_s;
  int   nlambda_t;

  int   nlambda_h_print3D;

  float *lambda_h;
  float *lambda_s;
  float *lambda_t;

  int   *map_e2h;

  int   ignore_solar_file;

  float *fbeam;
  float *filter;

  float *wvnmlo_t;
  float *wvnmhi_t;

  float *wvnmlo_h;
  float *wvnmhi_h;

  float delta_wvl_raman; /* The +- wavelengths to be considered for Raman shifts. */
  int   raman_start_id;
  int   raman_end_id;

  int  nlambda_rte_lower;
  int  nlambda_rte_upper;

  float *pfraction;
} wl_out_struct;


/**********************/
/* Altitude structure */
/**********************/

typedef struct {
  int    source;          /* where to get the altitude from (direct input or map)                        */
  float  altitude;        /* surface altitude over sea level                                             */
  float  scale_factor;    /* scale factor, e.g. 9.8 in order to convert from geopotential meter to meter */
  char  *netCDF_alt_name; /* name of the altitude variable in the netCDF file                            */
  float  altitude_dz;     /* fakes an increased vertical resolution, profiles are interpolated           */
} alt_inp_struct;

typedef struct {
  int    source;        /* where to get the altitude from (direct input, map, or atmosphere) */  
  float  altitude;      /* Same as altitude from input structure is set.
			   Otherwise obtained from atmosphere input file */
} alt_out_struct;

/*************************/
/* Atmosphere structures */
/*************************/

/* Atmosphere input data STRUCTURE */

typedef struct {
  char **filename;      /* Filename of density profile, input using dens_file  */
  int    rs_source;     /* where the radiosonde data comes from (file or ECMWF file) */
  int    n_rs_gas;      /* number of gases, which are given by radiosonde file */
  int   *rs_gas;        /* MOL-number for gases given by radiosonde data */
  int   *rs_unit;       /* specifies unit, in which radiosonde data is given in */  
  int    rs_add_upper_levels; /* marker in order to show, if levels above rs-levels are added or not */
  float *column;        /* Total column of a profile, input using dens_column  */
/*   float *scale_factor; /\* Factor to scale profile with, calculated           *\/ */
  int   *unit_column;   /* Unit of the column value, input using dens_column   */
  int   *unit_profile;  /* Unit of the profile, input using dens_file          */
  float *mol_mass;      /* molecular mass of one molucule in u                 */

  float *z_atm_forced_sea;  /* forced z_grid of the atmosphere relative to the sea surface */
  int    nz_atm_forced_sea; /* number of forced atmosphere levels                          */
  float *zout_sur;      /* output altitudes, relative to the surface           */
  float *zout_sea;      /* output altitudes, relative to sea level             */
  float *press_zout;    /* output levels in pressure  hPa                      */
  int    nzout;         /* number of output altitudes                          */
  int    zout_source;   /* input switch */

  float  sza;
  int    sza_source;    /* where to get sza from (direct input, file, or lat/lon/time)*/
  float  phi0;
  int    zout_interpolate; /* switch for zout-level interpolation */
  int    time_interpolate; /* switch for time interpolation (of netCDF data input) */ 
  int    z_interpolate;    /* group specification of the interpolation methods     */
                           /* of the atmospheric profiles                          */
  int    interpol_method_press;  /* individual z_interpolation method  */
  int    interpol_method_temper; /* individual z_interpolation method  */
  int    interpol_method_refind; /* individual z_interpolation method  */
  int   *interpol_method_gas;    /* individual z_interpolation methods */
  int   *well_mixed_gas;         /* specifies if a gas is well mixed   */

  int    ECMWF_new_format;
  int    ECMWF_ozone_climatology;

} atm_inp_struct;

/* Atmosphere output data STRUCTURE */

typedef struct {

  float *press;
  float *refind;
  float *temper;
  float *temper_avg;     /* average temperature temper_avg[lc] in the layer from z[lc] to z[lc+1]  */ 

  float **dens;          /* density profiles */
  float **dens_avg;      /* layer average density dens_avg[lc] in the layer from z[lc] to z[lc+1]  */
  float ***denstab;      /* density profiles as function of altitude and sza */
  float ***denstab_avg;  /* layer average density profiles as function of altitude and sza */
  float **denstab_amf;   /* density profile as function of altitude and sza, for sdisort */
  float **dens_zout;     /* gas number densities  on the zout - grid [MOL_GAS][lc] */
  float *press_zout;     /* pressure              on the zout - grid */
  float *temper_zout;    /* temperature           on the zout - grid */
  float *temper_d_zout;  /* dewpoint temperature  on the zout - grid */
  float *dtheta_dx;     /* derivative of temperature with respect to x, this only works with atm_maps */
  float *dtheta_dy;     /* derivative of temperature with respect to y, this only works with atm_maps */
  float *dtheta_dz;      /* derivative of potential temperature with respect to z */
  float *c_p;            /* specific heat capacity at constant pressure of moist air */
  float *theta;          /* potential temperature at atmospheric grid */
  float *theta_zout;     /* potential temperature on the zout - grid  */
  float *theta_e_zout;   /* equivalent potential temperature on the zout - grid */

  float *sza_denstab;    /* Ary for the sza in denstab */
  
  float z_cpt_sea;       /* height of the cold point tropopause above sea level */
  float temper_cpt;      /* temperature at the cold point tropopause */
  float press_cpt;       /* pressure at the cold point tropopause */

  int   denstab_id;      /* Density id for denstab, MOL_* */
  int   nsza_denstab;    /* Number of sza in denstab */

} atm_microphys_struct;


typedef struct {
  
  double ***tau_rayleigh_t;    /* wavelength-dependent Rayleigh scattering optical depth   */
  double ***tau_raman_t;       /* wavelength-dependent Raman scattering optical depth      */
  double ***tau_molabs_t;      /* wavelength-dependent molecular absorption optical depth  */
  double ***tau_molabs_md_t;   /* wavelength-dependent molecular absorption optical depth  */
			       /* with density-matrix absorption subtracted. For AMF       */
			       /* calculations                                             */
                               /* format: tau_molabs_t[lc][iv][iq]                         */
                               /* ??? this is not consistent with tau_molabs_user_t[][]    */
                               /* ??? and should be changed accordingly                    */
  double ****tau_molabs_t_ipa; /* h2o saturation inside of cloud dependent on IPA-clouds   */
                               /* format: tau_molabs_t_ipa[ic][lc][iv][iq]                 */

  /* The following are explicitely defined by the user and do not depend on subband iq     */
  /* they are copied to the above structures which are used as input to the solvers        */
  /* ATTENTION: These structures are not redistributed and must therefore not be used      */
  /* after the call to setup_redistribute(); optical_properties relies only on the above   */
  /* quantities.                                                                           */

  double  *tau_rayleigh_user;   /* monochromatic Rayleigh scattering optical depth         */
  double **tau_rayleigh_user_t; /* wavelength-dependend Rayleigh scattering optical depth  */
                                /* format: tau_rayleigh_user_t[iv][lc]                     */
  double  *tau_molabs_user;     /* monochromatic molecular absorption optical depth        */
  double **tau_molabs_user_t;   /* wavelength-dependend molecular absorption optical depth */
                                /* format: tau_molabs_user_t[iv][lc]                       */
} atm_optprop_struct;


typedef struct {
  char* pol_scat_files;
  float *cdento;

  float *temper_out;

  atm_microphys_struct microphys;
  atm_optprop_struct   optprop;

  int   molabs;                 /* molecular absorption flag         */
  int   rayleigh;               /* Rayleigh scattering flag          */
  int   raman;                  /* Raman scattering flag             */

  float *zd;
  int   nlev;
  int   nlyr;

  float *zout_sur;     /* output altitudes, relative to the surface    */
  float *zout_sur_org; /* first zout grid, which is in each case there */
  float *zout_sea;     /* output altitudes, relative to sea level      */
  float *zout_sea_org; /* first zout grid, which is in each case there */
  float *press_zout_org;    /* first pressure_out grid */
  int    nzout;        /* number of output altitudes                   */
  int    nzout_user;   /* number of output altitudes, same as nzout,   */
                       /* except when internally more zout are needed, */
                       /*eg. for raman scattering                      */
  int   *zout_user_index;/* indecies for the user output altitudes     */
  int   *zout_comp_index;/* indecies for the internal computational altitudes     */
  int    nzout_org;    /* number of output altitudes, of org grids     */
  int   *zout_index;   /* relationship between the zout indecies       */

  float *zd_common;
  int    nlev_common;
  int    nlyr_common;

  float *z_model_level;        /* altitudes from model data (ECMWF or ECHAM) */
  int   nz_model_level;       /* number of altitudes levels of the model data (ECMWF or ECHAM) */

  float *z_model_layer;        /* altitudes from model data (ECMWF or ECHAM) */
  int   nz_model_layer;       /* number of altitudes levels of the model data (ECMWF or ECHAM) */

  double **wght_t;        /* ck quadrature weight                   */
  double ***ipa_wght_t;   /* independent pixel ck quadrature weight */
  int    *nq_t;           /* number of ck quadrature weights        */

  float *file_phi0;
  float *file_sza;
  float *file_lambda_sza;
  float *sza_t;
  float *phi0_t;

  int    file_nlambda_sza;
  int    nlambda_sza;

  int    nmom;  

  int nthreed;   /* number of 3D layers   */
  int *threed;   /* flag if a layer is 3D */
  
  int Nxcld;     /* 3D horizontal cloud grid */
  int Nycld;     /* 3D horizontal cloud grid */
  int Nzcld;     /* 3D horizontal cloud grid */
} atm_out_struct;

/*******************************************************************/
/* The ck structure holds all the information required for         */
/* a correlated-k approximation.                                   */
/* Components:                                                     */
/*    scheme               CK_KATO, CK_FU                          */
/*    wvlc  [1 ... n_wvl]  central wavelength [nm]                 */                   
/*    wvnlo [1 ... n_wvl]  lower wavenumber [cm-1] for each band   */    
/*    wvnhi [1 ... n_wvl]  upper wavenumber [cm-1] for each band   */    
/*    n_wvl                number of wavelength bands              */
/*    comp  [1 ... n_comp][1 ... n_wvl]                            */
/*                         monochromatic cross sections bands      */
/*                         which don't require c-k but only a      */
/*                         monochromatic calculation               */
/*                         (e.g. ozone in the UV/VIS)              */
/*    wght  [1 ... n_comp][1 ... n_wvl]                            */
/*                         number of quadrature points             */
/*******************************************************************/

typedef struct {   /* correlated-k structure */
  int scheme;

  double *wvlc;
  double *wvnlo;
  double *wvnhi;
  int n_wvl;

  double **comp;
  int    **wght;

  void *x_thermal;  /* photon fraction; x_thermal[iv][iq1][iq2][...] */
  void *x_solar;    /* photon fraction; x_solar  [iv][iq1][iq2][...] */

  int n_comp;
} ck_struct;

/* Cross section output data STRUCTURE */

typedef struct {
  char  **filename;  /*  Filename of cross section versus wavelength, input using crs_file  */
  float ***crs;      /* 3D ary, for cross section[temperature (or layer), type, wavelength] */
  float **crs_amf;   /* 2D ary for amf calculations, needed by sdisort */
  int number_of_ramanshifts;      /* Total number of Raman transitions considered */
  int number_of_ramanwavelengths; /* Total number of Raman transitions considered plus the wanted wavelength*/
  float *wvl_of_ramanshifts; /* Wavelengths of Raman transitions considered for each user wavelength*/
  float **crs_raman_RL;   /* 2D ary for  Raman scattering cross section [layer][Raman shifted wavelengths]              */
                          /* weighted by N2 and O2 volume mixing ratio. Raman loss terms, shifted from wanted wavelength*/  
  float **crs_raman_RG;   /* 2D ary for  Raman scattering cross section [layer][Raman shifted wavelengths]              */
                          /* weighted by N2 and O2 volume mixing ratio. Raman gain terms, shifted to wanted wavelength*/  
} crs_out_struct;

typedef struct {
  float** kabs; /* 2D ary, for absorption coefficient[temperature (or layer), wavelength] */
} kabs_out_struct;

typedef struct {
  double **crs;      /* cross section [layer, quadrature point] */
  double *weight;    /* weight [quadrature point]               */
  int ngauss;        /* number of quadrature points             */
  int nlev;
} ck_profile;

typedef struct {
  ck_profile **profile;      /* 2D ary, for ck_profile [type, band] */
  ck_profile ***ipa_profile; /* same as profile for different independent columns [ipa][type][band] */
} crs_ck_out_struct;


typedef struct {
  float *cf;
  float cf_total;
  float *zd;
  int nlev;
} cf_out_struct;


/* Wind STRUCTURE */

typedef struct {
  /* quantities as defined by the user */
  float *u;   /* z-profile of the wind component in m/s */
  float *v;   /* z-profile of the wind component in m/s */
  float *w;   /* z-profile of the wind component in m/s */
  float *z;   /* z-levels in km */
  size_t nlev; /* number of levels from the arrays above */

  float *u_zout;   /* wind component in m/s at zout levels */
  float *v_zout;   /* wind component in m/s at zout levels */
  float *w_zout;   /* wind component in m/s at zout levels */
} wind_out_struct;

			    
/* INPUT STRUCTURE */

typedef struct {
  int quiet;
  int verbose;

  int          spectrum_unit; /* kind of units: W/(m**2 nm) or W/(m**2 cm-1) or W/(m**2 band) */

  char         **filename;    /* array of filenames */
  
  float        wc_cloudcover;
  float        ic_cloudcover;

  int          wcipa;         /* independent pixel for water clouds */
  int          icipa;         /* independent pixel for ice clouds   */
  int          cloud_overlap; 

  int          convolve;
  int          spline;
  int          header;
  int          output_format;
  int          source;
  double       bandwidth;
  int          bandwidth_unit;
  int          raman;
  int          n_raman_transitions_N2; /* Number of transitions considered for N2 */
  int          n_raman_transitions_O2; /* Number of transitions considered for O2 */
  int          rayleigh;
  int          rayleigh_crs;
  float        rayleigh_depol;
  int          molabs;
  int          absorption;
  int          scattering;
  float        spline_lambda_0;
  float        spline_lambda_1;
  float        spline_lambda_step;
  struct tm    UTC;  /* time as Universal Time Coordinated*/
  struct tm    LAT;  /* time as Local Apparent Time */
  int          time_zone;
  float        delta_time_start;   /* in seconds, for an sza averaged over an time interval */
  float        delta_time_end;     /* in seconds, for an sza averaged over an time interval */
  float        latitude;     
  float        longitude;   
  int          sat_pixel_x;
  int          sat_pixel_y;
  int          ozone_spectrum;
  int          co2_spectrum;
  int          o2_spectrum;
  float        no2_column;
  int          no2_spectrum;
  int          bro_spectrum;
  int          oclo_spectrum;
  int          hcho_spectrum;
  float        precip_water;
  float        pressure;
  float        surface_temperature;
  char        *netCDF_name_surf_T;
  float       *mixing_ratio;
  int          calibration;
  int          optimize_fortran;
  int          optimize_delta;

  int    brdf;
  /* float  u10; */  /* wind speed for Cox and Munk parameterization, probably not used */

  rte_struct     rte;
  atm_inp_struct atm;
  alt_inp_struct alt;
  alb_inp_struct alb;
  aer_inp_struct aer;

  rpv_inp_struct rpv;
  ambrals_inp_struct ambrals;
  cm_inp_struct cm;

  wl_inp_struct  wl;
  cld_inp_struct wc;
  cld_inp_struct ic;
  int            ck_scheme;
  int            ck_h2ocont;

  int            ck_o4abs;
  int            ck_n2abs;
  int            ck_coabs;
  int            ck_so2abs;
  int            ck_nh3abs;
  int            ck_noabs;
  int            ck_hno3abs;

  int            processing;
  int            output_unit;
  int            heating;

  int    n_output_user;
  int   *output_user;     /* output_user    [column of output] = kind of output */
  int   *output_user_gas; /* output_user_gas[column of output] = MOL_GAS        */
} input_struct;


/* OUTPUT STRUCTURE */

/* elements whose memory requirements have been free'ed are listed behind */
/* each declaration. Once they are all done rm this comment.              */

typedef struct {
  float       eccentricity_factor;
  float       sunshine_fraction;     /* part of the time interval, where sun is above horizon */
  int         spectrum_unit;         /* kind of units: W/(m**2 nm) or W/(m**2 cm-1) or W/(m**2 band) */
  int         bandwidth_unit;        /* kind of units: nm or cm-1 */

  float       *column;               /* Total column of a profile, input using dens_column */
  float       *scale_factor;         /* Factor to scale profile with, calculated           */
  float      **column_denstab;       /* Total column of a profile, input using dens_column [gas][sza] */
  float      **scale_factor_denstab; /* Factor to scale profile with, calculated           [gas][sza] */

  float       precip_water;
  float       water_scale_factor;

  float       *mixing_ratio;

  float       *rayleigh_depol;  /* wavelength-dependent depolarization */

  int         mc_photons;
  double      **mc_photons_t;   /* photon fraction for each wavelength and quadrature point */

  float       *dtauc;           /* free'ed */
  float       *dtauc_md;        /* free'ed */
  float       *ssalb;
  float       *ssalbR;
  float       *ssalbRL;
  float       *utau;
  float       ***pmom;

  float       surface_temperature;

  float        wc_cloudcover;
  float        ic_cloudcover;
  /* flux, actinic flux, and intensities on the output grid */ 
  float       **rfldir;     /* direct flux                    */
  float       **rfldn;      /* diffuse downward flux          */
  float       **flup;       /* diffuse upward flux            */ 

  float       **uavgso;     /* direct actinic flux            */
  float       **uavgdn;	    /* diffuse downward actinic flux  */
  float       **uavgup;	    /* diffuse upward actinic flux    */

  float       **uavg;       /* average intensity              */
  float       ***u0u;       /* azimutally averaged intensites */
  float       ****uu;       /* intensities                    */
  float       **heat;       /* heating rates,                   f(zout,lambda) */
  float       **emis;       /* thermal emitted energy in K/day, f(zout,lambda) */
  float       **w_zout;     /* vertical wind caused by radiative heating on the zout - grid */

  /* flux, actinic flux, and intensities on the internal transmittance grid */ 
  float       **rfldir_t;  /* direct flux                    */
  float       **rfldn_t;   /* diffuse downward flux          */
  float       **flup_t;	   /* diffuse upward flux            */
			                                       
  float       **uavgso_t;  /* direct actinic flux            */
  float       **uavgdn_t;  /* diffuse downward actinic flux  */
  float       **uavgup_t;  /* diffuse upward actinic flux    */
			                                       
  float       **uavg_t;	   /* average intensity              */
  float       ***u0u_t;	   /* azimutally averaged intensites */
  float       ****uu_t;	   /* intensities                    */
  float       **heat_t;    /* heating rates;                             depends on (zout,lambda) */
  float       **emis_t;    /* thermal emitted energy in K/day;           depends on (zout,lambda) */
  float       **w_zout_t;  /* vertical wind caused by radiative heating; depends on (zout,lambda) */

  /* wavelength integrated flux, actinic flux, and intensities on the output grid */
  double      *rfldir_int;  /* direct flux                    */
  double      *rfldn_int;   /* diffuse downward flux          */
  double      *flup_int;    /* diffuse upward flux            */ 

  double      *uavgso_int;  /* direct actinic flux            */
  double      *uavgdn_int;  /* diffuse downward actinic flux  */
  double      *uavgup_int;  /* diffuse upward actinic flux    */

  double      *uavg_int;    /* average intensity              */
  double      **u0u_int;    /* azimutally averaged intensites */
  double      ***uu_int;    /* intensities                    */  

  double      *heat_int;    /* heating rates f(z) */
  double      *emis_int;    /* heating rates f(z) */
  double      *w_zout_int;  /* vertical wind caused by radiative heating; f(z) */

  /* integrated properties */
  double      incident;


  /* MYSTIC 3D fields */
  int islower;  /* to save memory, only those pixels are allocated */
  int isupper;  /* which are actually sampled; in partcular for    */
  int jslower;  /* backward Monte Carlo only a small fraction of   */
  int jsupper;  /* the pixels is usually sampled .                 */

  /* MYSTIC 3D fields on the output grid */  
  float       ****rfldir3d; 
  float       ****rfldn3d;
  float       ****flup3d;
  float       ****uavgso3d; 
  float       ****uavgdn3d; 
  float       ****uavgup3d; 
  float       *****radiance3d;
  float       ****abs3d;
  float       ****absback3d;

  /* MYSTIC 3D fields on the internal transmittance grid */  
  float       ****rfldir3d_t;
  float       ****rfldn3d_t;
  float       ****flup3d_t;
  float       ****uavgso3d_t;
  float       ****uavgdn3d_t;
  float       ****uavgup3d_t;
  float       *****radiance3d_t;
  float       ****abs3d_t;
  float       ****absback3d_t;

  /* PolRadtran flux and intensities on the output grid */ 
  float       ***down_flux;     /* free'ed */
  float       ***up_flux;       /* free'ed */
  float       *****down_rad;    /* free'ed */
  float       *****up_rad;      /* free'ed */

  /* wavelength integrated PolRadtran flux and intensities on the output grid */ 
  double      **down_flux_int;     /* free'ed */
  double      **up_flux_int;       /* free'ed */
  double      ****down_rad_int;    /* free'ed */
  double      ****up_rad_int;      /* free'ed */

  /* PolRadtran flux and intensities on the internal transmittance grid */ 
  float       ***down_flux_t;   /* free'ed */
  float       ***up_flux_t;     /* free'ed */
  float       *****down_rad_t;  /* free'ed */
  float       *****up_rad_t;    /* free'ed */

  float       *mu_values;       /* free'ed */
  float       *sza_h;           /* free'ed */
  float       *sza_s;           /* free'ed */

  atm_out_struct      atm;
  alt_out_struct      alt;
  alb_out_struct      alb;
  rpv_out_struct      rpv;
  surfaces_out_struct surf;
  wind_out_struct     wind;

  crs_out_struct      crs;
  crs_ck_out_struct   crs_ck;
  kabs_out_struct     kabs;

  aer_out_struct      aer;

  cld_out_struct    wc;
  cld_out_struct    *wcipa;
  cld_out_struct    ic;
  cld_out_struct    *icipa;
  cf_out_struct     cf;

  mc_out_struct     mc;
  sssi_out_struct   sssi;

  float             *ipaweight;
  int               nipa;

  int               print_phi;

  wl_out_struct     wl;
  ck_struct         ck;
} output_struct;

int uvspec(input_struct input, output_struct *output);
int uvspec_check(input_struct input);


/* include all uvspec header files */
#include "albedo.h"
#include "aerosol.h"
#include "ancillary.h"
#include "atmosphere.h"
#include "ck.h"
#include "cloud.h"
#include "cloud3d.h"
#include "extraterrestrial.h"
#include "ipa.h"
#include "molecular.h"
#include "redistribute.h"
#include "solve_rte.h"
#include "sza.h"
#include "yang56.h"
#include "specrend_uvspec.h"

#if HAVE_SSS2
  #include "sss2.h"
#endif

#endif
